Package: mLLMCelltype
Type: Package
Title: Cell Type Annotation Using Large Language Models
Version: 2.0.7
Author: Chen Yang [aut, cre, cph]
Maintainer: Chen Yang <cafferychen777@tamu.edu>
Authors@R: 
    person("Chen", "Yang", email = "cafferychen777@tamu.edu",
           role = c("aut", "cre", "cph"))
Description: Automated cell type annotation for single-cell RNA sequencing data 
    using consensus predictions from multiple large language models. Integrates 
    with Seurat objects and provides uncertainty quantification for annotations.
    Supports various LLM providers including OpenAI, Anthropic, and Google.
    For details see Yang et al. (2026) <doi:10.1038/s42003-026-10420-8>.
License: MIT + file LICENSE
BugReports: https://github.com/cafferychen777/mLLMCelltype/issues
URL: https://cafferyang.com/mLLMCelltype/
Encoding: UTF-8
Imports: httr (>= 1.4.0), jsonlite (>= 1.7.0), R6 (>= 2.5.0), digest
        (>= 0.6.25), stats, tools, utils
Suggests: knitr, rmarkdown, Seurat, testthat
RoxygenNote: 7.3.3
Config/build/clean-inst-doc: TRUE
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-07-21 01:36:59 UTC; apple
Repository: CRAN
Date/Publication: 2026-07-21 08:20:09 UTC

# Generated by roxygen2: do not edit by hand

export(AnthropicProcessor)
export(BaseAPIProcessor)
export(CacheManager)
export(DeepSeekProcessor)
export(GeminiProcessor)
export(GrokProcessor)
export(KimiProcessor)
export(MinimaxProcessor)
export(OpenAIProcessor)
export(OpenRouterProcessor)
export(QwenProcessor)
export(StepFunProcessor)
export(UnifiedLogger)
export(ZhipuProcessor)
export(annotate_cell_types)
export(compare_model_predictions)
export(configure_logger)
export(create_annotation_prompt)
export(create_reasoning_annotation_prompt)
export(get_api_key)
export(get_logger)
export(get_provider)
export(interactive_consensus_annotation)
export(list_custom_models)
export(list_custom_providers)
export(log_debug)
export(log_error)
export(log_info)
export(log_warn)
export(mllmcelltype_cache_dir)
export(mllmcelltype_clear_cache)
export(register_custom_model)
export(register_custom_provider)
importFrom(R6,R6Class)
importFrom(digest,digest)
importFrom(jsonlite,toJSON)
importFrom(stats,setNames)
importFrom(tools,R_user_dir)
importFrom(utils,head)
importFrom(utils,tail)

# mLLMCelltype R Package

This directory contains additional resources for the mLLMCelltype R package.

## Package Structure

- `R/`: Contains R source code
- `man/`: Contains documentation files
- `tests/`: Contains test files
- `inst/`: Contains installed files (this directory)

## Usage

See the main README.md file in the project root directory for usage instructions.

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)

## -----------------------------------------------------------------------------
# # Conceptual representation of the initial annotation process
# initial_results <- list()
# for (model in models) {
#   initial_results[[model]] <- annotate_cell_types(
#     input = marker_data,
#     tissue_name = tissue_name,
#     model = model,
#     api_key = api_keys[[get_provider(model)]]
#   )
# }

## -----------------------------------------------------------------------------
# # Conceptual representation of controversial cluster identification
# controversial_clusters <- identify_controversial_clusters(
#   initial_results,
#   threshold = discussion_threshold
# )

## -----------------------------------------------------------------------------
# # Conceptual representation of the discussion process
# discussion_results <- facilitate_cluster_discussion(
#   controversial_clusters,
#   initial_results,
#   marker_data,
#   tissue_name,
#   discussion_model,
#   api_key
# )

## -----------------------------------------------------------------------------
# # Conceptual representation of consensus formation
# final_annotations <- combine_results(
#   initial_results,
#   discussion_results,
#   controversial_clusters
# )

## -----------------------------------------------------------------------------
# # Conceptual calculation of consensus proportion
# consensus_proportion <- sapply(clusters, function(cluster) {
#   annotations <- sapply(models, function(model) initial_results[[model]][cluster])
#   most_common <- names(which.max(table(annotations)))
#   sum(annotations == most_common) / length(annotations)
# })

## -----------------------------------------------------------------------------
# # Conceptual calculation of Shannon entropy
# shannon_entropy <- sapply(clusters, function(cluster) {
#   annotations <- sapply(models, function(model) initial_results[[model]][cluster])
#   p <- table(annotations) / length(annotations)
#   -sum(p * log2(p))
# })

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## ----setup--------------------------------------------------------------------
# library(mLLMCelltype)

## ----installation-------------------------------------------------------------
# install.packages("mLLMCelltype")

## ----installation-dev---------------------------------------------------------
# # install.packages("devtools")
# devtools::install_github("cafferychen777/mLLMCelltype")

## ----api-keys-----------------------------------------------------------------
# # Set API keys (recommended to use .Renviron or .env file)
# # OpenAI API key
# Sys.setenv(OPENAI_API_KEY = "your-openai-api-key")
# 
# # Anthropic API key
# Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-api-key")
# 
# # Google API key
# Sys.setenv(GOOGLE_API_KEY = "your-google-api-key")
# 
# # X.AI API key
# Sys.setenv(XAI_API_KEY = "your-xai-api-key")
# 
# # Alternatively, use .env file
# # dotenv::load_dot_env()

## ----basic-usage--------------------------------------------------------------
# # Load required packages
# library(mLLMCelltype)
# library(Seurat)
# library(dplyr)
# 
# # Load your preprocessed Seurat object
# # pbmc <- readRDS("your_seurat_object.rds")
# 
# # Find marker genes for each cluster
# # pbmc_markers <- FindAllMarkers(pbmc,
# #                           only.pos = TRUE,
# #                           min.pct = 0.25,
# #                           logfc.threshold = 0.25)
# 
# # Cache will be automatically managed using system cache directory
# 
# # Run LLMCelltype annotation with multiple LLM models
# consensus_results <- interactive_consensus_annotation(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",  # provide tissue context
#   models = c(
#     "claude-sonnet-4-6",  # Anthropic
#     "gpt-5.5",                   # OpenAI
#     "gemini-3.1-pro-preview"            # Google
#   ),
#   api_keys = list(
#     anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
#     openai = Sys.getenv("OPENAI_API_KEY"),
#     gemini = Sys.getenv("GOOGLE_API_KEY")
#   ),
#   top_gene_count = 10,
#   controversy_threshold = 1.0,
#   entropy_threshold = 1.0,
#   consensus_check_model = "claude-sonnet-4-6",  # Use a reliable model for consensus checking
#   cache_dir = NULL  # Uses default system cache directory
# )
# 
# # Add annotations to Seurat object
# # Get cell type annotations from consensus_results$final_annotations
# cluster_to_celltype_map <- consensus_results$final_annotations
# 
# # Create new cell type identifier column
# cell_types <- as.character(Idents(pbmc))
# for (cluster_id in names(cluster_to_celltype_map)) {
#   cell_types[cell_types == cluster_id] <- cluster_to_celltype_map[[cluster_id]]
# }
# 
# # Add cell types to Seurat object
# pbmc$mLLM_cell_type <- cell_types

## ----visualization------------------------------------------------------------
# # Plot UMAP with cell type annotations
# library(Seurat)
# library(ggplot2)
# library(cowplot)
# 
# # Basic visualization
# p1 <- DimPlot(pbmc, group.by = "mLLM_cell_type", label = TRUE) +
#   ggtitle("mLLMCelltype Consensus Annotations")
# 
# # Visualize consensus proportion (confidence metric)
# pbmc$consensus_proportion <- 0
# for (cluster_id in names(consensus_results$initial_results$consensus_results)) {
#   cluster_cells <- which(Idents(pbmc) == cluster_id)
#   pbmc$consensus_proportion[cluster_cells] <-
#     consensus_results$initial_results$consensus_results[[cluster_id]]$consensus_proportion
# }
# 
# p2 <- FeaturePlot(pbmc, features = "consensus_proportion",
#                  min.cutoff = 0, max.cutoff = 1) +
#   scale_color_gradientn(colors = c("blue", "green", "red")) +
#   ggtitle("Consensus Proportion") +
#   theme(legend.position = "right")
# 

## ----single-model-------------------------------------------------------------
# # Run annotation with a single model
# single_model_results <- annotate_cell_types(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#   top_gene_count = 10
# )
# 
# # Inspect the provider response lines. Parse or review this text before
# # mapping annotations back onto a Seurat object.
# cat(single_model_results)

## ----custom-consensus---------------------------------------------------------
# # Customize consensus parameters
# custom_consensus_results <- interactive_consensus_annotation(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
#   api_keys = list(
#     anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
#     openai = Sys.getenv("OPENAI_API_KEY"),
#     gemini = Sys.getenv("GOOGLE_API_KEY")
#   ),
#   top_gene_count = 15,                # Number of top marker genes to use
#   controversy_threshold = 0.7,        # Threshold for controversy detection
#   entropy_threshold = 0.7,            # Entropy threshold for controversy detection
#   max_discussion_rounds = 2,          # Maximum rounds of discussion
#   consensus_check_model = "claude-sonnet-4-6",  # Use a reliable model for consensus checking
#   cache_dir = NULL  # Uses default system cache directory
# )

## ----custom-provider----------------------------------------------------------
# # Register a custom provider
# register_custom_provider(
#   provider_name = "my_custom_provider",
#   process_fn = my_process_function,
#   description = "My custom LLM provider"
# )
# 
# # Register a model served by that provider
# register_custom_model(
#   model_name = "my-model-1",
#   provider_name = "my_custom_provider",
#   model_config = list()
# )
# 
# # Use the custom model
# results <- annotate_cell_types(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   model = "my-model-1",
#   api_key = Sys.getenv("MY_CUSTOM_PROVIDER_API_KEY")
# )

## ----caching------------------------------------------------------------------
# # Enable caching
# Sys.setenv(MLLM_CACHE_ENABLED = "TRUE")
# 
# # Set cache directory
# Sys.setenv(MLLM_CACHE_DIR = "path/to/cache")
# 
# # Clear cache if needed
# mllmcelltype_clear_cache()

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)

## -----------------------------------------------------------------------------
# library(mLLMCelltype)

## -----------------------------------------------------------------------------
# # Set API keys as environment variables
# Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-api-key")  # For Claude models
# Sys.setenv(OPENAI_API_KEY = "your-openai-api-key")        # For GPT models
# Sys.setenv(GEMINI_API_KEY = "your-gemini-api-key")        # For Gemini models
# Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key") # For OpenRouter models

## -----------------------------------------------------------------------------
# results <- annotate_cell_types(
#   input = markers,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = "your-anthropic-api-key",  # Direct API key
#   top_gene_count = 10
# )

## -----------------------------------------------------------------------------
# # Example marker data frame
# markers_df <- data.frame(
#   cluster = c(0, 0, 0, 1, 1, 1),
#   gene = c("CD3D", "CD3E", "CD2", "CD14", "LYZ", "CST3"),
#   avg_log2FC = c(2.5, 2.3, 2.1, 3.1, 2.8, 2.5),
#   p_val_adj = c(0.001, 0.001, 0.002, 0.0001, 0.0002, 0.0005)
# )

## -----------------------------------------------------------------------------
# # Assuming you have a Seurat object named 'seurat_obj'
# library(Seurat)
# all_markers <- FindAllMarkers(seurat_obj, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)

## -----------------------------------------------------------------------------
# # Path to your CSV file
# markers_file <- "path/to/markers.csv"

## -----------------------------------------------------------------------------
# # Example marker list
# markers_list <- list(
#   "0" = c("CD3D", "CD3E", "CD2", "IL7R", "LTB"),
#   "1" = c("CD14", "LYZ", "CST3", "MS4A7", "FCGR3A")
# )

## -----------------------------------------------------------------------------
# # Example marker data
# markers <- data.frame(
#   cluster = c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1),
#   gene = c("CD3D", "CD3E", "CD2", "IL7R", "LTB", "CD14", "LYZ", "CST3", "MS4A7", "FCGR3A"),
#   avg_log2FC = c(2.5, 2.3, 2.1, 1.8, 1.7, 3.1, 2.8, 2.5, 2.2, 2.0),
#   p_val_adj = c(0.001, 0.001, 0.002, 0.003, 0.005, 0.0001, 0.0002, 0.0005, 0.001, 0.002)
# )
# 
# # Run annotation with a single model
# results <- annotate_cell_types(
#   input = markers,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#   top_gene_count = 10,
#   debug = FALSE  # Set to TRUE for more detailed output
# )
# 
# # Print results
# print(results)

## -----------------------------------------------------------------------------
# # Define models to use
# models <- c(
#   "claude-sonnet-4-6",  # Anthropic
#   "gpt-5.5",                      # OpenAI
#   "gemini-3.1-pro-preview"               # Google
# )
# 
# # API keys for different providers
# api_keys <- list(
#   anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
#   openai = Sys.getenv("OPENAI_API_KEY"),
#   gemini = Sys.getenv("GEMINI_API_KEY")
# )
# 
# # Run annotation with multiple models
# results <- list()
# for (model in models) {
#   provider <- get_provider(model)
#   api_key <- api_keys[[provider]]
# 
#   results[[model]] <- annotate_cell_types(
#     input = markers,
#     tissue_name = "human PBMC",
#     model = model,
#     api_key = api_key,
#     top_gene_count = 10
#   )
# }
# 
# # Create consensus
# consensus_results <- interactive_consensus_annotation(
#   input = markers,
#   tissue_name = "human PBMC",
#   models = models,  # Use all the models defined above
#   api_keys = api_keys,
#   controversy_threshold = 0.7,
#   entropy_threshold = 1.0,
#   consensus_check_model = "claude-sonnet-4-6"
# )

## -----------------------------------------------------------------------------
# # Assuming you have a Seurat object named 'seurat_obj' and consensus results
# library(Seurat)
# 
# # Add consensus annotations to Seurat object
# seurat_obj$cell_type_consensus <- plyr::mapvalues(
#   x = as.character(Idents(seurat_obj)),
#   from = names(consensus_results$final_annotations),
#   to = consensus_results$final_annotations
# )
# 
# # Extract consensus metrics from the consensus results
# # Note: These metrics are available in the consensus_results$initial_results$consensus_results
# consensus_metrics <- lapply(names(consensus_results$initial_results$consensus_results), function(cluster_id) {
#   metrics <- consensus_results$initial_results$consensus_results[[cluster_id]]
#   return(list(
#     cluster = cluster_id,
#     consensus_proportion = metrics$consensus_proportion,
#     entropy = metrics$entropy
#   ))
# })
# 
# # Convert to data frame for easier handling
# metrics_df <- do.call(rbind, lapply(consensus_metrics, data.frame))
# 
# # Add consensus proportion to Seurat object
# seurat_obj$consensus_proportion <- plyr::mapvalues(
#   x = as.character(Idents(seurat_obj)),
#   from = metrics_df$cluster,
#   to = metrics_df$consensus_proportion
# )
# 
# # Add entropy to Seurat object
# seurat_obj$entropy <- plyr::mapvalues(
#   x = as.character(Idents(seurat_obj)),
#   from = metrics_df$cluster,
#   to = metrics_df$entropy
# )

## -----------------------------------------------------------------------------
# # Plot UMAP with cell type annotations
# DimPlot(seurat_obj, group.by = "cell_type_consensus", label = TRUE, repel = TRUE) +
#   ggtitle("Cell Type Annotations") +
#   theme(plot.title = element_text(hjust = 0.5))

## -----------------------------------------------------------------------------
# # Set OpenRouter API key
# Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key")
# 
# # Use a free model
# free_results <- annotate_cell_types(
#   input = markers,
#   tissue_name = "human PBMC",
#   model = "meta-llama/llama-4-maverick:free",  # Note the :free suffix
#   api_key = Sys.getenv("OPENROUTER_API_KEY"),
#   top_gene_count = 10
# )
# 
# # Print results
# print(free_results)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)

## -----------------------------------------------------------------------------
# # Install from CRAN
# install.packages("mLLMCelltype")

## -----------------------------------------------------------------------------
# # Install devtools if not already installed
# if (!requireNamespace("devtools", quietly = TRUE)) {
#   install.packages("devtools")
# }
# 
# # Install mLLMCelltype development version
# devtools::install_github("cafferychen777/mLLMCelltype", subdir = "R")

## -----------------------------------------------------------------------------
# # Assuming the package is in the current working directory
# devtools::install_local("path/to/mLLMCelltype/R")

## -----------------------------------------------------------------------------
# library(dotenv)
# dotenv::load_dot_env()

## -----------------------------------------------------------------------------
# library(mLLMCelltype)
# 
# results <- annotate_cell_types(
#   input = your_marker_data,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = "your-anthropic-key",
#   top_gene_count = 10
# )

## -----------------------------------------------------------------------------
# Sys.setenv(OPENAI_API_KEY = "your-openai-key")
# Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-key")
# # Set other API keys as needed

## -----------------------------------------------------------------------------
# library(mLLMCelltype)
# 
# # Check if the package is loaded correctly
# packageVersion("mLLMCelltype")
# 
# # Verify API key setup for a specific provider
# api_key <- get_api_key("anthropic")
# if (!is.null(api_key) && api_key != "") {
#   cat("Anthropic API key is set up correctly\n")
# } else {
#   cat("Anthropic API key is not set up\n")
# }

## -----------------------------------------------------------------------------
# # Example of setting proxy for httr
# httr::set_config(httr::use_proxy(url = "proxy_url", port = proxy_port))

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)

## -----------------------------------------------------------------------------
# library(mLLMCelltype)
# 
# results <- annotate_cell_types(
#   input,                # Marker gene data (data frame, list, or file path)
#   tissue_name,          # Tissue name (e.g., "human PBMC", "mouse brain")
#   model,                # LLM model to use
#   api_key = NA,         # API key (if not set in environment, NA returns prompt only)
#   top_gene_count = 10,  # Number of top genes per cluster to use
#   debug = FALSE         # Whether to print debugging information
# )

## ----eval=FALSE---------------------------------------------------------------
# consensus_results <- interactive_consensus_annotation(
#   input,                # Original marker gene data (Seurat FindAllMarkers result or list of genes)
#   tissue_name = NULL,   # Optional tissue name
#   models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),  # Models to use
#   api_keys,             # Named list of API keys
#   top_gene_count = 10,  # Number of top genes to use
#   controversy_threshold = 0.7,  # Threshold for identifying controversial clusters
#   entropy_threshold = 1.0,  # Entropy threshold for controversial clusters
#   max_discussion_rounds = 3,  # Maximum discussion rounds
#   consensus_check_model = NULL,  # Model to use for consensus checking (see recommendations below)
#   log_dir = "logs",     # Directory for logs
#   cache_dir = NULL,  # Uses default system cache directory
#   use_cache = TRUE      # Whether to use cache
# )

## -----------------------------------------------------------------------------
# # Load example data
# library(Seurat)
# data("pbmc_small")
# 
# # Find markers
# pbmc_markers <- FindAllMarkers(pbmc_small, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)
# 
# # Run annotation with a single model
# results <- annotate_cell_types(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#   top_gene_count = 10
# )
# 
# # Add annotations to Seurat object
# pbmc_small$cell_type_claude <- plyr::mapvalues(
#   x = as.character(Idents(pbmc_small)),
#   from = names(results),
#   to = results
# )
# 
# # Visualize
# DimPlot(pbmc_small, group.by = "cell_type_claude", label = TRUE)

## -----------------------------------------------------------------------------
# # Define multiple models to use
# models <- c(
#   "claude-sonnet-4-6",  # Anthropic
#   "gpt-5.5",                      # OpenAI
#   "gemini-3.1-pro-preview",              # Google
#   "grok-4.3"                       # X.AI
# )
# 
# # API keys for different providers
# api_keys <- list(
#   anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
#   openai = Sys.getenv("OPENAI_API_KEY"),
#   gemini = Sys.getenv("GEMINI_API_KEY"),
#   grok = Sys.getenv("GROK_API_KEY")
# )
# 
# # Run annotation with multiple models
# results <- list()
# for (model in models) {
#   provider <- get_provider(model)
#   api_key <- api_keys[[provider]]
# 
#   results[[model]] <- annotate_cell_types(
#     input = pbmc_markers,
#     tissue_name = "human PBMC",
#     model = model,
#     api_key = api_key,
#     top_gene_count = 10
#   )
# }
# 
# # Create consensus
# consensus_results <- interactive_consensus_annotation(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   models = models,  # Use all the models defined above
#   api_keys = api_keys,
#   controversy_threshold = 0.7,
#   entropy_threshold = 1.0,
#   consensus_check_model = "claude-sonnet-4-6"
# )
# 
# # View consensus results
# # You can access the final annotations with consensus_results$final_annotations
# 
# # Add consensus annotations and metrics to Seurat object
# pbmc_small$cell_type_consensus <- plyr::mapvalues(
#   x = as.character(Idents(pbmc_small)),
#   from = names(consensus_results$final_annotations),
#   to = consensus_results$final_annotations
# )
# 
# # Extract consensus metrics from the consensus results
# consensus_metrics <- lapply(names(consensus_results$initial_results$consensus_results), function(cluster_id) {
#   metrics <- consensus_results$initial_results$consensus_results[[cluster_id]]
#   return(list(
#     cluster = cluster_id,
#     consensus_proportion = metrics$consensus_proportion,
#     entropy = metrics$entropy
#   ))
# })
# 
# # Convert to data frame for easier handling
# metrics_df <- do.call(rbind, lapply(consensus_metrics, data.frame))
# 
# # Add consensus proportion to Seurat object
# pbmc_small$consensus_proportion <- plyr::mapvalues(
#   x = as.character(Idents(pbmc_small)),
#   from = metrics_df$cluster,
#   to = metrics_df$consensus_proportion
# )
# 
# # Add entropy to Seurat object
# pbmc_small$shannon_entropy <- plyr::mapvalues(
#   x = as.character(Idents(pbmc_small)),
#   from = metrics_df$cluster,
#   to = metrics_df$entropy
# )

## -----------------------------------------------------------------------------
# # Set OpenRouter API key
# openrouter_api_key <- Sys.getenv("OPENROUTER_API_KEY")
# 
# # Define free OpenRouter models to use
# free_models <- c(
#   "meta-llama/llama-4-maverick:free",                # Meta Llama 4 Maverick (free)
#   "meta-llama/llama-3.3-70b-instruct:free",    # Meta Llama 3.3 70B (free)
#   "deepseek/deepseek-v4-pro:free",             # DeepSeek V4 Pro (free)
#   "meta-llama/llama-3.3-70b-instruct:free"          # Meta Llama 3.3 70B (free)
# )
# 
# # Run annotation with free OpenRouter models
# free_results <- list()
# for (model in free_models) {
#   free_results[[model]] <- annotate_cell_types(
#     input = pbmc_markers,
#     tissue_name = "human PBMC",
#     model = model,  # OpenRouter models are automatically detected by format: 'provider/model-name:free'
#     api_key = openrouter_api_key,
#     top_gene_count = 10
#   )
# }
# 
# # Create consensus with free models
# free_consensus_results <- interactive_consensus_annotation(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   models = free_models,  # Use all the free models defined above
#   api_keys = list("openrouter" = openrouter_api_key),
#   controversy_threshold = 0.7,
#   entropy_threshold = 1.0,
#   consensus_check_model = "meta-llama/llama-4-maverick:free"  # Use a free model for consensus checking
# )
# 
# # View free model consensus results
# # You can access the final annotations with free_consensus_results$final_annotations
# 
# # Add free model consensus annotations to Seurat object
# pbmc_small$free_model_consensus <- plyr::mapvalues(
#   x = as.character(Idents(pbmc_small)),
#   from = names(free_consensus_results$final_annotations),
#   to = free_consensus_results$final_annotations
# )
# 
# # Compare paid vs. free model results
# comparison <- data.frame(
#   cluster = names(consensus_results$final_annotations),
#   paid_models = consensus_results$final_annotations,
#   free_models = free_consensus_results$final_annotations,
#   agreement = consensus_results$final_annotations == free_consensus_results$final_annotations
# )
# print(comparison)

## -----------------------------------------------------------------------------
# # Save markers to CSV
# write.csv(pbmc_markers, "pbmc_markers.csv", row.names = FALSE)
# 
# # Run annotation using the CSV file
# results <- annotate_cell_types(
#   input = "pbmc_markers.csv",
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )

## -----------------------------------------------------------------------------
# # Note: The annotate_cell_types function does not have built-in caching.
# # If you need caching, you can implement it separately.
# 
# # Run annotation
# results <- annotate_cell_types(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#   top_gene_count = 10,
#   debug = FALSE
# )
# 
# # If you need custom caching, you can implement it using your own cache manager
# # This is just a conceptual example and not part of the actual package
# # cache_manager <- YourCacheManager$new(cache_dir = "path/to/cache")
# # cache_manager$clear_cache()

## -----------------------------------------------------------------------------
# # Example of using a free model via OpenRouter
# # First, set your OpenRouter API key
# Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key")
# 
# # Then use a free model with the :free suffix
# free_model_results <- annotate_cell_types(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   model = "meta-llama/llama-4-maverick:free",  # Note the :free suffix
#   api_key = Sys.getenv("OPENROUTER_API_KEY")
#   # No need to specify provider - it's automatically detected from the model name format
# )

## -----------------------------------------------------------------------------
# library(Seurat)
# library(mLLMCelltype)
# library(ggplot2)
# 
# # Load data
# data("pbmc_small")
# 
# # Standard Seurat preprocessing
# pbmc_small <- NormalizeData(pbmc_small)
# pbmc_small <- FindVariableFeatures(pbmc_small)
# pbmc_small <- ScaleData(pbmc_small)
# pbmc_small <- RunPCA(pbmc_small)
# pbmc_small <- FindNeighbors(pbmc_small)
# pbmc_small <- FindClusters(pbmc_small, resolution = 0.5)
# pbmc_small <- RunUMAP(pbmc_small, dims = 1:10)
# 
# # Find markers for each cluster
# pbmc_markers <- FindAllMarkers(pbmc_small, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)
# 
# # Define models to use
# models <- c(
#   "claude-sonnet-4-6",
#   "gpt-5.5",
#   "gemini-3.1-pro-preview"
# )
# 
# # API keys
# api_keys <- list(
#   anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
#   openai = Sys.getenv("OPENAI_API_KEY"),
#   gemini = Sys.getenv("GEMINI_API_KEY")
# )
# 
# # Run annotation with multiple models
# results <- list()
# for (model in models) {
#   provider <- get_provider(model)
#   api_key <- api_keys[[provider]]
# 
#   results[[model]] <- annotate_cell_types(
#     input = pbmc_markers,
#     tissue_name = "human PBMC",
#     model = model,
#     api_key = api_key,
#     top_gene_count = 10
#   )
# 
#   # Add individual model results to Seurat object
#   column_name <- paste0("cell_type_", gsub("[^a-zA-Z0-9]", "_", model))
#   pbmc_small[[column_name]] <- plyr::mapvalues(
#     x = as.character(Idents(pbmc_small)),
#     from = names(results[[model]]),
#     to = results[[model]]
#   )
# }
# 
# # Create consensus
# consensus_results <- interactive_consensus_annotation(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   models = models,  # Use all the models defined above
#   api_keys = api_keys,
#   controversy_threshold = 0.7,
#   entropy_threshold = 1.0,
#   consensus_check_model = "claude-sonnet-4-6"
# )
# 
# # Add consensus results to Seurat object
# pbmc_small$cell_type_consensus <- plyr::mapvalues(
#   x = as.character(Idents(pbmc_small)),
#   from = names(consensus_results$final_annotations),
#   to = consensus_results$final_annotations
# )
# 
# # Extract consensus metrics from the consensus results
# consensus_metrics <- lapply(names(consensus_results$initial_results$consensus_results), function(cluster_id) {
#   metrics <- consensus_results$initial_results$consensus_results[[cluster_id]]
#   return(list(
#     cluster = cluster_id,
#     consensus_proportion = metrics$consensus_proportion,
#     entropy = metrics$entropy
#   ))
# })
# 
# # Convert to data frame for easier handling
# metrics_df <- do.call(rbind, lapply(consensus_metrics, data.frame))
# 
# # Add consensus proportion to Seurat object
# pbmc_small$consensus_proportion <- as.numeric(plyr::mapvalues(
#   x = as.character(Idents(pbmc_small)),
#   from = metrics_df$cluster,
#   to = metrics_df$consensus_proportion
# ))
# 
# # Add entropy to Seurat object
# pbmc_small$shannon_entropy <- as.numeric(plyr::mapvalues(
#   x = as.character(Idents(pbmc_small)),
#   from = metrics_df$cluster,
#   to = metrics_df$entropy
# ))
# 
# # Visualize results
# p1 <- DimPlot(pbmc_small, group.by = "cell_type_consensus", label = TRUE, repel = TRUE) +
#   ggtitle("Cell Type Annotations") +
#   theme(plot.title = element_text(hjust = 0.5))
# 
# p2 <- FeaturePlot(pbmc_small, features = "consensus_proportion", cols = c("yellow", "green", "blue")) +
#   ggtitle("Consensus Proportion") +
#   theme(plot.title = element_text(hjust = 0.5))
# 
# p3 <- FeaturePlot(pbmc_small, features = "shannon_entropy", cols = c("red", "orange")) +
#   ggtitle("Shannon Entropy") +
#   theme(plot.title = element_text(hjust = 0.5))
# 
# # Combine plots
# p1 | p2 | p3

## -----------------------------------------------------------------------------
# # Using more genes (better for well-characterized tissues)
# results_more_genes <- annotate_cell_types(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#   top_gene_count = 20  # Using more genes
# )
# 
# # Using fewer genes (better for noisy data)
# results_fewer_genes <- annotate_cell_types(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#   top_gene_count = 5   # Using fewer genes
# )

## ----eval=FALSE---------------------------------------------------------------
# # Example of using interactive_consensus_annotation with different controversy thresholds
# # Lower threshold (more clusters will be discussed)
# consensus_results_low_threshold <- interactive_consensus_annotation(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3-flash-preview"),
#   api_keys = list(
#     "anthropic" = Sys.getenv("ANTHROPIC_API_KEY"),
#     "openai" = Sys.getenv("OPENAI_API_KEY"),
#     "gemini" = Sys.getenv("GEMINI_API_KEY")
#   ),
#   controversy_threshold = 0.3  # Lower threshold - more clusters will be discussed
# )
# 
# # Higher threshold (fewer clusters will be discussed)
# consensus_results_high_threshold <- interactive_consensus_annotation(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3-flash-preview"),
#   api_keys = list(
#     "anthropic" = Sys.getenv("ANTHROPIC_API_KEY"),
#     "openai" = Sys.getenv("OPENAI_API_KEY"),
#     "gemini" = Sys.getenv("GEMINI_API_KEY")
#   ),
#   controversy_threshold = 0.7  # Higher threshold - fewer clusters will be discussed
# )

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)

## -----------------------------------------------------------------------------
# library(mLLMCelltype)
# library(Seurat)
# library(dplyr)
# 
# # Step 1: Perform initial high-level annotation
# high_level_results <- annotate_cell_types(
#   input = marker_data,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#   top_gene_count = 10
# )
# 
# # Step 2: Add high-level annotations to Seurat object
# seurat_obj$high_level_celltype <- plyr::mapvalues(
#   x = as.character(Idents(seurat_obj)),
#   from = names(high_level_results),
#   to = high_level_results
# )
# 
# # Step 3: Subset T cells for further annotation
# t_cells <- subset(seurat_obj, high_level_celltype == "T cells")
# 
# # Step 4: Find markers within T cells
# t_cell_markers <- FindAllMarkers(t_cells, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)
# 
# # Step 5: Perform T cell subtype annotation
# t_cell_subtypes <- annotate_cell_types(
#   input = t_cell_markers,
#   tissue_name = "human PBMC T cells",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#   top_gene_count = 10
# )
# 
# # Step 6: Add T cell subtypes back to original object
# t_cell_barcodes <- WhichCells(t_cells)
# seurat_obj$detailed_celltype <- seurat_obj$high_level_celltype
# seurat_obj$detailed_celltype[t_cell_barcodes] <- plyr::mapvalues(
#   x = as.character(Idents(t_cells)),
#   from = names(t_cell_subtypes),
#   to = paste0("T cells: ", t_cell_subtypes)
# )

## -----------------------------------------------------------------------------
# # Create a simple function to check parent-child consistency
# validate_hierarchy <- function(high_level, detailed_level) {
#   # Extract parent type from detailed annotation (before the colon)
#   parent_from_detailed <- sapply(strsplit(detailed_level, ": "), function(x) x[1])
# 
#   # Check if parent matches high-level annotation
#   consistent <- parent_from_detailed == high_level
# 
#   # Return consistency check results
#   data.frame(
#     high_level = high_level,
#     detailed_level = detailed_level,
#     consistent = consistent
#   )
# }
# 
# # Apply validation
# hierarchy_validation <- validate_hierarchy(
#   seurat_obj$high_level_celltype,
#   seurat_obj$detailed_celltype
# )
# 
# # Identify inconsistencies
# inconsistencies <- hierarchy_validation[!hierarchy_validation$consistent, ]
# print(inconsistencies)

## -----------------------------------------------------------------------------
# # For noisy data, use fewer top genes
# results_fewer_genes <- annotate_cell_types(
#   input = marker_data,  # Your marker gene data
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#   top_gene_count = 5  # Use fewer genes to focus on strongest signals
# )

## -----------------------------------------------------------------------------
# # Apply stricter filtering to marker genes
# filtered_markers <- marker_data %>%
#   filter(p_val_adj < 0.01, avg_log2FC > 1.0)  # Stricter thresholds
# 
# # Annotate with filtered markers
# results_filtered <- annotate_cell_types(
#   input = filtered_markers,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )

## -----------------------------------------------------------------------------
# # Set up API keys
# api_keys <- list(
#   anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
#   openai = Sys.getenv("OPENAI_API_KEY"),
#   gemini = Sys.getenv("GEMINI_API_KEY")
# )
# 
# # Define multiple models to use
# models <- c(
#   "claude-sonnet-4-6",
#   "gpt-5.5",
#   "gemini-3.1-pro-preview"
# )
# 
# # Create consensus using interactive_consensus_annotation
# consensus_results <- interactive_consensus_annotation(
#   input = marker_data,  # Your marker gene data
#   tissue_name = "human PBMC",
#   models = models,
#   api_keys = api_keys,
#   controversy_threshold = 0.7,
#   entropy_threshold = 1.0,
#   consensus_check_model = "claude-sonnet-4-6"
# )

## -----------------------------------------------------------------------------
# # For data with batch effects, use consensus with lower threshold
# batch_consensus <- interactive_consensus_annotation(
#   input = marker_data,  # Your marker gene data with batch effects
#   tissue_name = "mouse brain",
#   models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
#   api_keys = api_keys,
#   controversy_threshold = 0.4,  # Lower threshold to discuss more clusters
#   entropy_threshold = 0.8  # Lower entropy threshold
# )

## -----------------------------------------------------------------------------
# # Include batch information in the tissue context
# batch_aware_results <- annotate_cell_types(
#   input = marker_data,  # Your marker gene data with batch effects
#   tissue_name = "mouse brain with technical batch effects",  # Include batch context
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )

## -----------------------------------------------------------------------------
# # Basic annotation without specific tissue context
# basic_results <- annotate_cell_types(
#   input = marker_data,
#   tissue_name = "human sample",  # Generic context
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )
# 
# # Annotation with specific tissue context
# specific_results <- annotate_cell_types(
#   input = marker_data,
#   tissue_name = "human fetal liver at 20 weeks gestation",  # Detailed context
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )

## -----------------------------------------------------------------------------
# # Create a custom annotation prompt
# custom_prompt <- create_annotation_prompt(
#   input = marker_data,
#   tissue_name = "human PBMC",
#   top_gene_count = 10
# )
# 
# # Modify the prompt to include additional context
# modified_prompt <- paste0(
#   custom_prompt$prompt,
#   "\n\nAdditional context: This sample is from a patient with rheumatoid arthritis. ",
#   "Previous studies have identified activated T cells, B cells, and CXCR4-high monocytes in this condition."
# )
# 
# # Use the modified prompt directly
# custom_results <- get_model_response(
#   prompt = modified_prompt,
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )

## -----------------------------------------------------------------------------
# library(Seurat)
# library(dplyr)
# 
# # Example: Using CellMarker database information to validate annotations
# # This is a conceptual example - implementation would depend on your specific needs
# 
# # 1. Get annotations with mLLMCelltype
# annotations <- annotate_cell_types(
#   input = marker_data,
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )
# 
# # 2. Compare with known marker genes (conceptual)
# # In a real workflow, you would query a database or use a reference dataset
# known_markers <- list(
#   "T cells" = c("CD3D", "CD3E", "CD3G"),
#   "B cells" = c("CD19", "MS4A1", "CD79A"),
#   "Monocytes" = c("CD14", "LYZ", "CSF1R")
# )
# 
# # 3. Validate annotations against known markers
# # This is a simplified example of how you might validate annotations
# validate_annotations <- function(annotations, marker_data, known_markers) {
#   validation_results <- list()
# 
#   for (i in 1:length(annotations)) {
#     cluster_id <- i
#     predicted_type <- annotations[i]
# 
#     # Get markers for this cluster
#     cluster_markers <- marker_data %>%
#       filter(cluster == cluster_id) %>%
#       arrange(desc(avg_log2FC)) %>%
#       pull(gene) %>%
#       head(20)
# 
#     # Check overlap with known markers for this cell type
#     if (predicted_type %in% names(known_markers)) {
#       expected_markers <- known_markers[[predicted_type]]
#       overlap <- intersect(cluster_markers, expected_markers)
# 
#       validation_results[[i]] <- list(
#         cluster = cluster_id,
#         predicted_type = predicted_type,
#         overlap_count = length(overlap),
#         overlap_genes = paste(overlap, collapse = ", "),
#         confidence = length(overlap) / length(expected_markers)
#       )
#     } else {
#       validation_results[[i]] <- list(
#         cluster = cluster_id,
#         predicted_type = predicted_type,
#         overlap_count = 0,
#         overlap_genes = "",
#         confidence = 0
#       )
#     }
#   }
# 
#   return(validation_results)
# }
# 
# # This is a conceptual example of how you might validate annotations
# # validation_results <- validate_annotations(annotations, marker_data, known_markers)

## -----------------------------------------------------------------------------
# library(Seurat)
# library(mLLMCelltype)
# library(ggplot2)
# library(dplyr)
# 
# # Load example PBMC data
# # In a real workflow, you would use your own data
# data("pbmc_small")  # Example dataset from Seurat
# 
# # Find marker genes
# pbmc_markers <- FindAllMarkers(pbmc_small,
#                               only.pos = TRUE,
#                               min.pct = 0.25,
#                               logfc.threshold = 0.25)
# 
# # Set up API keys
# api_keys <- list(
#   anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
#   openai = Sys.getenv("OPENAI_API_KEY"),
#   gemini = Sys.getenv("GEMINI_API_KEY")
# )
# 
# # Use consensus annotation
# consensus_results <- interactive_consensus_annotation(
#   input = pbmc_markers,
#   tissue_name = "human PBMC",
#   models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
#   api_keys = api_keys,
#   controversy_threshold = 0.7,
#   entropy_threshold = 1.0,
#   consensus_check_model = "claude-sonnet-4-6"
# )
# 
# # Add results to Seurat object
# pbmc_small$cell_type <- plyr::mapvalues(
#   x = as.character(Idents(pbmc_small)),
#   from = names(consensus_results$final_annotations),
#   to = consensus_results$final_annotations
# )
# 
# # Visualize results
# # In a real workflow, you would create a UMAP or t-SNE plot
# # DimPlot(pbmc_small, group.by = "cell_type", label = TRUE) +
# #   ggtitle("PBMC Cell Types")

## -----------------------------------------------------------------------------
# # For rare cell types, use these strategies:
# 
# # 1. Increase the number of marker genes considered
# rare_cell_annotation <- annotate_cell_types(
#   input = marker_data,  # Your marker gene data
#   tissue_name = "human bone marrow",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#   top_gene_count = 20  # Use more genes for rare cell types
# )
# 
# # 2. Use consensus with lower thresholds to discuss more clusters
# rare_cell_consensus <- interactive_consensus_annotation(
#   input = marker_data,  # Your marker gene data
#   tissue_name = "human bone marrow",
#   models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
#   api_keys = api_keys,
#   controversy_threshold = 0.4,  # Lower threshold to discuss more clusters
#   entropy_threshold = 0.8,  # Lower entropy threshold
#   consensus_check_model = "claude-sonnet-4-6"
# )
# 
# # 3. Provide more specific tissue context
# specific_annotation <- annotate_cell_types(
#   input = marker_data,  # Your marker gene data
#   tissue_name = "human bone marrow with expected rare plasma cells and basophils",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )

## -----------------------------------------------------------------------------
# # Example workflow for cross-species comparison
# 
# # 1. Annotate human and mouse datasets separately
# # (Assuming you have marker data for both species)
# human_annotations <- annotate_cell_types(
#   input = human_marker_data,  # Your human marker data
#   tissue_name = "human brain cortex",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )
# 
# mouse_annotations <- annotate_cell_types(
#   input = mouse_marker_data,  # Your mouse marker data
#   tissue_name = "mouse brain cortex",
#   model = "claude-sonnet-4-6",
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )
# 
# # 2. Compare annotations
# # This is a conceptual example - in a real workflow, you would:
# # - Map annotations to Seurat objects
# # - Calculate proportions
# # - Create comparison visualizations
# # - Identify conserved and species-specific cell types
# 
# # Example comparison function (conceptual)
# compare_species_annotations <- function(human_annotations, mouse_annotations) {
#   # Get unique cell types from both species
#   human_types <- unique(human_annotations)
#   mouse_types <- unique(mouse_annotations)
# 
#   # Find common cell types
#   common_types <- intersect(human_types, mouse_types)
# 
#   # Find species-specific cell types
#   human_specific <- setdiff(human_types, mouse_types)
#   mouse_specific <- setdiff(mouse_types, human_types)
# 
#   # Return comparison results
#   list(
#     common_types = common_types,
#     human_specific = human_specific,
#     mouse_specific = mouse_specific
#   )
# }
# 
# # This is a conceptual example
# # comparison <- compare_species_annotations(human_annotations, mouse_annotations)

## -----------------------------------------------------------------------------
# # Example of cost-efficient model selection
# # Choose models based on your specific needs and budget
# 
# # For initial exploration or smaller datasets
# # Use more affordable models
# affordable_annotation <- annotate_cell_types(
#   input = marker_data,  # Your marker gene data
#   tissue_name = "human PBMC",
#   model = "claude-haiku-4-5-20251001",  # More affordable model
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )
# 
# # For final analysis or challenging datasets
# # Use larger models
# premium_annotation <- annotate_cell_types(
#   input = marker_data,  # Your marker gene data
#   tissue_name = "human PBMC",
#   model = "claude-sonnet-4-6",  # Larger model
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )
# 
# # Use OpenRouter for access to free models
# openrouter_annotation <- annotate_cell_types(
#   input = marker_data,  # Your marker gene data
#   tissue_name = "human PBMC",
#   model = "meta-llama/llama-3.3-70b-instruct:free",  # Free model via OpenRouter
#   api_key = Sys.getenv("OPENROUTER_API_KEY")
# )

## -----------------------------------------------------------------------------
# # 1. Use caching with interactive_consensus_annotation
# consensus_with_cache <- interactive_consensus_annotation(
#   input = marker_data,  # Your marker gene data
#   tissue_name = "human PBMC",
#   models = c("claude-sonnet-4-6", "gpt-5.5"),
#   api_keys = api_keys,
#   use_cache = TRUE,  # Enable caching
#   cache_dir = NULL  # Uses default system cache directory
# )
# 
# # 2. Process clusters in batches
# # This is a conceptual example - implementation would depend on your workflow
# process_in_batches <- function(marker_data, batch_size = 5) {
#   # Get unique clusters
#   clusters <- unique(marker_data$cluster)
# 
#   # Process in batches
#   results <- list()
#   for (i in seq(1, length(clusters), by = batch_size)) {
#     # Get current batch of clusters
#     batch_clusters <- clusters[i:min(i + batch_size - 1, length(clusters))]
# 
#     # Filter marker data for current batch
#     batch_data <- marker_data %>% filter(cluster %in% batch_clusters)
# 
#     # Process batch
#     batch_results <- annotate_cell_types(
#       input = batch_data,
#       tissue_name = "human PBMC",
#       model = "claude-sonnet-4-6",
#       api_key = Sys.getenv("ANTHROPIC_API_KEY")
#     )
# 
#     # Store results
#     results <- c(results, batch_results)
#   }
# 
#   return(results)
# }
# 
# # 3. Use faster models for initial exploration
# fast_annotation <- annotate_cell_types(
#   input = marker_data,  # Your marker gene data
#   tissue_name = "human PBMC",
#   model = "claude-haiku-4-5-20251001",  # Faster model
#   api_key = Sys.getenv("ANTHROPIC_API_KEY")
# )

## -----------------------------------------------------------------------------
# # Define a custom processing function
# # This function must accept prompt, model, and api_key parameters
# custom_process_fn <- function(prompt, model, api_key) {
#   # Custom implementation to process prompts and get responses
#   # This is a simplified example
#   cat("Processing prompt with custom provider\n")
#   cat("Model:", model, "\n")
# 
#   # In a real implementation, you would make API calls here
#   # For example:
#   # response <- httr::POST(
#   #   url = "https://api.custom-provider.com/v1/chat/completions",
#   #   body = list(prompt = prompt, model = model),
#   #   httr::add_headers(Authorization = paste("Bearer", api_key)),
#   #   encode = "json"
#   # )
#   # result <- httr::content(response)$choices[[1]]$text
# 
#   # For this example, just return a fixed response
#   result <- "T cells"
#   return(result)
# }
# 
# # Register the custom provider
# register_custom_provider(
#   provider_name = "custom_provider",
#   process_fn = custom_process_fn,
#   description = "My custom LLM provider"
# )
# 
# # Register a custom model
# register_custom_model(
#   model_name = "custom-model",
#   provider_name = "custom_provider",
#   model_config = list(
#     temperature = 0.7,
#     max_tokens = 2000
#   )
# )
# 
# # Use the custom model
# # custom_results <- annotate_cell_types(
# #   input = marker_data,
# #   tissue_name = "human PBMC",
# #   model = "custom-model",
# #   api_key = "your-custom-api-key"
# # )

## -----------------------------------------------------------------------------
# # Configure the global logger (recommended approach)
# configure_logger(level = "INFO", console_output = TRUE, json_format = TRUE)
# 
# # Use simple logging functions
# log_info("Starting analysis of cluster 0", list(
#   cluster_id = "0",
#   tissue_name = "human PBMC",
#   marker_genes = c("CD3D", "CD3E", "CD2", "IL7R", "LTB")
# ))
# 
# # Log API calls with performance tracking
# log_info("API call completed", list(
#   provider = "anthropic",
#   model = "claude-sonnet-4-6",
#   duration_seconds = 2.34,
#   success = TRUE
# ))
# 
# # Log warnings and errors
# log_warn("Model response had unusual format", list(
#   model = "gpt-5.5",
#   response_length = 50
# ))
# 
# log_error("API call failed", list(
#   provider = "openai",
#   error = "Rate limit exceeded"
# ))
# 
# # Alternatively, create a custom logger instance
# custom_logger <- UnifiedLogger$new(
#   base_dir = "custom_logs",
#   level = "DEBUG",
#   console_output = TRUE,
#   json_format = TRUE
# )
# 
# # Use the custom logger
# custom_logger$info("Custom log message", list(analysis_step = "preprocessing"))
# custom_logger$debug("Detailed debugging info", list(variable_state = "initialized"))
# 
# # Get performance summary
# performance <- get_logger()$get_performance_summary()
# print(performance)

## -----------------------------------------------------------------------------
# # Create a cache manager
# cache_manager <- CacheManager$new(cache_dir = NULL)
# 
# # Generate a cache key
# cache_key <- cache_manager$generate_key(
#   input = marker_data,
#   models = c("claude-sonnet-4-6", "gpt-5.5"),
#   cluster_id = "0"
# )
# 
# # Check if results exist in cache
# if (cache_manager$has_cache(cache_key)) {
#   # Load from cache
#   cached_results <- cache_manager$load_from_cache(cache_key)
# } else {
#   # Process and save to cache
#   # results <- process_cluster(...)
#   # cache_manager$save_to_cache(cache_key, results)
# }
# 
# # Get cache statistics
# cache_stats <- cache_manager$get_cache_stats()
# 
# # Clear cache (with confirmation)
# # cache_manager$clear_cache(confirm = TRUE)

## -----------------------------------------------------------------------------
# # Check cache location
# mllmcelltype_cache_dir()
# 
# # Use local cache
# mllmcelltype_cache_dir("local")
# 
# # Clear cache
# mllmcelltype_clear_cache()

#' Get mLLMCelltype cache location
#' 
#' @description Display the cache directory location
#'
#' @param cache_dir Cache directory specification. NULL uses system default, "local" uses current dir, "temp" uses temp dir, or custom path
#'
#' @return Invisible cache directory path
#' @export
#' @examples
#' \dontrun{
#' mllmcelltype_cache_dir()
#' mllmcelltype_cache_dir("local")
#' }
mllmcelltype_cache_dir <- function(cache_dir = NULL) {
  # Create temporary cache_manager to get actual path
  cm <- CacheManager$new(cache_dir)
  actual_path <- cm$get_cache_dir()
  
  message("Cache directory: ", actual_path)
  return(invisible(actual_path))
}

#' Clear mLLMCelltype cache
#' 
#' @description Clear the mLLMCelltype cache
#'
#' @param cache_dir Cache directory specification. NULL uses system default, "local" uses current dir, "temp" uses temp dir, or custom path
#'
#' @return Invisible NULL
#' @export
#' @examples
#' \dontrun{
#' mllmcelltype_clear_cache()
#' mllmcelltype_clear_cache("local")
#' }
mllmcelltype_clear_cache <- function(cache_dir = NULL) {
  cm <- CacheManager$new(cache_dir)
  stats <- cm$get_cache_stats()

  if (!stats$cache_exists || stats$cache_count == 0) {
    message("No cache files to clear.")
    return(invisible(NULL))
  }

  if (interactive()) {
    response <- readline(prompt = sprintf(
      "Delete %d cache files in %s? (yes/no): ",
      stats$cache_count, cm$get_cache_dir()
    ))

    if (!tolower(response) %in% c("yes", "y")) {
      message("Cache clearing cancelled.")
      return(invisible(NULL))
    }
  }

  cm$clear_cache(confirm = TRUE)
  invisible(NULL)
}

#' Print summary of consensus results
#' 
#' This function prints a detailed summary of the consensus building process,
#' including initial predictions from all models, uncertainty metrics, and final consensus
#' for each controversial cluster.
#' 
#
#'   \itemize{
#'     \item initial_results: A list containing individual_predictions, consensus_results, and controversial_clusters
#'     \item final_annotations: A list of final cell type annotations for each cluster
#'     \item controversial_clusters: A character vector of cluster IDs that were controversial
#'     \item discussion_logs: A list of discussion logs for each controversial cluster
#'   }
#
#' @keywords internal
print_consensus_summary <- function(results) {
  to_scalar_or_na <- function(x) {
    if (is.null(x) || length(x) == 0) return(NA_character_)
    scalar <- tryCatch(as.character(x[[1]]), error = function(e) NA_character_)
    if (is.na(scalar) || !nzchar(trimws(scalar))) return(NA_character_)
    scalar
  }

  to_display_text <- function(x, fallback = "No prediction provided") {
    scalar <- to_scalar_or_na(x)
    if (is.na(scalar)) fallback else scalar
  }

  # Print consensus building summary
  cat("\nConsensus Building Summary:\n")
  cat(sprintf("Total clusters analyzed: %d\n", length(results$final_annotations)))
  cat(sprintf("Controversial clusters requiring discussion: %d\n", 
              length(results$controversial_clusters)))
  
  # If there are controversial clusters, print detailed results
  if (length(results$controversial_clusters) > 0) {
    cat("\nDetailed results for controversial clusters:\n")
    
    # Iterate through each controversial cluster
    for (cluster_id in results$controversial_clusters) {
      # Ensure cluster_id is always a string type
      char_cluster_id <- as.character(cluster_id)
      
      cat(sprintf("\nCluster %s:\n", char_cluster_id))
      cat("Initial predictions:\n")
      
      # Prioritize initial predictions from discussion_logs, as these are the actual predictions used for discussion
      if (!is.null(results$discussion_logs) && 
          !is.null(results$discussion_logs[[char_cluster_id]]) && 
          !is.null(results$discussion_logs[[char_cluster_id]]$initial_predictions)) {
        
        # Use initial predictions from discussion logs
        initial_predictions <- results$discussion_logs[[char_cluster_id]]$initial_predictions
        
        # Iterate through each model's prediction
        for (model in names(initial_predictions)) {
          prediction <- initial_predictions[[model]]

          cat(sprintf("  %s: %s\n", model, to_display_text(prediction)))
        }
      } 
      # If no initial predictions in discussion logs, use initial_results
      else if (!is.null(results$initial_results) &&
               !is.null(results$initial_results$individual_predictions)) {
        # Derive sorted cluster order for positional fallback
        all_cluster_ids <- names(results$final_annotations)

        # Check naming convention once using the first model's predictions
        first_model_preds <- results$initial_results$individual_predictions[[1]]
        has_names <- !is.null(names(first_model_preds))

        for (model in names(results$initial_results$individual_predictions)) {
          model_preds <- results$initial_results$individual_predictions[[model]]

          if (has_names) {
            prediction <- model_preds[[char_cluster_id]]
          } else {
            # Positional fallback: find cluster's position among all clusters
            r_index <- match(char_cluster_id, all_cluster_ids)
            prediction <- if (!is.na(r_index) && r_index >= 1 && r_index <= length(model_preds)) {
              model_preds[r_index]
            } else {
              NA
            }
          }

          cat(sprintf("  %s: %s\n", model, to_display_text(prediction)))
        }
      } else {
        cat("  No initial predictions available\n")
      }
      
      # Print uncertainty metrics
      cat("\nUncertainty metrics:\n")
      if (!is.null(results$initial_results) && 
          !is.null(results$initial_results$consensus_results) && 
          !is.null(results$initial_results$consensus_results[[char_cluster_id]])) {
        
        consensus_result <- results$initial_results$consensus_results[[char_cluster_id]]
        
        # Print consensus proportion
        if (!is.null(consensus_result$consensus_proportion)) {
          cat(sprintf("  Consensus proportion: %.2f\n", consensus_result$consensus_proportion))
        }
        
        # Print Shannon entropy
        if (!is.null(consensus_result$entropy)) {
          cat(sprintf("  Shannon entropy: %.2f\n", consensus_result$entropy))
        }
      } else {
        cat("  Uncertainty metrics not available\n")
      }
      
      # Print final consensus
      if (!is.null(results$final_annotations) && 
          !is.null(results$final_annotations[[char_cluster_id]])) {
        
        final_annotation <- results$final_annotations[[char_cluster_id]]
        
        final_annotation_str <- to_scalar_or_na(final_annotation)
        if (is.na(final_annotation_str)) {
          final_annotation_str <- "Final_Annotation_Missing"
        }
        
        # Validate consistency between final consensus and initial predictions
        if (!is.null(results$initial_results) && 
            !is.null(results$initial_results$individual_predictions) &&
            length(names(results$initial_results$individual_predictions)) > 0) { # Ensure there are models
          
          # Validation using discussion log predictions
          discussion_log_predictions <- NULL
          if (!is.null(results$discussion_logs) && 
              !is.null(results$discussion_logs[[char_cluster_id]]) && 
              !is.null(results$discussion_logs[[char_cluster_id]]$initial_predictions)) {
            discussion_log_predictions <- results$discussion_logs[[char_cluster_id]]$initial_predictions
          }

          if (!is.null(discussion_log_predictions) && length(discussion_log_predictions) > 0) {
            all_predictions <- list()
            
            for (model in names(discussion_log_predictions)) {
              pred <- discussion_log_predictions[[model]]

              # Check if prediction is valid and not the placeholder
              pred_scalar <- to_scalar_or_na(pred)
              if (!is.na(pred_scalar) && pred_scalar != "No prediction provided") {
                all_predictions[[model]] <- pred_scalar
              }
            } # End model loop

            # Check if all collected models predicted the same result and compare with final consensus
            if (length(all_predictions) > 0) {
              unique_preds <- unique(unlist(all_predictions))
              if (length(unique_preds) == 1) {
                # We don't need to clean prefix here as discussion log preds should be clean
                clean_pred <- trimws(unique_preds[1]) # Just trim whitespace
                
                # If all models predicted same but differs from final consensus, add warning
                # Use normalize_annotation for robust comparison (handles punctuation, plurals, synonyms)
                if (normalize_annotation(clean_pred) != normalize_annotation(final_annotation_str)) {
                  cat(sprintf("WARNING: All models in discussion log predicted '%s' but final consensus is '%s'\n", 
                              clean_pred, final_annotation_str))
                }
              }
            }
          } # End check for discussion_log_predictions
        }
        
        cat(sprintf("Final consensus: %s\n", final_annotation_str))
      } else {
        cat("Final consensus: Not available\n")
      }
    }
  }
  
  # If no controversial clusters, print message
  if (length(results$controversial_clusters) == 0) {
    cat("\nNo controversial clusters found. All clusters reached consensus.\n")
  }
}

#' DeepSeek API Processor
#' 
#' Concrete implementation of BaseAPIProcessor for DeepSeek models.
#' Handles DeepSeek-specific API calls, authentication, and response parsing.
#'
#' @export
DeepSeekProcessor <- R6::R6Class("DeepSeekProcessor",
  inherit = BaseAPIProcessor,
  
  public = list(
    #' @description
    #' Initialize DeepSeek processor
    #' @param base_url Optional custom API endpoint
    initialize = function(base_url = NULL) {
      super$initialize("deepseek", base_url)
    },

    #' @description
    #' Get default DeepSeek API URL
    #
    get_default_api_url = function() {
      return("https://api.deepseek.com/v1/chat/completions")
    },
    
    #' @description
    #' Make API call to DeepSeek
    #' @param chunk_content Prompt text to send
    #' @param model Model identifier
    #' @param api_key DeepSeek API key
    make_api_call = function(chunk_content, model, api_key) {
      private$post_chat_completions_request(
        chunk_content,
        model,
        api_key,
        body_extra = list(
          temperature = 0.7,
          max_tokens = 4096,
          stream = FALSE,
          thinking = list(type = "disabled")
        )
      )
    },
    
    #' @description
    #' Extract response content from DeepSeek API response
    #' @param response HTTP response object
    #' @param model Model identifier
    extract_response_content = function(response, model) {
      private$extract_chat_completions_content(response, model)
    }
  )
)

# Tests for the Anthropic multi-block response parser.
#
# Anthropic Messages responses can contain multiple content blocks (e.g. a
# leading thinking/reasoning block). Hard-indexing the first block would discard
# the answer, so the parser iterates and concatenates the text blocks.

test_that("Anthropic extract_response_content concatenates text and skips thinking blocks", {
  quiet_logger <- list(
    info = function(...) NULL,
    debug = function(...) NULL,
    warn = function(...) NULL,
    error = function(...) NULL,
    log_api_call = function(...) NULL
  )

  extract <- function(blocks) {
    testthat::with_mocked_bindings(
      {
        processor <- AnthropicProcessor$new()
        testthat::with_mocked_bindings(
          processor$extract_response_content(
            structure(list(), class = "response"),
            "claude-opus-4-7"
          ),
          content = function(...) list(content = blocks),
          .package = "httr"
        )
      },
      get_logger = function() quiet_logger
    )
  }

  # A leading thinking block must not discard the answer.
  answer <- extract(list(
    list(type = "thinking", thinking = "reasoning..."),
    list(type = "text", text = "Cluster 1: T cells")
  ))
  expect_identical(answer, "Cluster 1: T cells")

  # Multiple text blocks are concatenated in order.
  multi <- extract(list(
    list(type = "text", text = "Cluster 1: T cells"),
    list(type = "text", text = "Cluster 2: B cells")
  ))
  expect_identical(multi, "Cluster 1: T cells\nCluster 2: B cells")

  # A response with no text block is rejected.
  err <- tryCatch(
    extract(list(list(type = "thinking", thinking = "reasoning..."))),
    error = identity
  )
  expect_s3_class(err, "error")
  expect_match(conditionMessage(err), "Unexpected response format from Anthropic")
})

test_that("Anthropic parser warns when the response is truncated at max_tokens", {
  warnings_seen <- character(0)
  recording_logger <- list(
    info = function(...) NULL,
    debug = function(...) NULL,
    warn = function(message, context = NULL) {
      warnings_seen <<- c(warnings_seen, message)
      invisible(NULL)
    },
    error = function(...) NULL,
    log_api_call = function(...) NULL
  )

  result <- testthat::with_mocked_bindings(
    {
      processor <- AnthropicProcessor$new()
      testthat::with_mocked_bindings(
        processor$extract_response_content(
          structure(list(), class = "response"),
          "claude-opus-4-7"
        ),
        content = function(...) {
          list(
            content = list(list(type = "text", text = "Cluster 1: T cells")),
            stop_reason = "max_tokens"
          )
        },
        .package = "httr"
      )
    },
    get_logger = function() recording_logger
  )

  expect_identical(result, "Cluster 1: T cells")
  expect_true(any(grepl("truncated", warnings_seen)))
})

test_that("Anthropic extraction tolerates a non-list content block", {
  quiet_logger <- list(
    info = function(...) NULL, debug = function(...) NULL,
    warn = function(...) NULL, error = function(...) NULL,
    log_api_call = function(...) NULL
  )
  result <- testthat::with_mocked_bindings(
    {
      processor <- AnthropicProcessor$new()
      testthat::with_mocked_bindings(
        processor$extract_response_content(
          structure(list(), class = "response"), "claude-opus-4-7"
        ),
        content = function(...) {
          list(content = list("bare string", list(type = "text", text = "Cluster 1: T cells")))
        },
        .package = "httr"
      )
    },
    get_logger = function() quiet_logger
  )
  expect_identical(result, "Cluster 1: T cells")
})

test_that("Anthropic parser stays silent on a normal stop_reason", {
  warnings_seen <- character(0)
  recording_logger <- list(
    info = function(...) NULL, debug = function(...) NULL,
    warn = function(message, context = NULL) {
      warnings_seen <<- c(warnings_seen, message)
      invisible(NULL)
    },
    error = function(...) NULL, log_api_call = function(...) NULL
  )
  testthat::with_mocked_bindings(
    {
      processor <- AnthropicProcessor$new()
      testthat::with_mocked_bindings(
        processor$extract_response_content(
          structure(list(), class = "response"), "claude-opus-4-7"
        ),
        content = function(...) {
          list(
            content = list(list(type = "text", text = "Cluster 1: T cells")),
            stop_reason = "end_turn"
          )
        },
        .package = "httr"
      )
    },
    get_logger = function() recording_logger
  )
  expect_false(any(grepl("truncated", warnings_seen)))
})

# Tests for MiniMax in-body business-error retry classification.
#
# MiniMax reports transient failures (rate limits, timeouts) via HTTP 200 with a
# non-zero base_resp$status_code, bypassing HTTP-status retry classification.
# These must surface as retryable mllm_api_error conditions so the shared retry
# loop backs off instead of dropping the model from the consensus panel.

test_that("MiniMax transient business errors are retryable, permanent ones are not", {
  quiet_logger <- list(
    info = function(...) NULL,
    debug = function(...) NULL,
    warn = function(...) NULL,
    error = function(...) NULL,
    log_api_call = function(...) NULL
  )

  classify <- function(status_code) {
    testthat::with_mocked_bindings(
      {
        processor <- MinimaxProcessor$new()
        testthat::with_mocked_bindings(
          tryCatch(
            processor$extract_response_content(
              structure(list(), class = "response"),
              "MiniMax-M2.7"
            ),
            error = identity
          ),
          content = function(...) {
            list(base_resp = list(status_code = status_code, status_msg = "boom"))
          },
          .package = "httr"
        )
      },
      get_logger = function() quiet_logger
    )
  }

  # 1002 (rate limit) and 1001 (timeout) are transient -> retryable.
  rate_limited <- classify(1002L)
  expect_s3_class(rate_limited, "mllm_api_error")
  expect_true(rate_limited$retryable)

  timed_out <- classify(1001L)
  expect_true(timed_out$retryable)

  # 1004 (auth) and 2013 (invalid params) are fatal -> not retryable.
  unauthorized <- classify(1004L)
  expect_s3_class(unauthorized, "mllm_api_error")
  expect_false(unauthorized$retryable)

  invalid_params <- classify(2013L)
  expect_false(invalid_params$retryable)
})

test_that("get_model_response retries a transient MiniMax business error end-to-end", {
  # Drives the real retry loop: two transient (1002) responses then a success.
  quiet_logger <- list(
    info = function(...) NULL, debug = function(...) NULL,
    warn = function(...) NULL, error = function(...) NULL,
    log_api_call = function(...) NULL,
    log_api_request_response = function(...) NULL
  )
  state <- new.env(parent = emptyenv())
  state$attempt <- 0L

  result <- testthat::with_mocked_bindings(
    testthat::with_mocked_bindings(
      get_model_response("genes", "MiniMax-M2.7", "test-key"),
      POST = function(...) {
        state$attempt <- state$attempt + 1L
        structure(list(), class = "response")
      },
      status_code = function(...) 200L,
      content = function(...) {
        if (state$attempt < 3L) {
          list(base_resp = list(status_code = 1002L, status_msg = "rate limit"))
        } else {
          list(choices = list(list(message = list(content = "Cluster 1: T cells"))))
        }
      },
      .package = "httr"
    ),
    get_logger = function() quiet_logger,
    wait_before_model_retry = function(seconds) invisible(NULL)
  )

  expect_identical(state$attempt, 3L)
  expect_identical(result, "Cluster 1: T cells")
})

# Tests for the shared OpenAI-compatible chat-completions truncation warning.
#
# When a provider stops because it hit max_tokens (finish_reason == "length"),
# the trailing clusters are silently dropped; the extractor must surface a
# warning so this is diagnosable, and must stay silent on a normal stop.

make_recording_logger <- function(sink) {
  list(
    info = function(...) NULL,
    debug = function(...) NULL,
    warn = function(message, context = NULL) {
      sink$messages <- c(sink$messages, message)
      invisible(NULL)
    },
    error = function(...) NULL,
    log_api_call = function(...) NULL
  )
}

test_that("chat-completions extraction warns when finish_reason is 'length'", {
  sink <- new.env()
  sink$messages <- character(0)
  logger <- make_recording_logger(sink)

  extract <- function(finish_reason) {
    testthat::with_mocked_bindings(
      {
        processor <- OpenAIProcessor$new()
        testthat::with_mocked_bindings(
          processor$extract_response_content(
            structure(list(), class = "response"), "gpt-5.5"
          ),
          content = function(...) {
            list(choices = list(list(
              message = list(content = "Cluster 1: T cells"),
              finish_reason = finish_reason
            )))
          },
          .package = "httr"
        )
      },
      get_logger = function() logger
    )
  }

  result <- extract("length")
  expect_identical(result, "Cluster 1: T cells")
  expect_true(any(grepl("truncated", sink$messages)))

  sink$messages <- character(0)
  extract("stop")
  expect_false(any(grepl("truncated", sink$messages)))
})

# Tests for the DeepSeek provider

make_deepseek_json_response <- function(payload, status_code = 200L) {
  structure(
    list(
      url = "https://api.deepseek.com/v1/chat/completions",
      status_code = status_code,
      headers = list("Content-Type" = "application/json"),
      all_headers = list(list(
        status = status_code,
        version = "HTTP/1.1",
        headers = list("Content-Type" = "application/json")
      )),
      cookies = data.frame(),
      content = charToRaw(jsonlite::toJSON(payload, auto_unbox = TRUE)),
      date = Sys.time(),
      times = c(),
      request = list(method = "POST")
    ),
    class = "response"
  )
}

quiet_deepseek_logger <- function() {
  list(
    info = function(...) NULL,
    debug = function(...) NULL,
    warn = function(...) NULL,
    error = function(...) NULL,
    log_api_call = function(...) NULL,
    log_api_request_response = function(...) NULL
  )
}

test_that("DeepSeekProcessor disables thinking for annotation requests", {
  captured <- new.env(parent = emptyenv())

  testthat::with_mocked_bindings({
    testthat::with_mocked_bindings({
      processor <- DeepSeekProcessor$new()
      response <- processor$make_api_call(
        "genes",
        "deepseek-v4-flash",
        "test-key"
      )

      expect_identical(captured$url, "https://api.deepseek.com/v1/chat/completions")
      expect_identical(captured$headers[["Authorization"]], "Bearer test-key")
      expect_identical(captured$body$model, "deepseek-v4-flash")
      expect_identical(
        captured$body$messages,
        list(list(role = "user", content = "genes"))
      )
      expect_identical(captured$body$temperature, 0.7)
      expect_identical(captured$body$max_tokens, 4096)
      expect_false(captured$body$stream)
      expect_identical(captured$body$thinking, list(type = "disabled"))
      expect_s3_class(response, "response")
    },
    POST = function(url, config = NULL, body = NULL, ...) {
      captured$url <- url
      captured$headers <- config$headers
      captured$body <- body
      make_deepseek_json_response(list(
        choices = list(list(message = list(content = "Cluster 1: T cells")))
      ))
    },
    .package = "httr")
  },
  get_logger = quiet_deepseek_logger)
})

#' Compare predictions from different models
#' 
#' This function runs the same input through multiple models and compares their predictions.
#' It provides both individual predictions and a consensus analysis.
#' 
#' @note This function uses create_standardization_prompt from prompt_templates.R
#
#
#
#'   Supported models:
#'   - OpenAI: 'gpt-5.5', 'gpt-5.4', 'gpt-5.4-mini'
#'   - Anthropic: 'claude-opus-4-7', 'claude-opus-4-6', 'claude-sonnet-4-6', 'claude-haiku-4-5-20251001'
#'   - DeepSeek: 'deepseek-v4-flash', 'deepseek-v4-pro'
#'   - Google: 'gemini-3.1-pro-preview', 'gemini-3-flash-preview', 'gemini-3.1-flash-lite'
#'   - Alibaba: 'qwen3.6-max-preview', 'qwen3.6-plus', 'qwen3.6-flash'
#'   - Stepfun: 'step-3.5-flash', 'step-3.5-flash-2603', 'step-3'
#'   - Zhipu/Z.AI: 'glm-5.1', 'glm-5-turbo', 'glm-5'
#'   - MiniMax: 'MiniMax-M2.7', 'MiniMax-M2.7-highspeed', 'MiniMax-M2.5'
#'   - X.AI: 'grok-4.3', 'grok-4.3-latest', 'grok-latest'
#'   - OpenRouter: Provides access to models from multiple providers through a single API. Format: 'provider/model-name'
#'     - OpenAI models: 'openai/gpt-5.5', 'openai/gpt-5.4-mini'
#'     - Anthropic models: 'anthropic/claude-opus-4.7', 'anthropic/claude-sonnet-4.6'
#'     - Google models: 'google/gemini-3.1-pro-preview', 'google/gemini-3-flash-preview'
#'     - X.AI models: 'x-ai/grok-4.3'
#'     - Stepfun models: 'stepfun/step-3.5-flash'
#
#'   1. With provider names as keys: `list("openai" = "sk-...", "anthropic" = "sk-ant-...", "openrouter" = "sk-or-...")`
#'   2. With model names as keys: `list("gpt-5.5" = "sk-...", "claude-sonnet-4-6" = "sk-ant-...")`
#'   
#'   The system first tries to find the API key using the provider name. If not found, it then tries using the model name.
#'   Example:
#'   ```r
#'   api_keys <- list(
#'     "openai" = Sys.getenv("OPENAI_API_KEY"),
#'     "anthropic" = Sys.getenv("ANTHROPIC_API_KEY"),
#'     "openrouter" = Sys.getenv("OPENROUTER_API_KEY"),
#'     "claude-opus-4-7" = "your-claude-opus-key"
#'   )
#'   ```
#'
#' @param input Either a data frame from Seurat's FindAllMarkers() containing columns 'cluster', 'gene', and 'avg_log2FC', or a list with 'genes' field for each cluster
#' @param tissue_name Tissue context (e.g., 'human PBMC', 'mouse brain') for more accurate annotations
#' @param models Vector of model names to use for comparison. Default includes top models from each provider
#' @param api_keys Named list of API keys for the models, with provider or model names as keys.
#'   Every model in \code{models} must resolve to a non-NULL API key.
#' @param top_gene_count Number of top genes to use per cluster when input is from Seurat. Default: 10
#' @param consensus_threshold Minimum agreement threshold for consensus (0-1). Default: 0.5.
#'   Consensus is only evaluated when at least two non-missing model predictions are available for a cluster.
#' @param base_urls Optional base URLs for API endpoints. Can be a string or named list for provider-specific custom endpoints.
#'
#' @return List containing individual model predictions and consensus analysis
#'   If a cluster has fewer than two valid predictions after alignment/padding,
#'   its consensus-related outputs are \code{NA}.
#' @export
#' @examples
#' \dontrun{
#' # Compare predictions using different models
#' api_keys <- list(
#'   "claude-sonnet-4-6" = "your-anthropic-key",
#'   "deepseek-v4-pro" = "your-deepseek-key",
#'   "gemini-3.1-pro-preview" = "your-gemini-key",
#'   "qwen3.6-plus" = "your-qwen-key"
#' )
#' 
#' results <- compare_model_predictions(
#'   input = list(gs1=c('CD4','CD3D'), gs2='CD14'),
#'   tissue_name = 'PBMC',
#'   api_keys = api_keys
#' )
#' }
compare_model_predictions <- function(input,
                                      tissue_name,
                                      models = c("claude-opus-4-7",
                                                 "gpt-5.5",
                                                 "gemini-3.1-pro-preview",
                                                 "deepseek-v4-flash",
                                                 "qwen3.6-plus",
                                                 "grok-4.3"),
                                      api_keys,
                                      top_gene_count = 10,
                                      consensus_threshold = 0.5,
                                      base_urls = NULL) {
  tissue_name <- .normalize_required_string(tissue_name, "tissue_name")
  models <- .normalize_model_vector(models, minimum_count = 2L)
  api_keys <- .normalize_api_keys(api_keys)
  top_gene_count <- .normalize_top_gene_count(top_gene_count)
  consensus_threshold <- .normalize_probability(
    consensus_threshold,
    "consensus_threshold"
  )
  
  # Validate model/API-key pairs without letting one bad model block the rest
  eligible_models <- character(0)
  for (m in models) {
    provider <- tryCatch(get_provider(m), error = function(e) NULL)
    api_key <- if (is.null(provider)) NULL else get_api_key(m, api_keys)
    if (is.null(provider)) {
      warning(sprintf("Skipping model '%s': unsupported model name", m))
    } else if (is.null(api_key)) {
      warning(sprintf("Skipping model '%s': no API key found for provider '%s' or model name", m, provider))
    } else {
      eligible_models <- c(eligible_models, m)
    }
  }
  if (length(eligible_models) < 2) {
    stop("At least two models must have both a supported provider and an API key")
  }
  models <- eligible_models

  # Extract cluster IDs from input for display
  prompt_result <- create_annotation_prompt(input, tissue_name, top_gene_count)
  cluster_ids <- names(prompt_result$gene_lists)

  # Initialize results storage
  all_predictions <- list()
  successful_models <- character(0)

  # Get predictions from each model
  for (model in models) {
    message(sprintf("\nRunning predictions with model: %s", model))
    tryCatch({
      api_key <- get_api_key(model, api_keys)

      predictions <- annotate_cell_types(
        input = input,
        tissue_name = tissue_name,
        model = model,
        api_key = api_key,
        top_gene_count = top_gene_count,
        base_urls = base_urls
      )
      all_predictions[[model]] <- align_model_predictions(predictions, cluster_ids)
      successful_models <- c(successful_models, model)
    }, error = function(e) {
      warning(sprintf("Error with model %s: %s", model, e$message))
      log_warn(sprintf("Model %s failed during prediction", model), list(model = model, error = e$message))
    })
  }
  
  # Check if we have any successful predictions
  if (length(successful_models) == 0) {
    stop("No models successfully completed predictions")
  }
  
  # Standardize cell type names using LLM
  message("\nStandardizing cell type names...")
  standardized_predictions <- standardize_cell_type_names(all_predictions, successful_models, api_keys, base_urls = base_urls)
  
  # Every model is aligned to the canonical cluster IDs before comparison.
  all_vectors <- all_predictions[successful_models]
  std_vectors <- standardized_predictions[successful_models]
  cluster_labels <- cluster_ids

  comparison_matrix <- do.call(cbind, std_vectors)
  colnames(comparison_matrix) <- successful_models
  rownames(comparison_matrix) <- cluster_labels
  raw_matrix <- do.call(cbind, all_vectors)
  colnames(raw_matrix) <- successful_models
  rownames(raw_matrix) <- cluster_labels
  n_clusters <- nrow(comparison_matrix)

  # Calculate consensus and agreement statistics
  consensus_results <- apply(comparison_matrix, 1, function(row) {
    # Remove NAs
    valid_predictions <- row[!is.na(row)]
    if (length(valid_predictions) < 2) {
      return(list(consensus = NA, consensus_proportion = NA, entropy = NA))
    }
    
    # Count occurrences of each prediction
    pred_table <- table(valid_predictions)
    max_count <- max(pred_table)
    max_agreement <- max_count / length(valid_predictions)
    has_unique_majority <- sum(pred_table == max_count) == 1
    
    # Get consensus if agreement meets threshold
    consensus <- if (has_unique_majority && max_agreement >= consensus_threshold) {
      names(pred_table)[which.max(pred_table)]
    } else {
      NA
    }
    
    # Calculate consensus proportion
    consensus_proportion <- if (!is.na(consensus)) {
      pred_table[consensus] / length(valid_predictions)
    } else {
      NA
    }
    
    # Calculate entropy
    entropy <- -sum(pred_table / length(valid_predictions) * log2(pred_table / length(valid_predictions)))
    
    list(
      consensus = consensus,
      consensus_proportion = consensus_proportion,
      entropy = entropy
    )
  })
  
  # Format results
  consensus_predictions <- sapply(consensus_results, function(x) x$consensus)
  consensus_proportions <- sapply(consensus_results, function(x) x$consensus_proportion)
  entropies <- sapply(consensus_results, function(x) x$entropy)
  names(consensus_predictions) <- cluster_labels
  names(consensus_proportions) <- cluster_labels
  names(entropies) <- cluster_labels

  # Calculate overall statistics
  model_agreement_matrix <- matrix(NA, 
                                   nrow = length(successful_models), 
                                   ncol = length(successful_models),
                                   dimnames = list(successful_models, successful_models))
  
  for (i in seq_along(successful_models)) {
    for (j in seq_along(successful_models)) {
      if (i != j) {
        valid_comparisons <- !is.na(comparison_matrix[, successful_models[i]]) &
          !is.na(comparison_matrix[, successful_models[j]])
        if (any(valid_comparisons)) {
          agreement <- mean(comparison_matrix[valid_comparisons, successful_models[i]] ==
                              comparison_matrix[valid_comparisons, successful_models[j]])
          model_agreement_matrix[i,j] <- agreement
        }
      }
    }
  }
  
  mean_or_na <- function(x) {
    if (all(is.na(x))) NA_real_ else mean(x, na.rm = TRUE)
  }

  # Prepare summary statistics
  summary_stats <- list(
    total_clusters = n_clusters,
    consensus_reached = sum(!is.na(consensus_predictions)),
    mean_consensus_proportion = mean_or_na(consensus_proportions),
    mean_entropy = mean_or_na(entropies),
    model_agreement_matrix = model_agreement_matrix
  )

  # Return results
  results <- list(
    individual_predictions = all_predictions[successful_models],
    standardized_predictions = standardized_predictions[successful_models],
    comparison_matrix = comparison_matrix,
    consensus_predictions = consensus_predictions,
    consensus_proportions = consensus_proportions,
    entropies = entropies,
    summary_stats = summary_stats
  )
  
  # Print summary
  message("\nModel Comparison Summary:")
  message(sprintf("Total clusters analyzed: %d", summary_stats$total_clusters))
  message(sprintf("Clusters with consensus: %d (%.1f%%)",
              summary_stats$consensus_reached,
              100 * summary_stats$consensus_reached / summary_stats$total_clusters))
  mean_consensus_label <- if (is.na(summary_stats$mean_consensus_proportion)) {
    "NA"
  } else {
    sprintf("%.2f", summary_stats$mean_consensus_proportion)
  }
  mean_entropy_label <- if (is.na(summary_stats$mean_entropy)) {
    "NA"
  } else {
    sprintf("%.2f", summary_stats$mean_entropy)
  }
  message(sprintf("Mean consensus proportion: %s", mean_consensus_label))
  message(sprintf("Mean entropy: %s", mean_entropy_label))

  message("\nPairwise Model Agreement:")
  message(paste(utils::capture.output(print(model_agreement_matrix)), collapse = "\n"))

  message("\nDetailed Results:")
  for (i in 1:n_clusters) {
    cluster_label <- cluster_labels[i]
    message(sprintf("\nCluster %s:", cluster_label))
    for (model in successful_models) {
      message(sprintf("  %s: %s (Standardized: %s)",
                model,
                raw_matrix[i, model],
                comparison_matrix[i, model]))
    }
    consensus_proportion_label <- if (is.na(consensus_proportions[i])) {
      "NA"
    } else {
      sprintf("%.2f", consensus_proportions[i])
    }
    entropy_label <- if (is.na(entropies[i])) {
      "NA"
    } else {
      sprintf("%.2f", entropies[i])
    }
    message(sprintf("  Consensus: %s (Consensus Proportion: %s, Entropy: %s)",
                consensus_predictions[i],
                consensus_proportion_label,
                entropy_label))
  }
  
  invisible(results)
}

#' Standardize cell type names using a language model
#' 
#' This function takes predictions from multiple models and standardizes the cell type
#' nomenclature to ensure consistent naming across different models' outputs.
#' 
#
#
#
#'   1. With provider names as keys: `list("openai" = "sk-...", "anthropic" = "sk-ant-...", "openrouter" = "sk-or-...")`
#'   2. With model names as keys: `list("gpt-5.5" = "sk-...", "claude-sonnet-4-6" = "sk-ant-...")`
#
#
#' @keywords internal
standardize_cell_type_names <- function(predictions,
                                       models,
                                       api_keys,
                                       standardization_model = NULL,
                                       base_urls = NULL) {
  models <- .normalize_model_vector(models)
  if (is.null(standardization_model)) {
    standardization_model <- models[[1]]
  } else {
    standardization_model <- .normalize_required_string(
      standardization_model,
      "standardization_model"
    )
  }

  # Get API key for standardization model
  api_key <- get_api_key(standardization_model, api_keys)
  
  if (is.null(api_key)) {
    warning(sprintf("No API key found for standardization model '%s'. Using the first available model instead.",
                  standardization_model))
    log_warn("No API key for standardization model, falling back",
             list(requested = standardization_model, fallback = models[1]))
    standardization_model <- models[1]
    api_key <- get_api_key(standardization_model, api_keys)
  }
  
  # Get unique cell type names from all predictions
  all_cell_types <- unique(unlist(predictions[models]))
  valid_cell_types <- vapply(
    as.list(all_cell_types),
    is_real_cell_type_annotation,
    logical(1)
  )
  all_cell_types <- all_cell_types[valid_cell_types]
  
  if (length(all_cell_types) == 0) {
    warning("No valid cell type predictions found to standardize")
    log_warn("No valid cell type predictions found to standardize")
    return(predictions)
  }
  
  # Create a mapping of original cell types to standardized names
  message(sprintf("Using %s to standardize %d unique cell type names", 
                standardization_model, length(all_cell_types)))
  
  # Use the standardization prompt template from prompt_templates.R
  prompt <- create_standardization_prompt(all_cell_types)
  
  # Call the LLM to get standardized names
  tryCatch({
    response <- get_model_response(
      prompt = prompt,
      model = standardization_model,
      api_key = api_key,
      base_urls = base_urls
    )
    
    # Parse the response to extract mappings
    # response may be a character vector (one element per line from get_model_response),
    # so collapse first to ensure all lines are processed
    mapping_lines <- strsplit(paste(response, collapse = "\n"), "\n")[[1]]
    mapping <- list()
    
    # Function to clean cell type names by removing prefixes, numbers, etc.
    clean_cell_type <- function(cell_type) {
      if (is.na(cell_type)) return(cell_type)
      
      # Remove various number/cluster prefixes
      cleaned <- cell_type
      # Remove "1: ", "Cluster 1: ", etc.
      cleaned <- gsub("^\\s*\\d+\\s*:\\s*", "", cleaned)
      cleaned <- gsub("^\\s*[Cc]luster\\s*\\d+\\s*:\\s*", "", cleaned)
      # Also remove any leading numbers with any separator
      cleaned <- gsub("^\\s*\\d+\\s*[.:-]?\\s*", "", cleaned)
      # Trim any leading/trailing whitespace
      cleaned <- trimws(cleaned)
      
      return(cleaned)
    }

    valid_mapping_keys <- unique(c(all_cell_types, vapply(all_cell_types, clean_cell_type, character(1))))
    valid_mapping_keys <- valid_mapping_keys[order(nchar(valid_mapping_keys), decreasing = TRUE)]
    for (line in mapping_lines) {
      line <- trimws(line)
      if (!nzchar(line)) next

      for (original in valid_mapping_keys) {
        prefix <- paste0(original, ":")
        if (startsWith(line, prefix)) {
          standardized <- trimws(substring(line, nchar(prefix) + 1))
          if (nzchar(standardized)) {
            mapping[[original]] <- standardized
          }
          break
        }
      }
    }

    # Apply standardization to all predictions
    standardized_predictions <- predictions
    for (model in models) {
      for (i in seq_along(predictions[[model]])) {
        original <- predictions[[model]][i]
        if (!is.na(original)) {
          # Clean the original name first
          cleaned_original <- clean_cell_type(original)
          
          # Try direct mapping with original or cleaned name
          if (original %in% names(mapping)) {
            standardized <- mapping[[original]]
          } else if (cleaned_original %in% names(mapping)) {
            standardized <- mapping[[cleaned_original]]
          } else {
            # If no mapping found, use the cleaned original
            standardized <- cleaned_original
          }
          
          # Clean the standardized result as well to remove any remaining prefixes
          standardized_predictions[[model]][i] <- clean_cell_type(standardized)
        }
      }
    }
    
    # Print standardization mapping for reference
    message("\nCell Type Standardization Mapping:\n")
    for (original in names(mapping)) {
      message(sprintf("  %s -> %s\n", original, mapping[[original]]))
    }
    
    return(standardized_predictions)
    
  }, error = function(e) {
    warning(sprintf("Error in standardization: %s\nReturning original predictions.", e$message))
    log_warn("Standardization failed, returning original predictions", list(error = e$message))
    return(predictions)
  })
}

# Centralized global variable declarations for R CMD check
utils::globalVariables(c("custom_models", "cluster", "avg_log2FC", "gene"))

#' Package startup message
#' @keywords internal
.onAttach <- function(libname, pkgname) {
  # Get package version
  version <- utils::packageDescription(pkgname, fields = "Version")
  
  # ASCII art logo
  logo <- paste0(
    "\n",
    "+-------------------------------------------------------------------------+\n",
    "|                             mLLMCelltype                               |\n",
    "|                    Cell Type Annotation with LLMs                      |\n",
    "+-------------------------------------------------------------------------+\n"
  )
  
  # Create startup message
  msg <- paste0(
    logo,
    "\n",
    "mLLMCelltype v", version, " loaded successfully!\n",
    "For more information, please visit:\n",
    "- Documentation: https://cafferychen777.github.io/mLLMCelltype/\n",
    "- GitHub: https://github.com/cafferychen777/mLLMCelltype\n",
    "- Paper: https://doi.org/10.1038/s42003-026-10420-8\n",
    "\n",
    "To cite this package in publications, use:\n",
    "  citation(\"mLLMCelltype\")"
  )
  
  
  # Display the message
  packageStartupMessage(msg)
  
  # Only show message when old cache is found
  old_cache <- file.path(".", "consensus_cache")
  if (dir.exists(old_cache)) {
    packageStartupMessage(
      "Note: Found old cache directory 'consensus_cache' in current directory.\n",
      "mLLMCelltype now uses system cache by default. To clean up:\n",
      "  unlink('consensus_cache', recursive = TRUE)\n",
      "For cache info, use: mllmcelltype_cache_dir()"
    )
  }
}

#' Package load message
#' @keywords internal
.onLoad <- function(libname, pkgname) {
  validate_builtin_provider_registry()

  # Set any package-wide options here if needed
  # For example, you could set default logging level
  op <- options()
  op.mLLMCelltype <- list(
    mLLMCelltype.log_level = "INFO"
  )
  toset <- !(names(op.mLLMCelltype) %in% names(op))
  if(any(toset)) options(op.mLLMCelltype[toset])
  
  
  invisible()
}

.BUILTIN_PROVIDER_SPECS <- list(
  openai = list(
    pattern = "^(gpt-|o[134]|chatgpt-|codex-)",
    processor_class = "OpenAIProcessor",
    display_name = "OpenAI"
  ),
  anthropic = list(
    pattern = "^claude-",
    processor_class = "AnthropicProcessor",
    display_name = "Anthropic"
  ),
  deepseek = list(
    pattern = "^deepseek-",
    processor_class = "DeepSeekProcessor",
    display_name = "DeepSeek"
  ),
  gemini = list(
    pattern = "^gemini-",
    processor_class = "GeminiProcessor",
    display_name = "Gemini",
    api_key_env_aliases = "GOOGLE_API_KEY"
  ),
  qwen = list(
    pattern = "^(qwen|qwq-)",
    processor_class = "QwenProcessor",
    display_name = "Qwen"
  ),
  stepfun = list(
    pattern = "^step-",
    processor_class = "StepFunProcessor",
    display_name = "StepFun"
  ),
  zhipu = list(
    pattern = "^(glm-|chatglm)",
    processor_class = "ZhipuProcessor",
    display_name = "Zhipu"
  ),
  minimax = list(
    pattern = "^minimax-",
    processor_class = "MinimaxProcessor",
    display_name = "MiniMax"
  ),
  grok = list(
    pattern = "^grok-",
    processor_class = "GrokProcessor",
    display_name = "Grok"
  ),
  kimi = list(
    pattern = "^(kimi-|moonshot-)",
    processor_class = "KimiProcessor",
    display_name = "Kimi",
    api_key_env_aliases = "MOONSHOT_API_KEY"
  ),
  openrouter = list(
    pattern = NULL,
    processor_class = "OpenRouterProcessor",
    display_name = "OpenRouter"
  )
)

# Derived compatibility view used by provider detection and error reporting.
.BUILTIN_PROVIDER_PATTERNS <- vapply(
  .BUILTIN_PROVIDER_SPECS[
    vapply(.BUILTIN_PROVIDER_SPECS, function(spec) !is.null(spec$pattern), logical(1))
  ],
  function(spec) spec$pattern,
  character(1)
)

get_builtin_provider_spec <- function(provider) {
  provider <- tolower(.normalize_required_string(provider, "provider"))
  spec <- .BUILTIN_PROVIDER_SPECS[[provider]]
  if (is.null(spec)) {
    stop("Unsupported model provider: ", provider)
  }
  spec
}

get_builtin_provider_display_name <- function(provider) {
  provider <- tolower(.normalize_required_string(provider, "provider"))
  spec <- .BUILTIN_PROVIDER_SPECS[[provider]]
  if (is.null(spec)) provider else spec$display_name
}

resolve_builtin_processor_class <- function(provider) {
  spec <- get_builtin_provider_spec(provider)
  class_name <- spec$processor_class
  valid_class_name <- is.character(class_name) &&
    length(class_name) == 1 &&
    !is.na(class_name) &&
    nzchar(class_name)
  if (!valid_class_name || !exists(class_name, envir = environment(), inherits = TRUE)) {
    stop("Invalid processor class for built-in provider: ", provider)
  }
  processor_class <- get(
    class_name,
    envir = environment(),
    inherits = TRUE
  )
  if (!inherits(processor_class, "R6ClassGenerator")) {
    stop("Invalid processor class for built-in provider: ", provider)
  }
  processor_class
}

new_builtin_provider_processor <- function(provider, base_url = NULL) {
  resolve_builtin_processor_class(provider)$new(base_url = base_url)
}

validate_builtin_provider_registry <- function() {
  required_fields <- c("processor_class", "display_name")
  for (provider in names(.BUILTIN_PROVIDER_SPECS)) {
    spec <- .BUILTIN_PROVIDER_SPECS[[provider]]
    if (!is.list(spec) || !all(required_fields %in% names(spec))) {
      stop("Invalid built-in provider specification: ", provider)
    }
    for (field in required_fields) {
      value <- spec[[field]]
      if (!is.character(value) || length(value) != 1 || is.na(value) || !nzchar(value)) {
        stop("Invalid ", field, " for built-in provider: ", provider)
      }
    }
    if (!is.null(spec$pattern)) {
      valid_pattern <- is.character(spec$pattern) &&
        length(spec$pattern) == 1 &&
        !is.na(spec$pattern) &&
        nzchar(spec$pattern)
      if (!valid_pattern) {
        stop("Invalid pattern for built-in provider: ", provider)
      }
      tryCatch(
        grepl(spec$pattern, ""),
        error = function(e) {
          stop("Invalid pattern for built-in provider: ", provider)
        }
      )
    }
    if (!is.null(spec$api_key_env_aliases)) {
      aliases <- spec$api_key_env_aliases
      valid_aliases <- is.character(aliases) &&
        length(aliases) > 0 &&
        !anyNA(aliases) &&
        all(grepl("^[A-Za-z_][A-Za-z0-9_]*$", aliases)) &&
        !anyDuplicated(aliases)
      if (!valid_aliases) {
        stop("Invalid API key environment aliases for built-in provider: ", provider)
      }
    }
    resolve_builtin_processor_class(provider)
  }
  invisible(TRUE)
}

#' Determine provider from model name
#'
#' This function determines the appropriate provider (e.g., OpenAI, Anthropic, Google, OpenRouter) based on the model name.
#' Uses prefix-based matching for efficient and maintainable provider detection.
#' New models following existing naming conventions are automatically supported.
#'
#' @param model Character string specifying the model name (e.g., "gpt-5.5", "claude-opus-4-7").
#' @return Character string of the provider name (e.g., "openai", "anthropic").
#' @details
#' Supported providers and model prefixes:
#' \itemize{
#'   \item OpenAI: gpt-*, o1*, o3*, o4*, chatgpt-*, codex-* (e.g., 'gpt-5.5', 'gpt-5.4-mini')
#'   \item Anthropic: claude-* (e.g., 'claude-opus-4-7', 'claude-sonnet-4-6')
#'   \item DeepSeek: deepseek-* (e.g., 'deepseek-v4-flash', 'deepseek-v4-pro')
#'   \item Google: gemini-* (e.g., 'gemini-3.1-pro-preview', 'gemini-3-flash-preview')
#'   \item Qwen: qwen*, qwq-* (e.g., 'qwen3.6-plus', 'qwen3.6-flash')
#'   \item Stepfun: step-* (e.g., 'step-3.5-flash', 'step-3')
#'   \item Zhipu: glm-*, chatglm* (e.g., 'glm-5.1', 'glm-5-turbo')
#'   \item MiniMax: minimax-* (e.g., 'MiniMax-M2.7', 'MiniMax-M2.5')
#'   \item Grok: grok-* (e.g., 'grok-4.3', 'grok-4.3-latest')
#'   \item Kimi (Moonshot AI Open Platform): kimi-*, moonshot-* (e.g., 'kimi-k2.6', 'moonshot-v1-8k')
#'   \item OpenRouter: Any model with '/' in the name (e.g., 'openai/gpt-5.5', 'anthropic/claude-opus-4.7')
#' }
#' @export
get_provider <- function(model) {
  # Normalize model name to lowercase for case-insensitive matching
  model_normalized <- .normalize_required_string(model, "model")
  model_lower <- tolower(model_normalized)

  # OpenRouter models always contain '/' (e.g., 'openai/gpt-5.5')
  if (grepl("/", model_lower)) {
    return("openrouter")
  }

  # Check for custom models
  if (exists(model_lower, envir = custom_models)) {
    model_data <- get(model_lower, envir = custom_models)
    return(model_data$provider)
  }

  # Prefix-based provider detection from the shared built-in registry.
  for (provider in names(.BUILTIN_PROVIDER_PATTERNS)) {
    if (grepl(.BUILTIN_PROVIDER_PATTERNS[[provider]], model_lower)) {
      return(provider)
    }
  }

  # No match — report error with supported prefixes
  supported <- paste(
    vapply(names(.BUILTIN_PROVIDER_PATTERNS), function(p) {
      sprintf("  %s: %s", p, .BUILTIN_PROVIDER_PATTERNS[[p]])
    }, character(1)),
    collapse = "\n"
  )
  stop(
    "Cannot determine provider for model: '", model_normalized, "'\n",
    "Supported model name patterns:\n", supported, "\n",
    "  openrouter: any model containing '/'\n",
    "Tip: Check for typos in the model name."
  )
}

# Internal failure markers that must never become user-facing annotations.
.SENTINEL_VALUES <- c(
  "Parsing_Failed",
  "Insufficient_Responses",
  "Prediction_Missing",
  "Annotation_Missing",
  "Unknown"
)

.UNKNOWN_ANNOTATION_TOKENS <- c(
  tolower(.SENTINEL_VALUES),
  "unk",
  "n/a",
  "na",
  "none",
  "null",
  "nan",
  "-",
  "--",
  "inconclusive"
)

.unwrap_annotation_wrappers <- function(annotation) {
  wrapper_pairs <- c(
    "[" = "]",
    "(" = ")",
    "{" = "}",
    "\"" = "\"",
    "'" = "'"
  )
  normalized <- trimws(annotation)

  while (nchar(normalized) >= 2) {
    opening <- substr(normalized, 1, 1)
    expected_closing <- unname(wrapper_pairs[opening])
    if (is.na(expected_closing) ||
        substr(normalized, nchar(normalized), nchar(normalized)) != expected_closing) {
      break
    }
    normalized <- trimws(substr(normalized, 2, nchar(normalized) - 1))
  }
  normalized
}

is_real_cell_type_annotation <- function(annotation) {
  is_text_scalar <- is.character(annotation) &&
    length(annotation) == 1 &&
    !is.na(annotation)
  if (!is_text_scalar) {
    return(FALSE)
  }

  normalized <- tolower(.unwrap_annotation_wrappers(annotation))
  if (!nzchar(normalized) || normalized %in% .UNKNOWN_ANNOTATION_TOKENS) {
    return(FALSE)
  }

  contextual_unknown <- grepl(
    "^(unknown|inconclusive)\\s*[\\(\\[\\{].*[\\)\\]\\}]$",
    normalized,
    perl = TRUE
  )
  error_sentinel <- grepl(
    "^error(\\s*:.*|\\s*\\(.*\\))?$",
    normalized,
    perl = TRUE
  )
  !contextual_unknown && !error_sentinel
}

is_error_response <- function(result) {
  if (!is.character(result) || length(result) == 0) {
    return(FALSE)
  }
  non_empty <- result[!is.na(result) & nzchar(trimws(result))]
  if (length(non_empty) == 0) {
    return(FALSE)
  }
  # A response is an error only when it carries an error marker AND offers no
  # usable annotation. This preserves a valid multi-cluster response that flags
  # a single uncertain cluster (e.g. "Error: markers ambiguous") instead of
  # discarding every annotation the model returned.
  has_error <- any(grepl("^\\s*error\\s*:", non_empty, ignore.case = TRUE))
  has_error && !any(vapply(non_empty, is_real_cell_type_annotation, logical(1)))
}

normalize_model_response_lines <- function(result) {
  if (!is.character(result) || length(result) == 0 || anyNA(result)) {
    stop("Model response must be non-missing text")
  }

  lines <- unlist(strsplit(result, "\n", fixed = TRUE), use.names = FALSE)
  lines <- trimws(lines)
  lines <- sub(",+$", "", lines)
  lines <- trimws(lines)
  lines <- unname(lines[nzchar(lines)])

  if (length(lines) == 0) {
    stop("Model response must contain non-empty text")
  }
  if (is_error_response(lines)) {
    stop("Model response contains an error sentinel")
  }
  lines
}

normalize_token_usage <- function(usage) {
  if (!is.list(usage)) {
    return(NULL)
  }

  normalize_count <- function(value) {
    valid <- is.numeric(value) && length(value) == 1 &&
      !is.na(value) && is.finite(value) && value >= 0 &&
      value == floor(value)
    if (valid) as.numeric(value) else NULL
  }

  normalized <- list(
    prompt_tokens = normalize_count(usage$prompt_tokens),
    completion_tokens = normalize_count(usage$completion_tokens),
    total_tokens = normalize_count(usage$total_tokens)
  )

  cost <- usage$cost
  if (is.numeric(cost) && length(cost) == 1 &&
      !is.na(cost) && is.finite(cost) && cost >= 0) {
    normalized$cost <- as.numeric(cost)
  }
  normalized
}

is_valid_model_response <- function(result) {
  isTRUE(tryCatch({
    normalize_model_response_lines(result)
    TRUE
  }, error = function(e) FALSE))
}

is_retryable_http_status <- function(status_code) {
  is.numeric(status_code) &&
    length(status_code) == 1 &&
    !is.na(status_code) &&
    is.finite(status_code) &&
    status_code == floor(status_code) &&
    (status_code %in% c(408, 425, 429) || status_code >= 500)
}

#' Extract the first JSON document from a model response
#'
#' Strips surrounding markdown fences and returns the first parseable JSON
#' array or object found in the text.
#'
#' @param text Single character string containing the model response.
#' @return The extracted JSON text as a character string.
#' @noRd
extract_json_from_response <- function(text) {
  if (!is.character(text) || length(text) != 1 || is.na(text)) {
    stop("Response text must be a single non-missing character string")
  }

  text <- paste(text, collapse = "\n")
  text <- trimws(text)

  # Strip markdown code fences if present
  text <- sub("^```[[:alnum:]]*\\s*\n?", "", text)
  text <- sub("\n?```\\s*$", "", text)
  text <- trimws(text)

  # Try parsing the whole text first
  parsed <- tryCatch(
    jsonlite::fromJSON(text, simplifyVector = FALSE),
    error = function(e) NULL
  )
  if (!is.null(parsed)) {
    return(text)
  }

  # Extract between first JSON delimiter and a matching closing delimiter
  first_bracket <- regexpr("\\[", text)
  first_brace <- regexpr("\\{", text)

  if (first_bracket > 0 && (first_brace < 0 || first_bracket < first_brace)) {
    start <- first_bracket
    end_positions <- gregexpr("\\]", text)[[1]]
  } else if (first_brace > 0) {
    start <- first_brace
    end_positions <- gregexpr("\\}", text)[[1]]
  } else {
    stop("No JSON array or object found in response")
  }

  end_positions <- end_positions[end_positions > start]
  if (length(end_positions) == 0) {
    stop("No complete JSON array or object found in response")
  }

  for (end in sort(end_positions, decreasing = TRUE)) {
    candidate <- substring(text, start, end)
    parsed <- tryCatch(
      jsonlite::fromJSON(candidate, simplifyVector = FALSE),
      error = function(e) NULL
    )
    if (!is.null(parsed)) {
      return(candidate)
    }
  }

  stop("Could not parse JSON array or object from response")
}

#' Parse reasoning annotation JSON into a structured named list
#'
#' @param json_text JSON text containing reasoning annotations.
#' @param expected_clusters Character vector of expected cluster IDs.
#' @return Named list mapping each expected cluster ID to a list with
#'   `cell_type`, `marker_genes`, and `gene_expression`.
#' @noRd
parse_reasoning_annotations <- function(json_text, expected_clusters) {
  json_text <- extract_json_from_response(json_text)

  parsed <- tryCatch(
    jsonlite::fromJSON(json_text, simplifyVector = FALSE),
    error = function(e) stop(paste("Failed to parse reasoning JSON response:", e$message))
  )

  if (!is.list(parsed)) {
    stop("Reasoning response must be a JSON array or object")
  }

  # Normalize array or object form into a list of records
  if (is.null(names(parsed))) {
    records <- parsed
  } else {
    records <- lapply(names(parsed), function(cid) {
      record <- parsed[[cid]]
      if (is.list(record)) {
        record$cluster_id <- cid
      }
      record
    })
  }

  result <- list()
  for (cid in expected_clusters) {
    result[[cid]] <- list(
      cell_type = "Unknown",
      marker_genes = "",
      gene_expression = ""
    )
  }

  for (record in records) {
    if (!is.list(record) || is.null(record$cluster_id)) {
      next
    }
    cid <- trimws(as.character(record$cluster_id))
    if (!(cid %in% expected_clusters)) {
      next
    }

    get_field <- function(field) {
      value <- record[[field]]
      if (!is.null(value) && is.character(value) && length(value) == 1 && !is.na(value)) {
        trimws(value)
      } else {
        ""
      }
    }

    result[[cid]] <- list(
      cell_type = get_field("cell_type"),
      marker_genes = get_field("marker_genes"),
      gene_expression = get_field("gene_expression")
    )
  }

  result
}

stop_api_request_error <- function(message, status_code = NULL,
                                   retryable = NULL) {
  if (is.null(retryable)) {
    retryable <- is_retryable_http_status(status_code)
  }
  condition <- structure(
    list(
      message = message,
      call = NULL,
      status_code = status_code,
      retryable = isTRUE(retryable)
    ),
    class = c("mllm_api_error", "error", "condition")
  )
  stop(condition)
}

#' Anthropic API Processor
#' 
#' Concrete implementation of BaseAPIProcessor for Anthropic models.
#' Handles Anthropic-specific API calls, authentication, and response parsing.
#'
#' @export
AnthropicProcessor <- R6::R6Class("AnthropicProcessor",
  inherit = BaseAPIProcessor,
  
  public = list(
    #' @description
    #' Initialize Anthropic processor
    #' @param base_url Optional custom API endpoint
    initialize = function(base_url = NULL) {
      super$initialize("anthropic", base_url)
    },

    #' @description
    #' Get default Anthropic API URL
    #
    get_default_api_url = function() {
      return("https://api.anthropic.com/v1/messages")
    },
    
    #' @description
    #' Make API call to Anthropic
    #' @param chunk_content Prompt text to send
    #' @param model Model identifier
    #' @param api_key Anthropic API key
    make_api_call = function(chunk_content, model, api_key) {
      # Prepare request body
      body <- list(
        model = model,
        max_tokens = 4096,
        messages = list(
          list(
            role = "user",
            content = chunk_content
          )
        )
      )
      
      self$logger$debug("Sending API request to Anthropic",
                       list(model = model, provider = self$provider_name))
      
      # Make the API request
      response <- httr::POST(
        url = self$get_api_url(),
        httr::add_headers(
          "x-api-key" = api_key,
          "anthropic-version" = "2023-06-01",
          "content-type" = "application/json"
        ),
        body = body,
        encode = "json",
        httr::timeout(30)
      )
      
      private$stop_for_http_error(response, model, "Anthropic")
      
      return(response)
    },
    
    #' @description
    #' Extract response content from Anthropic API response
    #' @param response HTTP response object
    #' @param model Model identifier
    extract_response_content = function(response, model) {
      private$extract_messages_content(response, model, "Anthropic")
    },

    #' @description
    #' Extract normalized Anthropic token usage
    #' @param response HTTP response object
    extract_usage = function(response) {
      private$extract_usage_fields(
        response,
        prompt_field = "input_tokens",
        completion_field = "output_tokens",
        total_field = NULL,
        cost_field = NULL,
        derive_total = TRUE
      )
    }
  )
)

# Normalize API key mappings before provider- or model-specific lookup.

.normalize_api_keys <- function(api_keys, allow_null = FALSE, validate_values = TRUE) {
  if (is.null(api_keys) && allow_null) {
    return(NULL)
  }
  if (!is.list(api_keys) || length(api_keys) == 0 || is.null(names(api_keys))) {
    stop("api_keys must be a named, non-empty list")
  }

  normalized_names <- trimws(names(api_keys))
  if (anyNA(names(api_keys)) || any(!nzchar(normalized_names))) {
    stop("api_keys names must be non-empty strings")
  }
  if (anyDuplicated(normalized_names)) {
    stop("api_keys names must be unique after whitespace normalization")
  }

  registered_providers <- c(
    names(.BUILTIN_PROVIDER_SPECS),
    ls(custom_providers, all.names = TRUE)
  )
  normalized_provider_names <- tolower(normalized_names)
  provider_keys <- normalized_provider_names[
    normalized_provider_names %in% registered_providers
  ]
  if (anyDuplicated(provider_keys)) {
    stop("api_keys provider names must be unique after case/whitespace normalization")
  }

  names(api_keys) <- normalized_names
  if (validate_values) {
    for (index in seq_along(api_keys)) {
      value <- api_keys[[index]]
      is_missing <- is.null(value) || (
        length(value) == 1 && is.atomic(value) && is.na(value)
      )
      is_character_scalar <- is.character(value) && length(value) == 1
      if (!is_missing && !is_character_scalar) {
        stop(
          "api_keys value for '", normalized_names[[index]],
          "' must be a character scalar or missing"
        )
      }
      if (is_character_scalar && !is.na(value)) {
        api_keys[[index]] <- trimws(value)
      }
    }
  }
  api_keys
}

.get_api_key_from_mapping <- function(model, provider, api_keys) {
  is_valid_key <- function(key) {
    is.character(key) && length(key) == 1 && !is.na(key) && nzchar(trimws(key))
  }
  get_named_key <- function(key_name, ignore_case = FALSE) {
    api_key_names <- names(api_keys)
    if (is.null(api_key_names)) {
      return(NULL)
    }

    match_index <- if (ignore_case) {
      match(tolower(key_name), tolower(api_key_names))
    } else {
      match(key_name, api_key_names)
    }
    if (is.na(match_index)) {
      return(NULL)
    }

    key <- api_keys[[match_index]]
    if (is_valid_key(key)) {
      return(trimws(key))
    }
    NULL
  }

  provider_key <- get_named_key(provider, ignore_case = TRUE)
  if (!is.null(provider_key)) {
    return(provider_key)
  }
  get_named_key(model)
}

.api_key_environment_variables <- function(provider) {
  provider <- tolower(.normalize_required_string(provider, "provider"))
  primary_variable <- paste0(toupper(provider), "_API_KEY")
  provider_spec <- .BUILTIN_PROVIDER_SPECS[[provider]]
  aliases <- if (is.null(provider_spec)) {
    character(0)
  } else {
    provider_spec$api_key_env_aliases
  }
  unique(c(primary_variable, aliases))
}

.resolve_model_api_key <- function(model, api_keys) {
  model <- .normalize_required_string(model, "model")
  api_keys <- .normalize_api_keys(
    api_keys,
    allow_null = TRUE,
    validate_values = FALSE
  )
  provider <- get_provider(model)
  api_key <- .get_api_key_from_mapping(model, provider, api_keys)
  if (!is.null(api_key)) {
    return(list(
      api_key = api_key,
      provider = provider,
      source = "api_keys"
    ))
  }

  for (variable in .api_key_environment_variables(provider)) {
    api_key <- trimws(Sys.getenv(variable, unset = ""))
    if (nzchar(api_key)) {
      return(list(
        api_key = api_key,
        provider = provider,
        source = "environment"
      ))
    }
  }

  list(
    api_key = NULL,
    provider = provider,
    source = "none"
  )
}

#' Get an API key for a model
#'
#' Retrieves a configured API key by checking the model's provider name first,
#' followed by the exact model name.
#'
#' @param model Model name to get API key for
#' @param api_keys Named list of API keys with provider or model names as keys
#'
#' @return A trimmed API key string, or `NULL` when no valid key is configured.
#' @export
get_api_key <- function(model, api_keys) {
  model_normalized <- .normalize_required_string(model, "model")
  api_keys <- .normalize_api_keys(
    api_keys,
    allow_null = TRUE,
    validate_values = FALSE
  )
  provider <- get_provider(model_normalized)
  .get_api_key_from_mapping(model_normalized, provider, api_keys)
}

.normalize_required_string <- function(value, field_name) {
  if (!is.character(value) || length(value) != 1 || is.na(value) ||
      !nzchar(trimws(value))) {
    stop(field_name, " must be a non-empty character scalar")
  }
  trimws(value)
}

.normalize_positive_integer <- function(value, field_name) {
  valid <- is.numeric(value) && length(value) == 1 &&
    !is.na(value) && is.finite(value) &&
    value >= 1 && value == as.integer(value)
  if (!valid) {
    stop(field_name, " must be a positive integer")
  }
  as.integer(value)
}

.normalize_probability <- function(value, field_name) {
  valid <- is.numeric(value) && length(value) == 1 &&
    !is.na(value) && is.finite(value) &&
    value >= 0 && value <= 1
  if (!valid) {
    stop(field_name, " must be a finite number between 0 and 1")
  }
  as.numeric(value)
}

.normalize_nonnegative_number <- function(value, field_name) {
  valid <- is.numeric(value) && length(value) == 1 &&
    !is.na(value) && is.finite(value) && value >= 0
  if (!valid) {
    stop(field_name, " must be a finite non-negative number")
  }
  as.numeric(value)
}

.normalize_positive_number <- function(value, field_name) {
  valid <- is.numeric(value) && length(value) == 1 &&
    !is.na(value) && is.finite(value) && value > 0
  if (!valid) {
    stop(field_name, " must be a finite positive number")
  }
  as.numeric(value)
}

.normalize_log_level <- function(level) {
  normalized <- toupper(.normalize_required_string(level, "level"))
  valid_levels <- c("DEBUG", "INFO", "WARN", "ERROR")
  if (!normalized %in% valid_levels) {
    stop("level must be one of: ", paste(valid_levels, collapse = ", "))
  }
  normalized
}

.normalize_flag <- function(value, field_name) {
  if (!is.logical(value) || length(value) != 1 || is.na(value)) {
    stop(field_name, " must be TRUE or FALSE")
  }
  value
}

.normalize_model_vector <- function(models, minimum_count = 1L) {
  valid <- is.character(models) &&
    length(models) >= minimum_count &&
    !anyNA(models) &&
    all(nzchar(trimws(models)))
  if (!valid) {
    stop(sprintf(
      "models must contain at least %d non-empty model name%s",
      minimum_count,
      if (minimum_count == 1L) "" else "s"
    ))
  }

  normalized <- trimws(models)
  if (anyDuplicated(normalized)) {
    stop("models must not contain duplicate names")
  }
  normalized
}

# Package-level cache for Qwen endpoint (persists across QwenProcessor instances)
.qwen_endpoint_cache <- new.env(parent = emptyenv())

#' Qwen API Processor
#'
#' Concrete implementation of BaseAPIProcessor for Qwen models.
#' Handles Qwen-specific API calls, authentication, and response parsing.
#'
#' @export
QwenProcessor <- R6::R6Class("QwenProcessor",
  inherit = BaseAPIProcessor,

  private = list(

    #' @description
    #' Test if an endpoint is accessible
    #
    #
    #
    test_endpoint = function(url, api_key) {
      tryCatch({
        # Simple OpenAI-compatible test payload. Region-specific DashScope
        # keys can return 401/403 on the wrong regional host, so keep probing
        # alternatives for those statuses.
        test_payload <- list(
          model = "qwen-turbo",
          messages = list(
            list(role = "user", content = "test")
          ),
          max_tokens = 1
        )

        response <- httr::POST(
          url = url,
          httr::add_headers(
            "Authorization" = paste("Bearer", api_key),
            "Content-Type" = "application/json"
          ),
          body = test_payload,
          encode = "json",
          httr::timeout(10)  # 10 second timeout for quick test
        )

        status_code <- httr::status_code(response)
        return(!is.null(status_code) &&
                 !(status_code %in% c(401, 403, 404)) &&
                 status_code < 500)
      }, error = function(e) {
        return(FALSE)
      })
    }
  ),

  public = list(
    #' @description
    #' Initialize Qwen processor
    #' @param base_url Optional custom API endpoint
    initialize = function(base_url = NULL) {
      super$initialize("qwen", base_url)
    },

    #' @description
    #' Get default Qwen OpenAI-compatible chat completions API URL
    #
    #' @details Qwen has OpenAI-compatible chat completions endpoints:
    #'   - International (US): https://dashscope-us.aliyuncs.com/compatible-mode/v1/chat/completions
    #'   - Domestic (China): https://dashscope.aliyuncs.com/compatible-mode/v1/chat/completions
    #'   - Legacy international: https://dashscope-intl.aliyuncs.com/compatible-mode/v1/chat/completions
    #'   The processor automatically tries international first, then domestic, then legacy international.
    get_default_api_url = function() {
      return("https://dashscope-us.aliyuncs.com/compatible-mode/v1/chat/completions")
    },

    #' @description
    #' Get working Qwen API URL with automatic endpoint detection
    #' @param api_key Qwen API key used for regional endpoint probing
    get_working_api_url = function(api_key) {
      cache_key <- digest::digest(api_key, algo = "xxhash64")
      if (exists(cache_key, envir = .qwen_endpoint_cache, inherits = FALSE)) {
        return(get(cache_key, envir = .qwen_endpoint_cache, inherits = FALSE))
      }

      endpoints <- list(
        international = "https://dashscope-us.aliyuncs.com/compatible-mode/v1/chat/completions",
        domestic = "https://dashscope.aliyuncs.com/compatible-mode/v1/chat/completions",
        legacy_international = "https://dashscope-intl.aliyuncs.com/compatible-mode/v1/chat/completions"
      )

      self$logger$debug("Testing Qwen endpoints for accessibility")

      for (endpoint_name in names(endpoints)) {
        endpoint_url <- endpoints[[endpoint_name]]
        if (private$test_endpoint(endpoint_url, api_key)) {
          self$logger$info(
            sprintf("Using Qwen %s endpoint", endpoint_name),
            list(url = endpoint_url)
          )
          assign(cache_key, endpoint_url, envir = .qwen_endpoint_cache)
          return(endpoint_url)
        }
      }

      # If all probes fail, return international as default and let the main call
      # handle the error.
      self$logger$warn(
        "All Qwen endpoints failed during testing, using international as default"
      )
      return(endpoints$international)
    },
    
    #' @description
    #' Make API call to Qwen
    #' @param chunk_content Prompt text to send
    #' @param model Model identifier
    #' @param api_key Qwen API key
    make_api_call = function(chunk_content, model, api_key) {
      # Get API URL: custom base_url takes priority, otherwise auto-detect endpoint
      api_url <- if (!is.null(self$base_url)) {
        self$base_url
      } else {
        self$get_working_api_url(api_key)
      }

      private$post_chat_completions_request(
        chunk_content,
        model,
        api_key,
        api_url = api_url,
        body_extra = list(
          temperature = 0.7,
          max_tokens = 4096
        )
      )
    },
    
    #' @description
    #' Extract response content from Qwen API response
    #' @param response HTTP response object
    #' @param model Model identifier
    extract_response_content = function(response, model) {
      self$logger$debug("Parsing Qwen API response",
                       list(provider = self$provider_name, model = model))

      # Parse the response
      content <- httr::content(response, "parsed")

      # Extract from the current OpenAI-compatible format first, then retain
      # legacy DashScope format support for older cached/mocked responses.
      if (!is.null(content$choices) &&
          length(content$choices) > 0 &&
          !is.null(content$choices[[1]]$message$content)) {
        response_content <- content$choices[[1]]$message$content
      } else if (!is.null(content$output$text)) {
        response_content <- content$output$text
      } else if (!is.null(content$output$choices) &&
                 length(content$output$choices) > 0 &&
                 !is.null(content$output$choices[[1]]$message$content)) {
        response_content <- content$output$choices[[1]]$message$content
      } else {
        self$logger$error("Unexpected response format from Qwen API",
                         list(provider = self$provider_name,
                              model = model,
                              content_structure = names(content),
                              output_keys = if (!is.null(content$output)) names(content$output) else NULL))
        stop("Unexpected response format from Qwen API")
      }

      return(response_content)
    }
  )
)

# =============================================================================
# API Rate Limit Testing Script
# =============================================================================
# Purpose: Test different parallelization strategies against API rate limits
# Author: Performance Analysis Team
# Date: 2025-11-14
# =============================================================================

library(mLLMCelltype)

# Load environment variables
if (file.exists(".env")) {
  readRenviron(".env")
}

# =============================================================================
# Test Configuration
# =============================================================================

# Test data: Small dataset for quick iteration
test_clusters <- list(
  "0" = list(genes = c("CD3D", "CD3E", "CD3G", "CD28", "TRAC", "TRBC1", "IL7R", "LCK", "CD2", "CD7")),
  "1" = list(genes = c("CD79A", "CD79B", "MS4A1", "CD19", "BANK1", "PAX5", "BLK", "CD22", "TCL1A", "IGHM")),
  "2" = list(genes = c("CD14", "CD68", "CSF1R", "FCGR3A", "LYZ", "S100A8", "S100A9", "VCAN", "FCN1", "MNDA"))
)

tissue_name <- "human PBMC"

# Models to test
models <- c(
  "claude-sonnet-4-6",
  "claude-haiku-4-5-20251001"
)

# API keys
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY")
)

# Verify API key
if (api_keys$anthropic == "") {
  stop("ANTHROPIC_API_KEY not found in environment. Please set it in .env file.")
}

cat("=============================================================================\n")
cat("API RATE LIMIT TESTING\n")
cat("=============================================================================\n")
cat(sprintf("Test dataset: %d clusters\n", length(test_clusters)))
cat(sprintf("Models to test: %d (%s)\n", length(models), paste(models, collapse = ", ")))
cat(sprintf("API Key found: %s...\n", substr(api_keys$anthropic, 1, 20)))
cat("=============================================================================\n\n")

# =============================================================================
# Test 1: Serial Execution (Baseline)
# =============================================================================

test_serial <- function() {
  cat("\n", strrep("=", 80), "\n")
  cat("TEST 1: SERIAL EXECUTION (BASELINE)\n")
  cat(strrep("=", 80), "\n\n")

  start_time <- Sys.time()
  results <- list()
  api_errors <- list()

  for (i in seq_along(models)) {
    model <- models[i]
    cat(sprintf("[%d/%d] Calling %s...\n", i, length(models), model))

    model_start <- Sys.time()

    result <- tryCatch({
      annotate_cell_types(
        input = test_clusters,
        tissue_name = tissue_name,
        model = model,
        api_key = api_keys$anthropic,
        top_gene_count = 10
      )
    }, error = function(e) {
      list(error = TRUE, message = e$message)
    })

    model_duration <- as.numeric(difftime(Sys.time(), model_start, units = "secs"))

    if (is.list(result) && !is.null(result$error)) {
      cat(sprintf("  ❌ ERROR: %s\n", result$message))
      api_errors[[model]] <- result$message
    } else {
      cat(sprintf("  ✓ Success (%.2f seconds)\n", model_duration))
      results[[model]] <- result
    }

    # Small delay between requests
    Sys.sleep(0.5)
  }

  total_duration <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

  cat("\n", strrep("-", 80), "\n")
  cat(sprintf("Total Duration: %.2f seconds\n", total_duration))
  cat(sprintf("Success Rate: %d/%d (%.0f%%)\n",
              length(results), length(models),
              100 * length(results) / length(models)))
  cat(sprintf("Average per model: %.2f seconds\n", total_duration / length(models)))

  if (length(api_errors) > 0) {
    cat("\nErrors encountered:\n")
    for (model in names(api_errors)) {
      cat(sprintf("  - %s: %s\n", model, api_errors[[model]]))
    }
  }

  return(list(
    duration = total_duration,
    results = results,
    errors = api_errors,
    success_rate = length(results) / length(models)
  ))
}

# =============================================================================
# Test 2: Parallel Execution (No Rate Limiting)
# =============================================================================

test_parallel_no_limit <- function() {
  cat("\n", strrep("=", 80), "\n")
  cat("TEST 2: PARALLEL EXECUTION (NO RATE LIMITING)\n")
  cat(strrep("=", 80), "\n\n")

  if (!requireNamespace("future", quietly = TRUE)) {
    cat("⚠️  'future' package not installed. Skipping test.\n")
    cat("Install with: install.packages('future')\n")
    return(NULL)
  }

  library(future)
  library(future.apply)

  # Setup parallel backend
  plan(multisession, workers = length(models))

  cat(sprintf("Parallel workers: %d\n", length(models)))
  cat("Launching all API calls simultaneously...\n\n")

  start_time <- Sys.time()

  results_list <- future_lapply(seq_along(models), function(i) {
    model <- models[i]

    model_start <- Sys.time()

    result <- tryCatch({
      annotate_cell_types(
        input = test_clusters,
        tissue_name = tissue_name,
        model = model,
        api_key = api_keys$anthropic,
        top_gene_count = 10
      )
    }, error = function(e) {
      list(error = TRUE, message = e$message, model = model)
    })

    model_duration <- as.numeric(difftime(Sys.time(), model_start, units = "secs"))

    return(list(
      model = model,
      result = result,
      duration = model_duration,
      error = !is.null(result$error) && result$error
    ))
  }, future.seed = TRUE)

  total_duration <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

  # Reset to sequential
  plan(sequential)

  # Analyze results
  results <- list()
  api_errors <- list()

  for (item in results_list) {
    cat(sprintf("[%s] ", item$model))
    if (item$error) {
      cat(sprintf("❌ ERROR: %s\n", item$result$message))
      api_errors[[item$model]] <- item$result$message
    } else {
      cat(sprintf("✓ Success (%.2f seconds)\n", item$duration))
      results[[item$model]] <- item$result
    }
  }

  cat("\n", strrep("-", 80), "\n")
  cat(sprintf("Total Duration: %.2f seconds\n", total_duration))
  cat(sprintf("Success Rate: %d/%d (%.0f%%)\n",
              length(results), length(models),
              100 * length(results) / length(models)))

  if (length(results) > 0) {
    avg_model_duration <- mean(sapply(results_list[!sapply(results_list, function(x) x$error)],
                                       function(x) x$duration))
    cat(sprintf("Average per model: %.2f seconds\n", avg_model_duration))
    cat(sprintf("Speedup vs Serial: %.2fx\n",
                (length(models) * avg_model_duration) / total_duration))
  }

  if (length(api_errors) > 0) {
    cat("\n⚠️  ERRORS ENCOUNTERED:\n")
    for (model in names(api_errors)) {
      cat(sprintf("  - %s: %s\n", model, api_errors[[model]]))
    }
  }

  return(list(
    duration = total_duration,
    results = results,
    errors = api_errors,
    success_rate = length(results) / length(models)
  ))
}

# =============================================================================
# Test 3: Parallel with Rate Limiting (Throttled)
# =============================================================================

test_parallel_with_throttle <- function(delay_seconds = 1.0) {
  cat("\n", strrep("=", 80), "\n")
  cat(sprintf("TEST 3: PARALLEL WITH RATE LIMITING (%.1f second delay)\n", delay_seconds))
  cat(strrep("=", 80), "\n\n")

  if (!requireNamespace("future", quietly = TRUE)) {
    cat("⚠️  'future' package not installed. Skipping test.\n")
    return(NULL)
  }

  library(future)
  library(future.apply)

  # Setup parallel backend with limited workers
  max_workers <- min(2, length(models))  # Limit concurrent requests
  plan(multisession, workers = max_workers)

  cat(sprintf("Parallel workers: %d (throttled)\n", max_workers))
  cat(sprintf("Delay between batches: %.1f seconds\n\n", delay_seconds))

  start_time <- Sys.time()

  results_list <- future_lapply(seq_along(models), function(i) {
    model <- models[i]

    # Stagger requests
    Sys.sleep((i - 1) * delay_seconds / max_workers)

    model_start <- Sys.time()

    result <- tryCatch({
      annotate_cell_types(
        input = test_clusters,
        tissue_name = tissue_name,
        model = model,
        api_key = api_keys$anthropic,
        top_gene_count = 10
      )
    }, error = function(e) {
      list(error = TRUE, message = e$message, model = model)
    })

    model_duration <- as.numeric(difftime(Sys.time(), model_start, units = "secs"))

    return(list(
      model = model,
      result = result,
      duration = model_duration,
      error = !is.null(result$error) && result$error
    ))
  }, future.seed = TRUE)

  total_duration <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

  # Reset to sequential
  plan(sequential)

  # Analyze results
  results <- list()
  api_errors <- list()

  for (item in results_list) {
    cat(sprintf("[%s] ", item$model))
    if (item$error) {
      cat(sprintf("❌ ERROR: %s\n", item$result$message))
      api_errors[[item$model]] <- item$result$message
    } else {
      cat(sprintf("✓ Success (%.2f seconds)\n", item$duration))
      results[[item$model]] <- item$result
    }
  }

  cat("\n", strrep("-", 80), "\n")
  cat(sprintf("Total Duration: %.2f seconds\n", total_duration))
  cat(sprintf("Success Rate: %d/%d (%.0f%%)\n",
              length(results), length(models),
              100 * length(results) / length(models)))

  if (length(results) > 0) {
    avg_model_duration <- mean(sapply(results_list[!sapply(results_list, function(x) x$error)],
                                       function(x) x$duration))
    cat(sprintf("Average per model: %.2f seconds\n", avg_model_duration))
  }

  if (length(api_errors) > 0) {
    cat("\n⚠️  ERRORS ENCOUNTERED:\n")
    for (model in names(api_errors)) {
      cat(sprintf("  - %s: %s\n", model, api_errors[[model]]))
    }
  }

  return(list(
    duration = total_duration,
    results = results,
    errors = api_errors,
    success_rate = length(results) / length(models),
    throttle_delay = delay_seconds
  ))
}

# =============================================================================
# Test 4: Adaptive Rate Limiting
# =============================================================================

test_adaptive_rate_limiting <- function() {
  cat("\n", strrep("=", 80), "\n")
  cat("TEST 4: ADAPTIVE RATE LIMITING (WITH BACKOFF)\n")
  cat(strrep("=", 80), "\n\n")

  start_time <- Sys.time()
  results <- list()
  api_errors <- list()

  # Exponential backoff parameters
  base_delay <- 0.5
  max_retries <- 3

  for (i in seq_along(models)) {
    model <- models[i]
    cat(sprintf("[%d/%d] Calling %s...\n", i, length(models), model))

    retry_count <- 0
    success <- FALSE

    while (!success && retry_count < max_retries) {
      if (retry_count > 0) {
        wait_time <- base_delay * (2 ^ (retry_count - 1))
        cat(sprintf("  ⏳ Retry %d after %.1f seconds...\n", retry_count, wait_time))
        Sys.sleep(wait_time)
      }

      model_start <- Sys.time()

      result <- tryCatch({
        annotate_cell_types(
          input = test_clusters,
          tissue_name = tissue_name,
          model = model,
          api_key = api_keys$anthropic,
          top_gene_count = 10
        )
      }, error = function(e) {
        list(error = TRUE, message = e$message)
      })

      model_duration <- as.numeric(difftime(Sys.time(), model_start, units = "secs"))

      if (is.list(result) && !is.null(result$error)) {
        # Check if it's a rate limit error
        if (grepl("rate|limit|429", result$message, ignore.case = TRUE)) {
          cat(sprintf("  ⚠️  Rate limit hit (%.2f seconds)\n", model_duration))
          retry_count <- retry_count + 1
        } else {
          cat(sprintf("  ❌ ERROR: %s\n", result$message))
          api_errors[[model]] <- result$message
          break
        }
      } else {
        cat(sprintf("  ✓ Success (%.2f seconds, %d retries)\n",
                   model_duration, retry_count))
        results[[model]] <- result
        success <- TRUE
      }
    }

    if (!success && retry_count >= max_retries) {
      cat(sprintf("  ❌ Failed after %d retries\n", max_retries))
      api_errors[[model]] <- "Max retries exceeded"
    }

    # Base delay between models
    if (i < length(models)) {
      Sys.sleep(base_delay)
    }
  }

  total_duration <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

  cat("\n", strrep("-", 80), "\n")
  cat(sprintf("Total Duration: %.2f seconds\n", total_duration))
  cat(sprintf("Success Rate: %d/%d (%.0f%%)\n",
              length(results), length(models),
              100 * length(results) / length(models)))

  if (length(api_errors) > 0) {
    cat("\nErrors encountered:\n")
    for (model in names(api_errors)) {
      cat(sprintf("  - %s: %s\n", model, api_errors[[model]]))
    }
  }

  return(list(
    duration = total_duration,
    results = results,
    errors = api_errors,
    success_rate = length(results) / length(models)
  ))
}

# =============================================================================
# Run All Tests
# =============================================================================

main <- function() {
  cat("\n")
  cat("╔════════════════════════════════════════════════════════════════════════════╗\n")
  cat("║                    API RATE LIMIT PERFORMANCE TESTING                     ║\n")
  cat("╚════════════════════════════════════════════════════════════════════════════╝\n")

  all_results <- list()

  # Test 1: Serial (Baseline)
  all_results$serial <- test_serial()

  # Test 2: Parallel (No limits)
  all_results$parallel_no_limit <- test_parallel_no_limit()

  # Test 3: Parallel with throttling (1 second delay)
  all_results$parallel_throttle_1s <- test_parallel_with_throttle(1.0)

  # Test 4: Parallel with throttling (0.5 second delay)
  all_results$parallel_throttle_0.5s <- test_parallel_with_throttle(0.5)

  # Test 5: Adaptive rate limiting
  all_results$adaptive <- test_adaptive_rate_limiting()

  # Summary
  cat("\n", strrep("=", 80), "\n")
  cat("SUMMARY OF ALL TESTS\n")
  cat(strrep("=", 80), "\n\n")

  summary_df <- data.frame(
    Test = character(),
    Duration_sec = numeric(),
    Success_Rate = numeric(),
    Speedup = numeric(),
    Recommendation = character(),
    stringsAsFactors = FALSE
  )

  baseline_duration <- all_results$serial$duration

  for (test_name in names(all_results)) {
    result <- all_results[[test_name]]
    if (!is.null(result)) {
      speedup <- baseline_duration / result$duration

      recommendation <- if (result$success_rate < 1.0) {
        "❌ Not Recommended (API errors)"
      } else if (speedup > 2.0) {
        "✅ Highly Recommended"
      } else if (speedup > 1.3) {
        "✓ Recommended"
      } else {
        "⚠️  Marginal benefit"
      }

      summary_df <- rbind(summary_df, data.frame(
        Test = test_name,
        Duration_sec = round(result$duration, 2),
        Success_Rate = sprintf("%.0f%%", result$success_rate * 100),
        Speedup = sprintf("%.2fx", speedup),
        Recommendation = recommendation,
        stringsAsFactors = FALSE
      ))
    }
  }

  print(summary_df, row.names = FALSE)

  cat("\n", strrep("=", 80), "\n")
  cat("RECOMMENDATIONS FOR IMPLEMENTATION\n")
  cat(strrep("=", 80), "\n\n")

  # Find best strategy
  successful_tests <- summary_df[grepl("100%", summary_df$Success_Rate), ]
  if (nrow(successful_tests) > 0) {
    best_test <- successful_tests[which.min(successful_tests$Duration_sec), ]

    cat("✅ RECOMMENDED STRATEGY:\n")
    cat(sprintf("   Strategy: %s\n", best_test$Test))
    cat(sprintf("   Duration: %.2f seconds\n", best_test$Duration_sec))
    cat(sprintf("   Speedup: %s\n", best_test$Speedup))
    cat(sprintf("   Success: %s\n\n", best_test$Success_Rate))

    if (grepl("throttle", best_test$Test)) {
      throttle_delay <- all_results[[best_test$Test]]$throttle_delay
      cat("   Implementation details:\n")
      cat(sprintf("   - Use parallel processing with max 2 workers\n"))
      cat(sprintf("   - Add %.1f second delay between batches\n", throttle_delay))
      cat(sprintf("   - This provides %.2fx speedup without API overload\n",
                 as.numeric(gsub("x", "", best_test$Speedup))))
    }
  } else {
    cat("⚠️  All parallel strategies encountered API errors.\n")
    cat("   RECOMMENDATION: Stick with serial execution\n")
    cat("   Consider:\n")
    cat("   - Increasing delays between requests\n")
    cat("   - Implementing request queuing\n")
    cat("   - Using different API keys for different models\n")
  }

  cat("\n", strrep("=", 80), "\n\n")

  return(all_results)
}

# Run if sourced directly
if (!interactive()) {
  results <- main()
}

#' Gemini API Processor
#' 
#' Concrete implementation of BaseAPIProcessor for Gemini models.
#' Handles Gemini-specific API calls, authentication, and response parsing.
#'
#' @export
GeminiProcessor <- R6::R6Class("GeminiProcessor",
  inherit = BaseAPIProcessor,
  
  public = list(
    #' @description
    #' Initialize Gemini processor
    #' @param base_url Optional custom API endpoint
    initialize = function(base_url = NULL) {
      super$initialize("gemini", base_url)
    },

    #' @description
    #' Get default Gemini API URL template
    #
    get_default_api_url = function() {
      return("https://generativelanguage.googleapis.com/v1beta/models")
    },

    #' @description
    #' Get API URL for specific model
    #' @param model Model identifier
    get_api_url_for_model = function(model) {
      return(paste0(self$get_api_url(), "/", model, ":generateContent"))
    },
    
    #' @description
    #' Make API call to Gemini
    #' @param chunk_content Prompt text to send
    #' @param model Model identifier
    #' @param api_key Gemini API key
    make_api_call = function(chunk_content, model, api_key) {
      # Build API URL with model
      api_url <- self$get_api_url_for_model(model)
      
      # Prepare request body
      body <- list(
        contents = list(
          list(
            parts = list(
              list(
                text = chunk_content
              )
            )
          )
        )
      )
      
      self$logger$debug("Sending API request to Gemini",
                       list(model = model, provider = self$provider_name))
      
      # Make the API request
      response <- httr::POST(
        url = api_url,
        httr::add_headers(
          "x-goog-api-key" = api_key,
          "Content-Type" = "application/json"
        ),
        body = body,
        encode = "json",
        httr::timeout(get_api_request_timeout())
      )
      
      private$stop_for_http_error(response, model, "Gemini")
      
      return(response)
    },
    
    #' @description
    #' Extract response content from Gemini API response
    #' @param response HTTP response object
    #' @param model Model identifier
    extract_response_content = function(response, model) {
      self$logger$debug("Parsing Gemini API response",
                       list(provider = self$provider_name, model = model))
      
      # Parse the response
      content <- httr::content(response, "parsed")
      
      # Check if response has the expected structure
      if (is.null(content) || is.null(content$candidates) || length(content$candidates) == 0 ||
          is.null(content$candidates[[1]]$content) || is.null(content$candidates[[1]]$content$parts) ||
          length(content$candidates[[1]]$content$parts) == 0 || is.null(content$candidates[[1]]$content$parts[[1]]$text)) {
        
        self$logger$error("Unexpected response format from Gemini API",
                         list(provider = self$provider_name,
                              model = model,
                              content_structure = names(content),
                              candidates_available = !is.null(content$candidates),
                              candidates_count = if(!is.null(content$candidates)) length(content$candidates) else 0))
        
        stop("Unexpected response format from Gemini API")
      }
      
      # Extract the response content
      response_content <- content$candidates[[1]]$content$parts[[1]]$text
      
      return(response_content)
    },

    #' @description
    #' Extract normalized Gemini token usage
    #' @param response HTTP response object
    extract_usage = function(response) {
      private$extract_usage_fields(
        response,
        usage_field = "usageMetadata",
        prompt_field = "promptTokenCount",
        completion_field = "candidatesTokenCount",
        total_field = "totalTokenCount",
        cost_field = NULL
      )
    }
  )
)

# Prompt templates for mLLMCelltype
#
# This file contains all prompt template functions used in mLLMCelltype.
# These functions create various prompts for different stages of the cell type annotation process.

.normalize_top_gene_count <- function(top_gene_count) {
  .normalize_positive_integer(top_gene_count, "top_gene_count")
}

.normalize_marker_gene_vector <- function(genes) {
  if (is.list(genes) || is.data.frame(genes)) {
    stop("marker genes must be an atomic vector")
  }
  normalized <- trimws(as.character(genes))
  unique(normalized[!is.na(normalized) & nzchar(normalized)])
}

extract_cluster_item_genes <- function(item) {
  if (is.list(item) && "genes" %in% names(item)) {
    return(.normalize_marker_gene_vector(item$genes))
  }
  if (is.atomic(item)) {
    return(.normalize_marker_gene_vector(item))
  }
  stop("When input is a list, each element must be a character vector of genes (or another atomic gene vector) or a list containing a 'genes' field")
}

select_cluster_marker_genes <- function(input, cluster_id, top_gene_count) {
  cluster_key <- trimws(as.character(cluster_id))
  top_gene_count <- .normalize_top_gene_count(top_gene_count)

  if (is.list(input) && !is.data.frame(input)) {
    cluster_names <- names(input)
    if (is.null(cluster_names)) {
      cluster_names <- as.character(seq_along(input) - 1)
    } else {
      cluster_names <- trimws(cluster_names)
    }
    if (anyNA(cluster_names) || any(!nzchar(cluster_names))) {
      stop("Cluster IDs must be non-empty strings")
    }
    if (anyDuplicated(cluster_names)) {
      stop("Duplicate cluster IDs detected after normalization")
    }

    cluster_index <- match(cluster_key, cluster_names)
    if (is.na(cluster_index)) {
      stop(sprintf("Cluster '%s' was not found in list input", cluster_key))
    }
    genes <- head(extract_cluster_item_genes(input[[cluster_index]]), top_gene_count)
  } else if (is.data.frame(input)) {
    required_columns <- c("cluster", "gene")
    if (!all(required_columns %in% names(input))) {
      stop("Data frame input must contain 'cluster' and 'gene' columns")
    }

    normalized_clusters <- trimws(as.character(input$cluster))
    cluster_rows <- input[
      !is.na(normalized_clusters) & normalized_clusters == cluster_key,
      ,
      drop = FALSE
    ]
    if ("avg_log2FC" %in% names(cluster_rows)) {
      rank_order <- order(cluster_rows$avg_log2FC, decreasing = TRUE, na.last = NA)
      cluster_rows <- cluster_rows[rank_order, , drop = FALSE]
    }
    genes <- head(.normalize_marker_gene_vector(cluster_rows$gene), top_gene_count)
  } else {
    stop("Input must be either a data.frame or a list of gene lists")
  }

  if (length(genes) == 0) {
    stop(sprintf("No genes found for cluster '%s'", cluster_key))
  }
  genes
}

#' Normalize list input into a canonical cluster->genes mapping
#'
#' For list input, each element can be either:
#' 1) a list containing a `genes` field, or
#' 2) a character vector of genes.
#'
#' Naming rules:
#' - unnamed lists are assigned 0-based IDs ("0", "1", ...)
#' - numeric names are preserved as-is (e.g., "1", "2", "3" stays unchanged)
#' - non-numeric names are preserved as-is
#'
#' @param input List input for cluster annotation
#' @return Named list of character vectors (cluster_id -> genes)
#' @keywords internal
normalize_cluster_gene_list <- function(input) {
  if (!is.list(input) || is.data.frame(input)) {
    stop("normalize_cluster_gene_list expects a list input")
  }
  if (length(input) == 0) {
    stop("input must contain at least one cluster")
  }

  names_vec <- names(input)
  gene_vectors <- lapply(input, extract_cluster_item_genes)

  if (is.null(names_vec)) {
    canonical_names <- as.character(seq_along(gene_vectors) - 1)
  } else {
    canonical_names <- trimws(names_vec)
    if (any(is.na(canonical_names)) || any(!nzchar(canonical_names))) {
      stop("Cluster IDs must be non-empty strings")
    }
  }

  if (anyDuplicated(canonical_names)) {
    stop("Duplicate cluster IDs detected after normalization")
  }

  names(gene_vectors) <- canonical_names
  empty_clusters <- canonical_names[lengths(gene_vectors) == 0]
  if (length(empty_clusters) > 0) {
    stop(
      "No genes found for cluster(s): ",
      paste(empty_clusters, collapse = ", ")
    )
  }
  gene_vectors
}

canonical_cluster_ids <- function(input) {
  cluster_ids <- if (is.list(input) && !is.data.frame(input)) {
    input_names <- names(input)
    if (is.null(input_names)) {
      as.character(seq_along(input) - 1)
    } else {
      trimws(input_names)
    }
  } else if (is.data.frame(input) && "cluster" %in% names(input)) {
    normalized <- trimws(as.character(input$cluster))
    unique(normalized[!is.na(normalized) & nzchar(normalized)])
  } else {
    stop("Input must provide cluster IDs through list names or a cluster column")
  }

  if (anyNA(cluster_ids) || any(!nzchar(cluster_ids))) {
    stop("Cluster IDs must be non-empty strings")
  }
  if (anyDuplicated(cluster_ids)) {
    stop("Duplicate cluster IDs detected after normalization")
  }
  if (length(cluster_ids) > 0 &&
      all(!is.na(suppressWarnings(as.numeric(cluster_ids))))) {
    cluster_ids <- cluster_ids[order(as.numeric(cluster_ids))]
  }
  cluster_ids
}

#' Extract marker genes for a discussion prompt
#'
#' @param input Marker gene input in list or data.frame format
#' @param cluster_id Cluster ID to extract
#' @param top_gene_count Maximum number of genes to include
#' @return Comma-separated marker genes for the requested cluster
#' @keywords internal
#' @noRd
extract_cluster_genes_for_discussion <- function(input, cluster_id, top_gene_count) {
  paste(
    select_cluster_marker_genes(input, cluster_id, top_gene_count),
    collapse = ","
  )
}

#' Create prompt for cell type annotation
#'
#' @param input Either a data frame from Seurat's FindAllMarkers() or a list for each cluster
#'   where each element is either a character vector of genes or a list containing a `genes` field
#'   Cluster IDs in named inputs are preserved as-is; unnamed list input receives sequential IDs starting at "0".
#' @param tissue_name Tissue context for the annotation (e.g., 'human PBMC', 'mouse brain')
#' @param top_gene_count Number of top genes to use per cluster when input is from Seurat. Default: 10
#'
#' @return A list with `prompt` (formatted prompt text), `expected_count`
#'   (number of clusters), and `gene_lists` (cluster ID to marker genes mapping).
#' @export
create_annotation_prompt <- function(input, tissue_name, top_gene_count = 10) {
  tissue_name <- .normalize_required_string(tissue_name, "tissue_name")
  top_gene_count <- .normalize_top_gene_count(top_gene_count)

  if (is.list(input) && !is.data.frame(input)) {
    normalized_input <- normalize_cluster_gene_list(input)

    gene_lists <- list()

    for (cluster_id in names(normalized_input)) {
      genes <- normalized_input[[cluster_id]]
      gene_lists[[cluster_id]] <- paste(genes, collapse = ", ")
    }
    
    expected_count <- length(normalized_input)
  } else if (is.data.frame(input)) {
    required_columns <- c("cluster", "gene", "avg_log2FC")
    column_counts <- vapply(required_columns, function(column) {
      sum(names(input) == column, na.rm = TRUE)
    }, integer(1))
    if (any(column_counts != 1)) {
      stop(
        "Data frame input must contain exactly one each of: ",
        paste(required_columns, collapse = ", ")
      )
    }
    if (!is.numeric(input$avg_log2FC)) {
      stop("avg_log2FC must be numeric")
    }

    normalized_clusters <- trimws(as.character(input$cluster))
    valid_cluster_rows <- !is.na(normalized_clusters) & nzchar(normalized_clusters)
    if (any(!valid_cluster_rows)) {
      warning("Skipping rows with missing or empty cluster IDs", call. = FALSE)
    }
    input <- input[valid_cluster_rows, , drop = FALSE]
    normalized_clusters <- normalized_clusters[valid_cluster_rows]
    if (nrow(input) == 0) {
      stop("input must contain at least one valid cluster")
    }

    cluster_names <- unique(normalized_clusters)
    gene_lists <- list()
    for (cluster_id in cluster_names) {
      genes <- select_cluster_marker_genes(input, cluster_id, top_gene_count)
      gene_lists[[cluster_id]] <- paste(genes, collapse = ",")
    }

    expected_count <- length(gene_lists)
  } else {
    stop("Input must be either a data.frame (from Seurat) or a list of gene lists")
  }
  
  cluster_names <- canonical_cluster_ids(gene_lists)
  gene_lists <- gene_lists[cluster_names]

  # Create formatted lines from gene_lists using the canonical cluster order.
  formatted_lines <- vapply(cluster_names, function(name) {
    paste0(name, ": ", gene_lists[[name]])
  }, character(1), USE.NAMES = FALSE)

  prompt <- paste0("You are a cell type annotation expert. Below are marker genes for different cell clusters in ", 
                  tissue_name, ".\n\n",
                  paste(formatted_lines, collapse = "\n"),
                  "\n\nReturn exactly one cell type name per line, in the same order as the clusters shown above, without cluster IDs or explanation.")
  
  return(list(
    prompt = prompt,
    expected_count = expected_count,
    gene_lists = gene_lists
  ))
}

#' Create reasoning-aware prompt for cell type annotation
#'
#' @param input Either a data frame from Seurat's FindAllMarkers() or a list for each cluster
#'   where each element is either a character vector of genes or a list containing a `genes` field.
#' @param tissue_name Tissue context for the annotation (e.g., 'human PBMC', 'mouse brain')
#' @param top_gene_count Number of top genes to use per cluster when input is from Seurat. Default: 10
#'
#' @return A list with `prompt` (formatted prompt text), `expected_count`
#'   (number of clusters), and `gene_lists` (cluster ID to marker genes mapping).
#' @export
create_reasoning_annotation_prompt <- function(input, tissue_name, top_gene_count = 10) {
  base_prompt <- create_annotation_prompt(input, tissue_name, top_gene_count)

  marker_section <- sub(
    "\n\nReturn exactly one cell type name per line, in the same order as the clusters shown above, without cluster IDs or explanation\\.$",
    "",
    base_prompt$prompt
  )

  reasoning_prompt <- paste0(
    marker_section,
    "\n\nFor each cluster, return a JSON object with the following fields:\n",
    "- \"cluster_id\": the cluster ID as shown above\n",
    "- \"cell_type\": the annotated cell type\n",
    "- \"marker_genes\": the marker genes from the provided gene set that support this annotation, ",
    "comma-separated, preserving the original case exactly as shown above\n",
    "- \"gene_expression\": a concise natural-language description of where each provided ",
    "differential gene is typically expressed across cell types\n\n",
    "Return a single JSON array containing one object per cluster, in the same order as the clusters shown above. ",
    "Do not include markdown formatting or any explanation outside the JSON array."
  )

  return(list(
    prompt = reasoning_prompt,
    expected_count = base_prompt$expected_count,
    gene_lists = base_prompt$gene_lists
  ))
}

#' Create prompt for checking consensus among model predictions
#
#
#
#
#' @keywords internal
create_consensus_check_prompt <- function(round_responses, controversy_threshold = 2/3, entropy_threshold = 1.0) {
  # Format the predictions for Claude
  paste(
    "You are a cell type annotation expert. Below are different models' predictions for the same cell cluster.",
    "Your task is to analyze these predictions and calculate uncertainty metrics.",
    "",
    "PREDICTIONS:",
    paste(paste("Model", seq_along(round_responses), ":", round_responses), collapse = "\n"),
    "",
    "IMPORTANT GUIDELINES:",
    "1. Consider predictions as matching only when they name the same biological entity, allowing:",
    "   - Synonymous nomenclature (e.g., 'NK cells' vs 'Natural Killer cells')",
    "   - Formatting and capitalization differences",
    "2. Preserve biologically meaningful subtype or state qualifiers such as activated, mature, memory, regulatory, CD4+, and CD8+; do not merge labels when doing so would lose specificity",
    "3. Group only predictions that remain biologically equivalent at the same level of granularity",
    "4. If any prediction is 'Unknown' or 'Unclear', treat it as a separate group",
    "",
    "CALCULATE THE FOLLOWING METRICS:",
    "1. Consensus Proportion = Number of models supporting the majority prediction / Total number of models",
    "2. Shannon Entropy = -sum(p_i * log2(p_i)) where p_i is the proportion of models predicting each unique cell type",
    sprintf("3. Determine if consensus is reached (Consensus Proportion >= %s AND Entropy <= %s)",
            format(controversy_threshold, nsmall=1), format(entropy_threshold, nsmall=1)),
    "",
    "RESPONSE FORMAT:",
    "Line 1: 1 if consensus is reached, 0 if not",
    "Line 2: Consensus Proportion (a decimal between 0 and 1)",
    "Line 3: Shannon Entropy (a decimal number)",
    "Line 4: The majority cell type prediction",
    "",
    "Example matches:",
    "- 'NK cells' = 'Natural Killer cells'",
    "- 'CD8+ T cells' = 'Cytotoxic T cells'",
    "- 'B cells' = 'B lymphocytes'",
    "",
    "RESPOND WITH EXACTLY FOUR LINES AS SPECIFIED ABOVE.",
    sep = "\n"
  )
}

format_named_model_responses <- function(responses,
                                         separator = ":\n",
                                         missing = "No prediction",
                                         validator = is_valid_model_response) {
  response_names <- names(responses)
  if (!is.list(responses) || is.null(response_names) ||
      anyNA(response_names) || any(!nzchar(trimws(response_names)))) {
    stop("responses must be a named list with non-empty model names")
  }

  vapply(seq_along(responses), function(index) {
    response <- responses[[index]]
    response_text <- if (isTRUE(validator(response))) {
      paste(response, collapse = "\n")
    } else {
      missing
    }
    paste0(trimws(response_names[[index]]), separator, response_text)
  }, character(1), USE.NAMES = FALSE)
}

#' Create prompt for additional discussion rounds
#
#
#
#
#
#
#' @keywords internal
create_discussion_prompt <- function(cluster_id,
                                     cluster_genes,
                                     tissue_name,
                                     previous_rounds,
                                     round_number) {
  # Compile previous discussion history
  discussion_history <- vapply(previous_rounds, function(round) {
    responses <- round$responses
    sprintf(
      "Round %d:\n%s",
      round$round_number,
      paste(format_named_model_responses(responses), collapse = "\n\n")
    )
  }, character(1), USE.NAMES = FALSE)
  
  sprintf(
    "We are continuing the discussion for cluster %s (marker genes: %s%s).
    
    Previous discussion:
    %s
    
    This is round %d of the discussion. 
    
    Using the Toulmin argumentation model, please structure your response as follows:
    
    1. CLAIM: State your clear cell type prediction (e.g., 'This is a T cell')
    2. GROUNDS/DATA: Present specific observable evidence that supports your claim (e.g., marker genes expressed in this cluster)
    3. WARRANT: Explain the logical reasoning that connects your grounds to your claim (e.g., why these genes imply this cell type)
    4. BACKING: Provide additional support for your warrant (e.g., citations, references, established knowledge in the field)
    5. QUALIFIER: Indicate any conditions or limitations affecting the certainty of your claim (e.g., 'probably', 'likely', 'almost certainly')
    6. REBUTTAL: Acknowledge possible counter-arguments or exceptions (including addressing other models' predictions)
    
    Based on previous discussion, also indicate:
    - Whether you agree or disagree with any emerging consensus
    - If you've revised your previous position, explain why
    
    Format your response as:
    CELL TYPE: [your current prediction]
    GROUNDS: [specific marker genes and expression evidence supporting your claim]
    WARRANT: [logical connection between your evidence and claim]
    BACKING: [additional support for your reasoning]
    QUALIFIER: [degree of certainty - definite, probable, possible, etc.]
    REBUTTAL: [addressing counter-arguments or alternative interpretations]
    CONSENSUS STATUS: [Agree/Disagree with emerging consensus]",
    cluster_id,
    cluster_genes,
    if (!is.null(tissue_name)) sprintf(" from %s", tissue_name) else "",
    paste(discussion_history, collapse = "\n\n"),
    round_number
  )
}

#' Create prompt for the initial round of discussion
#
#
#
#
#
#' @keywords internal
create_initial_discussion_prompt <- function(cluster_id,
                                             cluster_genes,
                                             tissue_name,
                                             initial_predictions) {
  sprintf(
    "We are analyzing cluster %s with the following marker genes: %s%s
    Different models have made different predictions:
    %s
    
    Please provide your cell type prediction using the Toulmin argumentation model to structure your response:
    
    1. CLAIM: State your clear cell type prediction (e.g., 'This is a T cell')
    2. GROUNDS/DATA: Present specific observable evidence that supports your claim (e.g., marker genes expressed in this cluster)
    3. WARRANT: Explain the logical reasoning that connects your grounds to your claim (e.g., why these genes imply this cell type)
    4. BACKING: Provide additional support for your warrant (e.g., citations, references, established knowledge in the field)
    5. QUALIFIER: Indicate any conditions or limitations affecting the certainty of your claim (e.g., 'probably', 'likely', 'almost certainly')
    6. REBUTTAL: Acknowledge possible counter-arguments or exceptions (including addressing other models' predictions)
    
    Format your response as:
    CELL TYPE: [your predicted cell type]
    GROUNDS: [specific marker genes and expression evidence supporting your claim]
    WARRANT: [logical connection between your evidence and claim]
    BACKING: [additional support for your reasoning]
    QUALIFIER: [degree of certainty - definite, probable, possible, etc.]
    REBUTTAL: [addressing counter-arguments or alternative interpretations]",
    cluster_id,
    cluster_genes,
    if (!is.null(tissue_name)) sprintf(" from %s", tissue_name) else "",
    paste(
      format_named_model_responses(
        initial_predictions,
        separator = ": ",
        missing = "No prediction",
        validator = is_real_cell_type_annotation
      ),
      collapse = "\n"
    )
  )
}

#' Create prompt for standardizing cell type names
#
#
#' @keywords internal
create_standardization_prompt <- function(all_cell_types) {
  paste0(
    "I need to standardize the following cell type names to ensure consistent nomenclature while preserving biological subtypes and specificity. \n\n",
    "IMPORTANT GUIDELINES:\n",
    "1. Preserve cell subtype information: Do NOT collapse specific subtypes into general categories (e.g., DO NOT convert 'Memory B cell' to just 'B cell').\n",
    "2. Standardize expression variations: Only standardize different ways of expressing the same biological entity (e.g., 'CD4+ T cell' and 'Helper T cell' can be standardized to 'CD4+ T cell').\n",
    "3. Maintain granularity: If a cell type has specific markers or functional designations, preserve that information.\n",
    "4. Use widely accepted nomenclature: When standardizing, use the most scientifically accepted term.\n",
    "5. Be consistent with surface markers: Use consistent formatting for surface markers (e.g., CD4+, CD8+).\n\n",
    "Examples of CORRECT standardization:\n",
    "- 'Helper T cell' -> 'CD4+ T cell' (same biological entity, standard nomenclature)\n",
    "- 'CD14+ monocyte' -> 'Classical monocyte' (if they are biologically equivalent)\n",
    "- 'B-lymphocyte' -> 'B cell' (expression variation)\n\n",
    "Examples of INCORRECT standardization:\n",
    "- 'Memory B cell' -> 'B cell' (loses subtype information)\n",
    "- 'Regulatory T cell' -> 'CD4+ T cell' (loses functional subtype)\n",
    "- 'Classical monocyte' -> 'Monocyte' (loses subtype information)\n\n",
    "Please provide a standardized name for each cell type in the exact format: 'ORIGINAL: STANDARDIZED'.\n",
    "Do not add any additional text, explanations, or formatting.\n\n",
    "Here are the cell types to standardize:\n\n",
    paste(all_cell_types, collapse = "\n")
  )
}

#' URL Utilities for Base URL Resolution
#'
#' This file contains utility functions for resolving custom base URLs
#' for different API providers.

#' Validate base URL
#'
#' @param url URL string to validate
#' @return TRUE if the URL is a safe HTTP(S) endpoint, otherwise FALSE
#' @keywords internal
#' @noRd
validate_base_url <- function(url) {
  if (!is.character(url) || length(url) != 1 || is.na(url)) {
    return(FALSE)
  }

  normalized <- trimws(url)
  if (!nzchar(normalized) || grepl("\\s", normalized)) {
    return(FALSE)
  }

  parsed <- tryCatch(
    httr::parse_url(normalized),
    error = function(e) NULL
  )
  if (is.null(parsed)) {
    return(FALSE)
  }

  if (is.null(parsed$scheme) || !parsed$scheme %in% c("http", "https")) {
    return(FALSE)
  }
  if (is.null(parsed$hostname) || !nzchar(parsed$hostname)) {
    return(FALSE)
  }
  if (!is.null(parsed$username) || !is.null(parsed$password)) {
    return(FALSE)
  }
  if (!is.null(parsed$query) && length(parsed$query) > 0) {
    return(FALSE)
  }
  if (!is.null(parsed$fragment) && nzchar(parsed$fragment)) {
    return(FALSE)
  }

  authority <- sub("^[A-Za-z][A-Za-z0-9+.-]*://", "", normalized)
  authority <- sub("[/?#].*$", "", authority)
  if (endsWith(authority, ":")) {
    return(FALSE)
  }
  if (!is.null(parsed$port)) {
    if (length(parsed$port) != 1 || !grepl("^[0-9]+$", parsed$port)) {
      return(FALSE)
    }
    port <- suppressWarnings(as.numeric(parsed$port))
    if (is.na(port) || port < 1 || port > 65535) {
      return(FALSE)
    }
  }

  TRUE
}

#' Resolve provider-specific base URL
#'
#' This is the single entry point for all base URL resolution. It resolves
#' the appropriate URL and normalizes it (strips trailing slashes).
#'
#' @param provider Provider name (e.g., "openai", "anthropic")
#' @param base_urls User-provided base URLs: NULL, a single string, or a named list
#' @return Resolved and normalized base URL, or NULL if not specified
#' @keywords internal
resolve_provider_base_url <- function(provider, base_urls) {
  normalize_value <- function(value, source) {
    if (is.null(value)) {
      return(NULL)
    }
    if (!is.character(value) || length(value) != 1 || is.na(value)) {
      stop(sprintf("base_urls%s must be a string URL", source))
    }

    normalized <- sub("/+$", "", trimws(value))
    if (!nzchar(normalized)) {
      return(NULL)
    }
    if (!validate_base_url(normalized)) {
      stop(sprintf("Invalid base URL in base_urls%s: %s", source, value))
    }
    normalized
  }

  if (is.null(base_urls) || is.null(provider)) {
    return(NULL)
  }

  if (!is.character(provider) || length(provider) != 1 || is.na(provider)) {
    return(NULL)
  }

  provider_normalized <- tolower(trimws(provider))
  if (!nzchar(provider_normalized)) {
    return(NULL)
  }

  if (is.character(base_urls) && length(base_urls) == 1) {
    return(normalize_value(base_urls, ""))
  }

  if (is.list(base_urls)) {
    base_url_names <- names(base_urls)
    if (is.null(base_url_names)) {
      stop("base_urls must be a named list")
    }

    normalized_names <- tolower(trimws(base_url_names))
    if (any(is.na(base_url_names)) || any(!nzchar(normalized_names))) {
      stop("base_urls provider names must be non-empty strings")
    }
    duplicate_index <- anyDuplicated(normalized_names)
    if (duplicate_index > 0) {
      duplicate_name <- normalized_names[[duplicate_index]]
      stop("Ambiguous base_urls provider names after case/whitespace normalization: ",
           duplicate_name)
    }
    match_idx <- match(provider_normalized, normalized_names)
    if (is.na(match_idx)) {
      return(NULL)
    }
    return(normalize_value(
      base_urls[[match_idx]],
      sprintf("[['%s']]", base_url_names[[match_idx]])
    ))
  }

  stop("base_urls must be NULL, a single string URL, or a named list")
}

#' Filter out error responses from model round responses
#' @keywords internal
filter_valid_responses <- function(responses, cluster_id, round = NULL) {
  valid <- list()
  for (model_name in names(responses)) {
    response <- responses[[model_name]]
    if (is_valid_model_response(response)) {
      valid[[model_name]] <- response
    } else {
      round_info <- if (!is.null(round)) sprintf(" in round %d", round) else ""
      log_warn(sprintf("Model %s failed to provide valid response for cluster %s%s",
                      model_name, cluster_id, round_info))
    }
  }
  valid
}

#' Facilitate discussion for a controversial cluster
#' @note This function uses create_initial_discussion_prompt and create_discussion_prompt from prompt_templates.R
#' @keywords internal
facilitate_cluster_discussion <- function(cluster_id,
                                          input,
                                          tissue_name,
                                          models,
                                          api_keys,
                                          initial_predictions,
                                          top_gene_count,
                                          max_rounds = 3,
                                          controversy_threshold = 0.7,
                                          entropy_threshold = 1.0,
                                          consensus_check_model = NULL,
                                          base_urls = NULL) {

  # Ensure cluster_id is always a string
  char_cluster_id <- as.character(cluster_id)

  # Get marker genes for this cluster
  cluster_genes <- tryCatch({
    extract_cluster_genes_for_discussion(input, char_cluster_id, top_gene_count)
  }, error = function(e) {
    warning("Error extracting genes for cluster ", char_cluster_id, ": ", e$message)
    log_warn("Error extracting genes", list(cluster_id = char_cluster_id, error = e$message))
    paste("Cluster", char_cluster_id, "- Error extracting genes:", e$message)
  })

  # Extract predictions for this cluster from each model
  # Parse against the full cluster sequence before selecting this cluster.
  all_cluster_ids <- tryCatch(
    canonical_cluster_ids(input),
    error = function(e) char_cluster_id
  )
  structured_predictions <- structure_cluster_predictions(
    initial_predictions,
    all_cluster_ids,
    char_cluster_id
  )

  # Create the discussion log with extracted predictions
  discussion_log <- list(
    cluster_id = char_cluster_id,
    initial_predictions = structured_predictions,
    rounds = list()
  )

  # Initialize clustering discussion log file
  get_logger()$log_discussion(char_cluster_id, "start", list(
    tissue_name = tissue_name,
    marker_genes = cluster_genes
  ))

  # First round: Initial reasoning
  first_round_prompt <- create_initial_discussion_prompt(
    cluster_id = char_cluster_id,
    cluster_genes = cluster_genes,
    tissue_name = tissue_name,
    initial_predictions = structured_predictions
  )

  # First round responses
  round1_responses <- list()
  for (model in models) {
    api_key <- get_api_key(model, api_keys)
    if (is.null(api_key)) {
      provider <- get_provider(model)
      warning(sprintf("No API key found for model '%s' (provider: %s). This model will be skipped.",
                   model, provider))
      log_warn("No API key found, skipping model", list(model = model, provider = provider))
      next
    }

    response <- tryCatch(
      get_model_response(
        prompt = first_round_prompt,
        model = model,
        api_key = api_key,
        base_urls = base_urls
      ),
      error = function(e) {
        log_warn(sprintf("Model %s failed in round 1 for cluster %s: %s",
                        model, char_cluster_id, e$message))
        paste0("Error: ", e$message)
      }
    )

    round1_responses[[model]] <- response

    # Log the model prediction to discussion file
    get_logger()$log_discussion(char_cluster_id, "prediction", list(
      model = model,
      round = 1,
      prediction = response
    ))
  }

  discussion_log$rounds[[1]] <- list(
    round_number = 1,
    responses = round1_responses
  )

  # Filter out error responses before consensus check
  valid_round1_responses <- filter_valid_responses(round1_responses, char_cluster_id)

  # Check if we have enough valid responses to continue
  if (length(valid_round1_responses) < 2) {
    log_warn(sprintf("Only %d valid responses received for cluster %s. Skipping discussion.",
                    length(valid_round1_responses), char_cluster_id))

    best_prediction <- if(length(valid_round1_responses) == 1) {
      extract_discussion_cell_type(valid_round1_responses[[1]])
    } else {
      "Unknown"
    }

    # Set a pseudo-consensus result so the caller gets a consistent structure
    discussion_log$rounds[[1]]$consensus_result <- list(
      reached = FALSE,
      consensus_proportion = 0,
      entropy = 0,
      majority_prediction = best_prediction
    )
    get_logger()$log_discussion(char_cluster_id, "consensus", discussion_log$rounds[[1]]$consensus_result)
    get_logger()$log_discussion(char_cluster_id, "end", list(
      final_result = best_prediction,
      rounds_completed = 1,
      consensus_reached = FALSE
    ))
    return(discussion_log)
  }

  # Check consensus after first round with valid responses only
  consensus_result <- check_consensus(valid_round1_responses, api_keys, controversy_threshold, entropy_threshold, consensus_check_model, base_urls)
  log_info("Consensus check completed", list(
    round = 1,
    consensus_reached = consensus_result$reached,
    consensus_proportion = consensus_result$consensus_proportion,
    entropy = consensus_result$entropy
  ))

  # Store consensus result in discussion log
  discussion_log$rounds[[1]]$consensus_result <- consensus_result
  
  # Log consensus result
  get_logger()$log_discussion(char_cluster_id, "consensus", consensus_result)

  if (consensus_result$reached && consensus_result$consensus_proportion >= controversy_threshold && consensus_result$entropy <= entropy_threshold) {
    consensus_reached <- TRUE
    # Log discussion end  
    get_logger()$log_discussion(char_cluster_id, "end", list(
      final_result = consensus_result$majority_prediction,
      rounds_completed = 1,
      consensus_reached = TRUE
    ))
    message(sprintf("Consensus reached in round 1 with consensus proportion %.2f and entropy %.2f. Stopping discussion.",
                   consensus_result$consensus_proportion, consensus_result$entropy))
    return(discussion_log)
  }

  # Additional rounds of discussion if needed
  round <- 1
  consensus_reached <- FALSE

  while (round < max_rounds && !consensus_reached) {
    round <- round + 1
    message(sprintf("\nStarting round %d of discussion...", round))

    # Create prompt that includes all previous responses
    discussion_prompt <- create_discussion_prompt(
      cluster_id = char_cluster_id,
      cluster_genes = cluster_genes,
      tissue_name = tissue_name,
      previous_rounds = discussion_log$rounds,
      round_number = round
    )

    round_responses <- list()
    for (model in models) {
      api_key <- get_api_key(model, api_keys)
      if (is.null(api_key)) {
        provider <- get_provider(model)
        warning(sprintf("No API key found for model '%s' (provider: %s). This model will be skipped.",
                     model, provider))
        log_warn("No API key found, skipping model", list(model = model, provider = provider, round = round))
        next
      }

      response <- tryCatch(
        get_model_response(
          prompt = discussion_prompt,
          model = model,
          api_key = api_key,
          base_urls = base_urls
        ),
        error = function(e) {
          log_warn(sprintf("Model %s failed in round %d for cluster %s: %s",
                          model, round, char_cluster_id, e$message))
          paste0("Error: ", e$message)
        }
      )

      round_responses[[model]] <- response

      # Log the model prediction to discussion file
      get_logger()$log_discussion(char_cluster_id, "prediction", list(
        model = model,
        round = round,
        prediction = response
      ))
    }

    discussion_log$rounds[[round]] <- list(
      round_number = round,
      responses = round_responses
    )

    # Filter out error responses before consensus check
    valid_round_responses <- filter_valid_responses(round_responses, char_cluster_id, round)
    
    # Check if we have enough valid responses to continue
    if (length(valid_round_responses) < 2) {
      log_warn(sprintf("Only %d valid responses in round %d for cluster %s. Ending discussion.", 
                      length(valid_round_responses), round, char_cluster_id))
      break  # Exit the discussion loop
    }

    # Check if consensus is reached with valid responses only
    consensus_result <- check_consensus(valid_round_responses, api_keys, controversy_threshold, entropy_threshold, consensus_check_model, base_urls)
    log_info("Consensus check completed", list(
      round = round,
      consensus_reached = consensus_result$reached,
      consensus_proportion = consensus_result$consensus_proportion,
      entropy = consensus_result$entropy
    ))

    # Store consensus result in discussion log
    discussion_log$rounds[[round]]$consensus_result <- consensus_result

    # Log consensus result
    get_logger()$log_discussion(char_cluster_id, "consensus", consensus_result)

    # Check if consensus conditions are met
    if (consensus_result$reached && consensus_result$consensus_proportion >= controversy_threshold && consensus_result$entropy <= entropy_threshold) {
      consensus_reached <- TRUE
      message(sprintf("Consensus reached in round %d with consensus proportion %.2f and entropy %.2f. Stopping discussion.",
                     round, consensus_result$consensus_proportion, consensus_result$entropy))
      break
    } else {
      message(sprintf("No strong consensus in round %d (consensus proportion: %.2f, entropy: %.2f). %s",
                     round, consensus_result$consensus_proportion, consensus_result$entropy,
                     if (round < max_rounds) "Continuing discussion..." else "Reached maximum rounds."))
    }
  }

  # Find the last round that has a consensus_result (the last round may have
  # broken early without one if valid responses were insufficient)
  final_result <- "Unknown"
  for (r in rev(seq_along(discussion_log$rounds))) {
    cr <- discussion_log$rounds[[r]]$consensus_result
    if (!is.null(cr)) {
      final_result <- cr$majority_prediction
      break
    }
  }
  
  get_logger()$log_discussion(char_cluster_id, "end", list(
    final_result = final_result,
    rounds_completed = length(discussion_log$rounds),
    consensus_reached = consensus_reached
  ))

  discussion_log
}

#' Unified Logger for mLLMCelltype Package
#' 
#' This logger provides centralized, multi-level logging with structured output,
#' log rotation, and performance monitoring capabilities.
#'
#' @export
UnifiedLogger <- R6::R6Class("UnifiedLogger",
  public = list(
    #' @field log_dir Directory for storing log files
    log_dir = NULL,
    
    #' @field log_level Current logging level
    log_level = NULL,
    
    #' @field session_id Unique identifier for the current session
    session_id = NULL,
    
    #' @field max_log_size Maximum log file size in MB (default: 10MB)
    max_log_size = 10,
    
    #' @field max_log_files Maximum number of log files to keep (default: 5)
    max_log_files = 5,
    
    #' @field enable_console Whether to output to console (default: FALSE)
    enable_console = FALSE,
    
    #' @field enable_json Whether to use JSON format (default: TRUE)
    enable_json = TRUE,
    
    #' @field performance_stats Performance monitoring statistics
    performance_stats = NULL,
    
    #' @description
    #' Initialize the unified logger
    #' @param base_dir Directory for log files
    #' @param level Minimum log level
    #' @param max_size Maximum main log size in megabytes
    #' @param max_files Maximum number of retained main logs
    #' @param console_output Whether to mirror logs to the console
    #' @param json_format Whether main logs use JSON format
    initialize = function(base_dir = "logs", level = "INFO", max_size = 10,
                         max_files = 5, console_output = FALSE, json_format = TRUE) {
      self$log_dir <- .normalize_required_string(base_dir, "base_dir")
      self$log_level <- .normalize_log_level(level)
      self$max_log_size <- .normalize_positive_number(max_size, "max_size")
      self$max_log_files <- .normalize_positive_integer(max_files, "max_files")
      self$enable_console <- .normalize_flag(console_output, "console_output")
      self$enable_json <- .normalize_flag(json_format, "json_format")
      
      # Only create log directory if we're not in R CMD check environment
      # This prevents the logs directory from being created during package checks
      if (!nzchar(Sys.getenv("_R_CHECK_PACKAGE_NAME_", "")) && !dir.exists(self$log_dir)) {
        created <- dir.create(self$log_dir, recursive = TRUE, showWarnings = FALSE)
        if (!created && !dir.exists(self$log_dir)) {
          stop("Failed to create log directory: ", self$log_dir)
        }
      }
      
      # Generate session ID with sub-second + random suffix to avoid collisions
      self$session_id <- private$create_session_id()
      
      # Initialize performance stats
      self$performance_stats <- list(
        session_start = Sys.time(),
        api_calls = 0,
        cache_hits = 0,
        cache_misses = 0,
        errors = 0,
        warnings = 0
      )
      
      # Log initialization
      self$info("Unified logger initialized", list(
        session_id = self$session_id,
        log_level = self$log_level,
        log_dir = self$log_dir
      ))
    },
    
    #' @description
    #' Log a debug message
    #' @param message Log message
    #' @param context Optional structured context
    debug = function(message, context = NULL) {
      tryCatch(
        private$log_message("DEBUG", message, context),
        warning = function(w) invisible(NULL),
        error = function(e) invisible(NULL)
      )
    },
    
    #' @description
    #' Log an info message
    #' @param message Log message
    #' @param context Optional structured context
    info = function(message, context = NULL) {
      tryCatch(
        private$log_message("INFO", message, context),
        warning = function(w) invisible(NULL),
        error = function(e) invisible(NULL)
      )
    },
    
    #' @description
    #' Log a warning message
    #' @param message Log message
    #' @param context Optional structured context
    warn = function(message, context = NULL) {
      self$performance_stats$warnings <- self$performance_stats$warnings + 1
      tryCatch(
        private$log_message("WARN", message, context),
        warning = function(w) invisible(NULL),
        error = function(e) invisible(NULL)
      )
    },
    
    #' @description
    #' Log an error message
    #' @param message Log message
    #' @param context Optional structured context
    error = function(message, context = NULL) {
      self$performance_stats$errors <- self$performance_stats$errors + 1
      tryCatch(
        private$log_message("ERROR", message, context),
        warning = function(w) invisible(NULL),
        error = function(e) invisible(NULL)
      )
    },
    
    #' @description
    #' Log API call performance
    #' @param provider Provider identifier
    #' @param model Model identifier
    #' @param duration Request duration in seconds
    #' @param success Whether the request succeeded
    #' @param tokens Optional token usage metadata
    log_api_call = function(provider, model, duration, success = TRUE, tokens = NULL) {
      self$performance_stats$api_calls <- self$performance_stats$api_calls + 1
      
      context <- list(
        provider = provider,
        model = model,
        duration_seconds = round(duration, 3),
        success = success,
        tokens = tokens,
        call_count = self$performance_stats$api_calls
      )
      
      if (success) {
        self$info(sprintf("API call completed: %s/%s (%.3fs)", provider, model, duration), context)
      } else {
        self$error(sprintf("API call failed: %s/%s (%.3fs)", provider, model, duration), context)
      }
    },
    
    #' @description
    #' Log complete API request and response for debugging and audit
    #' @param provider Provider identifier
    #' @param model Model identifier
    #' @param prompt_content Request prompt
    #' @param response_content Provider response or error text
    #' @param request_metadata Optional request metadata
    #' @param response_metadata Optional response metadata
    log_api_request_response = function(provider, model, prompt_content, response_content, 
                                      request_metadata = NULL, response_metadata = NULL) {
      # Skip during R CMD check
      if (nzchar(Sys.getenv("_R_CHECK_PACKAGE_NAME_", ""))) {
        return(invisible(NULL))
      }

      tryCatch({
      
      # Create API log directory if it doesn't exist
      api_log_dir <- file.path(self$log_dir, self$session_id, "api_logs")
      if (!dir.exists(api_log_dir)) {
        dir.create(api_log_dir, recursive = TRUE)
      }
      
      # Generate unique log file name for this API call
      timestamp <- format(Sys.time(), "%Y%m%d_%H%M%S")
      call_seq <- self$performance_stats$api_calls + 1
      provider_safe <- private$safe_filename_component(provider, "provider")
      model_safe <- private$safe_filename_component(model, "model")
      api_call_id <- sprintf(
        "%s_%s_%s_%03d",
        provider_safe,
        model_safe,
        timestamp,
        call_seq
      )
      
      # Create complete API log entry
      api_log_entry <- list(
        api_call_id = api_call_id,
        timestamp = format(Sys.time(), "%Y-%m-%d %H:%M:%S"),
        session_id = self$session_id,
        provider = provider,
        model = model,
        request = list(
          prompt_content = prompt_content,
          metadata = request_metadata
        ),
        response = list(
          content = response_content,
          metadata = response_metadata
        ),
        call_sequence = self$performance_stats$api_calls + 1
      )
      
      # Write detailed API log to separate file
      api_log_file <- file.path(api_log_dir, sprintf("%s.json", api_call_id))
      writeLines(jsonlite::toJSON(api_log_entry, auto_unbox = TRUE, pretty = TRUE), api_log_file)
      
      # Also create human-readable version
      readable_file <- file.path(api_log_dir, sprintf("%s.md", api_call_id))
      readable_content <- sprintf(
        "# API Call: %s\n\n**Provider:** %s\n**Model:** %s\n**Timestamp:** %s\n**Session:** %s\n\n## Request Prompt\n\n```\n%s\n```\n\n## Response\n\n```\n%s\n```\n\n## Metadata\n\n**Request Metadata:**\n```json\n%s\n```\n\n**Response Metadata:**\n```json\n%s\n```\n",
        api_call_id,
        provider,
        model, 
        api_log_entry$timestamp,
        self$session_id,
        prompt_content,
        if(is.character(response_content)) paste(response_content, collapse = "\n") else as.character(response_content),
        if(is.null(request_metadata)) "null" else jsonlite::toJSON(request_metadata, auto_unbox = TRUE, pretty = TRUE),
        if(is.null(response_metadata)) "null" else jsonlite::toJSON(response_metadata, auto_unbox = TRUE, pretty = TRUE)
      )
      
      writeLines(readable_content, readable_file)
      
      # Log summary to main log
      self$debug(sprintf("API request/response logged: %s", api_call_id), list(
        api_call_id = api_call_id,
        provider = provider,
        model = model,
        prompt_length = nchar(as.character(prompt_content)),
        response_length = if(is.character(response_content)) sum(nchar(response_content)) else nchar(as.character(response_content)),
        files_created = c(
          sprintf("%s.json", api_call_id),
          sprintf("%s.md", api_call_id)
        )
      ))
      
        api_call_id
      },
      warning = function(w) invisible(NULL),
      error = function(e) invisible(NULL))
    },
    
    #' @description
    #' Log cache operations
    #' @param operation Cache operation name
    #' @param key Cache key
    #' @param size Optional cache object size in bytes
    log_cache_operation = function(operation, key, size = NULL) {
      if (operation == "hit") {
        self$performance_stats$cache_hits <- self$performance_stats$cache_hits + 1
      } else if (operation == "miss") {
        self$performance_stats$cache_misses <- self$performance_stats$cache_misses + 1
      }
      
      total_cache_ops <- self$performance_stats$cache_hits + self$performance_stats$cache_misses
      context <- list(
        operation = operation,
        key = key,
        size_bytes = size,
        cache_hit_rate = if (total_cache_ops > 0) round(self$performance_stats$cache_hits / total_cache_ops, 3) else 0
      )
      
      self$debug(sprintf("Cache %s: %s", operation, key), context)
    },
    
    #' @description
    #' Log cluster annotation progress
    #' @param cluster_id Cluster identifier
    #' @param stage Processing stage
    #' @param progress Optional progress value
    log_cluster_progress = function(cluster_id, stage, progress = NULL) {
      context <- list(
        cluster_id = cluster_id,
        stage = stage,
        progress = progress,
        timestamp = format(Sys.time(), "%Y-%m-%d %H:%M:%S")
      )
      
      self$info(sprintf("Cluster %s: %s", cluster_id, stage), context)
    },
    
    #' @description
    #' Log detailed cluster discussion with complete model conversations
    #' @param cluster_id Cluster identifier
    #' @param event_type Discussion event type
    #' @param data Optional event payload
    log_discussion = function(cluster_id, event_type, data = NULL) {
      # Skip during R CMD check
      if (nzchar(Sys.getenv("_R_CHECK_PACKAGE_NAME_", ""))) {
        # Still log to console/main log but skip file creation
        context <- list(
          cluster_id = cluster_id,
          event_type = event_type,
          data = data
        )
        self$info(sprintf("Discussion %s for cluster %s", event_type, cluster_id), context)
        return(invisible(NULL))
      }

      tryCatch({
      
      # Create session directory if needed
      session_dir <- file.path(self$log_dir, self$session_id)
      if (!dir.exists(session_dir)) {
        dir.create(session_dir, recursive = TRUE)
      }
      
      # Log to both main log and cluster-specific file
      context <- list(
        cluster_id = cluster_id,
        event_type = event_type,
        data = data
      )
      
      self$info(sprintf("Discussion %s for cluster %s", event_type, cluster_id), context)
      
      # Also create detailed cluster log file for easy access
      cluster_safe <- private$safe_filename_component(cluster_id, "cluster")
      cluster_log_file <- file.path(
        session_dir,
        sprintf("cluster_%s_discussion.md", cluster_safe)
      )
      # Pure dialogue log: model outputs only (no timestamps/session metadata)
      dialogue_log_file <- file.path(
        session_dir,
        sprintf("cluster_%s_dialogue.txt", cluster_safe)
      )
      
      if (event_type == "start") {
        # Create markdown log for better readability
        content <- sprintf("# Cluster %s Discussion\n\n**Tissue:** %s\n**Markers:** %s\n**Started:** %s\n\n",
                          cluster_id, 
                          if(is.null(data$tissue_name)) "Unknown" else data$tissue_name,
                          if(is.null(data$marker_genes)) "Unknown" else data$marker_genes, 
                          format(Sys.time(), "%Y-%m-%d %H:%M:%S"))
        cat(content, file = cluster_log_file)
        cat("", file = dialogue_log_file)
        
      } else if (event_type == "prediction") {
        # Append model prediction in readable format
        # Handle both single strings and character vectors
        prediction_text <- if(is.vector(data$prediction) && length(data$prediction) > 1) {
          paste(data$prediction, collapse = "\n")
        } else {
          as.character(data$prediction)
        }
        
        content <- sprintf("## %s (Round %s)\n\n```\n%s\n```\n\n---\n\n",
                          data$model, 
                          if(is.null(data$round)) "1" else data$round, 
                          prediction_text)
        cat(content, file = cluster_log_file, append = TRUE)
        dialogue_content <- sprintf("%s (Round %s):\n%s\n\n",
                                   as.character(data$model),
                                   if (is.null(data$round)) "1" else as.character(data$round),
                                   prediction_text)
        cat(dialogue_content, file = dialogue_log_file, append = TRUE)
        
      } else if (event_type == "consensus") {
        # Log consensus result
        content <- sprintf("## Consensus Result\n\n**Reached:** %s\n**Proportion:** %.2f\n**Entropy:** %.2f\n**Decision:** %s\n\n---\n\n",
                          if (!is.null(data$reached)) data$reached else FALSE,
                          if (!is.null(data$consensus_proportion)) data$consensus_proportion else 0,
                          if (!is.null(data$entropy)) data$entropy else 0,
                          if (!is.null(data$majority_prediction)) data$majority_prediction else "Unknown")
        cat(content, file = cluster_log_file, append = TRUE)
        
      } else if (event_type == "end") {
        # Final summary
        content <- sprintf("## Discussion Complete\n\n**Final Result:** %s\n**Completed:** %s\n",
                          if (!is.null(data$final_result)) data$final_result else "Unknown",
                          format(Sys.time(), "%Y-%m-%d %H:%M:%S"))
        cat(content, file = cluster_log_file, append = TRUE)
      }
      },
      warning = function(w) invisible(NULL),
      error = function(e) invisible(NULL))
    },

    #' @description
    #' Log model response with concise summary in main log and full text in file
    #' @param provider Provider identifier
    #' @param model Model identifier
    #' @param response Model response
    #' @param stage Processing stage
    #' @param cluster_id Optional cluster identifier
    log_model_response = function(provider, model, response, stage = "annotation", cluster_id = NULL) {
      # Skip file writes during R CMD check
      if (nzchar(Sys.getenv("_R_CHECK_PACKAGE_NAME_", ""))) {
        response_text <- if (is.character(response)) paste(response, collapse = "\n") else as.character(response)
        self$info("Model response received", list(
          provider = provider,
          model = model,
          stage = stage,
          cluster_id = cluster_id,
          response_chars = nchar(response_text)
        ))
        return(invisible(NULL))
      }

      tryCatch({

      session_dir <- file.path(self$log_dir, self$session_id)
      if (!dir.exists(session_dir)) {
        dir.create(session_dir, recursive = TRUE)
      }
      response_dir <- file.path(session_dir, "model_responses")
      if (!dir.exists(response_dir)) {
        dir.create(response_dir, recursive = TRUE)
      }

      response_text <- if (is.character(response)) paste(response, collapse = "\n") else as.character(response)
      timestamp <- format(Sys.time(), "%Y%m%d_%H%M%S")
      model_safe <- gsub("[^A-Za-z0-9_-]", "_", model)
      provider_safe <- gsub("[^A-Za-z0-9_-]", "_", provider)
      stage_safe <- gsub("[^A-Za-z0-9_-]", "_", stage)
      cluster_safe <- if (is.null(cluster_id)) "na" else gsub("[^A-Za-z0-9_-]", "_", as.character(cluster_id))
      nonce <- sprintf("%06d", as.integer(stats::runif(1, 0, 999999)))
      response_file <- file.path(
        response_dir,
        sprintf("%s_%s_%s_cluster-%s_%s.txt", provider_safe, model_safe, stage_safe, cluster_safe, paste0(timestamp, "_", nonce))
      )

      writeLines(response_text, response_file)

      preview <- if (nchar(response_text) > 180) {
        paste0(substr(response_text, 1, 180), "...")
      } else {
        response_text
      }

      context <- list(
        provider = provider,
        model = model,
        stage = stage,
        response_chars = nchar(response_text),
        response_lines = length(strsplit(response_text, "\n", fixed = TRUE)[[1]]),
        response_file = response_file,
        response_preview = preview
      )
      if (!is.null(cluster_id)) {
        context$cluster_id <- cluster_id
      }

      self$info("Model response received", context)
        invisible(response_file)
      },
      warning = function(w) invisible(NULL),
      error = function(e) invisible(NULL))
    },
    
    #' @description
    #' Get performance summary
    #
    get_performance_summary = function() {
      session_duration <- as.numeric(difftime(Sys.time(), self$performance_stats$session_start, units = "secs"))
      
      summary <- list(
        session_id = self$session_id,
        session_duration_seconds = round(session_duration, 1),
        api_calls = self$performance_stats$api_calls,
        cache_hits = self$performance_stats$cache_hits,
        cache_misses = self$performance_stats$cache_misses,
        cache_hit_rate = if (self$performance_stats$cache_hits + self$performance_stats$cache_misses > 0) {
          round(self$performance_stats$cache_hits / (self$performance_stats$cache_hits + self$performance_stats$cache_misses), 3)
        } else { 0 },
        errors = self$performance_stats$errors,
        warnings = self$performance_stats$warnings
      )
      
      self$info("Performance summary generated", summary)
      return(summary)
    },
    
    #' @description
    #' Clean up old log files
    #' @param force Whether to remove every main log file
    cleanup_logs = function(force = FALSE) {
      log_files <- list.files(self$log_dir, pattern = "^mllm_.*\\.log$", full.names = TRUE)
      
      if (length(log_files) == 0) return(invisible(NULL))
      
      # Sort by modification time (newest first)
      file_info <- file.info(log_files)
      log_files <- log_files[order(file_info$mtime, decreasing = TRUE)]
      
      files_to_remove <- c()
      
      if (force) {
        files_to_remove <- log_files
      } else if (length(log_files) > self$max_log_files) {
        files_to_remove <- c(files_to_remove, log_files[(self$max_log_files + 1):length(log_files)])
      }
      
      # Check file sizes
      if (!force) {
        retained_files <- head(log_files, self$max_log_files)
        for (file in retained_files) {
          size_mb <- file.info(file)$size / (1024 * 1024)
          if (!is.na(size_mb) && size_mb > self$max_log_size) {
            files_to_remove <- c(files_to_remove, file)
          }
        }
      }
      
      if (length(files_to_remove) > 0) {
        files_to_remove <- unique(files_to_remove)
        for (file in files_to_remove) {
          if (file.exists(file)) {
            file.remove(file)
          }
        }
        if (!force) {
          self$info(sprintf("Cleaned up %d log files", length(files_to_remove)))
        }
      }
    },
    
    #' @description
    #' Set logging level
    #' @param level Minimum log level
    set_level = function(level) {
      normalized_level <- .normalize_log_level(level)
      old_level <- self$log_level
      self$log_level <- normalized_level
      self$info(sprintf("Log level changed from %s to %s", old_level, self$log_level))
    }
  ),
  
  private = list(
    safe_filename_component = function(value, fallback) {
      scalar <- if (is.null(value) || length(value) == 0 || is.na(value[[1]])) {
        fallback
      } else {
        as.character(value[[1]])
      }
      safe <- gsub("[^A-Za-z0-9_-]", "_", scalar)
      if (!nzchar(safe)) fallback else safe
    },

    create_session_id = function() {
      ts <- format(Sys.time(), "%Y%m%d_%H%M%S")
      subsec <- sprintf("%03d", as.integer((as.numeric(Sys.time()) %% 1) * 1000))
      nonce <- sprintf("%06d", as.integer(stats::runif(1, 0, 999999)))
      sprintf("%s%s_%d_%s", ts, subsec, Sys.getpid(), nonce)
    },

    # Internal method to write log messages
    # @param level Log level
    # @param message Log message  
    # @param context Additional context
    log_message = function(level, message, context) {
      # Check if we should log this level
      level_priority <- list("DEBUG" = 1, "INFO" = 2, "WARN" = 3, "ERROR" = 4)
      if (level_priority[[level]] < level_priority[[self$log_level]]) {
        return(invisible(NULL))
      }
      
      timestamp <- format(Sys.time(), "%Y-%m-%d %H:%M:%S")
      
      if (self$enable_json) {
        log_entry <- list(
          timestamp = timestamp,
          session_id = self$session_id,
          level = level,
          message = message,
          context = context,
          pid = Sys.getpid()
        )
        
        formatted_message <- jsonlite::toJSON(log_entry, auto_unbox = TRUE)
      } else {
        formatted_message <- sprintf("[%s] [%s] [%s] %s", 
                                   timestamp, self$session_id, level, message)
        if (!is.null(context)) {
          formatted_message <- paste(formatted_message, 
                                   sprintf("Context: %s", jsonlite::toJSON(context, auto_unbox = TRUE)))
        }
      }
      
      # Write to file
      private$write_to_file(formatted_message)
      
      # Output to console if enabled
      if (self$enable_console) {
        if (level == "ERROR") {
          cat(formatted_message, "\n", file = stderr())
        } else {
          cat(formatted_message, "\n")
        }
      }
    },
    
    # Write message to log file
    # @param message Formatted log message
    write_to_file = function(message) {
      # Skip file writing during R CMD check
      if (nzchar(Sys.getenv("_R_CHECK_PACKAGE_NAME_", ""))) {
        return(invisible(NULL))
      }
      
      log_file <- file.path(self$log_dir, sprintf("mllm_%s.log", self$session_id))
      
      # Ensure log directory exists before writing
      if (!dir.exists(self$log_dir)) {
        dir.create(self$log_dir, recursive = TRUE)
      }
      
      # Check if log rotation is needed
      if (file.exists(log_file)) {
        size_mb <- file.info(log_file)$size / (1024 * 1024)
        if (size_mb > self$max_log_size) {
          # Rotate log file
          rotated_file <- file.path(self$log_dir, 
                                  sprintf("mllm_%s_%s.log", self$session_id, 
                                         format(Sys.time(), "%H%M%S")))
          file.rename(log_file, rotated_file)
        }
      }
      
      # Write to log file
      cat(message, "\n", file = log_file, append = TRUE)
    }
  )
)

# Package-level environment for logger storage (avoids modifying .GlobalEnv)
.mllm_pkg_env <- new.env(parent = emptyenv())

#' Get the global logger instance
#
#' @export
get_logger <- function() {
  if (is.null(.mllm_pkg_env$.mllm_logger)) {
    .mllm_pkg_env$.mllm_logger <- UnifiedLogger$new()
  }
  return(.mllm_pkg_env$.mllm_logger)
}

#' Reinitialize global logger with a specific directory
#'
#' Preserves the current logger configuration (level, size, retention, console/json)
#' while changing the log directory for a new annotation session.
#'
#' @param log_dir Directory for log files
#' @return Invisible logger object
#' @keywords internal
initialize_logger <- function(log_dir = "logs") {
  log_dir <- .normalize_required_string(log_dir, "log_dir")
  current_logger <- if (!is.null(.mllm_pkg_env$.mllm_logger)) .mllm_pkg_env$.mllm_logger else NULL

  level <- "INFO"
  max_size <- 10
  max_files <- 5
  console_output <- FALSE
  json_format <- TRUE

  has_logger_fields <- function(x) {
    is.environment(x) &&
      !is.null(x$log_level) &&
      !is.null(x$max_log_size) &&
      !is.null(x$max_log_files) &&
      !is.null(x$enable_console) &&
      !is.null(x$enable_json)
  }

  if (!is.null(current_logger) && has_logger_fields(current_logger)) {
    if (!is.null(current_logger$log_level)) level <- current_logger$log_level
    if (!is.null(current_logger$max_log_size)) max_size <- current_logger$max_log_size
    if (!is.null(current_logger$max_log_files)) max_files <- current_logger$max_log_files
    if (!is.null(current_logger$enable_console)) console_output <- current_logger$enable_console
    if (!is.null(current_logger$enable_json)) json_format <- current_logger$enable_json
  }

  .mllm_pkg_env$.mllm_logger <- UnifiedLogger$new(
    base_dir = log_dir,
    level = level,
    max_size = max_size,
    max_files = max_files,
    console_output = console_output,
    json_format = json_format
  )

  invisible(.mllm_pkg_env$.mllm_logger)
}

#' Set global logger configuration
#'
#' @param level Logging level: "DEBUG", "INFO", "WARN", or "ERROR". Default: "INFO"
#' @param console_output Whether to enable console output. Default: FALSE
#' @param json_format Whether to use JSON format for log messages. Default: TRUE
#'
#' @return Invisible logger object
#' @export
configure_logger <- function(level = "INFO", console_output = FALSE, json_format = TRUE) {
  level <- .normalize_log_level(level)
  console_output <- .normalize_flag(console_output, "console_output")
  json_format <- .normalize_flag(json_format, "json_format")
  logger <- get_logger()
  logger$set_level(level)
  logger$enable_console <- console_output
  logger$enable_json <- json_format
  
  invisible(logger)
}

#' Convenience functions for logging
#'
#' @param message Log message string
#' @param context Optional context information (list or character)
#'
#' @return Invisible NULL
#' @name logging_functions
NULL

#' @rdname logging_functions
#' @export
log_debug <- function(message, context = NULL) {
  get_logger()$debug(message, context)
}

#' @rdname logging_functions
#' @export
log_info <- function(message, context = NULL) {
  get_logger()$info(message, context)
}

#' @rdname logging_functions
#' @export
log_warn <- function(message, context = NULL) {
  get_logger()$warn(message, context)
}

#' @rdname logging_functions
#' @export
log_error <- function(message, context = NULL) {
  get_logger()$error(message, context)
}

# Input contract consistency tests

make_test_json_response <- function(payload,
                                    status_code = 200L,
                                    url = "https://example.test/v1/chat/completions") {
  structure(
    list(
      url = url,
      status_code = status_code,
      headers = list("Content-Type" = "application/json"),
      all_headers = list(list(
        status = status_code,
        version = "HTTP/1.1",
        headers = list("Content-Type" = "application/json")
      )),
      cookies = data.frame(),
      content = charToRaw(jsonlite::toJSON(payload, auto_unbox = TRUE)),
      date = Sys.time(),
      times = c(),
      request = list(method = "POST")
    ),
    class = "response"
  )
}

enable_logger_file_writes <- function() {
  variable <- "_R_CHECK_PACKAGE_NAME_"
  previous <- Sys.getenv(variable, unset = NA_character_)
  Sys.unsetenv(variable)

  function() {
    if (is.na(previous)) {
      Sys.unsetenv(variable)
    } else {
      do.call(Sys.setenv, setNames(list(previous), variable))
    }
  }
}

preserve_environment_variables <- function(variables) {
  previous <- Sys.getenv(variables, unset = NA_character_)

  function() {
    Sys.unsetenv(variables)
    configured <- !is.na(previous)
    if (any(configured)) {
      do.call(
        Sys.setenv,
        setNames(as.list(previous[configured]), variables[configured])
      )
    }
  }
}

make_mock_builtin_processor_factory <- function(process_request) {
  force(process_request)
  function(provider, base_url = NULL) {
    list(process_request = function(prompt, model, api_key, ...) {
      process_request(prompt, model, api_key)
    })
  }
}

test_that("normalize_cluster_gene_list canonicalizes list inputs", {
  unnamed <- normalize_cluster_gene_list(list(c("G1"), c("G2")))
  expect_identical(names(unnamed), c("0", "1"))

  numeric_named <- normalize_cluster_gene_list(
    list("1" = c("A"), "2" = list(genes = c("B")))
  )
  expect_identical(names(numeric_named), c("1", "2"))

  non_numeric_named <- normalize_cluster_gene_list(
    list("t_cells" = c("CD3D"), "b_cells" = list(genes = c("MS4A1")))
  )
  expect_identical(names(non_numeric_named), c("t_cells", "b_cells"))
})

test_that("normalize_cluster_gene_list cleans genes and rejects ambiguous cluster IDs", {
  normalized <- normalize_cluster_gene_list(list(
    " cluster_a " = list(genes = c(" CD3D ", NA, "", "CD3D", "CD3E"))
  ))
  expect_identical(normalized, list(cluster_a = c("CD3D", "CD3E")))

  expect_error(normalize_cluster_gene_list(list()), "at least one cluster")
  expect_error(
    normalize_cluster_gene_list(setNames(list(c("G1")), "")),
    "Cluster IDs must be non-empty strings"
  )
  expect_error(
    normalize_cluster_gene_list(list("a" = list(genes = list("G1")))),
    "marker genes must be an atomic vector"
  )
  expect_error(
    normalize_cluster_gene_list(list("a" = c(NA, "   "))),
    "No genes found for cluster"
  )
})

test_that("create_annotation_prompt accepts mixed list element formats", {
  prompt_result <- create_annotation_prompt(
    input = list(
      gs1 = c("CD4", "CD3D"),
      gs2 = list(genes = c("CD14"))
    ),
    tissue_name = "PBMC"
  )

  expect_true(all(c("gs1", "gs2") %in% names(prompt_result$gene_lists)))
  expect_match(prompt_result$prompt, "gs1: CD4, CD3D")
  expect_match(prompt_result$prompt, "gs2: CD14")
})

test_that("create_annotation_prompt returns the exact numeric display order", {
  prompt_result <- create_annotation_prompt(
    list("10" = "G10", "2" = "G2", "1" = "G1"),
    "PBMC"
  )

  expect_identical(names(prompt_result$gene_lists), c("1", "2", "10"))
  prompt_lines <- strsplit(prompt_result$prompt, "\n", fixed = TRUE)[[1]]
  cluster_lines <- prompt_lines[grepl("^(1|2|10):", prompt_lines)]
  expect_identical(cluster_lines, c("1: G1", "2: G2", "10: G10"))
  expect_match(prompt_result$prompt, "in the same order", fixed = TRUE)

  aligned <- align_model_predictions(
    c("Type 1", "Type 2", "Type 10"),
    names(prompt_result$gene_lists)
  )
  expect_identical(
    aligned,
    c("1" = "Type 1", "2" = "Type 2", "10" = "Type 10")
  )
})

test_that("create_annotation_prompt validates exported inputs and data frame schema", {
  expect_error(
    create_annotation_prompt(list("0" = c("G1")), "   "),
    "tissue_name must be a non-empty character scalar"
  )
  expect_error(
    create_annotation_prompt(list("0" = c("G1")), "PBMC", top_gene_count = 1.5),
    "top_gene_count must be a positive integer"
  )
  expect_error(
    create_annotation_prompt(data.frame(cluster = "0", gene = "G1"), "PBMC"),
    "exactly one each"
  )
  expect_error(
    create_annotation_prompt(
      data.frame(cluster = "0", gene = "G1", avg_log2FC = "high"),
      "PBMC"
    ),
    "avg_log2FC must be numeric"
  )

  result <- create_annotation_prompt(
    data.frame(
      cluster = c(" 1 ", "1", "2"),
      gene = c(" G1 ", "G1", "G2"),
      avg_log2FC = c(2, 1, 3)
    ),
    " PBMC ",
    top_gene_count = 2
  )
  expect_identical(names(result$gene_lists), c("1", "2"))
  expect_identical(result$gene_lists[["1"]], "G1")
  expect_match(result$prompt, "in PBMC")
})

test_that("annotation, discussion, and cache marker selection stay aligned", {
  markers <- data.frame(
    cluster = c(" 1 ", "1", "1", NA),
    gene = c(" LOW ", "HIGH", "HIGH", "IGNORED"),
    avg_log2FC = c(-1, 3, 2, 10)
  )

  prompt_result <- suppressWarnings(
    create_annotation_prompt(markers, "PBMC", top_gene_count = 2)
  )
  discussion_genes <- extract_cluster_genes_for_discussion(markers, "1", 2)

  expect_identical(prompt_result$gene_lists[["1"]], "HIGH,LOW")
  expect_identical(discussion_genes, "HIGH,LOW")

  cache_manager <- CacheManager$new("temp")
  expect_false(identical(
    cache_manager$generate_key(markers, "gpt-5.5", "1", "PBMC", 1),
    cache_manager$generate_key(markers, "gpt-5.5", "1", "PBMC", 2)
  ))
})

test_that("CacheManager generate_key supports character-vector list clusters", {
  cache_manager <- CacheManager$new(cache_dir = "temp")
  input_list <- list(
    "0" = c("CD3D", "CD3E"),
    "1" = list(genes = c("MS4A1"))
  )

  key_0 <- cache_manager$generate_key(input_list, c("gpt-5.5"), "0")
  key_1 <- cache_manager$generate_key(input_list, c("gpt-5.5"), "1")

  expect_true(is.character(key_0))
  expect_gt(nchar(key_0), 10)
  expect_false(identical(key_0, key_1))
})

test_that("parse_text_predictions resolves cluster aliases without positional reassignment", {
  parsed <- parse_text_predictions(
    c("Cluster 2: B cell"),
    all_clusters = c("1", "2")
  )

  expect_null(parsed[["1"]])
  expect_identical(parsed[["2"]], "B cell")

  complete <- parse_text_predictions(
    c("Cluster_1: T cell", "2) B cell"),
    all_clusters = c("1", "2")
  )
  expect_identical(complete, list("1" = "T cell", "2" = "B cell"))
})

test_that("parse_text_predictions uses positions only for unlabeled responses", {
  parsed <- parse_text_predictions(
    c("T-cell", "B cell"),
    all_clusters = c("alpha", "beta")
  )

  expect_identical(parsed, list(alpha = "T-cell", beta = "B cell"))
})

test_that("parse_text_predictions preserves exact IDs over ambiguous aliases", {
  exact <- parse_text_predictions(
    "Cluster 1: Named cluster",
    all_clusters = c("1", "Cluster 1")
  )
  expect_identical(exact, list("Cluster 1" = "Named cluster"))

  ambiguous <- parse_text_predictions(
    "Cluster_1: Ambiguous",
    all_clusters = c("1", "Cluster 1")
  )
  expect_length(ambiguous, 0)
})

test_that("parse_text_predictions maps a numbered list onto 0-based clusters without shifting", {
  # Regression: models commonly number their answers ("1.", "2.", ...). On
  # Seurat 0-based clusters the ordinals are list indices, not cluster IDs;
  # trusting them shifts every annotation one cluster and drops cluster 0.
  parsed <- parse_text_predictions(
    c("1. T cells", "2. B cells", "3. NK cells"),
    all_clusters = c("0", "1", "2")
  )
  expect_identical(parsed, list("0" = "T cells", "1" = "B cells", "2" = "NK cells"))
})

test_that("parse_text_predictions keeps a positional response despite a preamble header", {
  parsed <- parse_text_predictions(
    c("Here are the cell type annotations:", "T cells", "B cells", "NK cells"),
    all_clusters = c("0", "1", "2")
  )
  expect_identical(parsed, list("0" = "T cells", "1" = "B cells", "2" = "NK cells"))
})

test_that("parse_text_predictions keeps an annotation-internal colon whole", {
  parsed <- parse_text_predictions(
    c("Astrocytes", "Neurons: excitatory", "Microglia"),
    all_clusters = c("0", "1", "2")
  )
  expect_identical(
    parsed,
    list("0" = "Astrocytes", "1" = "Neurons: excitatory", "2" = "Microglia")
  )
})

test_that("parse_text_predictions keeps a mid-list Unknown in its own slot", {
  # Regression: dropping a sentinel line must not shift later clusters up one.
  parsed <- parse_text_predictions(
    c("T cells", "Unknown", "B cells"),
    all_clusters = c("0", "1", "2")
  )
  # cluster 1 is left unassigned (-> Unknown downstream); cluster 2 keeps B cells.
  expect_identical(parsed, list("0" = "T cells", "2" = "B cells"))
})

test_that("parse_text_predictions ignores a stray Summary: line and keeps labels", {
  parsed <- parse_text_predictions(
    c("Summary: three immune populations", "0: T cells", "1: B cells", "2: NK cells"),
    all_clusters = c("0", "1", "2")
  )
  expect_identical(parsed, list("0" = "T cells", "1" = "B cells", "2" = "NK cells"))
})

test_that("parse_text_predictions maps out-of-order labels by cluster ID, ignoring hallucinated ones", {
  parsed <- parse_text_predictions(
    c("2: NK cells", "1: B cells", "0: T cells", "5: Doublets"),
    all_clusters = c("0", "1", "2")
  )
  expect_identical(parsed, list("0" = "T cells", "1" = "B cells", "2" = "NK cells"))
})

test_that("parse_text_predictions does not strip a hyphenated cell type as an ordinal", {
  parsed <- parse_text_predictions(
    c("5 - HT neurons"),
    all_clusters = c("0")
  )
  expect_identical(parsed, list("0" = "5 - HT neurons"))
})

test_that("compare_model_predictions pads unequal lengths without recycling", {
  res <- testthat::with_mocked_bindings({
    capture.output({
      result <- suppressMessages(compare_model_predictions(
        input = list(
          "0" = list(genes = c("G1")),
          "1" = list(genes = c("G2")),
          "2" = list(genes = c("G3"))
        ),
        tissue_name = "PBMC",
        models = c("m1", "m2"),
        api_keys = list(openai = "k")
      ))
    })
    result
  },
  get_api_key = function(model, api_keys) "k",
  get_provider = function(model) "openai",
  annotate_cell_types = function(input, tissue_name, model, api_key, top_gene_count = 10, ...) {
    if (identical(model, "m1")) c("A", "B", "C") else c("A", "B")
  },
  standardize_cell_type_names = function(predictions, models, api_keys, standardization_model = "x", base_urls = NULL) {
    predictions
  })

  expect_identical(dim(res$comparison_matrix), c(3L, 2L))
  expect_true(is.na(res$comparison_matrix[3, "m2"]))
  expect_true(is.na(res$consensus_predictions[3]))
  expect_true(is.na(res$consensus_proportions[3]))
  expect_true(is.na(res$entropies[3]))
  expect_identical(rownames(res$comparison_matrix), c("0", "1", "2"))
  expect_identical(names(res$consensus_predictions), c("0", "1", "2"))
  expect_false(is.nan(res$summary_stats$mean_consensus_proportion))
  expect_false(is.nan(res$summary_stats$mean_entropy))
})

test_that("compare_model_predictions aligns explicit labels by cluster ID", {
  res <- testthat::with_mocked_bindings({
    capture.output({
      result <- suppressMessages(compare_model_predictions(
        input = list(
          alpha = list(genes = c("G1")),
          beta = list(genes = c("G2"))
        ),
        tissue_name = "PBMC",
        models = c("m1", "m2"),
        api_keys = list(openai = "k")
      ))
    })
    result
  },
  get_api_key = function(model, api_keys) "k",
  get_provider = function(model) "openai",
  annotate_cell_types = function(input, tissue_name, model, api_key, top_gene_count = 10, ...) {
    if (identical(model, "m1")) {
      c("beta: B cell", "alpha: T cell")
    } else {
      c("alpha: T cell", "beta: B cell")
    }
  },
  standardize_cell_type_names = function(predictions, models, api_keys, standardization_model = "x", base_urls = NULL) {
    predictions
  })

  expect_identical(rownames(res$comparison_matrix), c("alpha", "beta"))
  expect_identical(unname(res$comparison_matrix[, "m1"]), c("T cell", "B cell"))
  expect_identical(unname(res$consensus_predictions), c("T cell", "B cell"))
})

test_that("compare_model_predictions does not report a tied vote as consensus", {
  result <- testthat::with_mocked_bindings({
    capture.output({
      comparison <- suppressMessages(compare_model_predictions(
        input = list("0" = "G1"),
        tissue_name = "PBMC",
        models = c("m1", "m2"),
        api_keys = list(openai = "key"),
        consensus_threshold = 0.5
      ))
    })
    comparison
  },
  get_api_key = function(...) "key",
  get_provider = function(...) "openai",
  annotate_cell_types = function(input, tissue_name, model, ...) {
    if (model == "m1") "B cell" else "T cell"
  },
  standardize_cell_type_names = function(predictions, ...) predictions)

  expect_true(is.na(result$consensus_predictions[["0"]]))
  expect_true(is.na(result$consensus_proportions[["0"]]))
})

test_that("align_model_predictions honors names and marks missing clusters", {
  aligned <- align_model_predictions(
    c(beta = "B cell", alpha = "T cell"),
    c("alpha", "beta", "gamma")
  )

  expect_identical(
    aligned,
    c(alpha = "T cell", beta = "B cell", gamma = NA_character_)
  )
})

test_that("align_model_predictions treats internal sentinels as missing", {
  aligned <- align_model_predictions(
    c("0: Unknown", "1: Prediction_Missing", "2: T cell"),
    c("0", "1", "2")
  )

  expect_true(is.na(aligned[["0"]]))
  expect_true(is.na(aligned[["1"]]))
  expect_identical(aligned[["2"]], "T cell")
})

test_that("consensus phases share canonical numeric cluster order", {
  input <- list("10" = c("G10"), "2" = c("G2"), "1" = c("G1"))
  predictions <- list(
    m1 = c("Type 1", "Type 2", "Type 10"),
    m2 = c("Type 1", "Type 2", "Type 10")
  )

  result <- testthat::with_mocked_bindings({
    suppressMessages(identify_controversial_clusters(
      input = input,
      individual_predictions = predictions,
      controversy_threshold = 0.5,
      entropy_threshold = 1,
      api_keys = list(openai = "k")
    ))
  },
  check_consensus = function(round_responses, ...) {
    list(
      reached = TRUE,
      consensus_proportion = 1,
      entropy = 0,
      majority_prediction = unname(round_responses[[1]])
    )
  },
  log_info = function(...) NULL,
  log_warn = function(...) NULL)

  expect_identical(
    result$final_annotations,
    list("1" = "Type 1", "2" = "Type 2", "10" = "Type 10")
  )
})

test_that("standardize_cell_type_names ignores prose and accepts numbered mapping keys", {
  result <- testthat::with_mocked_bindings({
    suppressMessages(standardize_cell_type_names(
      predictions = list(m1 = c("1: T cell", "B cell")),
      models = "m1",
      api_keys = list(openai = "k"),
      standardization_model = "gpt-5.5"
    ))
  },
  get_api_key = function(...) "k",
  get_model_response = function(...) c(
    "Here is the requested mapping:",
    "1: T cell: CD4 T cell",
    "B cell: B cell",
    "Note: keep broad immune labels unchanged"
  ),
  log_warn = function(...) NULL)

  expect_identical(result$m1, c("CD4 T cell", "B cell"))
})

test_that("standardization defaults to the first requested model", {
  requested_model <- NULL
  result <- testthat::with_mocked_bindings(
    standardize_cell_type_names(
      predictions = list(m1 = "T cell", m2 = "T cell"),
      models = c("m1", "m2"),
      api_keys = list(openai = "key", anthropic = "other-key")
    ),
    get_api_key = function(...) "key",
    get_model_response = function(prompt, model, ...) {
      requested_model <<- model
      "T cell: T cell"
    },
    log_warn = function(...) NULL
  )

  expect_identical(requested_model, "m1")
  expect_identical(result, list(m1 = "T cell", m2 = "T cell"))
})

test_that("compare_model_predictions requires two eligible models", {
  expect_error(
    suppressWarnings(compare_model_predictions(
      input = list("0" = "G1"),
      tissue_name = "PBMC",
      models = c("gpt-5.5", "bad-model"),
      api_keys = list(openai = "key")
    )),
    "At least two models"
  )
})

test_that("compare_model_predictions skips invalid models and uses valid remaining models", {
  res <- testthat::with_mocked_bindings({
    capture.output({
      result <- suppressWarnings(suppressMessages(compare_model_predictions(
        input = list("0" = list(genes = c("G1"))),
        tissue_name = "PBMC",
        models = c("bad-model", "m1", "m2"),
        api_keys = list(openai = "k")
      )))
    })
    result
  },
  get_api_key = function(model, api_keys) {
    if (identical(model, "bad-model")) stop("unknown model") else "k"
  },
  get_provider = function(model) {
    if (identical(model, "bad-model")) stop("unknown model") else "openai"
  },
  annotate_cell_types = function(input, tissue_name, model, api_key, top_gene_count = 10, ...) {
    c("T cell")
  },
  standardize_cell_type_names = function(predictions, models, api_keys, standardization_model = "x", base_urls = NULL) {
    predictions
  })

  expect_identical(colnames(res$comparison_matrix), c("m1", "m2"))
})

test_that("interactive_consensus_annotation supports mapped cluster selectors", {
  mocked_call <- function(selector) {
    testthat::with_mocked_bindings({
      suppressMessages(interactive_consensus_annotation(
        input = list(
          "1" = c("CD3D"),
          "2" = list(genes = c("MS4A1"))
        ),
        tissue_name = "PBMC",
        models = c("gpt-5.5", "grok-4.3"),
        api_keys = list(openai = "k", grok = "k"),
        use_cache = FALSE,
        clusters_to_analyze = selector
      ))
    },
    log_info = function(...) NULL,
    print_consensus_summary = function(...) NULL,
    get_initial_predictions = function(input, ...) {
      # Predictions keyed by original cluster names (no 0-based conversion)
      cluster_ids <- names(input)
      preds <- setNames(rep("T", length(cluster_ids)), cluster_ids)
      list(
        individual_predictions = list(m1 = as.list(preds), m2 = as.list(preds)),
        successful_models = c("m1", "m2")
      )
    },
    identify_controversial_clusters = function(input, ...) {
      list(
        consensus_results = list(),
        controversial_clusters = character(0),
        final_annotations = setNames(rep("X", length(input)), names(input))
      )
    },
    process_controversial_clusters = function(...) {
      list(discussion_logs = list(), final_annotations = list())
    },
    combine_results = function(initial_results, controversy_results, discussion_results) {
      list(final_annotations = controversy_results$final_annotations)
    })
  }

  # Cluster names are preserved as-is (no 0-based conversion)
  result_single <- mocked_call("1")
  expect_identical(names(result_single$final_annotations), "1")

  result_both <- mocked_call(c("1", "2"))
  expect_identical(sort(names(result_both$final_annotations)), c("1", "2"))
})

test_that("parse_consensus_response falls back to flexible parsing for short responses", {
  result <- testthat::with_mocked_bindings({
    parse_consensus_response(
      c(
        "Consensus Proportion = 0.75",
        "Entropy = 0.81",
        "B cell"
      )
    )
  },
  get_logger = function() {
    list(
      warn = function(...) NULL,
      info = function(...) NULL,
      debug = function(...) NULL,
      error = function(...) NULL
    )
  })

  expect_identical(result$reached, FALSE)
  expect_equal(result$consensus_proportion, 0.75)
  expect_equal(result$entropy, 0.81)
  expect_identical(result$majority_prediction, "B cell")
})

test_that("get_initial_predictions skips invalid model names without aborting", {
  result <- testthat::with_mocked_bindings({
    suppressWarnings(get_initial_predictions(
      input = list("0" = list(genes = c("CD3D"))),
      tissue_name = "PBMC",
      models = c("bad-model", "gpt-5.5"),
      api_keys = list(openai = "k"),
      top_gene_count = 10
    ))
  },
  get_provider = function(model) {
    if (identical(model, "bad-model")) stop("unknown model")
    "openai"
  },
  get_api_key = function(model, api_keys) {
    provider <- get_provider(model)
    if (provider %in% names(api_keys)) api_keys[[provider]] else NULL
  },
  annotate_cell_types = function(...) c("T cell"),
  log_info = function(...) NULL,
  log_warn = function(...) NULL)

  expect_identical(result$successful_models, "gpt-5.5")
  expect_true("gpt-5.5" %in% names(result$individual_predictions))
})

test_that("interactive_consensus_annotation consumes log_dir via initialize_logger", {
  seen_log_dir <- NULL

  testthat::with_mocked_bindings({
    suppressMessages(interactive_consensus_annotation(
      input = list("0" = list(genes = c("CD3D"))),
      tissue_name = "PBMC",
      models = c("gpt-5.5", "grok-4.3"),
      api_keys = list(openai = "k", grok = "k"),
      use_cache = FALSE,
      log_dir = "custom_logs_dir"
    ))
  },
  initialize_logger = function(log_dir = "logs") {
    seen_log_dir <<- log_dir
    invisible(NULL)
  },
  log_info = function(...) NULL,
  print_consensus_summary = function(...) NULL,
  get_initial_predictions = function(input, ...) {
    list(
      individual_predictions = list(m1 = c("0:T"), m2 = c("0:T")),
      successful_models = c("m1", "m2")
    )
  },
  identify_controversial_clusters = function(input, ...) {
    list(
      consensus_results = list(),
      controversial_clusters = character(0),
      final_annotations = setNames(rep("X", length(input)), names(input))
    )
  },
  process_controversial_clusters = function(...) {
    list(discussion_logs = list(), final_annotations = list())
  },
  combine_results = function(initial_results, controversy_results, discussion_results) {
    list(final_annotations = controversy_results$final_annotations)
  })

  expect_identical(seen_log_dir, "custom_logs_dir")
})

test_that("combine_results provides stable aliases for return fields", {
  result <- testthat::with_mocked_bindings({
    combine_results(
      initial_results = list(individual_predictions = list(m1 = "A")),
      controversy_results = list(
        consensus_results = list("0" = list(reached = TRUE)),
        controversial_clusters = c("1"),
        final_annotations = list("0" = "T cell")
      ),
      discussion_results = list(
        discussion_logs = list("1" = list(rounds = list())),
        final_annotations = list("1" = "B cell")
      )
    )
  },
  get_logger = function() list(session_id = "session-test"))

  expect_identical(result$voting_results, result$initial_results)
  expect_identical(result$discussion_results, result$discussion_logs)
  expect_identical(result$final_consensus, result$final_annotations)
})

test_that("annotation cleanup is scalar-safe and consistent across result phases", {
  expect_identical(clean_annotation(c("T cell", "B cell")), "Unknown")
  expect_identical(clean_annotation(1), "Unknown")
  expect_identical(clean_annotation("cell type: T cell"), "T cell")
  expect_identical(clean_annotation("unknown"), "Unknown")

  result <- testthat::with_mocked_bindings(
    combine_results(
      initial_results = list(individual_predictions = list()),
      controversy_results = list(
        consensus_results = list(),
        controversial_clusters = "1",
        final_annotations = list("0" = "CELL TYPE: T cell")
      ),
      discussion_results = list(
        discussion_logs = list(),
        final_annotations = list("1" = c("B cell", "NK cell"))
      )
    ),
    get_logger = function() list(session_id = "session-test")
  )

  expect_identical(result$final_annotations[["0"]], "T cell")
  expect_identical(result$final_annotations[["1"]], "Unknown")
})

test_that("select_best_prediction uses model order rather than label length for ties", {
  result <- select_best_prediction(
    list(majority_prediction = "Parsing_Failed"),
    c("T cell", "longer unsupported label")
  )

  expect_identical(result, "T cell")
})

capture_discussion_genes <- function(input, cluster_id = "0", top_gene_count = 5) {
  captured_cluster_genes <- NULL
  testthat::with_mocked_bindings({
    suppressWarnings(suppressMessages(facilitate_cluster_discussion(
      cluster_id = cluster_id,
      input = input,
      tissue_name = "PBMC",
      models = c("m1", "m2"),
      api_keys = list(openai = "k"),
      initial_predictions = list(m1 = c("0:T"), m2 = c("0:T")),
      top_gene_count = top_gene_count,
      max_rounds = 2
    )))
  },
  create_initial_discussion_prompt = function(cluster_id, cluster_genes, tissue_name, initial_predictions) {
    captured_cluster_genes <<- cluster_genes
    "prompt"
  },
  get_model_response = function(...) "CELL TYPE: T cell",
  check_consensus = function(...) {
    list(reached = TRUE, consensus_proportion = 1, entropy = 0, majority_prediction = "T cell")
  },
  get_api_key = function(model, api_keys) "k",
  get_provider = function(model) "openai",
  get_logger = function() {
    list(log_discussion = function(...) NULL)
  },
  log_warn = function(...) NULL,
  log_info = function(...) NULL)
  captured_cluster_genes
}

test_that("facilitate_cluster_discussion extracts data-frame genes by marker rank", {
  markers <- data.frame(
    cluster = c(0, 0, 0, 1),
    gene = c("low", "top", "negative", "other"),
    avg_log2FC = c(1, 3, -2, 5)
  )

  expect_identical(capture_discussion_genes(markers, cluster_id = "0", top_gene_count = 2), "top,low")
})

test_that("facilitate_cluster_discussion preserves data-frame row order without avg_log2FC", {
  markers <- data.frame(
    cluster = c("0", "0", "0"),
    gene = c("G1", "G2", "G3")
  )

  expect_identical(capture_discussion_genes(markers, cluster_id = "0", top_gene_count = 2), "G1,G2")
})

test_that("facilitate_cluster_discussion keeps fallback cluster_genes on extraction error", {
  captured_cluster_genes <- capture_discussion_genes(42, cluster_id = "0")
  expect_true(grepl("Error extracting genes", captured_cluster_genes, fixed = TRUE))
})

test_that("facilitate_cluster_discussion falls back for missing clusters and malformed target entries", {
  missing_cluster_genes <- capture_discussion_genes(list("1" = c("G1")), cluster_id = "0")
  malformed_entry_genes <- capture_discussion_genes(list("0" = list(markers = c("G1"))), cluster_id = "0")

  expect_true(grepl("Cluster '0' was not found", missing_cluster_genes, fixed = TRUE))
  expect_true(grepl("character vector of genes", malformed_entry_genes, fixed = TRUE))
})


test_that("facilitate_cluster_discussion ignores malformed unrelated list entries", {
  cluster_genes <- capture_discussion_genes(
    list("0" = c("G1", "G2"), "1" = list(markers = c("bad"))),
    cluster_id = "0"
  )

  expect_identical(cluster_genes, "G1,G2")
})

test_that("is_error_response flags pure errors but preserves partial annotations", {
  expect_true(is_error_response("Error: request failed"))
  expect_true(is_error_response("  ERROR: request failed"))
  # Every non-empty line is an error -> the whole response is an error.
  expect_true(is_error_response(c("Error: rate limit", "error : try again")))
  # A usable annotation alongside an error marker is NOT a pure error: a single
  # flagged/uncertain cluster must not discard the model's other annotations.
  expect_false(is_error_response(c("CD8+ T cells", "Error: markers ambiguous", "B cells")))
  expect_false(is_error_response("CELL TYPE: T cell"))
  expect_false(is_error_response(list(message = "Error: request failed")))
  expect_false(is_error_response(NA_character_))
  expect_false(is_error_response(character(0)))
})

test_that("model response normalization gives every provider the same line contract", {
  expect_identical(
    normalize_model_response_lines(c("  Cluster 1: T cells,  ", "", "Cluster 2: B cells\n")),
    c("Cluster 1: T cells", "Cluster 2: B cells")
  )

  invalid_responses <- list(NULL, character(), NA_character_, 123, "  ", "Error: failed")
  for (response in invalid_responses) {
    expect_error(normalize_model_response_lines(response), "Model response")
    expect_false(is_valid_model_response(response))
  }
})

test_that("token usage normalization rejects malformed external telemetry", {
  expect_identical(
    normalize_token_usage(list(
      prompt_tokens = 0,
      completion_tokens = 2L,
      total_tokens = 2,
      cost = 1
    )),
    list(
      prompt_tokens = 0,
      completion_tokens = 2,
      total_tokens = 2,
      cost = 1
    )
  )

  invalid <- normalize_token_usage(list(
    prompt_tokens = -1,
    completion_tokens = TRUE,
    total_tokens = "3",
    cost = Inf
  ))
  expect_null(invalid$prompt_tokens)
  expect_null(invalid$completion_tokens)
  expect_null(invalid$total_tokens)
  expect_false("cost" %in% names(invalid))
  expect_null(normalize_token_usage("not usage"))
})

test_that("provider usage extractors share one canonical schema", {
  openai_response <- make_test_json_response(list(
    usage = list(
      prompt_tokens = 10,
      completion_tokens = 4,
      total_tokens = 14,
      cost = 0.001
    )
  ))
  anthropic_response <- make_test_json_response(list(
    usage = list(input_tokens = 10, output_tokens = 4)
  ))
  gemini_response <- make_test_json_response(list(
    usageMetadata = list(
      promptTokenCount = 10,
      candidatesTokenCount = 4,
      totalTokenCount = 14
    )
  ))

  expect_identical(
    OpenAIProcessor$new()$extract_usage(openai_response),
    list(prompt_tokens = 10, completion_tokens = 4, total_tokens = 14, cost = 0.001)
  )
  expect_identical(
    AnthropicProcessor$new()$extract_usage(anthropic_response),
    list(prompt_tokens = 10, completion_tokens = 4, total_tokens = 14)
  )
  expect_identical(
    GeminiProcessor$new()$extract_usage(gemini_response),
    list(prompt_tokens = 10, completion_tokens = 4, total_tokens = 14)
  )
})

test_that("BaseAPIProcessor forwards normalized usage to API call logs", {
  captured_tokens <- NULL
  logger <- list(
    info = function(...) NULL,
    debug = function(...) NULL,
    warn = function(...) NULL,
    error = function(...) NULL,
    log_api_request_response = function(...) NULL,
    log_api_call = function(provider, model, duration, success = TRUE, tokens = NULL) {
      captured_tokens <<- tokens
    }
  )

  testthat::with_mocked_bindings({
    UsageProcessor <- R6::R6Class(
      "UsageProcessor",
      inherit = BaseAPIProcessor,
      public = list(
        initialize = function() super$initialize("openai", NULL),
        get_default_api_url = function() "https://example.test",
        make_api_call = function(chunk_content, model, api_key) {
          make_test_json_response(list(
            usage = list(
              prompt_tokens = 5,
              completion_tokens = 2,
              total_tokens = 7
            )
          ))
        },
        extract_response_content = function(response, model) "T cell"
      )
    )

    expect_identical(
      UsageProcessor$new()$process_request("prompt", "model", "key"),
      "T cell"
    )
  }, get_logger = function() logger)

  expect_identical(
    captured_tokens,
    list(prompt_tokens = 5, completion_tokens = 2, total_tokens = 7)
  )
})

test_that("filter_valid_responses excludes case-insensitive error sentinels", {
  result <- testthat::with_mocked_bindings({
    filter_valid_responses(
      list(
        good = "CELL TYPE: T cell",
        bad_upper = "ERROR: provider failed",
        bad_spaced = " error : provider failed",
        missing = NA_character_,
        blank = "   ",
        empty = character(0)
      ),
      cluster_id = "0"
    )
  },
  log_warn = function(...) NULL)

  expect_identical(names(result), "good")
})

test_that("facilitate_cluster_discussion selects the current cluster prediction", {
  captured_predictions <- NULL

  testthat::with_mocked_bindings({
    suppressMessages(facilitate_cluster_discussion(
      cluster_id = "1",
      input = list("0" = c("G1"), "1" = c("G2")),
      tissue_name = "PBMC",
      models = c("m1", "m2"),
      api_keys = list(openai = "k"),
      initial_predictions = list(
        m1 = c("T cell", "B cell"),
        m2 = c("T cell", "B cell")
      ),
      top_gene_count = 5,
      max_rounds = 1
    ))
  },
  create_initial_discussion_prompt = function(cluster_id, cluster_genes, tissue_name,
                                               initial_predictions) {
    captured_predictions <<- initial_predictions
    "prompt"
  },
  get_model_response = function(...) "CELL TYPE: B cell",
  check_consensus = function(...) {
    list(reached = TRUE, consensus_proportion = 1, entropy = 0, majority_prediction = "B cell")
  },
  get_api_key = function(...) "k",
  get_logger = function() list(log_discussion = function(...) NULL),
  log_info = function(...) NULL,
  log_warn = function(...) NULL)

  expect_identical(captured_predictions, list(m1 = "B cell", m2 = "B cell"))
})

test_that("initial discussion prompt renders canonical scalar predictions", {
  prompt <- create_initial_discussion_prompt(
    cluster_id = "1",
    cluster_genes = "MS4A1,CD79A",
    tissue_name = "PBMC",
    initial_predictions = list(m1 = "B cell", m2 = "Memory B cell")
  )

  expect_match(prompt, "m1: B cell", fixed = TRUE)
  expect_match(prompt, "m2: Memory B cell", fixed = TRUE)
  expect_false(grepl("No prediction", prompt, fixed = TRUE))
})

test_that("discussion prompt preserves multi-line responses per model", {
  prompt <- create_discussion_prompt(
    cluster_id = "1",
    cluster_genes = "MS4A1,CD79A",
    tissue_name = "PBMC",
    previous_rounds = list(list(
      round_number = 1,
      responses = list(
        m1 = c("CELL TYPE: B cell", "GROUNDS: MS4A1"),
        m2 = c("CELL TYPE: B cell", "GROUNDS: CD79A")
      )
    )),
    round_number = 2
  )

  expect_match(prompt, "m1:\nCELL TYPE: B cell\nGROUNDS: MS4A1", fixed = TRUE)
  expect_match(prompt, "m2:\nCELL TYPE: B cell\nGROUNDS: CD79A", fixed = TRUE)
  expect_false(grepl('c\\("CELL TYPE', prompt))
})

test_that("annotate_cell_types supports custom providers", {
  custom_env <- new.env(parent = emptyenv())
  assign("customx", TRUE, envir = custom_env)

  result <- testthat::with_mocked_bindings({
    annotate_cell_types(
      input = list("0" = list(genes = c("CD3D", "CD3E"))),
      tissue_name = "PBMC",
      model = "custom-model",
      api_key = "k"
    )
  },
  get_provider = function(model) "customx",
  resolve_provider_base_url = function(provider, base_urls) NULL,
  create_annotation_prompt = function(input, tissue_name, top_gene_count = 10) {
    list(prompt = "prompt", expected_count = 1L, gene_lists = list("0" = "CD3D, CD3E"))
  },
  process_custom = function(prompt, model, api_key) "custom-ok",
  custom_providers = custom_env,
  log_info = function(...) NULL,
  log_debug = function(...) NULL)

  expect_identical(result, "custom-ok")
})

test_that("get_model_response enforces the response contract for custom providers", {
  custom_env <- new.env(parent = emptyenv())
  assign("customx", TRUE, envir = custom_env)
  invalid_responses <- list(NULL, "", NA_character_, 123, "Error: failed")

  for (invalid_response in invalid_responses) {
    expect_error(
      testthat::with_mocked_bindings(
        get_model_response("prompt", "custom-model", "key"),
        get_provider = function(...) "customx",
        resolve_provider_base_url = function(...) NULL,
        process_custom = function(...) invalid_response,
        custom_providers = custom_env,
        log_debug = function(...) NULL
      ),
      "Invalid response format from provider 'customx'"
    )
  }
})

test_that("get_model_response normalizes built-in and custom provider output identically", {
  custom_env <- new.env(parent = emptyenv())
  assign("customx", TRUE, envir = custom_env)

  result <- testthat::with_mocked_bindings(
    get_model_response("prompt", "custom-model", "key"),
    get_provider = function(...) "customx",
    resolve_provider_base_url = function(...) NULL,
    process_custom = function(...) c("  Cluster 1: T cells,", "Cluster 2: B cells\n"),
    custom_providers = custom_env,
    log_debug = function(...) NULL
  )

  expect_identical(result, c("Cluster 1: T cells", "Cluster 2: B cells"))
})

test_that("process_custom forwards model_config when provider supports it", {
  skip_if_not_installed("testthat", minimum_version = "3.2.0")
  original_providers <- as.list(custom_providers)
  original_models <- as.list(custom_models)
  rm(list = ls(envir = custom_providers), envir = custom_providers)
  rm(list = ls(envir = custom_models), envir = custom_models)
  on.exit({
    rm(list = ls(envir = custom_providers), envir = custom_providers)
    rm(list = ls(envir = custom_models), envir = custom_models)
    list2env(original_providers, envir = custom_providers)
    list2env(original_models, envir = custom_models)
  }, add = TRUE)

  captured_config <- NULL
  assign("customx",
         list(
           process_fn = function(prompt, model, api_key, model_config) {
             captured_config <<- model_config
             "custom-ok"
           },
           description = NULL,
           models = "custom-model"
         ),
         envir = custom_providers)
  assign("custom-model",
         list(provider = "customx", config = list(temperature = 0.2, max_tokens = 100)),
         envir = custom_models)

  result <- testthat::with_mocked_bindings({
    suppressMessages(process_custom("prompt", "custom-model", "k"))
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      error = function(...) NULL
    )
  })

  expect_identical(result, "custom-ok")
  expect_identical(captured_config, list(temperature = 0.2, max_tokens = 100))

  captured_config <- NULL
  custom_providers$customx$process_fn <- function(prompt, model, api_key, ...) {
    captured_config <<- list(...)[["model_config"]]
    "custom-ok"
  }

  result <- testthat::with_mocked_bindings({
    suppressMessages(process_custom("prompt", "custom-model", "k"))
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      error = function(...) NULL
    )
  })

  expect_identical(result, "custom-ok")
  expect_identical(captured_config, list(temperature = 0.2, max_tokens = 100))

  captured_base_url <- NULL
  custom_providers$customx$process_fn <- function(prompt, model, api_key,
                                                   model_config, base_url) {
    captured_config <<- model_config
    captured_base_url <<- base_url
    "custom-ok"
  }

  result <- testthat::with_mocked_bindings({
    suppressMessages(process_custom(
      "prompt",
      "custom-model",
      "k",
      "https://custom.example.test/v1"
    ))
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      error = function(...) NULL
    )
  })

  expect_identical(result, "custom-ok")
  expect_identical(captured_config, list(temperature = 0.2, max_tokens = 100))
  expect_identical(captured_base_url, "https://custom.example.test/v1")

  custom_providers$customx$process_fn <- function(prompt, model, api_key) {
    stop_api_request_error("rate limited", status_code = 429L)
  }
  api_error <- tryCatch(
    suppressMessages(process_custom("prompt", "custom-model", "k")),
    error = identity
  )
  expect_s3_class(api_error, "mllm_api_error")
  expect_identical(api_error$status_code, 429L)
  expect_true(api_error$retryable)

  custom_providers$customx$process_fn <- function(prompt, model, api_key) "custom-ok"
  expect_error(
    suppressMessages(process_custom(
      "prompt",
      "custom-model",
      "k",
      "https://custom.example.test/v1"
    )),
    "does not accept base_url"
  )
})

test_that("custom provider identifiers are normalized and cannot shadow built-ins", {
  expect_error(
    register_custom_provider(" openai ", function(prompt, model, api_key) "x"),
    "reserved for a built-in provider"
  )
  expect_error(
    register_custom_provider("openrouter", function(prompt, model, api_key) "x"),
    "reserved for a built-in provider"
  )
  expect_error(
    register_custom_provider("   ", function(prompt, model, api_key) "x"),
    "provider_name must be a non-empty character scalar"
  )
  expect_error(
    register_custom_model(NA_character_, "customx"),
    "model_name must be a non-empty character scalar"
  )
  expect_error(
    register_custom_provider("../unsafe", function(prompt, model, api_key) "x"),
    "must contain only letters, digits"
  )
})

test_that("check_consensus normalizes discussion cell type labels", {
  result <- testthat::with_mocked_bindings({
    suppressMessages(check_consensus(
      list(
        m1 = "CELL TYPE: T cell",
        m2 = "Reasoning\nCELL TYPE: T cell"
      ),
      controversy_threshold = 1,
      entropy_threshold = 0
    ))
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      debug = function(...) NULL,
      warn = function(...) NULL,
      error = function(...) NULL
    )
  })

  expect_true(result$reached)
  expect_identical(result$majority_prediction, "T cell")
})

test_that("check_consensus rejects unparseable or missing responses", {
  result <- testthat::with_mocked_bindings({
    check_consensus(
      list(
        m1 = "Reasoning without a final label",
        m2 = "More reasoning\nwithout a final label",
        m3 = NA_character_
      )
    )
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      debug = function(...) NULL,
      warn = function(...) NULL,
      error = function(...) NULL
    )
  })

  expect_false(result$reached)
  expect_identical(result$majority_prediction, "Insufficient_Responses")
})


test_that("consensus prompt documents inclusive consensus threshold", {
  prompt <- create_consensus_check_prompt(
    c("T cell", "B cell"),
    controversy_threshold = 0.5,
    entropy_threshold = 1
  )

  expect_true(grepl("Consensus Proportion >= 0.5", prompt, fixed = TRUE))
  expect_true(grepl("Preserve biologically meaningful subtype or state qualifiers", prompt, fixed = TRUE))
  expect_true(grepl("do not merge labels when doing so would lose specificity", prompt, fixed = TRUE))
  expect_false(grepl("ignoring differences in.*Additional qualifiers", prompt))
})


test_that("parse_consensus_response handles NA_character_ safely", {
  result <- testthat::with_mocked_bindings({
    parse_consensus_response(NA_character_)
  },
  get_logger = function() {
    list(
      warn = function(...) NULL,
      info = function(...) NULL,
      debug = function(...) NULL,
      error = function(...) NULL
    )
  })

  expect_identical(result, list(
    reached = FALSE,
    consensus_proportion = 0,
    entropy = 0,
    majority_prediction = "Unknown"
  ))
})

test_that("print_consensus_summary handles vector predictions without crashing", {
  results <- list(
    final_annotations = list("0" = c("Consensus A", "Consensus B")),
    controversial_clusters = c("0"),
    discussion_logs = list(
      "0" = list(initial_predictions = list(m1 = c("Pred A", "Pred B")))
    )
  )

  expect_no_error(capture.output(print_consensus_summary(results)))
})

test_that("QwenProcessor caches endpoint selection per API key", {
  original_cache <- as.list(.qwen_endpoint_cache)
  rm(list = ls(envir = .qwen_endpoint_cache), envir = .qwen_endpoint_cache)
  on.exit({
    rm(list = ls(envir = .qwen_endpoint_cache), envir = .qwen_endpoint_cache)
    list2env(original_cache, envir = .qwen_endpoint_cache)
  }, add = TRUE)

  key_a <- digest::digest("key-a", algo = "xxhash64")
  key_b <- digest::digest("key-b", algo = "xxhash64")
  assign(key_a, "https://intl.example.test", envir = .qwen_endpoint_cache)
  assign(key_b, "https://domestic.example.test", envir = .qwen_endpoint_cache)

  processor <- QwenProcessor$new()

  expect_identical(processor$get_working_api_url("key-a"), "https://intl.example.test")
  expect_identical(processor$get_working_api_url("key-b"), "https://domestic.example.test")
})

test_that("QwenProcessor does not cache an unverified fallback endpoint", {
  original_cache <- as.list(.qwen_endpoint_cache)
  rm(list = ls(envir = .qwen_endpoint_cache), envir = .qwen_endpoint_cache)
  on.exit({
    rm(list = ls(envir = .qwen_endpoint_cache), envir = .qwen_endpoint_cache)
    list2env(original_cache, envir = .qwen_endpoint_cache)
  }, add = TRUE)

  testthat::with_mocked_bindings({
    ProbeFailingQwenProcessor <- R6::R6Class(
      "ProbeFailingQwenProcessor",
      inherit = QwenProcessor,
      public = list(probe_count = 0L),
      private = list(
        test_endpoint = function(url, api_key) {
          self$probe_count <- self$probe_count + 1L
          FALSE
        }
      )
    )

    processor <- ProbeFailingQwenProcessor$new()
    expected <- processor$get_default_api_url()
    expect_identical(processor$get_working_api_url("key-a"), expected)
    expect_identical(processor$get_working_api_url("key-a"), expected)
    expect_identical(processor$probe_count, 6L)
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      warn = function(...) NULL,
      debug = function(...) NULL,
      error = function(...) NULL
    )
  })
})

test_that("resolve_provider_base_url matches Python URL contract", {
  expect_null(resolve_provider_base_url("openai", NULL))
  expect_null(resolve_provider_base_url("openai", list(anthropic = "https://anthropic.example.test/v1")))

  expect_identical(
    resolve_provider_base_url("  OPENAI  ", "  https://proxy.example.test/v1/  "),
    "https://proxy.example.test/v1"
  )
  expect_identical(
    resolve_provider_base_url(
      "openai",
      list("  OpenAI  " = "https://openai-proxy.example.test/v1/chat/completions/")
    ),
    "https://openai-proxy.example.test/v1/chat/completions"
  )

  expect_error(resolve_provider_base_url("openai", "not-a-url"), "Invalid base URL")
  expect_error(resolve_provider_base_url("openai", "https://api.example.test/v1?token=abc"), "Invalid base URL")
  expect_error(resolve_provider_base_url("openai", list(openai = 123)), "must be a string URL")
  expect_error(resolve_provider_base_url("openai", c("https://a.example", "https://b.example")), "base_urls must be")
  expect_error(
    resolve_provider_base_url("openai", list("OpenAI" = "https://a.example", " openai " = "https://b.example")),
    "Ambiguous base_urls provider names"
  )
  expect_error(resolve_provider_base_url("openai", list("https://a.example")), "named list")
  expect_error(resolve_provider_base_url("openai", setNames(list("https://a.example"), "")), "non-empty strings")

  invalid_ports <- c(
    "https://api.example.test:invalid/v1",
    "https://api.example.test:0/v1",
    "https://api.example.test:99999/v1",
    "https://api.example.test:/v1"
  )
  for (url in invalid_ports) {
    expect_false(validate_base_url(url))
  }
  expect_true(validate_base_url("https://api.example.test:65535/v1"))
})

test_that("provider processors normalize and validate direct base URLs", {
  processor <- OpenAIProcessor$new("  https://proxy.example.test/v1/chat/completions/  ")
  expect_identical(
    processor$get_api_url(),
    "https://proxy.example.test/v1/chat/completions"
  )

  invalid_urls <- list(123, character(), NA_character_, "not-a-url")
  for (base_url in invalid_urls) {
    expect_error(OpenAIProcessor$new(base_url), "base_urls|Invalid base URL")
  }
})

test_that("retryable HTTP status policy rejects malformed status values", {
  expect_true(all(vapply(c(408, 425, 429, 500, 503), is_retryable_http_status, logical(1))))
  invalid_statuses <- list(NULL, TRUE, NA_real_, NaN, Inf, 500.5, 200, 302, 400, 404)
  expect_false(any(vapply(invalid_statuses, is_retryable_http_status, logical(1))))
})

test_that("provider default endpoints use current OpenAI-compatible chat APIs", {
  expect_identical(
    QwenProcessor$new()$get_default_api_url(),
    "https://dashscope-us.aliyuncs.com/compatible-mode/v1/chat/completions"
  )
  expect_identical(
    MinimaxProcessor$new()$get_default_api_url(),
    "https://api.minimax.io/v1/chat/completions"
  )
})


test_that("BaseAPIProcessor enforces and normalizes required string inputs", {
  TP <- R6::R6Class(
    "TP",
    inherit = BaseAPIProcessor,
    public = list(
      get_default_api_url = function() "x",
      make_api_call = function(chunk_content, model, api_key) {
        expect_identical(chunk_content, "prompt")
        expect_identical(model, "model")
        expect_identical(api_key, "key")
        list()
      },
      extract_response_content = function(response, model) "ok"
    )
  )

  p <- TP$new("tp")
  expect_error(
    p$process_request(c("a", "b"), "m", "k"),
    "Prompt is required but not provided"
  )
  expect_error(p$process_request(123, "m", "k"), "Prompt is required but not provided")
  expect_error(p$process_request("p", 123, "k"), "Model is required but not provided")
  expect_error(p$process_request("p", "m", 123), "API key is required but not provided")
  expect_identical(p$process_request("  prompt  ", "  model  ", "  key  "), "ok")
})

test_that("BaseAPIProcessor accepts only HTTP 200 responses", {
  testthat::with_mocked_bindings({
    StatusCheckingProcessor <- R6::R6Class(
      "StatusCheckingProcessor",
      inherit = BaseAPIProcessor,
      public = list(
        initialize = function() super$initialize("openai", NULL),
        get_default_api_url = function() "https://example.test",
        make_api_call = function(chunk_content, model, api_key) NULL,
        extract_response_content = function(response, model) NULL,
        check_status = function(status_code) {
          private$stop_for_http_error(
            make_test_json_response(list(message = "redirected"), status_code),
            "gpt-5.5"
          )
        }
      )
    )

    processor <- StatusCheckingProcessor$new()
    expect_no_error(processor$check_status(200L))
    redirect_error <- tryCatch(
      processor$check_status(302L),
      error = identity
    )
    expect_s3_class(redirect_error, "mllm_api_error")
    expect_identical(redirect_error$status_code, 302L)
    expect_false(redirect_error$retryable)
    expect_identical(
      conditionMessage(redirect_error),
      "OpenAI API request failed: redirected"
    )
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      debug = function(...) NULL,
      error = function(...) NULL
    )
  })
})

test_that("BaseAPIProcessor rejects empty, missing, and error text responses", {
  invalid_results <- list("", "   ", NA_character_, character(0), "Error: failed")

  for (invalid_result in invalid_results) {
    testthat::with_mocked_bindings({
      InvalidResponseProcessor <- R6::R6Class(
        "InvalidResponseProcessor",
        inherit = BaseAPIProcessor,
        public = list(
          initialize = function() super$initialize("test", NULL),
          get_default_api_url = function() "https://example.test",
          make_api_call = function(chunk_content, model, api_key) list(),
          extract_response_content = function(response, model) invalid_result
        )
      )

      expect_error(
        InvalidResponseProcessor$new()$process_request("prompt", "model", "key"),
        "Invalid response format from API"
      )
    },
    get_logger = function() {
      list(
        info = function(...) NULL,
        debug = function(...) NULL,
        error = function(...) NULL,
        log_api_call = function(...) NULL,
        log_api_request_response = function(...) NULL
      )
    })
  }
})

test_that("BaseAPIProcessor does not duplicate provider HTTP error context", {
  testthat::with_mocked_bindings({
    FailingHTTPProcessor <- R6::R6Class(
      "FailingHTTPProcessor",
      inherit = BaseAPIProcessor,
      public = list(
        initialize = function() super$initialize("openai", NULL),
        get_default_api_url = function() "https://example.test",
        make_api_call = function(chunk_content, model, api_key) {
          private$stop_for_http_error(
            make_test_json_response(list(error = "bad request"), 400L),
            model,
            "OpenAI"
          )
        },
        extract_response_content = function(response, model) "unused"
      )
    )

    error <- tryCatch(
      FailingHTTPProcessor$new()$process_request("prompt", "model", "key"),
      error = identity
    )
    expect_s3_class(error, "mllm_api_error")
    expect_identical(error$status_code, 400L)
    expect_false(error$retryable)
    expect_identical(
      conditionMessage(error),
      "OpenAI API request failed: bad request"
    )
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      debug = function(...) NULL,
      error = function(...) NULL,
      log_api_call = function(...) NULL,
      log_api_request_response = function(...) NULL
    )
  })
})

test_that("BaseAPIProcessor shared chat completion helpers keep canonical contract", {
  testthat::with_mocked_bindings({
    HelperProcessor <- R6::R6Class(
      "HelperProcessor",
      inherit = BaseAPIProcessor,
      public = list(
        initialize = function() super$initialize("openai", NULL),
        get_default_api_url = function() "https://example.test/v1/chat/completions",
        make_api_call = function(chunk_content, model, api_key) NULL,
        extract_response_content = function(response, model) NULL,
        build_body_for_test = function() {
          private$build_chat_completions_body(
            "genes",
            "gpt-5.5",
            extra = list(stream = FALSE)
          )
        },
        extract_content_for_test = function(payload) {
          private$extract_chat_completions_content(
            make_test_json_response(payload),
            "gpt-5.5"
          )
        }
      )
    )

    processor <- HelperProcessor$new()
    body <- processor$build_body_for_test()
    expect_identical(body$model, "gpt-5.5")
    expect_identical(body$messages[[1]], list(role = "user", content = "genes"))
    expect_identical(body$stream, FALSE)
    expect_identical(
      processor$extract_content_for_test(list(
        choices = list(list(message = list(content = "T cell\nB cell")))
      )),
      "T cell\nB cell"
    )
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      debug = function(...) NULL,
      error = function(...) NULL
    )
  })
})

test_that("BaseAPIProcessor extracts provider error messages from common shapes", {
  testthat::with_mocked_bindings({
    HelperProcessor <- R6::R6Class(
      "HelperProcessor",
      inherit = BaseAPIProcessor,
      public = list(
        initialize = function() super$initialize("openai", NULL),
        get_default_api_url = function() "https://example.test/v1/chat/completions",
        make_api_call = function(chunk_content, model, api_key) NULL,
        extract_response_content = function(response, model) NULL,
        error_for_test = function(payload) {
          private$extract_error_message(make_test_json_response(payload, status_code = 400L))
        }
      )
    )

    processor <- HelperProcessor$new()
    expect_identical(
      processor$error_for_test(list(error = list(message = "nested error"))),
      "nested error"
    )
    expect_identical(
      processor$error_for_test(list(error = "flat error")),
      "flat error"
    )
    expect_identical(
      processor$error_for_test(list(message = "top-level error")),
      "top-level error"
    )
    expect_identical(
      processor$error_for_test(list(error = list(message = NA_character_))),
      "HTTP 400 error"
    )
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      debug = function(...) NULL,
      error = function(...) NULL
    )
  })
})

test_that("MiniMax business error without status message is stable", {
  response <- make_test_json_response(
    list(base_resp = list(status_code = 1001, status_msg = NA_character_))
  )

  testthat::with_mocked_bindings({
    expect_error(
      MinimaxProcessor$new()$extract_response_content(response, "MiniMax-M2.7"),
      "MiniMax API error: Unknown MiniMax API error"
    )
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      debug = function(...) NULL,
      error = function(...) NULL
    )
  })
})

test_that("MiniMax strips think blocks from OpenAI-compatible content", {
  response <- make_test_json_response(
    list(
      choices = list(list(message = list(
        content = "<think>internal reasoning</think>\nCELL TYPE: T cell"
      )))
    )
  )

  testthat::with_mocked_bindings({
    result <- MinimaxProcessor$new()$extract_response_content(response, "MiniMax-M2.7")
    expect_identical(result, "CELL TYPE: T cell")
  },
  get_logger = function() {
    list(
      info = function(...) NULL,
      debug = function(...) NULL,
      error = function(...) NULL
    )
  })
})

test_that("get_provider validates model as non-empty scalar", {
  expect_identical(get_provider("  gpt-5.5  "), "openai")
  expect_identical(get_provider("  anthropic/claude-opus-4.7  "), "openrouter")
  expect_error(get_provider(character(0)), "model must be a non-empty character scalar")
  expect_error(get_provider(c("gpt-5.5", "grok-4.3")), "model must be a non-empty character scalar")
  expect_error(get_provider(NA_character_), "model must be a non-empty character scalar")
  expect_error(get_provider("o2-preview"), "Cannot determine provider")
  expect_error(get_provider("o9-preview"), "Cannot determine provider")
  expect_identical(get_provider("o1-preview"), "openai")
  expect_identical(get_provider("o3-mini"), "openai")
  expect_identical(get_provider("o4-mini"), "openai")
})

test_that("built-in provider registry owns detection and processor construction", {
  expect_setequal(
    names(.BUILTIN_PROVIDER_SPECS),
    c(names(.BUILTIN_PROVIDER_PATTERNS), "openrouter")
  )
  expect_silent(validate_builtin_provider_registry())

  processor <- new_builtin_provider_processor(
    "openai",
    "https://proxy.example.test/v1/chat/completions/"
  )
  expect_s3_class(processor, "OpenAIProcessor")
  expect_identical(processor$provider_name, "openai")
  expect_identical(
    processor$base_url,
    "https://proxy.example.test/v1/chat/completions"
  )
  expect_error(
    new_builtin_provider_processor("unsupported"),
    "Unsupported model provider"
  )
})

test_that("get_model_response validates and normalizes prompt text", {
  invalid_prompts <- list(NULL, character(0), c("a", "b"), NA_character_, "   ", 123)
  for (prompt in invalid_prompts) {
    expect_error(
      get_model_response(prompt, "gpt-5.5", "key"),
      "prompt must be a non-empty character scalar"
    )
  }

  result <- testthat::with_mocked_bindings({
    get_model_response("  prompt  ", "gpt-5.5", "key")
  },
  new_builtin_provider_processor = make_mock_builtin_processor_factory(
    function(prompt, model, api_key) {
      expect_identical(prompt, "prompt")
      "ok"
    }
  ))
  expect_identical(result, "ok")
})

test_that("annotate_cell_types validates public scalar controls", {
  base_args <- list(
    input = list("0" = list(genes = c("CD3D"))),
    tissue_name = "PBMC",
    model = "gpt-5.5",
    api_key = NA
  )

  expect_error(
    do.call(annotate_cell_types, modifyList(base_args, list(tissue_name = c("PBMC", "blood")))),
    "tissue_name is required"
  )
  expect_error(
    do.call(annotate_cell_types, modifyList(base_args, list(top_gene_count = 1.5))),
    "top_gene_count must be a positive integer"
  )
  expect_error(
    do.call(annotate_cell_types, modifyList(base_args, list(debug = NA))),
    "debug must be TRUE or FALSE"
  )
})

test_that("annotate_cell_types validates api_key scalar contract", {
  expect_error(
    annotate_cell_types(
      input = list("0" = list(genes = c("CD3D"))),
      tissue_name = "PBMC",
      model = "gpt-5.5",
      api_key = c("a", "b")
    ),
    "api_key must be a non-empty character scalar, or NA to return prompt only"
  )
  expect_error(
    annotate_cell_types(
      input = list("0" = list(genes = c("CD3D"))),
      tissue_name = "PBMC",
      model = "gpt-5.5",
      api_key = 123
    ),
    "api_key must be a non-empty character scalar, or NA to return prompt only"
  )
})

test_that("interactive_consensus_annotation validates api_keys list contract", {
  expect_error(
    interactive_consensus_annotation(
      input = list("0" = list(genes = c("CD3D"))),
      tissue_name = "PBMC",
      models = c("gpt-5.5", "grok-4.3"),
      api_keys = "bad",
      use_cache = FALSE
    ),
    "api_keys must be a named, non-empty list"
  )
  expect_error(
    interactive_consensus_annotation(
      input = list("0" = list(genes = c("CD3D"))),
      tissue_name = "PBMC",
      models = c("gpt-5.5", "grok-4.3"),
      api_keys = list(openai = c("first", "second"), grok = "key"),
      use_cache = FALSE
    ),
    "api_keys value for 'openai' must be a character scalar or missing"
  )
})

test_that("model identity normalization trims without rewriting case", {
  expect_identical(
    .normalize_model_vector(c(" MiniMax-M2.7 ", "minimax-m2.7")),
    c("MiniMax-M2.7", "minimax-m2.7")
  )

  cache_manager <- CacheManager$new("temp")
  input <- list("0" = "CD3D")
  expect_false(identical(
    cache_manager$generate_key(input, "MiniMax-M2.7", "0"),
    cache_manager$generate_key(input, "minimax-m2.7", "0")
  ))
})

test_that("multi-model entry points share strict control validation", {
  expect_error(
    compare_model_predictions(
      list("0" = c("G1")),
      "PBMC",
      models = c("gpt-5.5", "gpt-5.5"),
      api_keys = list(openai = "k")
    ),
    "models must not contain duplicate names"
  )
  expect_error(
    compare_model_predictions(
      list("0" = c("G1")),
      "PBMC",
      models = c("gpt-5.5", "grok-4.3"),
      api_keys = list(openai = "k", grok = "k"),
      consensus_threshold = 1.1
    ),
    "consensus_threshold must be a finite number between 0 and 1"
  )

  consensus_args <- list(
    input = list("0" = c("G1")),
    tissue_name = "PBMC",
    models = c("gpt-5.5", "grok-4.3"),
    api_keys = list(openai = "k", grok = "k"),
    use_cache = FALSE
  )
  expect_error(
    do.call(interactive_consensus_annotation, modifyList(consensus_args, list(use_cache = NA))),
    "use_cache must be TRUE or FALSE"
  )
  expect_error(
    do.call(interactive_consensus_annotation, modifyList(consensus_args, list(entropy_threshold = -1))),
    "entropy_threshold must be a finite non-negative number"
  )
  expect_error(
    do.call(interactive_consensus_annotation, modifyList(consensus_args, list(max_discussion_rounds = 1.5))),
    "max_discussion_rounds must be a positive integer"
  )
})

test_that("api key names reject ambiguous normalized duplicates", {
  ambiguous_provider_keys <- list("OpenAI" = "first", " openai " = "second")
  expect_error(
    .normalize_api_keys(ambiguous_provider_keys),
    "api_keys provider names must be unique after case/whitespace normalization"
  )
  expect_error(
    get_api_key(
      "gpt-5.5",
      ambiguous_provider_keys
    ),
    "api_keys provider names must be unique after case/whitespace normalization"
  )
})

test_that("model-specific api key names preserve model identity case", {
  api_keys <- list(
    "MiniMax-M2.7" = "canonical-key",
    "minimax-m2.7" = "different-key"
  )

  expect_identical(get_api_key("MiniMax-M2.7", api_keys), "canonical-key")
  expect_identical(get_api_key("minimax-m2.7", api_keys), "different-key")
  expect_null(get_api_key("MINIMAX-M2.7", api_keys))
})

test_that("get_api_key ignores empty/NA/non-scalar keys and falls back correctly", {
  expect_null(get_api_key("gpt-5.5", list(openai = "")))
  expect_null(get_api_key("gpt-5.5", list(openai = "   ")))
  expect_null(get_api_key("gpt-5.5", list(openai = NA_character_)))
  expect_null(get_api_key("gpt-5.5", list(openai = c("a", "b"))))
  expect_identical(get_api_key("gpt-5.5", list(openai = "", "gpt-5.5" = "k2")), "k2")
  expect_identical(get_api_key("gpt-5.5", list(openai = "  k1  ")), "k1")
  expect_identical(get_api_key("  gpt-5.5  ", list(" OpenAI " = "  k1  ")), "k1")
  expect_null(get_api_key("  gpt-5.5  ", list(" GPT-5.5 " = "  k2  ")))
  expect_identical(get_api_key("  GPT-5.5  ", list(" GPT-5.5 " = "  k2  ")), "k2")
})

test_that("consensus credentials use Gemini environment aliases in priority order", {
  variables <- c("GEMINI_API_KEY", "GOOGLE_API_KEY")
  restore_environment <- preserve_environment_variables(variables)
  on.exit(restore_environment(), add = TRUE)
  Sys.unsetenv(variables)
  Sys.setenv(GOOGLE_API_KEY = "  alias-key  ")

  alias_credential <- .resolve_model_api_key(
    "gemini-3.1-pro-preview",
    api_keys = NULL
  )
  expect_identical(alias_credential$api_key, "alias-key")
  expect_identical(alias_credential$provider, "gemini")
  expect_identical(alias_credential$source, "environment")

  observed_key <- NULL
  result <- testthat::with_mocked_bindings(
    execute_consensus_check(
      "prompt",
      api_keys = NULL,
      models_to_try = "gemini-3.1-pro-preview"
    ),
    get_model_response = function(prompt, model, api_key, base_urls = NULL) {
      observed_key <<- api_key
      "1\n1\n0\nT cell"
    }
  )
  expect_true(result$success)
  expect_identical(observed_key, "alias-key")

  Sys.setenv(GEMINI_API_KEY = "  primary-key  ")
  primary_credential <- .resolve_model_api_key(
    "gemini-3.1-pro-preview",
    api_keys = NULL
  )
  expect_identical(primary_credential$api_key, "primary-key")

  explicit_credential <- .resolve_model_api_key(
    "gemini-3.1-pro-preview",
    api_keys = list(gemini = "  explicit-key  ")
  )
  expect_identical(explicit_credential$api_key, "explicit-key")
  expect_identical(explicit_credential$source, "api_keys")
})

test_that("discussion cache context uses the resolved credential source", {
  variables <- c("GEMINI_API_KEY", "GOOGLE_API_KEY")
  restore_environment <- preserve_environment_variables(variables)
  on.exit(restore_environment(), add = TRUE)
  Sys.unsetenv(variables)
  Sys.setenv(GOOGLE_API_KEY = "alias-key")

  context <- build_discussion_cache_context(
    input = list("0" = "CD3D"),
    cluster_id = "0",
    models = "gemini-3.1-pro-preview",
    api_keys = NULL,
    initial_predictions = list(
      "gemini-3.1-pro-preview" = "Cluster 0: T cell"
    ),
    max_discussion_rounds = 2,
    controversy_threshold = 0.7,
    entropy_threshold = 1,
    consensus_check_model = "gemini-3.1-pro-preview",
    base_urls = list(gemini = "https://proxy.example.test/v1/")
  )

  expect_identical(
    context$requests[[1]][c("provider", "credential_source", "base_url")],
    list(
      provider = "gemini",
      credential_source = "environment",
      base_url = "https://proxy.example.test/v1"
    )
  )
})

test_that("model and api_key are trimmed before R provider dispatch", {
  direct_result <- testthat::with_mocked_bindings({
    get_model_response("prompt", "  gpt-5.5  ", "  key  ")
  },
  new_builtin_provider_processor = make_mock_builtin_processor_factory(
    function(prompt, model, api_key) {
      expect_identical(model, "gpt-5.5")
      expect_identical(api_key, "key")
      "ok"
    }
  ))
  expect_identical(direct_result, "ok")

  public_result <- testthat::with_mocked_bindings({
    suppressMessages(annotate_cell_types(
      input = list("0" = list(genes = c("CD3D"))),
      tissue_name = "PBMC",
      model = "  gpt-5.5  ",
      api_key = "  key  "
    ))
  },
  get_model_response = function(prompt, model, api_key, base_urls = NULL, normalize = TRUE) {
    expect_identical(model, "gpt-5.5")
    expect_identical(api_key, "key")
    c("T cell")
  })
  expect_identical(public_result, c("T cell"))
})

test_that("facilitate_cluster_discussion uses last available consensus on early break", {
  logger_env <- new.env(parent = emptyenv())
  logger_env$events <- list()
  logger <- list(
    log_discussion = function(cluster_id, event_type, data = NULL) {
      logger_env$events <- c(logger_env$events, list(list(event = event_type, data = data)))
    }
  )

  testthat::with_mocked_bindings({
    suppressMessages(facilitate_cluster_discussion(
      cluster_id = "0",
      input = list("0" = list(genes = c("G1"))),
      tissue_name = "PBMC",
      models = c("m1", "m2"),
      api_keys = list(openai = "k"),
      initial_predictions = list(m1 = c("0:T"), m2 = c("0:T")),
      top_gene_count = 5,
      max_rounds = 3
    ))
  },
  create_initial_discussion_prompt = function(...) "p1",
  create_discussion_prompt = function(...) "p2",
  get_model_response = local({
    cnt <- 0
    function(...) {
      cnt <<- cnt + 1
      if (cnt <= 2) "CELL TYPE: T cell" else if (cnt == 3) "CELL TYPE: T cell" else "Error: failed"
    }
  }),
  check_consensus = local({
    k <- 0
    function(...) {
      k <<- k + 1
      list(reached = FALSE, consensus_proportion = 0.5, entropy = 1.2, majority_prediction = paste0("M", k))
    }
  }),
  get_api_key = function(...) "k",
  get_provider = function(...) "openai",
  get_logger = function() logger,
  log_warn = function(...) NULL,
  log_info = function(...) NULL)

  end_events <- Filter(function(x) identical(x$event, "end"), logger_env$events)
  expect_equal(length(end_events), 1)
  expect_identical(end_events[[1]]$data$final_result, "M1")
})

test_that("process_controversial_clusters backtracks final prediction when last round lacks consensus", {
  out <- testthat::with_mocked_bindings({
    process_controversial_clusters(
      controversial_clusters = c("0"),
      input = list("0" = list(genes = c("G1"))),
      tissue_name = "PBMC",
      successful_models = c("m1", "m2"),
      api_keys = list(openai = "k"),
      individual_predictions = list(m1 = c("0:T"), m2 = c("0:T")),
      top_gene_count = 5,
      controversy_threshold = 0.7,
      entropy_threshold = 1.0,
      max_discussion_rounds = 3,
      cache_manager = list(
        generate_key = function(...) "k",
        save_to_cache = function(...) NULL,
        has_cache = function(...) FALSE
      ),
      use_cache = FALSE
    )
  },
  facilitate_cluster_discussion = function(...) {
    list(
      cluster_id = "0",
      rounds = list(
        list(consensus_result = list(majority_prediction = "T cell")),
        list(responses = list(m1 = "..."))
      )
    )
  },
  log_info = function(...) NULL,
  log_warn = function(...) NULL)

  expect_identical(out$final_annotations[["0"]], "T cell")
})

test_that("process_controversial_clusters ignores invalid cache payloads", {
  discussion_calls <- 0L
  result <- testthat::with_mocked_bindings(
    suppressMessages(process_controversial_clusters(
      controversial_clusters = "0",
      input = list("0" = "G1"),
      tissue_name = "PBMC",
      successful_models = c("m1", "m2"),
      api_keys = list(openai = "key"),
      individual_predictions = list(m1 = "T cell", m2 = "T cell"),
      top_gene_count = 5,
      controversy_threshold = 0.7,
      entropy_threshold = 1,
      max_discussion_rounds = 3,
      cache_manager = list(
        generate_key = function(...) "key",
        load_from_cache = function(...) list(
          annotation = c("T cell", "B cell"),
          discussion_log = list(rounds = list())
        ),
        save_to_cache = function(...) TRUE
      ),
      use_cache = TRUE
    )),
    build_discussion_cache_context = function(...) list(),
    facilitate_cluster_discussion = function(...) {
      discussion_calls <<- discussion_calls + 1L
      list(
        cluster_id = "0",
        rounds = list(list(consensus_result = list(
          majority_prediction = "T cell"
        )))
      )
    },
    log_info = function(...) NULL,
    log_warn = function(...) NULL,
    log_debug = function(...) NULL
  )

  expect_identical(discussion_calls, 1L)
  expect_identical(result$final_annotations[["0"]], "T cell")
})

test_that("check_consensus rejects an LLM indicator that contradicts its metrics", {
  result <- testthat::with_mocked_bindings(
    check_consensus(
      round_responses = c(model_a = "T cell", model_b = "NK cell"),
      api_keys = list(openai = "key"),
      controversy_threshold = 0.7,
      entropy_threshold = 0.8,
      consensus_check_model = "gpt-5.5"
    ),
    execute_consensus_check = function(...) {
      list(success = TRUE, response = "1\n0.5\n1.0\nT cell")
    }
  )

  expect_false(result$reached)
  expect_equal(result$consensus_proportion, 0.5)
  expect_equal(result$entropy, 1)
})

test_that("consensus filtering uses the canonical unknown-annotation contract", {
  invalid_annotations <- c(
    "Unknown",
    "unknown (low confidence)",
    "[Inconclusive]",
    "error: upstream timeout",
    "N/A",
    "--"
  )
  expect_false(any(vapply(
    as.list(invalid_annotations),
    is_real_cell_type_annotation,
    logical(1)
  )))
  expect_true(is_real_cell_type_annotation("CD4+ T cell"))
  expect_identical(clean_annotation("unknown (low confidence)"), "Unknown")

  result <- testthat::with_mocked_bindings(
    suppressMessages(check_consensus(
      round_responses = c(first_model = "unknown", second_model = "UNKNOWN")
    )),
    execute_consensus_check = function(...) {
      stop("unknown annotations must be filtered before an LLM request")
    }
  )

  expect_false(result$reached)
  expect_identical(result$majority_prediction, "Insufficient_Responses")
})

test_that("check_consensus does not expose an invalid LLM majority label", {
  result <- testthat::with_mocked_bindings(
    suppressMessages(check_consensus(
      round_responses = c(first_model = "T cell", second_model = "B cell"),
      api_keys = list(openai = "key"),
      controversy_threshold = 0.5,
      entropy_threshold = 1,
      consensus_check_model = "gpt-5.5"
    )),
    execute_consensus_check = function(...) {
      list(
        success = TRUE,
        response = "1\n0.5\n1\nunknown (low confidence)"
      )
    }
  )

  expect_false(result$reached)
  expect_identical(result$majority_prediction, "T cell")
  expect_equal(result$consensus_proportion, 0.5)
  expect_equal(result$entropy, 1)
})

test_that("check_consensus rejects metrics that cannot arise from the model votes", {
  result <- testthat::with_mocked_bindings(
    suppressMessages(check_consensus(
      round_responses = c(model_a = "T cell", model_b = "NK cell"),
      api_keys = list(openai = "key"),
      controversy_threshold = 0.5,
      entropy_threshold = 1,
      consensus_check_model = "gpt-5.5"
    )),
    execute_consensus_check = function(...) {
      list(success = TRUE, response = "1\n0.9\n0.2\nT cell")
    }
  )

  expect_false(result$reached)
  expect_identical(result$majority_prediction, "T cell")
  expect_equal(result$consensus_proportion, 0.5)
  expect_equal(result$entropy, 1)
})

test_that("a tied vote is not deterministic consensus at a 0.5 threshold", {
  result <- testthat::with_mocked_bindings(
    suppressMessages(check_consensus(
      round_responses = c(first_model = "B cell", second_model = "T cell"),
      api_keys = list(openai = "key"),
      controversy_threshold = 0.5,
      entropy_threshold = 1
    )),
    execute_consensus_check = function(...) {
      list(success = FALSE, response = NULL)
    }
  )

  expect_false(result$reached)
  expect_identical(result$majority_prediction, "B cell")
  expect_equal(result$consensus_proportion, 0.5)
})

test_that("consensus response parsing rejects non-text and ignores blank lines", {
  expect_identical(
    parse_consensus_response(list("1", "1", "0", "T cell")),
    .DEFAULT_CONSENSUS_RESULT
  )
  expect_identical(
    parse_consensus_response("1\n  \n1\n0\nT cell"),
    list(
      reached = TRUE,
      consensus_proportion = 1,
      entropy = 0,
      majority_prediction = "T cell"
    )
  )
})

test_that("flexible consensus parsing requires complete metrics and extracts labels", {
  structured <- parse_consensus_response(paste(
    "Consensus: 1",
    "Consensus Proportion: 0.5",
    "Entropy: 1",
    "Majority Prediction: T cell",
    sep = "\n"
  ))
  expect_identical(structured, list(
    reached = TRUE,
    consensus_proportion = 0.5,
    entropy = 1,
    majority_prediction = "T cell"
  ))

  negative <- parse_consensus_response(c(
    "1",
    "Consensus Proportion = -0.5",
    "Entropy = 1",
    "T cell"
  ))
  oversized <- parse_consensus_response(c(
    "1",
    "Consensus Proportion = 1.5",
    "Entropy = 1",
    "T cell"
  ))
  expect_identical(negative$consensus_proportion, 0)
  expect_identical(oversized$consensus_proportion, 0)
})

test_that("prepare_models_list only adds valid direct-provider fallbacks", {
  openai_models <- prepare_models_list("openai/gpt-5.5")
  unsupported_models <- prepare_models_list("meta-llama/llama-3.3-70b")

  expect_identical(openai_models[1:2], c("openai/gpt-5.5", "gpt-5.5"))
  expect_identical(unsupported_models[[1]], "meta-llama/llama-3.3-70b")
  expect_false("llama-3.3-70b" %in% unsupported_models)
})

test_that("model requests retry only transient API failures", {
  run_failure <- function(status_code) {
    state <- new.env(parent = emptyenv())
    state$calls <- 0L
    state$waits <- numeric(0)
    error <- tryCatch(testthat::with_mocked_bindings(
      get_model_response("prompt", "gpt-5.5", "key"),
      new_builtin_provider_processor = make_mock_builtin_processor_factory(
        function(...) {
          state$calls <- state$calls + 1L
          stop_api_request_error("request failed", status_code = status_code)
        }
      ),
      wait_before_model_retry = function(seconds) {
        state$waits <- c(state$waits, seconds)
      }
    ), error = identity)
    list(error = error, calls = state$calls, waits = state$waits)
  }

  permanent <- run_failure(401L)
  transient <- run_failure(429L)

  expect_s3_class(permanent$error, "mllm_api_error")
  expect_identical(permanent$calls, 1L)
  expect_length(permanent$waits, 0)
  expect_s3_class(transient$error, "mllm_api_error")
  expect_identical(transient$calls, 3L)
  expect_identical(transient$waits, c(5, 10))
})

test_that("model request retry returns the first successful transient recovery", {
  calls <- 0L
  waits <- numeric(0)

  result <- testthat::with_mocked_bindings(
    get_model_response("prompt", "gpt-5.5", "key"),
    new_builtin_provider_processor = make_mock_builtin_processor_factory(
      function(...) {
        calls <<- calls + 1L
        if (calls == 1L) {
          stop_api_request_error("rate limited", status_code = 429L)
        }
        "T cell"
      }
    ),
    wait_before_model_retry = function(seconds) {
      waits <<- c(waits, seconds)
    }
  )

  expect_identical(result, "T cell")
  expect_identical(calls, 2L)
  expect_identical(waits, 5)
})

test_that("consensus candidate orchestration does not add another retry layer", {
  calls <- 0L
  result <- suppressWarnings(testthat::with_mocked_bindings(
    execute_consensus_check(
      "prompt",
      list(openai = "key"),
      "gpt-5.5"
    ),
    get_model_response = function(...) {
      calls <<- calls + 1L
      stop_api_request_error("request failed", status_code = 429L)
    }
  ))

  expect_false(result$success)
  expect_identical(calls, 1L)
})

test_that("facilitate_cluster_discussion logs consensus and end on first-round early return", {
  logger_env <- new.env(parent = emptyenv())
  logger_env$events <- list()
  logger <- list(
    log_discussion = function(cluster_id, event_type, data = NULL) {
      logger_env$events <- c(logger_env$events, list(list(event = event_type, data = data)))
    }
  )

  testthat::with_mocked_bindings({
    facilitate_cluster_discussion(
      cluster_id = "0",
      input = list("0" = list(genes = c("G1"))),
      tissue_name = "PBMC",
      models = c("m1", "m2"),
      api_keys = list(openai = "k"),
      initial_predictions = list(m1 = c("0:T"), m2 = c("0:T")),
      top_gene_count = 5,
      max_rounds = 3
    )
  },
  create_initial_discussion_prompt = function(...) "p1",
  create_discussion_prompt = function(...) "p2",
  get_model_response = function(...) "Error: failed",
  check_consensus = function(...) list(reached = FALSE, consensus_proportion = 0, entropy = 0, majority_prediction = "Unknown"),
  get_api_key = function(...) "k",
  get_provider = function(...) "openai",
  get_logger = function() logger,
  log_warn = function(...) NULL,
  log_info = function(...) NULL)

  event_names <- vapply(logger_env$events, function(x) x$event, character(1))
  expect_true("consensus" %in% event_names)
  expect_true("end" %in% event_names)
})

test_that("BaseAPIProcessor marks semantic error result as failed API call", {
  logger_env <- new.env(parent = emptyenv())
  logger_env$api_call_success <- logical(0)
  logger_env$failed_audit_logs <- 0L

  logger <- list(
    info = function(...) NULL,
    debug = function(...) NULL,
    error = function(...) NULL,
    log_api_call = function(provider, model, duration, success = TRUE, tokens = NULL) {
      logger_env$api_call_success <- c(logger_env$api_call_success, success)
    },
    log_api_request_response = function(...) {
      logger_env$failed_audit_logs <- logger_env$failed_audit_logs + 1L
    }
  )

  testthat::with_mocked_bindings({
    TestProcessor <- R6::R6Class("TestProcessor",
      inherit = BaseAPIProcessor,
      public = list(
        initialize = function() super$initialize("testp", NULL),
        get_default_api_url = function() "https://example.test",
        make_api_call = function(chunk_content, model, api_key) stop("simulated failure"),
        extract_response_content = function(response, model) "unused"
      )
    )

    proc <- TestProcessor$new()
    expect_error(proc$process_request("prompt", "test-model", "test-key"),
                 "simulated failure")
  },
  get_logger = function() logger)

  expect_true(length(logger_env$api_call_success) >= 1)
  expect_identical(tail(logger_env$api_call_success, 1), FALSE)
  # Failed requests should still produce audit logs for traceability
  expect_true(logger_env$failed_audit_logs >= 1L)
})

test_that("initialize_logger preserves logger configuration and regenerates unique session IDs", {
  initialize_logger("logs")
  configure_logger(level = "DEBUG", console_output = FALSE, json_format = FALSE)
  logger_cfg <- get_logger()
  logger_cfg$max_log_size <- 42
  logger_cfg$max_log_files <- 7

  initialize_logger("logs")
  logger_after_reinit <- get_logger()
  expect_identical(logger_after_reinit$log_level, "DEBUG")
  expect_identical(logger_after_reinit$enable_console, FALSE)
  expect_identical(logger_after_reinit$enable_json, FALSE)
  expect_identical(logger_after_reinit$max_log_size, 42)
  expect_equal(logger_after_reinit$max_log_files, 7)

  id1 <- logger_after_reinit$session_id
  initialize_logger("logs")
  id2 <- get_logger()$session_id
  expect_false(identical(id1, id2))
})

test_that("UnifiedLogger validates configuration before mutating state", {
  log_dir <- tempfile("mllm-logger-")
  logger <- UnifiedLogger$new(base_dir = log_dir, level = "INFO")

  expect_error(logger$set_level("verbose"), "level must be one of")
  expect_identical(logger$log_level, "INFO")
  expect_error(
    UnifiedLogger$new(base_dir = log_dir, max_files = 0),
    "max_files must be a positive integer"
  )
  expect_error(
    UnifiedLogger$new(base_dir = log_dir, console_output = NA),
    "console_output must be TRUE or FALSE"
  )
})

test_that("UnifiedLogger counts cache misses once and sanitizes discussion paths", {
  restore_check_environment <- enable_logger_file_writes()
  on.exit(restore_check_environment(), add = TRUE)
  log_dir <- tempfile("mllm-logger-")
  logger <- UnifiedLogger$new(base_dir = log_dir)

  logger$log_cache_operation("miss", "key")
  expect_identical(logger$performance_stats$cache_misses, 1)

  logger$log_discussion(
    "../../outside",
    "start",
    list(tissue_name = "PBMC", marker_genes = "G1")
  )
  created_files <- list.files(log_dir, recursive = TRUE, full.names = TRUE)
  expect_true(any(grepl("cluster_.*outside_discussion\\.md$", created_files)))
  expect_true(all(startsWith(normalizePath(created_files), normalizePath(log_dir))))
})

test_that("UnifiedLogger force cleanup is stable for fewer than max files", {
  restore_check_environment <- enable_logger_file_writes()
  on.exit(restore_check_environment(), add = TRUE)
  log_dir <- tempfile("mllm-logger-")
  logger <- UnifiedLogger$new(base_dir = log_dir, max_files = 5)
  expect_true(length(list.files(log_dir, pattern = "^mllm_.*\\.log$")) >= 1)

  expect_no_error(logger$cleanup_logs(force = TRUE))
  expect_length(list.files(log_dir, pattern = "^mllm_.*\\.log$"), 0)
})

test_that("UnifiedLogger file failures do not change business control flow", {
  restore_check_environment <- enable_logger_file_writes()
  on.exit(restore_check_environment(), add = TRUE)
  log_dir <- tempfile("mllm-logger-")
  logger <- UnifiedLogger$new(base_dir = log_dir)
  blocked_path <- tempfile("mllm-log-file-")
  writeLines("not a directory", blocked_path)
  logger$log_dir <- blocked_path

  expect_no_error(logger$info("non-fatal"))
  expect_null(logger$log_model_response("openai", "gpt-5.5", "T cell"))
  expect_null(logger$log_api_request_response(
    "../provider",
    "gpt-5.5",
    "prompt",
    "T cell"
  ))
  expect_null(logger$log_discussion(
    "../../cluster",
    "start",
    list(tissue_name = "PBMC", marker_genes = "G1")
  ))
})

# MiniMax reports business errors via HTTP 200 + base_resp$status_code. The
# codes below are transient per the MiniMax docs (1000 unknown error, 1001
# timeout, 1002 rate limit, 1039 token limit) and should be retried; all other
# non-zero codes (1004 auth, 1008 balance, 1027 content policy, 2013 invalid
# params) are fatal.
.MINIMAX_TRANSIENT_STATUS_CODES <- c(1000L, 1001L, 1002L, 1039L)

#' Minimax API Processor
#'
#' Concrete implementation of BaseAPIProcessor for Minimax models.
#' Handles Minimax-specific API calls, authentication, and response parsing.
#'
#' @export
MinimaxProcessor <- R6::R6Class("MinimaxProcessor",
  inherit = BaseAPIProcessor,
  
  public = list(
    #' @description
    #' Initialize Minimax processor
    #' @param base_url Optional custom API endpoint
    initialize = function(base_url = NULL) {
      super$initialize("minimax", base_url)
    },

    #' @description
    #' Get default MiniMax OpenAI-compatible chat completions API URL
    #
    get_default_api_url = function() {
      return("https://api.minimax.io/v1/chat/completions")
    },
    
    #' @description
    #' Make API call to Minimax
    #' @param chunk_content Prompt text to send
    #' @param model Model identifier
    #' @param api_key MiniMax API key
    make_api_call = function(chunk_content, model, api_key) {
      private$post_chat_completions_request(chunk_content, model, api_key)
    },
    
    #' @description
    #' Extract response content from Minimax API response
    #' @param response HTTP response object
    #' @param model Model identifier
    extract_response_content = function(response, model) {
      self$logger$debug("Parsing MiniMax API response",
                       list(provider = self$provider_name, model = model))
      
      # Parse the response
      content <- httr::content(response, "parsed")
      
      base_resp <- content$base_resp
      if (is.list(base_resp) &&
          !is.null(base_resp$status_code) &&
          !identical(as.integer(base_resp$status_code), 0L)) {
        status_msg_available <- is.character(base_resp$status_msg) &&
          length(base_resp$status_msg) == 1 &&
          !is.na(base_resp$status_msg) &&
          nzchar(base_resp$status_msg)
        error_message <- if (status_msg_available) {
          base_resp$status_msg
        } else {
          "Unknown MiniMax API error"
        }
        status_code_int <- as.integer(base_resp$status_code)
        stop_api_request_error(
          sprintf("MiniMax API error: %s", error_message),
          retryable = status_code_int %in% .MINIMAX_TRANSIENT_STATUS_CODES
        )
      }

      if (!is.null(content$choices) &&
          length(content$choices) > 0 &&
          !is.null(content$choices[[1]]$message$content)) {
        response_content <- content$choices[[1]]$message$content
      } else if (!is.null(content$choices) &&
                 length(content$choices) > 0 &&
                 !is.null(content$choices[[1]]$messages) &&
                 length(content$choices[[1]]$messages) > 0 &&
                 !is.null(content$choices[[1]]$messages[[1]]$text)) {
        # Legacy MiniMax text/chatcompletion_v2 shape.
        response_content <- content$choices[[1]]$messages[[1]]$text
      } else {
        
        self$logger$error("Unexpected response format from MiniMax API",
                         list(provider = self$provider_name,
                              model = model,
                              content_structure = names(content),
                              choices_available = !is.null(content$choices),
                              choices_count = if (!is.null(content$choices)) {
                                length(content$choices)
                              } else {
                                0
                              }))
        
        stop("Unexpected response format from MiniMax API")
      }

      response_content <- gsub(
        "<think>[\\s\\S]*?</think>\\s*",
        "",
        response_content,
        perl = TRUE
      )

      return(response_content)
    }
  )
)

is_valid_consensus_cache_data <- function(data) {
  if (!is.list(data) ||
      !all(c("annotation", "discussion_log") %in% names(data))) {
    return(FALSE)
  }
  annotation_is_valid <- is.character(data$annotation) &&
    length(data$annotation) == 1 &&
    !is.na(data$annotation) &&
    nzchar(trimws(data$annotation))
  discussion_is_valid <- is.list(data$discussion_log) &&
    "rounds" %in% names(data$discussion_log) &&
    is.list(data$discussion_log$rounds)
  annotation_is_valid && discussion_is_valid
}

#' Cache Manager Class
#' @description Manages caching of consensus analysis results
#' @export
CacheManager <- R6::R6Class(
  "CacheManager",
  
  public = list(
    #' @field cache_dir Directory to store cache files. Options:
    #'   - NULL (default): Uses system cache directory
    #'   - "local": Uses .mllmcelltype_cache in current directory
    #'   - "temp": Uses temporary directory
    #'   - Custom path: Any other string is used as directory path
    cache_dir = NULL,
    
    #' @field cache_version Current cache version
    cache_version = "1.2",
    
    #' @description Initialize cache manager
    #
    #'   - NULL (default): Uses system cache directory via \code{tools::R_user_dir()}
    #'   - "local": Uses .mllmcelltype_cache in current directory  
    #'   - "temp": Uses temporary directory (cleared on R restart)
    #'   - Custom path: Any other string is used as directory path
    #' @param cache_dir Cache directory selector or custom path
    initialize = function(cache_dir = NULL) {
      if (is.null(cache_dir)) {
        # Default: use system cache directory
        cache_dir <- file.path(
          tools::R_user_dir("mLLMCelltype", which = "cache"),
          "consensus_cache"
        )
      } else {
        cache_dir <- .normalize_required_string(cache_dir, "cache_dir")
      }

      if (identical(cache_dir, "local")) {
        # Special value: use project local cache
        cache_dir <- file.path(".", ".mllmcelltype_cache")
      } else if (identical(cache_dir, "temp")) {
        # Special value: use temporary directory
        cache_dir <- file.path(tempdir(), "mllmcelltype_cache")
      }
      # Other values: directly use user specified path
      
      self$cache_dir <- cache_dir
      
      if (!dir.exists(cache_dir)) {
        created <- dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
        if (!created && !dir.exists(cache_dir)) {
          stop("Failed to create cache directory: ", cache_dir)
        }
      }
    },
    
    #' @description Get actual cache directory path
    #
    get_cache_dir = function() {
      return(self$cache_dir)
    },
    
    #' @description Generate cache key from input parameters (improved version)
    #' @param input Marker gene input
    #' @param models Model identifiers contributing to the result
    #' @param cluster_id Cluster identifier
    #' @param tissue_name Tissue context included in the cache key
    #' @param top_gene_count Number of marker genes used
    #' @param discussion_context Optional normalized inputs that affect a discussion
    generate_key = function(input, models, cluster_id, tissue_name = "", top_gene_count = 10,
                            discussion_context = NULL) {
      models <- .normalize_model_vector(models)
      cluster_id <- .normalize_required_string(as.character(cluster_id), "cluster_id")
      top_gene_count <- .normalize_top_gene_count(top_gene_count)
      if (!is.character(tissue_name) || length(tissue_name) != 1 || is.na(tissue_name)) {
        stop("tissue_name must be a character scalar")
      }
      tissue_name <- trimws(tissue_name)
      if (!is.null(discussion_context) && !is.list(discussion_context)) {
        stop("discussion_context must be NULL or a list")
      }

      genes <- select_cluster_marker_genes(
        input,
        cluster_id,
        top_gene_count
      )

      # Create standardized components with input context for empty genes
      genes_hash <- private$create_genes_hash(genes)
      models_hash <- private$create_models_hash(models)
      cluster_hash <- private$create_cluster_hash(cluster_id)
      context_hash <- private$create_context_hash(
        tissue_name,
        top_gene_count,
        discussion_context
      )

      # Combine into final key with version prefix
      key <- paste("v", self$cache_version, genes_hash, models_hash, cluster_hash, context_hash, sep = "_")

      return(key)
    },
    
    #' @description Save results to cache
    #' @param key Valid cache key
    #' @param data Result object to persist
    save_to_cache = function(key, data) {
      cache_file <- private$cache_file_path(key)

      # Ensure cache directory exists
      if (!dir.exists(self$cache_dir)) {
        created <- dir.create(self$cache_dir, recursive = TRUE, showWarnings = FALSE)
        if (!created && !dir.exists(self$cache_dir)) {
          warning("Failed to create cache directory: ", self$cache_dir, call. = FALSE)
          return(invisible(FALSE))
        }
      }

      saved <- tryCatch({
        private$atomic_save_rds(data, cache_file)
        get_logger()$log_cache_operation("store", key, file.size(cache_file))
        TRUE
      }, error = function(e) {
        get_logger()$log_cache_operation("store_failed", key, NULL)
        warning(paste("Failed to save cache file:", e$message), call. = FALSE)
        FALSE
      })
      invisible(saved)
    },
    
    #' @description Load results from cache
    #' @param key Valid cache key
    load_from_cache = function(key) {
      cache_file <- private$cache_file_path(key)
      if (!file.exists(cache_file) || dir.exists(cache_file)) {
        get_logger()$log_cache_operation("miss", key)
        return(NULL)
      }

      tryCatch({
        data <- readRDS(cache_file)
        get_logger()$log_cache_operation("hit", key, file.size(cache_file))
        data
      }, error = function(e) {
        get_logger()$log_cache_operation("load_failed", key, file.size(cache_file))
        warning(paste("Failed to load cache file:", e$message), call. = FALSE)
        NULL
      })
    },
    
    #' @description Check if results exist in cache
    #' @param key Valid cache key
    has_cache = function(key) {
      cache_file <- private$cache_file_path(key)
      return(file.exists(cache_file) && !dir.exists(cache_file))
    },
    
    #' @description Get cache statistics
    #
    get_cache_stats = function() {
      if (!dir.exists(self$cache_dir)) {
        return(list(
          cache_exists = FALSE,
          cache_count = 0,
          cache_size_mb = 0,
          cache_files = character()
        ))
      }
      
      cache_files <- private$list_cache_files()
      cache_sizes <- file.size(cache_files)
      
      return(list(
        cache_exists = TRUE,
        cache_count = length(cache_files),
        cache_size_mb = sum(cache_sizes, na.rm = TRUE) / (1024 * 1024),
        cache_files = cache_files
      ))
    },
    
    #' @description Clear all cache
    #' @param confirm Whether deletion is explicitly confirmed
    clear_cache = function(confirm = FALSE) {
      confirm <- .normalize_flag(confirm, "confirm")
      if (!dir.exists(self$cache_dir)) {
        message("Cache directory does not exist.")
        return(invisible(NULL))
      }
      
      cache_files <- private$list_cache_files()
      
      if (length(cache_files) == 0) {
        message("No cache files to clear.")
        return(invisible(NULL))
      }
      
      if (!confirm) {
        message(sprintf("This will delete %d cache files. Set confirm=TRUE to proceed.", length(cache_files)))
        return(invisible(NULL))
      }
      
      unlink(cache_files)
      removed_count <- sum(!file.exists(cache_files))
      get_logger()$log_cache_operation("clear", "all_cache", removed_count)
      message(sprintf("Cleared %d cache files.", removed_count))
      if (removed_count < length(cache_files)) {
        warning(sprintf("Failed to remove %d cache files.", length(cache_files) - removed_count),
                call. = FALSE)
      }
      
      return(invisible(NULL))
    },
    
    #' @description Validate cache content
    #' @param key Valid cache key
    validate_cache = function(key) {
      if (!self$has_cache(key)) {
        return(FALSE)
      }
      
      tryCatch({
        cache_data <- self$load_from_cache(key)
        is_valid_consensus_cache_data(cache_data)
      }, error = function(e) {
        message("Error validating cache: ", e$message)
        return(FALSE)
      })
    }
  ),
  
  private = list(
    #' Return deterministic paths for regular RDS cache files
    list_cache_files = function() {
      cache_files <- sort(list.files(
        self$cache_dir,
        pattern = "\\.rds$",
        full.names = TRUE
      ))
      if (length(cache_files) == 0) {
        return(character())
      }
      file_info <- file.info(cache_files)
      cache_files[!is.na(file_info$isdir) & !file_info$isdir]
    },

    #' Resolve a validated cache key to a path inside the cache directory
    cache_file_path = function(key) {
      valid_key <- is.character(key) && length(key) == 1 && !is.na(key) &&
        nzchar(key) && nchar(key, type = "bytes") <= 240 &&
        grepl("^[A-Za-z0-9._-]+$", key) && !key %in% c(".", "..")
      if (!valid_key) {
        stop("cache key must contain 1-240 ASCII letters, digits, dots, underscores, or hyphens")
      }
      file.path(self$cache_dir, paste0(key, ".rds"))
    },

    #' Write one R object to a path; separated for deterministic failure testing
    write_cache_data = function(data, path) {
      saveRDS(data, path)
    },

    #' Persist cache data through a same-directory temporary file
    atomic_save_rds = function(data, cache_file) {
      temp_file <- tempfile(
        pattern = paste0(".", basename(cache_file), "."),
        tmpdir = dirname(cache_file),
        fileext = ".tmp"
      )
      backup_file <- NULL
      committed <- FALSE

      on.exit({
        if (file.exists(temp_file)) {
          unlink(temp_file)
        }
        if (!is.null(backup_file) && file.exists(backup_file)) {
          if (committed) {
            unlink(backup_file)
          } else {
            if (file.exists(cache_file)) {
              unlink(cache_file)
            }
            if (!file.rename(backup_file, cache_file)) {
              warning("Failed to restore previous cache file after write failure", call. = FALSE)
            }
          }
        }
      }, add = TRUE)

      private$write_cache_data(data, temp_file)

      if (file.exists(cache_file)) {
        backup_file <- tempfile(
          pattern = paste0(".", basename(cache_file), "."),
          tmpdir = dirname(cache_file),
          fileext = ".bak"
        )
        if (!file.rename(cache_file, backup_file)) {
          stop("Failed to prepare existing cache file for replacement")
        }
      }

      if (!file.rename(temp_file, cache_file)) {
        stop("Failed to atomically replace cache file")
      }
      committed <- TRUE
      invisible(TRUE)
    },

    #' Create stable hash from genes list
    create_genes_hash = function(genes) {
      # Marker rank is part of the prompt, so preserve the selected gene order.
      digest::digest(genes, algo = "xxhash64")
    },
    
    #' Create stable hash from models list
    create_models_hash = function(models) {
      # Model order controls response ordering in later discussion prompts.
      digest::digest(as.character(models), algo = "xxhash64")
    },
    
    #' Create stable hash from tissue_name and top_gene_count
    create_context_hash = function(tissue_name, top_gene_count, discussion_context) {
      context <- list(
        tissue_name = tissue_name,
        top_gene_count = top_gene_count,
        discussion = discussion_context
      )
      digest::digest(context, algo = "xxhash64")
    },

    #' Create stable hash from cluster ID
    create_cluster_hash = function(cluster_id) {
      # Keep only short path-safe IDs readable; hash every other value.
      if (grepl("^[A-Za-z0-9_-]{1,8}$", cluster_id)) {
        return(paste0("c", cluster_id))
      } else {
        return(paste0("c", substr(digest::digest(cluster_id, algo = "xxhash64"), 1, 8)))
      }
    }
  )
)

# Simplified Cache Test

test_that("Cache directory creation works with simple test", {
  skip_if_not_cran_env()

  # Test system cache directory creation
  system_cache <- mllmcelltype_cache_dir()
  expect_true(dir.exists(system_cache))
  expect_true(grepl("mLLMCelltype", system_cache))
  expect_true(grepl("consensus_cache", system_cache))
  
  # Test local cache directory
  local_cache <- mllmcelltype_cache_dir("local") 
  expect_equal(local_cache, file.path(".", ".mllmcelltype_cache"))
  
  # Test temp cache directory
  temp_cache <- mllmcelltype_cache_dir("temp")
  expect_true(grepl("mllmcelltype_cache", temp_cache))
  
  # Test cache manager with different modes
  cm_system <- CacheManager$new(cache_dir = NULL)
  expect_true(dir.exists(cm_system$get_cache_dir()))
  
  cm_local <- CacheManager$new(cache_dir = "local")
  expect_equal(cm_local$get_cache_dir(), file.path(".", ".mllmcelltype_cache"))
  
  cm_temp <- CacheManager$new(cache_dir = "temp") 
  expect_true(dir.exists(cm_temp$get_cache_dir()))
})

test_that("Cache key generation works correctly", {
  skip_if_not_cran_env()
  
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  
  # Test data in expected format
  test_data <- data.frame(
    cluster = c(0, 0, 1, 1),
    gene = c("CD3D", "CD8A", "CD14", "LYZ"),
    avg_log2FC = c(2.5, 2.1, 1.8, 1.6)
  )
  
  # Generate keys for different scenarios
  key1 <- cache_manager$generate_key(test_data, "openai/gpt-4o-mini", "0")
  key2 <- cache_manager$generate_key(test_data, "openai/gpt-4o-mini", "1") 
  key3 <- cache_manager$generate_key(test_data, "claude-haiku-4-5-20251001", "0")
  
  # Keys should be different
  expect_true(nchar(key1) > 0)
  expect_true(nchar(key2) > 0)
  expect_true(nchar(key3) > 0)
  expect_false(key1 == key2)  # Different cluster
  expect_false(key1 == key3)  # Different model
  
  # Same parameters should generate same key
  key1_repeat <- cache_manager$generate_key(test_data, "openai/gpt-4o-mini", "0")
  expect_equal(key1, key1_repeat)
})

test_that("Cache save and load operations work", {
  skip_if_not_cran_env()
  
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  
  # Mock result data
  test_result <- list(
    annotation = "T cell",
    discussion_log = list(
      rounds = list(
        list(model = "gpt-5.5", response = "T cell")
      )
    ),
    confidence = 0.95
  )
  
  key <- "test_save_load"
  
  # Test save
  cache_manager$save_to_cache(key, test_result)
  expect_true(cache_manager$has_cache(key))
  
  # Test load
  loaded_result <- cache_manager$load_from_cache(key)
  expect_equal(loaded_result$annotation, "T cell")
  expect_equal(loaded_result$confidence, 0.95)
  
  # Test validation
  expect_true(cache_manager$validate_cache(key))
})

# Mock API Response Tests

test_that("Cache saves and retrieves mock consensus results", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  
  # Mock input data
  mock_input <- data.frame(
    gene = c("CD3D", "CD8A", "IL7R"),
    cluster = c("T_cells", "CD8_T", "Memory_T")
  )
  
  mock_consensus <- list(
    consensus_annotation = "T cell",
    confidence_score = 0.95,
    models_used = c("gpt-5.5", "claude-sonnet-4-6"),
    timestamp = Sys.time()
  )
  
  # Generate cache key
  key <- cache_manager$generate_key(mock_input, c("gpt-5.5", "claude-sonnet-4-6"), "T_cells")
  
  # Test saving
  cache_manager$save_to_cache(key, mock_consensus)
  expect_true(cache_manager$has_cache(key))
  
  # Test loading
  loaded_data <- cache_manager$load_from_cache(key)
  expect_equal(loaded_data$consensus_annotation, "T cell")
  expect_equal(loaded_data$confidence_score, 0.95)
})

test_that("Cache treats equivalent marker input formats consistently", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  
  # Test with data.frame format that matches expected structure
  df_input <- data.frame(
    cluster = c(0, 0, 1, 1),
    gene = c("CD3D", "CD8A", "CD14", "LYZ"),
    avg_log2FC = c(2.1, 1.5, 0.8, 1.2)
  )
  
  # Test with list format
  simple_input <- list("0" = c("CD3D", "CD8A"))
  
  # Both should generate valid cache keys
  key1 <- cache_manager$generate_key(df_input, "gpt-5.5", "0")
  key2 <- cache_manager$generate_key(simple_input, "gpt-5.5", "0")
  
  expect_true(nchar(key1) > 0)
  expect_true(nchar(key2) > 0)
  # Equivalent prompts should share a cache key regardless of input container.
  expect_equal(key1, key2)
})

test_that("Cache validation works", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  
  # Test with valid data that matches expected structure
  valid_result <- list(
    annotation = "T cell",
    discussion_log = list(
      rounds = list(
        list(model = "gpt-5.5", response = "T cell")
      )
    ),
    confidence_score = 0.95
  )
  
  key <- "test_validation_key"
  cache_manager$save_to_cache(key, valid_result)
  
  # Validation should pass
  expect_true(cache_manager$validate_cache(key))
  
  # Test with invalid data structure
  invalid_result <- list(
    consensus_annotation = "T cell",  # Wrong field name
    confidence_score = 0.95
  )
  
  invalid_key <- "test_invalid_key"
  cache_manager$save_to_cache(invalid_key, invalid_result)
  
  # Validation should fail
  expect_false(cache_manager$validate_cache(invalid_key))
  
  # Test with non-existent key
  expect_false(cache_manager$validate_cache("non_existent_key"))
})

test_that("Cache validation rejects malformed discussion payloads", {
  cache_manager <- CacheManager$new(cache_dir = tempfile("invalid_cache_"))
  malformed_values <- list(
    list(annotation = c("T cell", "B cell"), discussion_log = list(rounds = list())),
    list(annotation = "T cell", discussion_log = "not-a-log"),
    list(annotation = "T cell", discussion_log = list(rounds = "not-a-list"))
  )

  for (index in seq_along(malformed_values)) {
    key <- paste0("malformed_", index)
    expect_true(cache_manager$save_to_cache(key, malformed_values[[index]]))
    expect_false(cache_manager$validate_cache(key))
  }
})

# Core Cache Logic Tests

test_that("Cache directory resolution works correctly", {
  # Test NULL (system cache)
  cache_manager <- CacheManager$new(cache_dir = NULL)
  actual_dir <- cache_manager$get_cache_dir()
  expected_pattern <- "mLLMCelltype.*consensus_cache"
  expect_true(grepl(expected_pattern, actual_dir))
  
  # Test "local" option
  cache_manager <- CacheManager$new(cache_dir = "local")
  expect_equal(cache_manager$get_cache_dir(), file.path(".", ".mllmcelltype_cache"))
  
  # Test "temp" option
  cache_manager <- CacheManager$new(cache_dir = "temp")
  expect_true(grepl("mllmcelltype_cache", cache_manager$get_cache_dir()))
  
  # Test custom path
  custom_path <- file.path(tempdir(), "custom_cache_test")
  cache_manager <- CacheManager$new(cache_dir = custom_path)
  expect_equal(cache_manager$get_cache_dir(), custom_path)
})

test_that("CacheManager rejects ambiguous cache directory inputs", {
  invalid_dirs <- list(character(0), c("one", "two"), NA_character_, "")

  for (cache_dir in invalid_dirs) {
    expect_error(
      CacheManager$new(cache_dir = cache_dir),
      "cache_dir must be a non-empty character scalar"
    )
  }
})

test_that("CacheManager normalizes cache directory selectors", {
  sandbox <- tempfile("cache_selector_")
  dir.create(sandbox)
  original_wd <- setwd(sandbox)
  on.exit({
    setwd(original_wd)
    unlink(sandbox, recursive = TRUE)
  }, add = TRUE)

  cache_manager <- CacheManager$new(cache_dir = "  temp  ")

  expect_identical(
    cache_manager$get_cache_dir(),
    file.path(tempdir(), "mllmcelltype_cache")
  )
  expect_false(dir.exists("  temp  "))
})

test_that("Cache statistics keep one schema when the directory is absent", {
  cache_dir <- tempfile("missing_cache_")
  cache_manager <- CacheManager$new(cache_dir = cache_dir)
  unlink(cache_dir, recursive = TRUE)

  expect_identical(
    cache_manager$get_cache_stats(),
    list(
      cache_exists = FALSE,
      cache_count = 0,
      cache_size_mb = 0,
      cache_files = character()
    )
  )
})

test_that("Cache directory creation follows CRAN policies", {
  # Save current working directory
  original_wd <- getwd()
  
  # Create a temporary directory to test in
  test_dir <- tempfile()
  dir.create(test_dir)
  setwd(test_dir)
  
  # Verify no directory created in current working directory by default
  initial_files <- list.files(".", all.files = TRUE)
  cache_manager <- CacheManager$new()
  after_files <- list.files(".", all.files = TRUE)
  
  expect_false("consensus_cache" %in% after_files)
  expect_equal(length(initial_files), length(after_files))
  
  # Restore working directory
  setwd(original_wd)
  unlink(test_dir, recursive = TRUE)
})

test_that("Utility functions work correctly", {
  # Test mllmcelltype_cache_dir()
  system_cache <- mllmcelltype_cache_dir()
  expect_true(dir.exists(system_cache))
  
  local_cache <- mllmcelltype_cache_dir("local")
  expect_equal(local_cache, file.path(".", ".mllmcelltype_cache"))
  
  # Test with temp option
  temp_cache <- mllmcelltype_cache_dir("temp")
  expect_true(grepl("mllmcelltype_cache", temp_cache))
})

test_that("Cache key generation works", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  
  # Test with simple data
  test_data <- data.frame(
    cluster = c("cluster_0", "cluster_0", "cluster_1"),
    gene = c("CD3D", "CD8A", "IL7R"),
    avg_log2FC = c(2.1, 1.5, 0.8)
  )
  
  key1 <- cache_manager$generate_key(test_data, "gpt-5.5", "cluster_0")
  key2 <- cache_manager$generate_key(test_data, "gpt-5.5", "cluster_1")
  key3 <- cache_manager$generate_key(test_data, "claude-sonnet-4-6", "cluster_0")

  # Keys should be different for different inputs
  expect_true(nchar(key1) > 0)
  expect_false(key1 == key2)  # Different cluster
  expect_false(key1 == key3)  # Different model
})

test_that("Cache key varies with tissue_name and top_gene_count", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())

  test_data <- data.frame(
    cluster = c(0, 0, 1, 1),
    gene = c("CD3D", "CD8A", "CD14", "LYZ"),
    avg_log2FC = c(2.5, 2.1, 1.8, 1.6)
  )

  base_key <- cache_manager$generate_key(test_data, "gpt-5.5", "0",
                                          tissue_name = "human PBMC", top_gene_count = 10)
  diff_tissue <- cache_manager$generate_key(test_data, "gpt-5.5", "0",
                                             tissue_name = "mouse brain", top_gene_count = 10)
  diff_topn <- cache_manager$generate_key(test_data, "gpt-5.5", "0",
                                           tissue_name = "human PBMC", top_gene_count = 5)
  same_key <- cache_manager$generate_key(test_data, "gpt-5.5", "0",
                                          tissue_name = "human PBMC", top_gene_count = 10)

  expect_false(base_key == diff_tissue)  # Different tissue
  expect_false(base_key == diff_topn)    # Different top_gene_count
  expect_equal(base_key, same_key)       # Same parameters = same key
})

test_that("Cache key preserves marker rank and model order", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  ranked_markers <- data.frame(
    cluster = c(0, 0),
    gene = c("CD3D", "IL7R"),
    avg_log2FC = c(2, 1)
  )
  reversed_rank <- ranked_markers
  reversed_rank$avg_log2FC <- rev(reversed_rank$avg_log2FC)

  ranked_key <- cache_manager$generate_key(
    ranked_markers,
    c("gpt-5.5", "claude-sonnet-4-6"),
    "0"
  )
  reversed_genes_key <- cache_manager$generate_key(
    reversed_rank,
    c("gpt-5.5", "claude-sonnet-4-6"),
    "0"
  )
  reversed_models_key <- cache_manager$generate_key(
    ranked_markers,
    c("claude-sonnet-4-6", "gpt-5.5"),
    "0"
  )
  named_models_key <- cache_manager$generate_key(
    ranked_markers,
    c(primary = "gpt-5.5", secondary = "claude-sonnet-4-6"),
    "0"
  )

  expect_false(ranked_key == reversed_genes_key)
  expect_false(ranked_key == reversed_models_key)
  expect_equal(ranked_key, named_models_key)
  expect_match(ranked_key, "^v_1\\.2_")
})

test_that("Cache key includes discussion inputs and rejects malformed markers", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  markers <- list("0" = c("CD3D", "IL7R"))
  context <- list(
    initial_predictions = list(model_a = "T cell", model_b = "NK cell"),
    max_discussion_rounds = 3L,
    controversy_threshold = 0.7,
    entropy_threshold = 1
  )

  base_key <- cache_manager$generate_key(
    markers,
    c("gpt-5.5", "claude-sonnet-4-6"),
    "0",
    discussion_context = context
  )
  changed_context <- context
  changed_context$max_discussion_rounds <- 4L
  changed_key <- cache_manager$generate_key(
    markers,
    c("gpt-5.5", "claude-sonnet-4-6"),
    "0",
    discussion_context = changed_context
  )

  expect_false(base_key == changed_key)
  expect_error(
    cache_manager$generate_key(data.frame(CD3D = 1), "gpt-5.5", "0"),
    "must contain 'cluster' and 'gene'"
  )
})

test_that("Cache keys cannot escape the configured cache directory", {
  cache_dir <- tempfile("safe_cache_")
  cache_manager <- CacheManager$new(cache_dir = cache_dir)
  unsafe_keys <- c("../outside", "nested/key", "nested\\key", "", paste(rep("a", 241), collapse = ""))

  for (key in unsafe_keys) {
    expect_error(cache_manager$save_to_cache(key, list(value = 1)), "cache key must contain")
    expect_error(cache_manager$load_from_cache(key), "cache key must contain")
    expect_error(cache_manager$has_cache(key), "cache key must contain")
  }
})

test_that("Generated keys remain path-safe for arbitrary cluster IDs", {
  cache_dir <- tempfile("generated_key_cache_")
  cache_manager <- CacheManager$new(cache_dir = cache_dir)
  key <- cache_manager$generate_key(
    list("../x" = list(genes = c("CD3D"))),
    "gpt-5.5",
    "../x"
  )

  expect_false(grepl("/|\\\\", key))
  expect_true(cache_manager$save_to_cache(key, list(value = "safe")))
  expect_equal(cache_manager$load_from_cache(key)$value, "safe")
})

test_that("Failed cache writes preserve the previous value", {
  cache_dir <- tempfile("atomic_cache_")
  cache_manager <- CacheManager$new(cache_dir = cache_dir)
  key <- "atomic_key"
  expect_true(cache_manager$save_to_cache(key, list(value = "original")))

  FailingCacheManager <- R6::R6Class(
    "FailingCacheManager",
    inherit = CacheManager,
    private = list(
      write_cache_data = function(data, path) {
        stop("simulated serialization failure")
      }
    )
  )
  failing_manager <- FailingCacheManager$new(cache_dir = cache_dir)

  expect_warning(
    expect_false(failing_manager$save_to_cache(key, list(value = "replacement"))),
    "simulated serialization failure"
  )
  expect_equal(cache_manager$load_from_cache(key)$value, "original")
  expect_length(list.files(cache_dir, pattern = "\\.(tmp|bak)$"), 0)
})

test_that("Corrupted cache files are treated as misses", {
  cache_dir <- tempfile("corrupt_cache_")
  cache_manager <- CacheManager$new(cache_dir = cache_dir)
  key <- "corrupt_key"
  writeLines("not an RDS file", file.path(cache_dir, paste0(key, ".rds")))

  expect_warning(
    expect_null(cache_manager$load_from_cache(key)),
    "Failed to load cache file"
  )
})

test_that("Cache clearing requires an explicit logical confirmation", {
  cache_dir <- tempfile("clear_cache_")
  cache_manager <- CacheManager$new(cache_dir = cache_dir)
  expect_true(cache_manager$save_to_cache("keep_me", list(value = "safe")))

  for (confirm in list(1, 0, NA, NULL, "TRUE", logical(0), c(TRUE, FALSE))) {
    expect_error(cache_manager$clear_cache(confirm), "confirm must be TRUE or FALSE")
    expect_true(cache_manager$has_cache("keep_me"))
  }

  expect_message(cache_manager$clear_cache(FALSE), "Set confirm=TRUE")
  expect_true(cache_manager$has_cache("keep_me"))
  expect_message(cache_manager$clear_cache(TRUE), "Cleared 1 cache files")
  expect_false(cache_manager$has_cache("keep_me"))
})

test_that("Cache operations ignore directories with an RDS suffix", {
  cache_dir <- tempfile("regular_cache_files_")
  cache_manager <- CacheManager$new(cache_dir = cache_dir)
  fake_cache_dir <- file.path(cache_dir, "fake.rds")
  dir.create(fake_cache_dir)

  expect_false(cache_manager$has_cache("fake"))
  expect_null(cache_manager$load_from_cache("fake"))
  expect_equal(cache_manager$get_cache_stats()$cache_count, 0)
  expect_message(cache_manager$clear_cache(TRUE), "No cache files to clear")
  expect_true(dir.exists(fake_cache_dir))
})

# CRAN Compliance and Cross-Platform Tests

test_that("No directory pollution in current working directory", {
  # Save current working directory
  original_wd <- getwd()
  
  # Create a clean temporary directory to test in
  test_dir <- tempfile()
  dir.create(test_dir)
  setwd(test_dir)
  
  # Record initial files
  initial_files <- list.files(".", all.files = TRUE)
  initial_files <- initial_files[!initial_files %in% c(".", "..")]
  
  # Load the package and create cache manager with default settings
  # This should NOT create any files in current directory
  cache_manager <- CacheManager$new()  # Default: NULL cache_dir
  
  # Check current directory is clean
  after_files <- list.files(".", all.files = TRUE) 
  after_files <- after_files[!after_files %in% c(".", "..")]
  
  # Should be no new files in current directory
  expect_equal(length(after_files), length(initial_files))
  expect_false("consensus_cache" %in% after_files)
  expect_false(".mllmcelltype_cache" %in% after_files)
  
  # Verify cache was created in system location instead
  system_cache <- cache_manager$get_cache_dir()
  expect_true(dir.exists(system_cache))
  expect_false(startsWith(system_cache, test_dir))  # Not in test directory
  
  # Test with utility functions too
  system_cache2 <- mllmcelltype_cache_dir()
  expect_true(dir.exists(system_cache2))
  expect_false(startsWith(system_cache2, test_dir))  # Not in test directory
  
  # Cleanup and restore working directory
  setwd(original_wd)
  unlink(test_dir, recursive = TRUE)
})

test_that("System cache directories follow OS conventions", {
  system_cache <- mllmcelltype_cache_dir()
  
  # Check that system cache follows OS conventions
  os_type <- Sys.info()["sysname"]
  
  if (os_type == "Darwin") {  # macOS
    expect_true(grepl("Library/Caches", system_cache) || grepl("\\.cache", system_cache))
  } else if (os_type == "Linux") {
    expect_true(grepl("\\.cache", system_cache) || grepl("/cache/", system_cache))
  } else if (os_type == "Windows") {
    expect_true(grepl("AppData.*Local", system_cache) || grepl("cache", system_cache, ignore.case = TRUE))
  }
  
  # Should always contain package identifier
  expect_true(grepl("mLLMCelltype", system_cache, ignore.case = TRUE))
  expect_true(grepl("consensus_cache", system_cache))
})

test_that("Package startup message handles old cache detection", {
  # Save original working directory
  original_wd <- getwd()

  # Create temporary test directory
  test_dir <- tempfile()
  dir.create(test_dir)
  setwd(test_dir)

  # Create old cache directory
  old_cache_dir <- file.path(".", "consensus_cache")
  dir.create(old_cache_dir)
  writeLines("old cache content", file.path(old_cache_dir, "test.txt"))

  # Capture packageStartupMessage conditions directly
  # (capture.output doesn't reliably capture packageStartupMessage
  # because testthat's condition handler intercepts them first)
  captured_messages <- character(0)
  withCallingHandlers(
    .onAttach("mLLMCelltype", "mLLMCelltype"),
    packageStartupMessage = function(m) {
      captured_messages <<- c(captured_messages, conditionMessage(m))
      invokeRestart("muffleMessage")
    }
  )

  # Should contain old cache detection message
  all_messages <- paste(captured_messages, collapse = " ")
  expect_true(grepl("Found old cache directory", all_messages))
  expect_true(grepl("mllmcelltype_cache_dir", all_messages))

  # Cleanup
  setwd(original_wd)
  unlink(test_dir, recursive = TRUE)
})

test_that("Build ignore patterns are correct", {
  rbuildignore_path <- file.path("R", ".Rbuildignore")
  
  if (file.exists(rbuildignore_path)) {
    rbuildignore <- readLines(rbuildignore_path)
    
    # Should ignore various cache directories
    cache_patterns <- c(
      "mllmcelltype_cache",
      "consensus_cache"
    )
    
    for (pattern in cache_patterns) {
      pattern_found <- any(grepl(pattern, rbuildignore))
      expect_true(pattern_found, info = paste("Pattern", pattern, "not found in .Rbuildignore"))
    }
  }
})

test_that("Cache directories are properly isolated", {
  # Test that different cache modes create separate directories
  cache_modes <- list(
    system = NULL,
    local = "local", 
    temp = "temp",
    custom = file.path(tempdir(), "custom_isolation_test")
  )
  
  cache_dirs <- list()
  cache_managers <- list()
  
  for (mode_name in names(cache_modes)) {
    cache_dir <- cache_modes[[mode_name]]
    cache_manager <- CacheManager$new(cache_dir = cache_dir)
    
    cache_managers[[mode_name]] <- cache_manager
    cache_dirs[[mode_name]] <- cache_manager$get_cache_dir()
    
    # Each should create a separate directory
    expect_true(dir.exists(cache_dirs[[mode_name]]))
  }
  
  # All cache directories should be different
  unique_dirs <- unique(unlist(cache_dirs))
  expect_equal(length(unique_dirs), length(cache_dirs))
  
  # Test that cache operations are isolated
  for (mode_name in names(cache_managers)) {
    manager <- cache_managers[[mode_name]]
    
    test_data <- list(
      annotation = paste("Test", mode_name),
      mode = mode_name
    )
    key <- paste0("isolation_test_", mode_name)
    
    manager$save_to_cache(key, test_data)
    expect_true(manager$has_cache(key))
    
    # Other managers should not see this cache
    for (other_mode in names(cache_managers)) {
      if (other_mode != mode_name) {
        other_manager <- cache_managers[[other_mode]]
        expect_false(other_manager$has_cache(key), 
                    info = paste("Cache leaked from", mode_name, "to", other_mode))
      }
    }
  }
})

test_that("Cache operations work across different R sessions", {
  # Simulate cross-session cache usage
  temp_cache_dir <- file.path(tempdir(), "cross_session_test")

  # Session 1: Create cache with data matching validate_cache() schema
  # (validate_cache requires: annotation + discussion_log with rounds)
  cache_manager1 <- CacheManager$new(cache_dir = temp_cache_dir)

  test_data <- list(
    annotation = "Cross-session test",
    discussion_log = list(
      cluster_id = "0",
      rounds = list(
        list(round_number = 1, responses = list(model1 = "T cells"))
      )
    )
  )

  key <- "cross_session_key"
  cache_manager1$save_to_cache(key, test_data)
  expect_true(cache_manager1$has_cache(key))

  # Session 2: Create new manager instance (simulates new R session)
  cache_manager2 <- CacheManager$new(cache_dir = temp_cache_dir)

  # Should be able to access cache from previous "session"
  expect_true(cache_manager2$has_cache(key))

  loaded_data <- cache_manager2$load_from_cache(key)
  expect_equal(loaded_data$annotation, test_data$annotation)
  expect_equal(loaded_data$discussion_log$cluster_id, "0")

  # Should be able to validate cache
  expect_true(cache_manager2$validate_cache(key))
})

test_that("Memory usage is reasonable for cache operations", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())

  # Create test data of various sizes
  small_data <- list(annotation = "Small test")
  medium_data <- list(
    annotation = "Medium test",
    details = rep("test data", 100)
  )
  large_data <- list(
    annotation = "Large test",
    details = rep("test data with more content", 1000)
  )

  test_cases <- list(
    small = small_data,
    medium = medium_data,
    large = large_data
  )

  for (size_name in names(test_cases)) {
    data <- test_cases[[size_name]]
    key <- paste0("memory_test_", size_name)

    # Cross-platform memory measurement using gc()
    # gc()[,2] returns "used (Mb)" for Ncells and Vcells
    gc(verbose = FALSE, reset = TRUE)
    mem_before <- sum(gc(verbose = FALSE)[, 2])

    cache_manager$save_to_cache(key, data)
    loaded_data <- cache_manager$load_from_cache(key)

    gc(verbose = FALSE)
    mem_after <- sum(gc(verbose = FALSE)[, 2])

    # Memory increase should be reasonable (less than 10MB for our test data)
    mem_increase <- mem_after - mem_before
    expect_true(mem_increase < 10,
                info = sprintf("Memory increased by %.1f MB for %s data", mem_increase, size_name))

    # Loaded data should match original
    expect_equal(loaded_data$annotation, data$annotation)
  }
})

# Cache Performance Tests

test_that("Cache key generation performance", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  
  # Test with moderately large dataset
  large_data <- data.frame(
    cluster = sample(0:9, 1000, replace = TRUE),  # 10 clusters
    gene = paste0("Gene_", sample(1:500, 1000, replace = TRUE)),  # 500 unique genes
    avg_log2FC = rnorm(1000, mean = 1.5, sd = 0.8)
  )
  
  # Key generation should be fast even for large data
  start_time <- Sys.time()
  key <- cache_manager$generate_key(large_data, "gpt-5.5", "0")
  key_time <- as.numeric(Sys.time() - start_time)
  
  expect_true(key_time < 2.0)  # Should take less than 2 seconds
  expect_true(nchar(key) > 0)
  expect_true(grepl("^v_", key))  # Should start with version prefix
})

test_that("Cache I/O performance", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  
  # Create moderately large result object
  large_result <- list(
    annotation = "T cell",
    discussion_log = list(
      rounds = replicate(3, list(
        model_responses = replicate(5, list(
          response = paste(sample(letters, 200, replace = TRUE), collapse = ""),
          reasoning = paste(sample(LETTERS, 300, replace = TRUE), collapse = "")
        ), simplify = FALSE),
        consensus_result = list(majority_prediction = "T cell")
      ), simplify = FALSE)
    ),
    confidence_score = 0.95,
    metadata = list(
      processing_time = 45.2,
      models_used = c("gpt-5.5", "claude-sonnet-4-6", "gemini-3.1-pro-preview"),
      gene_count = 50
    )
  )
  
  key <- "performance_test_key"
  
  # Save performance
  start_time <- Sys.time()
  cache_manager$save_to_cache(key, large_result)
  save_time <- as.numeric(Sys.time() - start_time)
  
  # Load performance  
  start_time <- Sys.time()
  loaded <- cache_manager$load_from_cache(key)
  load_time <- as.numeric(Sys.time() - start_time)
  
  expect_true(save_time < 1.0)  # Save should be fast
  expect_true(load_time < 0.2)  # Load should be very fast
  expect_equal(loaded$annotation, large_result$annotation)
  expect_equal(length(loaded$discussion_log$rounds), 3)
})

test_that("Cache statistics and management performance", {
  # Use a unique cache directory to avoid interference
  unique_cache_dir <- file.path(tempdir(), paste0("perf_test_", Sys.getpid(), "_", as.numeric(Sys.time())))
  cache_manager <- CacheManager$new(cache_dir = unique_cache_dir)
  
  # Create multiple cache entries
  for (i in 1:20) {
    test_result <- list(
      annotation = paste("Cell_type", i),
      discussion_log = list(rounds = list(list(response = paste("Response", i)))),
      confidence = runif(1)
    )
    key <- paste0("perf_test_", i)
    cache_manager$save_to_cache(key, test_result)
  }
  
  # Test cache statistics performance
  start_time <- Sys.time()
  stats <- cache_manager$get_cache_stats()
  stats_time <- as.numeric(Sys.time() - start_time)
  
  expect_true(stats_time < 0.5)  # Should be fast
  expect_equal(stats$cache_count, 20)
  expect_true(stats$cache_size_mb > 0)
  expect_true(stats$cache_exists)
  
  # Test cache clearing performance
  start_time <- Sys.time()
  cache_manager$clear_cache(confirm = TRUE)
  clear_time <- as.numeric(Sys.time() - start_time)
  
  expect_true(clear_time < 1.0)  # Should be fast
  
  # Verify cache is cleared
  final_stats <- cache_manager$get_cache_stats()
  expect_equal(final_stats$cache_count, 0)
})

test_that("Different cache directory modes performance", {
  # Test all cache directory modes
  cache_modes <- list(
    system = NULL,
    local = "local",
    temp = "temp",
    custom = file.path(tempdir(), "custom_perf_test")
  )
  
  results <- list()
  
  for (mode_name in names(cache_modes)) {
    cache_dir <- cache_modes[[mode_name]]
    
    # Test cache manager creation time
    start_time <- Sys.time()
    cache_manager <- CacheManager$new(cache_dir = cache_dir)
    creation_time <- as.numeric(Sys.time() - start_time)
    
    # Cache manager creation should be fast
    expect_true(creation_time < 0.5)
    
    # Test directory exists
    actual_dir <- cache_manager$get_cache_dir()
    expect_true(dir.exists(actual_dir))
    
    # Test basic operation
    test_data <- list(annotation = paste("Test", mode_name))
    key <- paste0("mode_test_", mode_name)
    
    cache_manager$save_to_cache(key, test_data)
    expect_true(cache_manager$has_cache(key))
    
    loaded <- cache_manager$load_from_cache(key)
    expect_equal(loaded$annotation, test_data$annotation)
    
    results[[mode_name]] <- list(
      creation_time = creation_time,
      cache_dir = actual_dir
    )
  }
  
  # All modes should have reasonable performance
  creation_times <- sapply(results, function(x) x$creation_time)
  expect_true(all(creation_times < 0.5))
  
  # All cache directories should exist and be different
  cache_dirs <- sapply(results, function(x) x$cache_dir)
  expect_equal(length(unique(cache_dirs)), length(cache_dirs))
})

test_that("Cache validation performance with various data sizes", {
  cache_manager <- CacheManager$new(cache_dir = tempdir())
  
  # Test validation with different sized objects
  sizes <- c(
    small = 10,    # Small result
    medium = 100,  # Medium result  
    large = 500    # Large result
  )
  
  validation_times <- list()
  
  for (size_name in names(sizes)) {
    size_val <- sizes[[size_name]]
    
    # Create test data of specified size
    test_result <- list(
      annotation = paste("Test annotation", size_name),
      discussion_log = list(
        rounds = replicate(3, list(
          model_responses = replicate(size_val, list(
            response = paste(sample(letters, 20, replace = TRUE), collapse = "")
          ), simplify = FALSE),
          consensus_result = list(majority_prediction = "Test cell")
        ), simplify = FALSE)
      )
    )
    
    key <- paste0("validation_test_", size_name)
    cache_manager$save_to_cache(key, test_result)
    
    # Test validation performance
    start_time <- Sys.time()
    is_valid <- cache_manager$validate_cache(key)
    validation_time <- as.numeric(Sys.time() - start_time)
    
    expect_true(is_valid)
    expect_true(validation_time < 0.1)  # Should be very fast
    
    validation_times[[size_name]] <- validation_time
  }
  
  # Validation should be consistently fast regardless of data size
  expect_true(all(unlist(validation_times) < 0.1))
})

.MODEL_REQUEST_RETRY <- list(
  MAX_ATTEMPTS = 3L,
  BASE_DELAY_SECONDS = 5
)

wait_before_model_retry <- function(seconds) {
  Sys.sleep(seconds)
}

dispatch_model_request_once <- function(prompt, model, api_key, provider,
                                        provider_base_url, normalize = TRUE) {
  if (exists(provider, envir = custom_providers)) {
    log_debug("Using custom provider", list(provider = provider))
    response <- if (is.null(provider_base_url)) {
      process_custom(prompt, model, api_key)
    } else {
      process_custom(prompt, model, api_key, provider_base_url)
    }
  } else {
    processor <- new_builtin_provider_processor(provider, provider_base_url)
    response <- processor$process_request(prompt, model, api_key, normalize = normalize)
  }

  if (isTRUE(normalize)) {
    tryCatch(
      normalize_model_response_lines(response),
      error = function(e) {
        stop("Invalid response format from provider '", provider, "'")
      }
    )
  } else {
    response
  }
}

#' Get response from a specific model
#'
#' @param prompt Non-empty prompt string
#' @param model Non-empty model name
#' @param api_key Non-empty API key
#' @param base_urls Optional shared or provider-specific base URL configuration
#' @param normalize Logical; if \code{TRUE} (default) the response is normalized
#'   into non-empty lines, otherwise the raw response string is returned.
#' @return Provider response as a character vector
#' @keywords internal
get_model_response <- function(prompt, model, api_key, base_urls = NULL,
                               normalize = TRUE) {
  prompt <- .normalize_required_string(prompt, "prompt")
  model <- .normalize_required_string(model, "model")
  api_key <- .normalize_required_string(api_key, "api_key")

  # Get the provider for the model
  provider <- get_provider(model)

  # Resolve provider-specific base URL (same pattern as annotate_cell_types)
  provider_base_url <- resolve_provider_base_url(provider, base_urls)

  log_debug("Starting model request", list(
    provider = provider,
    model = model,
    prompt_length = nchar(prompt),
    custom_url = !is.null(provider_base_url),
    normalize = normalize
  ))

  for (attempt in seq_len(.MODEL_REQUEST_RETRY$MAX_ATTEMPTS)) {
    result <- tryCatch(
      list(
        success = TRUE,
        response = dispatch_model_request_once(
          prompt,
          model,
          api_key,
          provider,
          provider_base_url,
          normalize = normalize
        )
      ),
      error = function(error) list(success = FALSE, error = error)
    )

    if (isTRUE(result$success)) {
      return(result$response)
    }

    retryable <- inherits(result$error, "mllm_api_error") &&
      isTRUE(result$error$retryable)
    if (!retryable || attempt == .MODEL_REQUEST_RETRY$MAX_ATTEMPTS) {
      stop(result$error)
    }

    wait_seconds <- .MODEL_REQUEST_RETRY$BASE_DELAY_SECONDS * 2^(attempt - 1)
    get_logger()$warn("Retrying transient model request failure", list(
      provider = provider,
      model = model,
      attempt = attempt,
      max_attempts = .MODEL_REQUEST_RETRY$MAX_ATTEMPTS,
      wait_seconds = wait_seconds,
      error = result$error$message
    ))
    wait_before_model_retry(wait_seconds)
  }

  stop("Model request retry loop ended unexpectedly")
}

# Custom model manager for user-defined providers and models.

.normalize_custom_identifier <- function(value, field_name) {
  normalized <- tolower(.normalize_required_string(value, field_name))
  if (!grepl("^[a-z0-9][a-z0-9._-]*$", normalized)) {
    stop(
      field_name,
      " must contain only letters, digits, dots, underscores, or hyphens"
    )
  }
  normalized
}

# Environments storing custom provider and model registrations.
custom_providers <- new.env(parent = emptyenv())
custom_models <- new.env(parent = emptyenv())

#' Register a custom LLM provider
#'
#' @param provider_name Unique name for the custom provider
#' @param process_fn Function that processes LLM requests. Must accept parameters:
#'   prompt, model, api_key; may optionally accept model_config and base_url
#' @param description Optional description of the provider
#'
#' @return Invisible NULL
#' @export
#'
#' @examples
#' \dontrun{
#' register_custom_provider(
#'   provider_name = "my_provider",
#'   process_fn = function(prompt, model, api_key) {
#'     # Custom implementation
#'     response <- httr::POST(
#'       url = "your_api_endpoint",
#'       body = list(prompt = prompt),
#'       encode = "json"
#'     )
#'     return(httr::content(response)$choices[[1]]$text)
#'   }
#' )
#' }
register_custom_provider <- function(provider_name, process_fn,
                                     description = NULL) {
  # Input validation
  if (!is.function(process_fn)) {
    stop("process_fn must be a function")
  }

  # Normalize provider name to lowercase for consistent lookup
  provider_name <- .normalize_custom_identifier(provider_name, "provider_name")
  if (provider_name %in% names(.BUILTIN_PROVIDER_SPECS)) {
    stop("Provider '", provider_name, "' is reserved for a built-in provider")
  }

  # Check if provider already exists
  if (exists(provider_name, envir = custom_providers)) {
    stop("Provider '", provider_name, "' already exists")
  }

  # Validate process_fn arguments
  fn_args <- names(formals(process_fn))
  required_args <- c("prompt", "model", "api_key")
  if (!all(required_args %in% fn_args)) {
    stop("process_fn must accept parameters: ",
         paste(required_args, collapse = ", "))
  }

  # Store provider configuration
  assign(provider_name,
         list(
           process_fn = process_fn,
           description = description,
           models = character(0)
         ),
         envir = custom_providers)

  get_logger()$info("Registered custom provider", list(provider = provider_name))
  invisible(TRUE)
}

#' Register a custom model for a provider
#'
#' @param model_name Unique name for the custom model
#' @param provider_name Name of the provider this model belongs to
#' @param model_config List of configuration parameters for the model (e.g., temperature, max_tokens)
#'
#' @return Invisible TRUE on success
#' @export
#'
#' @examples
#' \dontrun{
#' register_custom_model(
#'   model_name = "my_model",
#'   provider_name = "my_provider",
#'   model_config = list(
#'     temperature = 0.7,
#'     max_tokens = 2000
#'   )
#' )
#' }
register_custom_model <- function(model_name, provider_name,
                                  model_config = list()) {
  # Input validation
  if (!is.list(model_config)) {
    stop("model_config must be a list")
  }

  # Normalize names to lowercase for consistent lookup with get_provider()
  model_name <- .normalize_custom_identifier(model_name, "model_name")
  provider_name <- .normalize_custom_identifier(provider_name, "provider_name")

  # Check if provider exists
  if (!exists(provider_name, envir = custom_providers)) {
    stop("Provider '", provider_name, "' does not exist")
  }

  # Check if model already exists
  if (exists(model_name, envir = custom_models)) {
    stop("Model '", model_name, "' already exists")
  }

  # Store model configuration
  assign(model_name,
         list(
           provider = provider_name,
           config = model_config
         ),
         envir = custom_models)

  # Update provider's model list
  provider_data <- get(provider_name, envir = custom_providers)
  provider_data$models <- c(provider_data$models, model_name)
  assign(provider_name, provider_data, envir = custom_providers)

  get_logger()$info("Registered custom model for provider",
           list(model = model_name, provider = provider_name))
  invisible(TRUE)
}

#' Process request using custom provider
#' @keywords internal
process_custom <- function(prompt, model, api_key, base_url = NULL) {
  # Normalize to lowercase for consistent lookup with get_provider()
  model_lower <- .normalize_custom_identifier(model, "model")

  # Check if model exists
  if (!exists(model_lower, envir = custom_models)) {
    stop("Model '", model, "' not found")
  }

  # Get model and provider data
  model_data <- get(model_lower, envir = custom_models)
  provider_data <- get(model_data$provider, envir = custom_providers)

  # Call provider's process function
  get_logger()$info("Processing request with custom model", list(model = model))
  tryCatch({
    process_args <- names(formals(provider_data$process_fn))
    accepts_dots <- "..." %in% process_args
    call_args <- list(prompt = prompt, model = model, api_key = api_key)

    if ("model_config" %in% process_args || "..." %in% process_args) {
      call_args$model_config <- model_data$config
    }
    if (!is.null(base_url)) {
      if (!("base_url" %in% process_args || accepts_dots)) {
        stop("Custom provider process_fn does not accept base_url")
      }
      call_args$base_url <- base_url
    }

    response <- do.call(provider_data$process_fn, call_args)
    get_logger()$info("Custom model request processed successfully")
    return(response)
  }, error = function(e) {
    get_logger()$error("Error processing custom model request",
              list(error = e$message))
    if (inherits(e, "mllm_api_error")) {
      stop(e)
    }
    stop("Failed to process request with custom model: ", e$message)
  })
}

#' Get list of registered custom providers
#
#' @export
list_custom_providers <- function() {
  ls(envir = custom_providers)
}

#' Get list of registered custom models
#
#' @export
list_custom_models <- function() {
  ls(envir = custom_models)
}

#' Normalize annotation for comparison
#
#
#' @keywords internal
normalize_annotation <- function(annotation) {
  # Convert to lowercase
  normalized <- tolower(trimws(annotation))
  
  # Replace punctuation with space to preserve word boundaries (e.g., "T-cell" -> "T cell")
  normalized <- gsub("[[:punct:]]", " ", normalized)
  # Normalize whitespace
  normalized <- gsub("\\s+", " ", normalized)
  normalized <- trimws(normalized)
  
  # Handle plurals (simple approach)
  normalized <- gsub("cells$", "cell", normalized)
  normalized <- gsub("cytes$", "cyte", normalized)
  
  # Common cell type variations
  normalized <- gsub("t lymphocyte", "t cell", normalized)
  normalized <- gsub("b lymphocyte", "b cell", normalized)
  normalized <- gsub("natural killer", "nk", normalized)
  
  return(normalized)
}

#' Extract discussion cell type label
#'
#' @keywords internal
#' @noRd
extract_discussion_cell_type <- function(response, default = "Unknown") {
  lines <- tryCatch(
    normalize_model_response_lines(response),
    error = function(e) NULL
  )
  if (is.null(lines)) {
    return(default)
  }
  cell_type_lines <- grep("^CELL TYPE:", trimws(lines), value = TRUE, ignore.case = TRUE)
  if (length(cell_type_lines) > 0) {
    annotation <- trimws(sub("^CELL TYPE:\\s*", "", cell_type_lines[1], ignore.case = TRUE))
    if (nzchar(annotation)) {
      return(annotation)
    }
    return(default)
  }
  if (length(lines) == 1) {
    return(lines)
  }

  default
}

#' Calculate simple consensus without LLM
#
#
#' @keywords internal
calculate_simple_consensus <- function(round_responses) {
  valid_responses <- vapply(
    as.list(round_responses),
    is_valid_model_response,
    logical(1)
  )
  round_responses <- round_responses[valid_responses]
  if (length(round_responses) == 0) {
    return(list(
      consensus_proportion = 0,
      entropy = 0,
      majority_prediction = "Insufficient_Responses",
      has_unique_majority = FALSE
    ))
  }

  # Normalize all annotations
  normalized_responses <- sapply(round_responses, normalize_annotation)
  
  # Count occurrences
  response_counts <- table(normalized_responses)
  total_responses <- length(round_responses)
  
  
  # Find most common response
  most_common_idx <- which.max(response_counts)
  most_common_count <- as.numeric(response_counts[most_common_idx])
  majority_groups <- names(response_counts)[response_counts == most_common_count]
  has_unique_majority <- length(majority_groups) == 1
  
  # Preserve model order when multiple groups are tied.
  matching_idx <- which(normalized_responses %in% majority_groups)[1]
  majority_prediction <- unname(round_responses[matching_idx])
  
  # Calculate consensus proportion
  consensus_proportion <- most_common_count / total_responses
  
  # Calculate Shannon entropy
  proportions <- as.numeric(response_counts) / total_responses
  entropy <- -sum(proportions * log2(proportions))
  
  
  return(list(
    consensus_proportion = consensus_proportion,
    entropy = entropy,
    majority_prediction = majority_prediction,
    has_unique_majority = has_unique_majority
  ))
}

simple_consensus_reached <- function(result, controversy_threshold,
                                     entropy_threshold) {
  isTRUE(result$has_unique_majority) &&
    result$consensus_proportion >= controversy_threshold &&
    result$entropy <= entropy_threshold
}

# Validate that consensus metrics can arise from a finite set of votes.
consensus_metrics_plausible <- function(result, vote_count, tolerance = 0.02) {
  consensus_proportion <- result$consensus_proportion
  entropy <- result$entropy

  if (length(vote_count) != 1 || !is.numeric(vote_count) ||
      is.na(vote_count) || !is.finite(vote_count) || vote_count < 1) {
    return(FALSE)
  }
  if (length(consensus_proportion) != 1 || !is.numeric(consensus_proportion) ||
      is.na(consensus_proportion) || !is.finite(consensus_proportion) ||
      consensus_proportion < 0 || consensus_proportion > 1) {
    return(FALSE)
  }
  if (length(entropy) != 1 || !is.numeric(entropy) ||
      is.na(entropy) || !is.finite(entropy) || entropy < 0) {
    return(FALSE)
  }

  supporting_votes <- consensus_proportion * vote_count
  nearest_vote_count <- round(supporting_votes)
  if (nearest_vote_count < 1 || nearest_vote_count > vote_count ||
      abs(supporting_votes - nearest_vote_count) > tolerance) {
    return(FALSE)
  }

  group_sizes <- nearest_vote_count
  remaining_votes <- vote_count - nearest_vote_count
  while (remaining_votes > 0) {
    next_group_size <- min(nearest_vote_count, remaining_votes)
    group_sizes <- c(group_sizes, next_group_size)
    remaining_votes <- remaining_votes - next_group_size
  }

  group_proportions <- group_sizes / vote_count
  minimum_entropy <- -sum(group_proportions * log2(group_proportions))
  majority_share <- nearest_vote_count / vote_count
  maximum_entropy <- -(majority_share * log2(majority_share))
  if (nearest_vote_count < vote_count) {
    singleton_share <- 1 / vote_count
    maximum_entropy <- maximum_entropy -
      (vote_count - nearest_vote_count) * singleton_share * log2(singleton_share)
  }

  entropy >= minimum_entropy - tolerance &&
    entropy <= maximum_entropy + tolerance
}

# Constants for consensus checking
.CONSENSUS_CONSTANTS <- list(
  DEFAULT_RESPONSE = "0\n0\n0\nUnknown",
  FALLBACK_MODELS = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3-flash-preview", "qwen3.6-flash", "deepseek-v4-flash"),
  NUMBER_PATTERN = "^[+-]?([0-9]+(\\.[0-9]+)?|\\.[0-9]+)$",
  SEPARATOR_PATTERN = "(?:[:=]|\uFF1A)",
  SEARCH_PATTERNS = list(
    CONSENSUS_LABEL = "(?:consensus\\s+proportion(?:\\s*\\(\\s*cp\\s*\\))?|cp)",
    ENTROPY_LABEL = "(?:(?:shannon\\s+)?entropy(?:\\s*\\(\\s*h\\s*\\))?|h)"
  )
)

# Default values for failed parsing
.DEFAULT_CONSENSUS_RESULT <- list(
  reached = FALSE,
  consensus_proportion = 0,
  entropy = 0,
  majority_prediction = "Unknown"
)

parse_consensus_number <- function(value) {
  value <- trimws(value)
  if (length(value) != 1 || is.na(value) ||
      !grepl(.CONSENSUS_CONSTANTS$NUMBER_PATTERN, value)) {
    return(NULL)
  }

  number <- suppressWarnings(as.numeric(value))
  if (!is.finite(number)) NULL else number
}

#' Prepare list of models to try for consensus checking
#
#
#' @keywords internal
prepare_models_list <- function(consensus_check_model = NULL) {
  models_to_try <- c()
  
  if (!is.null(consensus_check_model)) {
    get_logger()$info("Using specified consensus check model", list(model = consensus_check_model))
    
    if (grepl("/", consensus_check_model)) {
      parts <- strsplit(consensus_check_model, "/")[[1]]
      if (length(parts) > 1) {
        base_model <- paste(parts[-1], collapse = "/")
        direct_provider <- if (grepl("/", base_model)) {
          NULL
        } else {
          tryCatch(get_provider(base_model), error = function(e) NULL)
        }
        if (!is.null(direct_provider)) {
          get_logger()$info("Detected OpenRouter model. Using base model name", list(base_model = base_model))
          models_to_try <- c(consensus_check_model, base_model)
        } else {
          models_to_try <- c(consensus_check_model)
        }
      } else {
        models_to_try <- c(consensus_check_model)
      }
    } else {
      models_to_try <- c(consensus_check_model)
    }
  }
  
  fallback_models <- .CONSENSUS_CONSTANTS$FALLBACK_MODELS[!.CONSENSUS_CONSTANTS$FALLBACK_MODELS %in% models_to_try]
  c(models_to_try, fallback_models)
}

#' Parse standard 4-line consensus response format
#
#
#' @keywords internal
parse_standard_format <- function(result_lines) {
  if (length(result_lines) != 4) return(NULL)

  indicator <- trimws(result_lines[[1]])
  consensus_proportion <- parse_consensus_number(result_lines[[2]])
  entropy <- parse_consensus_number(result_lines[[3]])
  valid_metrics <- !is.null(consensus_proportion) &&
    consensus_proportion >= 0 &&
    consensus_proportion <= 1 &&
    !is.null(entropy) &&
    entropy >= 0
  if (!indicator %in% c("0", "1") || !valid_metrics) {
    return(NULL)
  }

  get_logger()$debug("Detected standard 4-line format")

  list(
    reached = indicator == "1",
    consensus_proportion = consensus_proportion,
    entropy = entropy,
    majority_prediction = trimws(result_lines[[4]])
  )
}

#' Extract numeric value from line containing a label
#
#
#
#
#' @keywords internal
extract_labeled_value <- function(lines, label_pattern) {
  line_pattern <- paste0(
    "^\\s*",
    label_pattern,
    "\\s*",
    .CONSENSUS_CONSTANTS$SEPARATOR_PATTERN,
    "\\s*(.*?)\\s*$"
  )
  for (line in lines) {
    match <- regexec(line_pattern, line, ignore.case = TRUE, perl = TRUE)
    captures <- regmatches(line, match)[[1]]
    if (length(captures) != 2) {
      next
    }

    value <- parse_consensus_number(captures[[2]])
    if (!is.null(value)) {
      get_logger()$debug("Found value in line", list(value = value, line = line))
      return(value)
    }
  }
  NULL
}

extract_consensus_indicator <- function(lines) {
  first_line <- trimws(lines[[1]])
  if (first_line %in% c("0", "1")) {
    return(first_line == "1")
  }

  indicator_pattern <- paste0(
    "^\\s*consensus(?:\\s+reached)?\\s*",
    .CONSENSUS_CONSTANTS$SEPARATOR_PATTERN,
    "\\s*",
    "(0|1|true|false|yes|no)\\s*$"
  )
  for (line in lines) {
    match <- regexec(indicator_pattern, line, ignore.case = TRUE, perl = TRUE)
    captures <- regmatches(line, match)[[1]]
    if (length(captures) == 2) {
      return(tolower(captures[[2]]) %in% c("1", "true", "yes"))
    }
  }
  FALSE
}

#' Find majority prediction from response lines
#
#
#' @keywords internal
find_majority_prediction <- function(lines) {
  structured_labels <- c(
    "majority(?:\\s+prediction|\\s+cell\\s*type)?",
    "cell\\s*type",
    "claim"
  )
  separators <- paste0(
    "\\s*",
    .CONSENSUS_CONSTANTS$SEPARATOR_PATTERN,
    "\\s*"
  )

  for (line in lines) {
    for (label_pattern in structured_labels) {
      pattern <- paste0("^\\s*", label_pattern, separators, "(.*?)\\s*$")
      match <- regexec(pattern, line, ignore.case = TRUE, perl = TRUE)
      captures <- regmatches(line, match)[[1]]
      if (length(captures) == 2 && nzchar(trimws(captures[[2]]))) {
        return(trimws(captures[[2]]))
      }
    }
  }

  metadata_labels <- c(
    .CONSENSUS_CONSTANTS$SEARCH_PATTERNS$CONSENSUS_LABEL,
    .CONSENSUS_CONSTANTS$SEARCH_PATTERNS$ENTROPY_LABEL,
    "consensus(?:\\s+reached)?",
    structured_labels
  )
  for (line in lines) {
    line_clean <- trimws(line)
    if (!nzchar(line_clean) || !is.null(parse_consensus_number(line_clean))) {
      next
    }

    is_metadata <- any(vapply(metadata_labels, function(label_pattern) {
      grepl(
        paste0("^\\s*", label_pattern, separators),
        line_clean,
        ignore.case = TRUE,
        perl = TRUE
      )
    }, logical(1)))
    if (is_metadata) {
      next
    }

    return(line_clean)
  }

  "Parsing_Failed"
}

#' Parse flexible format consensus response
#
#
#' @keywords internal
parse_flexible_format <- function(lines) {
  result <- .DEFAULT_CONSENSUS_RESULT
  result$reached <- extract_consensus_indicator(lines)

  proportion_value <- extract_labeled_value(
    lines,
    .CONSENSUS_CONSTANTS$SEARCH_PATTERNS$CONSENSUS_LABEL
  )
  if (!is.null(proportion_value) && proportion_value >= 0 && proportion_value <= 1) {
    result$consensus_proportion <- proportion_value
  }

  entropy_value <- extract_labeled_value(
    lines,
    .CONSENSUS_CONSTANTS$SEARCH_PATTERNS$ENTROPY_LABEL
  )
  if (!is.null(entropy_value) && entropy_value >= 0) {
    result$entropy <- entropy_value
  }

  result$majority_prediction <- find_majority_prediction(lines)
  result
}

#' Parse consensus response from model
#
#
#' @keywords internal
parse_consensus_response <- function(response) {
  lines <- tryCatch(
    normalize_model_response_lines(response),
    error = function(e) {
      get_logger()$warn("Invalid consensus response", list(error = e$message))
      NULL
    }
  )
  if (is.null(lines)) {
    return(.DEFAULT_CONSENSUS_RESULT)
  }
  
  if (length(lines) < 4) {
    get_logger()$warn("Not enough lines for standard format, trying flexible parsing", list(line_count = length(lines)))
  } else {
    # Try standard format first
    result_lines <- tail(lines, 4)
    standard_result <- parse_standard_format(result_lines)
    
    if (!is.null(standard_result)) {
      get_logger()$info("Parsed standard format", list(
        consensus = standard_result$reached,
        proportion = standard_result$consensus_proportion,
        entropy = standard_result$entropy
      ))
      return(standard_result)
    }
  }

  # Fall back to flexible parsing
  get_logger()$debug("Using flexible format parsing")
  parse_flexible_format(lines)
}

#' Execute consensus check across ordered model candidates
#' @param formatted_responses Prompt containing model responses
#' @param api_keys Named API key list
#' @param models_to_try Ordered model candidates
#' @param base_urls Optional provider base URLs
#' @return A list containing success status and response text
#' @keywords internal
execute_consensus_check <- function(formatted_responses, api_keys, models_to_try, base_urls = NULL) {
  for (model_name in models_to_try) {
    get_logger()$info("Trying model for consensus check", list(model = model_name))
    
    credential <- .resolve_model_api_key(model_name, api_keys)
    if (is.null(credential$api_key)) {
      get_logger()$warn("No API key available for model, skipping", list(model = model_name))
      next
    }
    
    result <- tryCatch({
      temp_response <- get_model_response(
        formatted_responses,
        model_name,
        credential$api_key,
        base_urls
      )

      if (is.character(temp_response) && length(temp_response) > 1) {
        temp_response <- paste(temp_response, collapse = "\n")
      }

      get_logger()$info("Successfully got response from model", list(model = model_name))
      list(success = TRUE, response = temp_response)
    }, error = function(error) {
      get_logger()$error("Consensus model request failed", list(
        model = model_name,
        error = error$message
      ))
      list(success = FALSE, response = NULL)
    })

    if (result$success) {
      return(result)
    }
  }
  
  # All attempts failed
  get_logger()$warn("All model attempts failed, consensus check could not be performed")
  warning("All available models failed for consensus check. Please ensure at least one model API key is valid.")
  list(success = FALSE, response = .CONSENSUS_CONSTANTS$DEFAULT_RESPONSE)
}

#' Check if consensus is reached among models
#
#
#
#
#
#' @note This function uses create_consensus_check_prompt from prompt_templates.R
#' @importFrom utils tail
#' @keywords internal
check_consensus <- function(round_responses, api_keys = NULL, controversy_threshold = 2/3, entropy_threshold = 1.0, consensus_check_model = NULL, base_urls = NULL) {
  # Initialize logging
  get_logger()$info("Starting check_consensus function")
  get_logger()$debug("Input responses", list(responses = round_responses))

  controversy_threshold <- .normalize_probability(
    controversy_threshold,
    "controversy_threshold"
  )
  entropy_threshold <- .normalize_nonnegative_number(
    entropy_threshold,
    "entropy_threshold"
  )

  response_list <- as.list(round_responses)
  valid_response_mask <- vapply(
    response_list,
    is_valid_model_response,
    logical(1)
  )
  response_list <- response_list[valid_response_mask]

  if (length(response_list) < 2) {
    get_logger()$warn("Not enough responses to check consensus", list(response_count = length(response_list)))
    return(list(reached = FALSE, consensus_proportion = 0, entropy = 0, majority_prediction = "Insufficient_Responses"))
  }

  # OPTIMIZATION: First try simple consensus calculation
  get_logger()$info("Starting with simple consensus calculation")
  
  # Extract cell types from responses if they are discussion format
  extracted_cell_types <- vapply(
    response_list,
    extract_discussion_cell_type,
    character(1),
    default = NA_character_
  )
  real_annotation_mask <- vapply(
    as.list(extracted_cell_types),
    is_real_cell_type_annotation,
    logical(1)
  )
  extracted_cell_types <- extracted_cell_types[real_annotation_mask]
  if (length(extracted_cell_types) < 2) {
    get_logger()$warn("Not enough parseable cell type responses to check consensus", list(
      response_count = length(extracted_cell_types)
    ))
    return(list(reached = FALSE, consensus_proportion = 0, entropy = 0, majority_prediction = "Insufficient_Responses"))
  }
  
  
  # Calculate simple consensus
  simple_result <- calculate_simple_consensus(extracted_cell_types)
  deterministic_result <- list(
    reached = simple_consensus_reached(
      simple_result,
      controversy_threshold,
      entropy_threshold
    ),
    consensus_proportion = simple_result$consensus_proportion,
    entropy = simple_result$entropy,
    majority_prediction = simple_result$majority_prediction
  )

  if (deterministic_result$reached) {
    # Simple consensus is sufficient
    get_logger()$info("CONSENSUS ACHIEVED WITH SIMPLE CHECK - NO LLM NEEDED", list(
      proportion = simple_result$consensus_proportion,
      entropy = simple_result$entropy,
      controversy_threshold = controversy_threshold,
      entropy_threshold = entropy_threshold,
      majority = simple_result$majority_prediction
    ))
    
    message(sprintf("Cluster consensus achieved via SIMPLE CHECK: %s (CP=%.2f, H=%.2f)",
                    simple_result$majority_prediction,
                    simple_result$consensus_proportion,
                    simple_result$entropy))
    
    return(deterministic_result)
  }

  # Simple consensus didn't meet thresholds, use LLM for double-checking
  failure_reasons <- c(
    if (!isTRUE(simple_result$has_unique_majority)) "no unique majority",
    if (simple_result$consensus_proportion < controversy_threshold) {
      "consensus proportion below threshold"
    },
    if (simple_result$entropy > entropy_threshold) "entropy above threshold"
  )
  get_logger()$info("Simple consensus BELOW threshold, REQUIRING LLM double-check", list(
    proportion = simple_result$consensus_proportion,
    entropy = simple_result$entropy,
    controversy_threshold = controversy_threshold,
    entropy_threshold = entropy_threshold,
    reasons = failure_reasons
  ))

  message(sprintf("Simple check insufficient (%s; CP=%.2f, H=%.2f) - Using LLM",
                  paste(failure_reasons, collapse = "; "),
                  simple_result$consensus_proportion,
                  simple_result$entropy))

  # Get the formatted prompt from the dedicated function
  formatted_responses <- create_consensus_check_prompt(extracted_cell_types, controversy_threshold, entropy_threshold)

  # Prepare models and execute consensus check
  models_to_try <- prepare_models_list(consensus_check_model)
  execution_result <- execute_consensus_check(formatted_responses, api_keys, models_to_try, base_urls)

  # Handle execution failure - fall back to simple consensus
  if (!execution_result$success) {
    get_logger()$error("All model attempts failed, using simple consensus results")
    return(deterministic_result)
  }

  # Parse the response using the new modular approach
  result <- parse_consensus_response(execution_result$response)

  if (!consensus_metrics_plausible(result, length(extracted_cell_types))) {
    get_logger()$warn("Ignoring impossible LLM consensus metrics", list(
      consensus_proportion = result$consensus_proportion,
      entropy = result$entropy,
      vote_count = length(extracted_cell_types)
    ))
    return(deterministic_result)
  }

  reported_majority <- result$majority_prediction
  valid_majority <- is_real_cell_type_annotation(reported_majority)
  metrics_reach_consensus <- result$consensus_proportion >= controversy_threshold &&
    result$entropy <= entropy_threshold
  llm_claimed_consensus <- isTRUE(result$reached)
  result$reached <- llm_claimed_consensus && metrics_reach_consensus && valid_majority
  if (!valid_majority) {
    result$majority_prediction <- deterministic_result$majority_prediction
  }

  if (llm_claimed_consensus && !result$reached) {
    get_logger()$warn("Ignoring inconsistent LLM consensus indicator", list(
      consensus_proportion = result$consensus_proportion,
      entropy = result$entropy,
      majority_prediction = reported_majority
    ))
  }
  
  # Compare LLM result with simple consensus
  if (result$reached) {
    get_logger()$info("LLM confirmed consensus", list(
      majority_prediction = result$majority_prediction,
      consensus_proportion = result$consensus_proportion,
      entropy = result$entropy
    ))
    message(sprintf("LLM CONFIRMED consensus: %s (CP=%.2f, H=%.2f)",
                    result$majority_prediction,
                    result$consensus_proportion,
                    result$entropy))
  } else {
    get_logger()$info("LLM rejected consensus", list(
      simple_suggestion = simple_result$majority_prediction,
      llm_consensus_proportion = result$consensus_proportion,
      llm_entropy = result$entropy
    ))
    message(sprintf("LLM REJECTED consensus: Simple suggested %s but LLM found CP=%.2f, H=%.2f",
                    simple_result$majority_prediction,
                    result$consensus_proportion,
                    result$entropy))
  }
  

  return(result)
}

#' OpenAI API Processor
#' 
#' Concrete implementation of BaseAPIProcessor for OpenAI models.
#' Handles OpenAI-specific API calls, authentication, and response parsing.
#'
#' @export
OpenAIProcessor <- R6::R6Class("OpenAIProcessor",
  inherit = BaseAPIProcessor,
  
  public = list(
    #' @description
    #' Initialize OpenAI processor
    #' @param base_url Optional custom API endpoint
    initialize = function(base_url = NULL) {
      super$initialize("openai", base_url)
    },

    #' @description
    #' Get default OpenAI API URL
    #
    get_default_api_url = function() {
      return("https://api.openai.com/v1/chat/completions")
    },
    
    #' @description
    #' Make API call to OpenAI
    #' @param chunk_content Prompt text to send
    #' @param model Model identifier
    #' @param api_key OpenAI API key
    make_api_call = function(chunk_content, model, api_key) {
      private$post_chat_completions_request(chunk_content, model, api_key)
    },
    
    #' @description
    #' Extract response content from OpenAI API response
    #' @param response HTTP response object
    #' @param model Model identifier
    extract_response_content = function(response, model) {
      private$extract_chat_completions_content(response, model)
    }
  )
)

#' Kimi API Processor
#'
#' Concrete implementation of BaseAPIProcessor for Kimi models. By default it
#' targets the Moonshot AI Open Platform over the OpenAI-compatible Chat
#' Completions protocol, with k2 thinking mode disabled for deterministic
#' output. A custom `base_url` may instead point at the Kimi Code platform
#' (api.kimi.com/coding), which speaks both protocols; the protocol is
#' inferred from the effective endpoint URL. URLs ending in '/messages' use
#' the Anthropic Messages protocol; the Kimi Code base
#' 'https://api.kimi.com/coding' and URLs ending in '/chat/completions' use
#' OpenAI-compatible Chat Completions.
#'
#' @export
KimiProcessor <- R6::R6Class("KimiProcessor",
  inherit = BaseAPIProcessor,

  public = list(
    #' @description
    #' Initialize Kimi processor
    #' @param base_url Optional custom API endpoint
    initialize = function(base_url = NULL) {
      super$initialize("kimi", base_url)
      private$endpoint <- private$resolve_kimi_endpoint(self$get_api_url())
    },

    #' @description
    #' Get default Kimi API URL
    #
    get_default_api_url = function() {
      return("https://api.moonshot.cn/v1/chat/completions")
    },

    #' @description
    #' Make API call to Kimi
    #' @param chunk_content Prompt text to send
    #' @param model Model identifier (e.g., 'kimi-k2.6', 'moonshot-v1-8k')
    #' @param api_key Moonshot API key
    make_api_call = function(chunk_content, model, api_key) {
      if (identical(private$endpoint$mode, "anthropic")) {
        return(private$post_messages_request(chunk_content, model, api_key))
      }
      private$post_chat_completions_request(
        chunk_content,
        model,
        api_key,
        api_url = private$endpoint$url,
        body_extra = list(
          temperature = 0.6,
          max_tokens = 4096,
          thinking = list(type = "disabled")
        )
      )
    },

    #' @description
    #' Extract response content from Kimi API response
    #' @param response HTTP response object
    #' @param model Model identifier
    extract_response_content = function(response, model) {
      if (identical(private$endpoint$mode, "anthropic")) {
        return(private$extract_messages_content(response, model, "Kimi"))
      }
      private$extract_chat_completions_content(response, model)
    },

    #' @description
    #' Extract normalized token usage from a Kimi API response
    #' @param response HTTP response object
    extract_usage = function(response) {
      if (identical(private$endpoint$mode, "anthropic")) {
        return(private$extract_usage_fields(
          response,
          prompt_field = "input_tokens",
          completion_field = "output_tokens",
          total_field = NULL,
          cost_field = NULL,
          derive_total = TRUE
        ))
      }
      private$extract_usage_fields(response)
    }
  ),

  private = list(
    # Resolved endpoint contract: list(url = <final URL>, mode = "openai"|"anthropic")
    endpoint = NULL,

    # Classify the effective endpoint URL into its final form and protocol.
    # Kimi Code base URLs are completed to their canonical endpoints; every
    # other URL is used as-is in OpenAI-compatible mode.
    resolve_kimi_endpoint = function(url) {
      lower <- tolower(url)
      if (grepl("/messages$", lower)) {
        return(list(url = url, mode = "anthropic"))
      }
      if (grepl("/chat/completions$", lower)) {
        return(list(url = url, mode = "openai"))
      }
      if (grepl("/coding/v1$", lower)) {
        completed <- paste0(url, "/chat/completions")
        self$logger$debug(
          sprintf("Using Kimi Code OpenAI-compatible endpoint: %s", completed),
          list(provider = self$provider_name)
        )
        return(list(url = completed, mode = "openai"))
      }
      if (grepl("/coding$", lower)) {
        completed <- paste0(url, "/v1/chat/completions")
        self$logger$debug(
          sprintf("Using Kimi Code OpenAI-compatible endpoint: %s", completed),
          list(provider = self$provider_name)
        )
        return(list(url = completed, mode = "openai"))
      }
      list(url = url, mode = "openai")
    },

    # Anthropic Messages protocol request (Kimi Code endpoints)
    post_messages_request = function(chunk_content, model, api_key) {
      body <- list(
        model = model,
        max_tokens = 4096,
        temperature = 0.6,
        messages = list(
          list(
            role = "user",
            content = chunk_content
          )
        )
      )

      self$logger$debug("Sending API request to Kimi (Anthropic-compatible mode)",
                       list(model = model, provider = self$provider_name))

      response <- httr::POST(
        url = private$endpoint$url,
        httr::add_headers(
          "x-api-key" = api_key,
          "anthropic-version" = "2023-06-01",
          "content-type" = "application/json"
        ),
        body = body,
        encode = "json",
        httr::timeout(get_api_request_timeout())
      )

      private$stop_for_http_error(response, model)
      response
    }
  )
)

skip_if_not_cran_env <- function() {
  if (!identical(Sys.getenv("NOT_CRAN"), "true")) {
    testthat::skip("Set NOT_CRAN=true to run this test")
  }
}

skip_if_not_integration <- function() {
  enabled <- tolower(Sys.getenv("MLLMCELLTYPE_RUN_INTEGRATION"))
  if (!enabled %in% c("1", "true", "yes", "on")) {
    testthat::skip("Set MLLMCELLTYPE_RUN_INTEGRATION=true to run real API integration tests")
  }
}

skip_if_no_openrouter_key <- function() {
  if (!nzchar(Sys.getenv("OPENROUTER_API_KEY"))) {
    testthat::skip("No OpenRouter API key available")
  }
}

skip_real_api_test <- function() {
  skip_if_not_cran_env()
  skip_if_not_integration()
  skip_if_no_openrouter_key()
}

openrouter_test_models <- c("meta-llama/llama-4-maverick:free", "meta-llama/llama-3.3-70b-instruct:free")

openrouter_api_keys <- function() {
  list(openrouter = Sys.getenv("OPENROUTER_API_KEY"))
}

minimal_pbmc_markers <- function() {
  data.frame(
    cluster = c(0, 0, 0, 1, 1, 1),
    gene = c("CD3D", "CD8A", "IL7R", "CD14", "LYZ", "S100A8"),
    avg_log2FC = c(3.2, 2.1, 1.2, 2.8, 2.3, 1.9)
  )
}

# Real API Integration Tests

test_that("Real API integration with system cache", {
  skip_real_api_test()

  formatted_data <- minimal_pbmc_markers()

  # Test with system cache (NULL)
  result1 <- interactive_consensus_annotation(
    input = formatted_data,
    tissue_name = "human PBMC",
    cache_dir = NULL,
    models = openrouter_test_models,
    api_keys = openrouter_api_keys()
  )

  expect_true(!is.null(result1))
  expect_true("final_annotations" %in% names(result1))

  # Test cache was created in system directory
  cache_dir <- mllmcelltype_cache_dir()
  expect_true(dir.exists(cache_dir))

  # Test local cache in an isolated working directory so the repository is not polluted.
  old_wd <- getwd()
  isolated_wd <- tempfile("mllmcelltype_local_cache_")
  dir.create(isolated_wd)
  on.exit(setwd(old_wd), add = TRUE)
  setwd(isolated_wd)

  result2 <- interactive_consensus_annotation(
    input = formatted_data,
    tissue_name = "human PBMC",
    cache_dir = "local",
    models = openrouter_test_models,
    api_keys = openrouter_api_keys()
  )

  expect_true(!is.null(result2))
  expect_true("final_annotations" %in% names(result2))
  expect_true(dir.exists(file.path(isolated_wd, ".mllmcelltype_cache")))
})

test_that("Cache reuse works with real API", {
  skip_real_api_test()

  test_data <- minimal_pbmc_markers()
  temp_cache <- tempfile("cache_reuse_test_")

  # First call - should hit API
  start_time1 <- Sys.time()
  result1 <- interactive_consensus_annotation(
    input = test_data,
    tissue_name = "human PBMC",
    cache_dir = temp_cache,
    models = openrouter_test_models,
    api_keys = openrouter_api_keys()
  )
  time1 <- as.numeric(Sys.time() - start_time1)

  # Second call - should use cache
  start_time2 <- Sys.time()
  result2 <- interactive_consensus_annotation(
    input = test_data,
    tissue_name = "human PBMC",
    cache_dir = temp_cache,
    models = openrouter_test_models,
    api_keys = openrouter_api_keys()
  )
  time2 <- as.numeric(Sys.time() - start_time2)

  # Cache call should be much faster
  expect_true(time2 < time1 * 0.5)
  expect_equal(result1$final_annotations, result2$final_annotations)
})

test_that("Different cache directories work independently", {
  skip_real_api_test()

  test_data <- data.frame(
    cluster = c(0, 0),
    gene = c("CD3D", "CD8A"),
    avg_log2FC = c(2.5, 2.0)
  )

  cache_dirs <- list(
    temp1 = tempfile("cache_test_1_"),
    temp2 = tempfile("cache_test_2_")
  )

  results <- list()

  for (i in seq_along(cache_dirs)) {
    cache_name <- names(cache_dirs)[i]
    cache_dir <- cache_dirs[[i]]

    result <- interactive_consensus_annotation(
      input = test_data,
      tissue_name = "human PBMC",
      cache_dir = cache_dir,
      models = openrouter_test_models,
      api_keys = openrouter_api_keys()
    )

    results[[cache_name]] <- result

    expect_true(dir.exists(cache_dir))

    cache_files <- list.files(cache_dir, recursive = TRUE)
    expect_true(length(cache_files) > 0)
  }

  expect_equal(results$temp1$final_annotations, results$temp2$final_annotations)
})

#' Base API Processor Class
#'
#' Abstract base class for API processors that provides common functionality
#' including unified logging, error handling, input processing, and response validation.
#' This eliminates code duplication across all provider-specific processors.
#'
#' @export
BaseAPIProcessor <- R6::R6Class("BaseAPIProcessor",
  public = list(
    #' @field provider_name Name of the API provider
    provider_name = NULL,

    #' @field logger Unified logger instance
    logger = NULL,

    #' @field base_url Custom base URL for API endpoints
    base_url = NULL,

    #' @description
    #' Initialize the base API processor
    #' @param provider_name Provider identifier used for logging and dispatch
    #' @param base_url Optional custom API endpoint
    initialize = function(provider_name, base_url = NULL) {
      self$provider_name <- .normalize_required_string(provider_name, "provider_name")
      self$base_url <- resolve_provider_base_url(self$provider_name, base_url)
      self$logger <- get_logger()
      self$logger$info(sprintf("Initialized %s processor", self$provider_name),
                      list(provider = self$provider_name,
                           custom_url = !is.null(self$base_url)))
    },
    
    #' @description
    #' Main entry point for processing API requests
    #' @param prompt Prompt text to send
    #' @param model Model identifier
    #' @param api_key Provider API key
    #' @param normalize Logical. If \code{TRUE} (default), the response is
    #'   normalized into non-empty trimmed lines. If \code{FALSE}, the raw
    #'   response string is returned instead. Set to \code{FALSE} when the
    #'   caller needs the original text (e.g., JSON parsing in reasoning mode).
    process_request = function(prompt, model, api_key, normalize = TRUE) {
      start_time <- Sys.time()

      tryCatch({
        inputs <- private$validate_inputs(prompt, model, api_key)
        prompt <- inputs$prompt
        model <- inputs$model
        api_key <- inputs$api_key

        self$logger$info(sprintf("Starting %s API request", self$provider_name),
                        list(model = model, provider = self$provider_name))

        # Make the API call and extract response
        call_result <- private$call_and_extract(prompt, model, api_key, normalize = normalize)
        final_result <- call_result$response
        
        # Log final status using semantic success (not just exception status)
        duration <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
        semantic_success <- private$is_successful_result(final_result)
        self$logger$log_api_call(
          self$provider_name,
          model,
          duration,
          semantic_success,
          tokens = call_result$usage
        )
        
        return(final_result)
        
      }, error = function(e) {
        duration <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
        self$logger$log_api_call(self$provider_name, model, duration, FALSE)
        self$logger$error(sprintf("%s API request failed: %s", self$provider_name, e$message),
                         list(provider = self$provider_name, model = model, error = e$message))
        if (inherits(e, "mllm_api_error")) {
          stop(e)
        }
        if (inherits(e, "curl_error")) {
          stop_api_request_error(
            sprintf("%s API request failed: %s", self$provider_name, e$message),
            retryable = TRUE
          )
        }
        stop(sprintf("%s API request failed: %s", self$provider_name, e$message))
      })
    },

    #' @description
    #' Get the API URL to use for requests
    #
    get_api_url = function() {
      if (!is.null(self$base_url)) {
        self$logger$debug("Using custom base URL",
                         list(provider = self$provider_name, url = self$base_url))
        return(self$base_url)
      }
      return(self$get_default_api_url())
    },

    #' @description
    #' Abstract method to be implemented by subclasses for getting default API URL
    #
    get_default_api_url = function() {
      stop("get_default_api_url must be implemented by subclass")
    },

    #' @description
    #' Abstract method to be implemented by subclasses for making the actual API call
    #' @param chunk_content Prompt text to send
    #' @param model Model identifier
    #' @param api_key Provider API key
    make_api_call = function(chunk_content, model, api_key) {
      stop("make_api_call must be implemented by subclass")
    },
    
    #' @description
    #' Abstract method to be implemented by subclasses for extracting content from response
    #' @param response HTTP response object
    #' @param model Model identifier
    extract_response_content = function(response, model) {
      stop("extract_response_content must be implemented by subclass")
    },

    #' @description
    #' Extract normalized token usage from a provider response
    #' @param response HTTP response object
    extract_usage = function(response) {
      private$extract_usage_fields(response)
    }
  ),
  
  private = list(
    normalize_processor_input = function(value, argument_name, error_message) {
      tryCatch(
        .normalize_required_string(value, argument_name),
        error = function(e) {
          self$logger$error(
            sprintf("%s is missing, empty, or not a character scalar", argument_name),
            list(provider = self$provider_name, argument = argument_name)
          )
          stop(error_message)
        }
      )
    },

    # Validate and normalize input parameters
    validate_inputs = function(prompt, model, api_key) {
      normalized <- list(
        api_key = private$normalize_processor_input(
          api_key,
          "API key",
          sprintf("%s API key is required but not provided", self$provider_name)
        ),
        prompt = private$normalize_processor_input(
          prompt,
          "Prompt",
          "Prompt is required but not provided"
        ),
        model = private$normalize_processor_input(
          model,
          "Model",
          "Model is required but not provided"
        )
      )
      
      self$logger$debug("Input validation passed",
                       list(provider = self$provider_name, model = normalized$model))
      normalized
    },
    
    #' Make API call and extract response content
    #
    #
    #
    call_and_extract = function(prompt, model, api_key, normalize = TRUE) {
      # Track progress through stages so the error handler knows what failed
      response <- NULL

      tryCatch({
        # Stage 1: API call
        response <- self$make_api_call(prompt, model, api_key)
        # Stage 2: Response extraction
        content <- self$extract_response_content(response, model)
      }, error = function(e) {
        # Unified audit log for failures at any stage
        self$logger$log_api_request_response(
          provider = self$provider_name,
          model = model,
          prompt_content = prompt,
          response_content = paste0("ERROR: ", e$message),
          request_metadata = list(provider = self$provider_name, failed = TRUE),
          response_metadata = list(
            error = e$message,
            stage = if (is.null(response)) "api_call" else "response_extraction"
          )
        )
        stop(e)
      })

      normalized_content <- tryCatch(
        normalize_model_response_lines(content),
        error = function(e) NULL
      )
      if (is.null(normalized_content)) {
        response_type <- typeof(content)
        self$logger$log_api_request_response(
          provider = self$provider_name,
          model = model,
          prompt_content = prompt,
          response_content = paste0("ERROR: Invalid model response (", response_type, ")"),
          request_metadata = list(provider = self$provider_name, failed = TRUE),
          response_metadata = list(
            error = "Invalid response format",
            stage = "response_validation",
            response_type = response_type
          )
        )
        stop("Invalid response format from API")
      }

      # Log successful request and response for audit/debugging
      self$logger$log_api_request_response(
        provider = self$provider_name,
        model = model,
        prompt_content = prompt,
        response_content = content,
        request_metadata = list(provider = self$provider_name),
        response_metadata = list(
          raw_response_class = class(response),
          extracted_content_length = sum(nchar(content))
        )
      )

      self$logger$debug(sprintf("Processed response from %s", self$provider_name),
                       list(provider = self$provider_name,
                            model = model,
                            lines_count = length(normalized_content),
                            response_length = sum(nchar(content))))
      usage <- tryCatch(
        self$extract_usage(response),
        error = function(e) {
          if (is.function(self$logger$warn)) {
            self$logger$warn("Failed to extract token usage", list(
              provider = self$provider_name,
              error = e$message
            ))
          }
          NULL
        }
      )
      final_response <- if (isTRUE(normalize)) normalized_content else content
      return(list(response = final_response, usage = usage))
    },

    build_chat_completions_body = function(chunk_content, model, extra = list()) {
      body <- list(
        model = model,
        messages = list(
          list(
            role = "user",
            content = chunk_content
          )
        )
      )

      if (length(extra) > 0) {
        for (field in names(extra)) {
          body[[field]] <- extra[[field]]
        }
      }

      body
    },

    extract_error_message = function(response) {
      error_content <- tryCatch(
        httr::content(response, "parsed"),
        error = function(e) NULL
      )
      is_non_empty_string <- function(value) {
        is.character(value) && length(value) == 1 && !is.na(value) && nzchar(value)
      }

      if (is.list(error_content) &&
          is.list(error_content$error) &&
          is_non_empty_string(error_content$error$message)) {
        return(error_content$error$message)
      }
      if (is.list(error_content) &&
          is_non_empty_string(error_content$error)) {
        return(error_content$error)
      }
      if (is.list(error_content) &&
          is_non_empty_string(error_content$message)) {
        return(error_content$message)
      }

      sprintf("HTTP %d error", httr::status_code(response))
    },

    stop_for_http_error = function(
      response,
      model,
      provider_label = get_builtin_provider_display_name(self$provider_name)
    ) {
      status_code <- httr::status_code(response)
      if (identical(as.integer(status_code), 200L)) {
        return(invisible(NULL))
      }

      error_message <- private$extract_error_message(response)

      self$logger$error(sprintf("%s API request failed", provider_label),
                       list(error = error_message,
                            provider = self$provider_name,
                            model = model,
                            status_code = status_code))

      stop_api_request_error(
        sprintf("%s API request failed: %s", provider_label, error_message),
        status_code = status_code
      )
    },

    post_chat_completions_request = function(chunk_content,
                                            model,
                                            api_key,
                                            api_url = self$get_api_url(),
                                            body_extra = list(),
                                            headers = list(),
                                            provider_label = get_builtin_provider_display_name(
                                              self$provider_name
                                            )) {
      body <- private$build_chat_completions_body(
        chunk_content = chunk_content,
        model = model,
        extra = body_extra
      )

      self$logger$debug(sprintf("Sending API request to %s", provider_label),
                       list(model = model, provider = self$provider_name))

      request_headers <- c(
        list(
          "Authorization" = paste("Bearer", api_key),
          "Content-Type" = "application/json"
        ),
        headers
      )

      response <- httr::POST(
        url = api_url,
        do.call(httr::add_headers, request_headers),
        body = body,
        encode = "json",
        httr::timeout(get_api_request_timeout())
      )

      private$stop_for_http_error(response, model, provider_label)
      response
    },

    extract_chat_completions_content = function(response,
                                               model,
                                               provider_label = get_builtin_provider_display_name(
                                                 self$provider_name
                                               )) {
      self$logger$debug(sprintf("Parsing %s API response", provider_label),
                       list(provider = self$provider_name, model = model))

      content <- httr::content(response, "parsed")

      if (is.null(content) ||
          is.null(content$choices) ||
          length(content$choices) == 0 ||
          is.null(content$choices[[1]]$message) ||
          is.null(content$choices[[1]]$message$content)) {

        self$logger$error(sprintf("Unexpected response format from %s API", provider_label),
                         list(provider = self$provider_name,
                              model = model,
                              content_structure = names(content),
                              choices_available = !is.null(content$choices),
                              choices_count = if (!is.null(content$choices)) {
                                length(content$choices)
                              } else {
                                0
                              }))

        stop(sprintf("Unexpected response format from %s API", provider_label))
      }

      if (identical(content$choices[[1]]$finish_reason, "length")) {
        self$logger$warn(
          sprintf(
            "%s response was truncated (finish_reason='length'); trailing clusters may be marked Unknown",
            provider_label
          ),
          list(provider = self$provider_name, model = model)
        )
      }

      content$choices[[1]]$message$content
    },

    extract_messages_content = function(response,
                                        model,
                                        provider_label = get_builtin_provider_display_name(
                                          self$provider_name
                                        )) {
      self$logger$debug(sprintf("Parsing %s Messages API response", provider_label),
                       list(provider = self$provider_name, model = model))

      content <- httr::content(response, "parsed")

      if (is.null(content) || is.null(content$content) || length(content$content) == 0) {
        self$logger$error(sprintf("Unexpected response format from %s Messages API", provider_label),
                         list(provider = self$provider_name,
                              model = model,
                              content_structure = names(content),
                              content_available = !is.null(content$content)))
        stop(sprintf("Unexpected response format from %s Messages API", provider_label))
      }

      text_blocks <- vapply(
        content$content,
        function(block) {
          if (is.list(block) && !is.null(block$text) && is.character(block$text) &&
              length(block$text) == 1 &&
              (is.null(block$type) || identical(block$type, "text"))) {
            return(block$text)
          }
          NA_character_
        },
        character(1)
      )
      text_blocks <- text_blocks[!is.na(text_blocks)]

      if (length(text_blocks) == 0) {
        block_types <- vapply(
          content$content,
          function(block) {
            if (!is.list(block) || is.null(block$type)) "unknown" else as.character(block$type)
          },
          character(1)
        )
        self$logger$error(
          sprintf("Unexpected response format from %s Messages API: no text block found", provider_label),
          list(provider = self$provider_name,
               model = model,
               content_structure = names(content),
               content_count = length(content$content),
               content_types = block_types)
        )
        stop(sprintf("Unexpected response format from %s Messages API", provider_label))
      }

      if (identical(content$stop_reason, "max_tokens")) {
        self$logger$warn(
          sprintf(
            "%s response was truncated (stop_reason='max_tokens'); trailing clusters may be marked Unknown",
            provider_label
          ),
          list(provider = self$provider_name, model = model)
        )
      }

      paste(text_blocks, collapse = "\n")
    },

    extract_usage_fields = function(response,
                                    usage_field = "usage",
                                    prompt_field = "prompt_tokens",
                                    completion_field = "completion_tokens",
                                    total_field = "total_tokens",
                                    cost_field = "cost",
                                    derive_total = FALSE) {
      if (!inherits(response, "response")) {
        return(NULL)
      }
      content <- tryCatch(
        httr::content(response, "parsed"),
        error = function(e) NULL
      )
      if (!is.list(content)) {
        return(NULL)
      }
      usage <- content[[usage_field, exact = TRUE]]
      if (!is.list(usage)) {
        return(NULL)
      }

      field_value <- function(field) {
        if (is.null(field)) NULL else usage[[field, exact = TRUE]]
      }
      normalized <- normalize_token_usage(list(
        prompt_tokens = field_value(prompt_field),
        completion_tokens = field_value(completion_field),
        total_tokens = field_value(total_field),
        cost = field_value(cost_field)
      ))
      if (isTRUE(derive_total) &&
          !is.null(normalized$prompt_tokens) &&
          !is.null(normalized$completion_tokens)) {
        normalized$total_tokens <- normalized$prompt_tokens +
          normalized$completion_tokens
      }
      normalized
    },

    is_successful_result = function(result) {
      is_valid_model_response(result)
    }
  )
)

# Tests for the Kimi provider (Moonshot AI Open Platform)

make_kimi_json_response <- function(payload,
                                    status_code = 200L,
                                    url = "https://api.moonshot.cn/v1/chat/completions") {
  structure(
    list(
      url = url,
      status_code = status_code,
      headers = list("Content-Type" = "application/json"),
      all_headers = list(list(
        status = status_code,
        version = "HTTP/1.1",
        headers = list("Content-Type" = "application/json")
      )),
      cookies = data.frame(),
      content = charToRaw(jsonlite::toJSON(payload, auto_unbox = TRUE)),
      date = Sys.time(),
      times = c(),
      request = list(method = "POST")
    ),
    class = "response"
  )
}

preserve_kimi_env_variables <- function(variables) {
  previous <- Sys.getenv(variables, unset = NA_character_)

  function() {
    Sys.unsetenv(variables)
    configured <- !is.na(previous)
    if (any(configured)) {
      do.call(
        Sys.setenv,
        setNames(as.list(previous[configured]), variables[configured])
      )
    }
  }
}

quiet_kimi_logger <- function() {
  list(
    info = function(...) NULL,
    debug = function(...) NULL,
    warn = function(...) NULL,
    error = function(...) NULL,
    log_api_call = function(...) NULL,
    log_api_request_response = function(...) NULL
  )
}

test_that("get_provider detects Kimi Open Platform model prefixes", {
  expect_identical(get_provider("kimi-k2.6"), "kimi")
  expect_identical(get_provider("moonshot-v1-8k"), "kimi")
  expect_identical(get_provider("KIMI-K2.6"), "kimi")
  expect_identical(get_provider("  Moonshot-v1-8k  "), "kimi")
})

test_that("kimi registry entry owns pattern, processor, display name, and env aliases", {
  spec <- .BUILTIN_PROVIDER_SPECS$kimi
  expect_identical(spec$pattern, "^(kimi-|moonshot-)")
  expect_identical(spec$processor_class, "KimiProcessor")
  expect_identical(spec$display_name, "Kimi")
  expect_identical(spec$api_key_env_aliases, "MOONSHOT_API_KEY")
  expect_identical(.BUILTIN_PROVIDER_PATTERNS[["kimi"]], "^(kimi-|moonshot-)")
  expect_identical(get_builtin_provider_display_name("kimi"), "Kimi")
  expect_silent(validate_builtin_provider_registry())

  processor <- new_builtin_provider_processor("kimi")
  expect_s3_class(processor, "KimiProcessor")
  expect_identical(processor$provider_name, "kimi")
})

test_that("KimiProcessor targets the Moonshot Open Platform endpoint by default", {
  processor <- KimiProcessor$new()
  expect_identical(
    processor$get_default_api_url(),
    "https://api.moonshot.cn/v1/chat/completions"
  )
  expect_identical(
    processor$get_api_url(),
    "https://api.moonshot.cn/v1/chat/completions"
  )
})

test_that("KimiProcessor honors and normalizes a custom base URL", {
  processor <- KimiProcessor$new("  https://proxy.example.test/v1/chat/completions/  ")
  expect_identical(
    processor$get_api_url(),
    "https://proxy.example.test/v1/chat/completions"
  )
})

test_that("kimi credentials resolve KIMI_API_KEY before the MOONSHOT_API_KEY alias", {
  variables <- c("KIMI_API_KEY", "MOONSHOT_API_KEY")
  restore_environment <- preserve_kimi_env_variables(variables)
  on.exit(restore_environment(), add = TRUE)
  Sys.unsetenv(variables)

  Sys.setenv(MOONSHOT_API_KEY = "  alias-key  ")
  alias_credential <- .resolve_model_api_key("kimi-k2.6", api_keys = NULL)
  expect_identical(alias_credential$api_key, "alias-key")
  expect_identical(alias_credential$provider, "kimi")
  expect_identical(alias_credential$source, "environment")

  Sys.setenv(KIMI_API_KEY = " primary-key ")
  primary_credential <- .resolve_model_api_key("kimi-k2.6", api_keys = NULL)
  expect_identical(primary_credential$api_key, "primary-key")

  explicit_credential <- .resolve_model_api_key(
    "kimi-k2.6",
    api_keys = list(kimi = "  explicit-key  ")
  )
  expect_identical(explicit_credential$api_key, "explicit-key")
  expect_identical(explicit_credential$source, "api_keys")
})

test_that("KimiProcessor sends Bearer auth, model, messages, and disabled thinking", {
  captured <- new.env(parent = emptyenv())

  testthat::with_mocked_bindings({
    testthat::with_mocked_bindings({
      processor <- KimiProcessor$new()
      response <- processor$make_api_call("genes", "kimi-k2.6", "test-key")

      expect_identical(captured$url, "https://api.moonshot.cn/v1/chat/completions")
      expect_identical(captured$headers[["Authorization"]], "Bearer test-key")
      expect_identical(captured$headers[["Content-Type"]], "application/json")
      expect_identical(captured$body$model, "kimi-k2.6")
      expect_identical(
        captured$body$messages,
        list(list(role = "user", content = "genes"))
      )
      expect_identical(captured$body$temperature, 0.6)
      expect_identical(captured$body$max_tokens, 4096)
      expect_identical(captured$body$thinking, list(type = "disabled"))
      expect_s3_class(response, "response")
    },
    POST = function(url, config = NULL, body = NULL, ...) {
      captured$url <- url
      captured$headers <- config$headers
      captured$body <- body
      make_kimi_json_response(list(
        choices = list(list(message = list(content = "Cluster 1: T cells")))
      ))
    },
    .package = "httr")
  },
  get_logger = quiet_kimi_logger)
})

test_that("KimiProcessor parses OpenAI-compatible content and normalized usage", {
  payload <- list(
    choices = list(list(message = list(
      content = "Cluster 1: T cells\nCluster 2: B cells"
    ))),
    usage = list(prompt_tokens = 30, completion_tokens = 6, total_tokens = 36)
  )
  response <- make_kimi_json_response(payload)

  testthat::with_mocked_bindings({
    processor <- KimiProcessor$new()
    expect_identical(
      processor$extract_response_content(response, "kimi-k2.6"),
      "Cluster 1: T cells\nCluster 2: B cells"
    )
    expect_identical(
      processor$extract_usage(response),
      list(prompt_tokens = 30, completion_tokens = 6, total_tokens = 36)
    )
    expect_null(processor$extract_usage(make_kimi_json_response(list(choices = list()))))
  },
  get_logger = quiet_kimi_logger)
})

test_that("KimiProcessor normalizes response lines through process_request", {
  testthat::with_mocked_bindings({
    testthat::with_mocked_bindings({
      processor <- KimiProcessor$new()
      result <- processor$process_request("genes", "kimi-k2.6", "test-key")
      expect_identical(result, c("Cluster 1: T cells", "Cluster 2: B cells"))
    },
    POST = function(...) {
      make_kimi_json_response(list(
        choices = list(list(message = list(
          content = "  Cluster 1: T cells,  \n\nCluster 2: B cells,\n"
        ))),
        usage = list(prompt_tokens = 30, completion_tokens = 6, total_tokens = 36)
      ))
    },
    .package = "httr")
  },
  get_logger = quiet_kimi_logger)
})

test_that("KimiProcessor surfaces provider error messages for non-200 responses", {
  testthat::with_mocked_bindings({
    testthat::with_mocked_bindings({
      processor <- KimiProcessor$new()
      error <- tryCatch(
        processor$make_api_call("genes", "kimi-k2.6", "test-key"),
        error = identity
      )
      expect_s3_class(error, "mllm_api_error")
      expect_identical(error$status_code, 401L)
      expect_false(error$retryable)
      expect_identical(
        conditionMessage(error),
        "Kimi API request failed: Invalid authentication credentials"
      )
    },
    POST = function(...) {
      make_kimi_json_response(
        list(error = list(message = "Invalid authentication credentials")),
        status_code = 401L
      )
    },
    .package = "httr")
  },
  get_logger = quiet_kimi_logger)
})

test_that("KimiProcessor rejects malformed response payloads", {
  testthat::with_mocked_bindings({
    processor <- KimiProcessor$new()
    expect_error(
      processor$extract_response_content(
        make_kimi_json_response(list(invalid = "payload")),
        "kimi-k2.6"
      ),
      "Unexpected response format from Kimi API"
    )
    expect_error(
      processor$extract_response_content(
        make_kimi_json_response(list(choices = list(list()))),
        "kimi-k2.6"
      ),
      "Unexpected response format from Kimi API"
    )
  },
  get_logger = quiet_kimi_logger)
})

test_that("get_model_response dispatches kimi models through KimiProcessor", {
  testthat::with_mocked_bindings({
    testthat::with_mocked_bindings({
      result <- get_model_response("genes", "kimi-k2.6", "test-key")
      expect_identical(result, "Cluster 1: T cells")
    },
    POST = function(url, ...) {
      expect_identical(url, "https://api.moonshot.cn/v1/chat/completions")
      make_kimi_json_response(list(
        choices = list(list(message = list(content = "Cluster 1: T cells")))
      ))
    },
    .package = "httr")
  },
  get_logger = quiet_kimi_logger)
})

test_that("KimiProcessor switches to the Anthropic protocol for Messages endpoints", {
  captured <- new.env(parent = emptyenv())

  testthat::with_mocked_bindings({
    testthat::with_mocked_bindings({
      processor <- KimiProcessor$new("https://api.kimi.com/coding/v1/messages")
      response <- processor$make_api_call("genes", "kimi-for-coding", "test-key")

      expect_identical(captured$url, "https://api.kimi.com/coding/v1/messages")
      expect_identical(captured$headers[["x-api-key"]], "test-key")
      expect_identical(captured$headers[["anthropic-version"]], "2023-06-01")
      expect_false("Authorization" %in% names(captured$headers))
      expect_identical(captured$body$model, "kimi-for-coding")
      expect_identical(
        captured$body$messages,
        list(list(role = "user", content = "genes"))
      )
      expect_identical(captured$body$max_tokens, 4096)
      expect_null(captured$body$thinking)
      expect_s3_class(response, "response")
    },
    POST = function(url, config = NULL, body = NULL, ...) {
      captured$url <- url
      captured$headers <- config$headers
      captured$body <- body
      make_kimi_json_response(list(content = list(list(text = "Cluster 1: T cells"))))
    },
    .package = "httr")
  },
  get_logger = quiet_kimi_logger)
})

test_that("KimiProcessor completes Kimi Code base URLs to the OpenAI protocol by default", {
  captured <- new.env(parent = emptyenv())

  testthat::with_mocked_bindings({
    testthat::with_mocked_bindings({
      coding_processor <- KimiProcessor$new("https://api.kimi.com/coding/")
      coding_processor$make_api_call("genes", "kimi-for-coding", "test-key")
      expect_identical(captured$url, "https://api.kimi.com/coding/v1/chat/completions")
      expect_identical(captured$headers[["Authorization"]], "Bearer test-key")
      expect_identical(captured$body$thinking, list(type = "disabled"))

      openai_processor <- KimiProcessor$new("https://api.kimi.com/coding/v1")
      openai_processor$make_api_call("genes", "kimi-for-coding", "test-key")
      expect_identical(captured$url, "https://api.kimi.com/coding/v1/chat/completions")
      expect_identical(captured$headers[["Authorization"]], "Bearer test-key")
      expect_identical(captured$body$thinking, list(type = "disabled"))
    },
    POST = function(url, config = NULL, body = NULL, ...) {
      captured$url <- url
      captured$headers <- config$headers
      captured$body <- body
      make_kimi_json_response(list(
        choices = list(list(message = list(content = "Cluster 1: T cells")))
      ))
    },
    .package = "httr")
  },
  get_logger = quiet_kimi_logger)
})

test_that("KimiProcessor parses Anthropic-mode content and derives total usage", {
  processor <- KimiProcessor$new("https://api.kimi.com/coding/v1/messages")
  response <- make_kimi_json_response(list(
    content = list(list(text = "Cluster 1: T cells\nCluster 2: B cells")),
    usage = list(input_tokens = 20, output_tokens = 4)
  ))

  testthat::with_mocked_bindings({
    expect_identical(
      processor$extract_response_content(response, "kimi-for-coding"),
      "Cluster 1: T cells\nCluster 2: B cells"
    )
    expect_identical(
      processor$extract_usage(response),
      list(prompt_tokens = 20, completion_tokens = 4, total_tokens = 24)
    )
    expect_error(
      processor$extract_response_content(
        make_kimi_json_response(list(invalid = "payload")),
        "kimi-for-coding"
      ),
      "Unexpected response format from Kimi Messages API"
    )
  },
  get_logger = quiet_kimi_logger)
})

test_that("KimiProcessor parses multi-block Anthropic-mode responses", {
  processor <- KimiProcessor$new("https://api.kimi.com/coding/v1/messages")

  thinking_then_text <- make_kimi_json_response(list(
    content = list(
      list(type = "thinking", thinking = "reasoning..."),
      list(type = "text", text = "Cluster 1: T cells")
    )
  ))

  multiple_text <- make_kimi_json_response(list(
    content = list(
      list(type = "text", text = "Cluster 1: T cells"),
      list(type = "text", text = "Cluster 2: B cells")
    )
  ))

  only_thinking <- make_kimi_json_response(list(
    content = list(list(type = "thinking", thinking = "reasoning..."))
  ))

  testthat::with_mocked_bindings({
    expect_identical(
      processor$extract_response_content(thinking_then_text, "kimi-for-coding"),
      "Cluster 1: T cells"
    )
    expect_identical(
      processor$extract_response_content(multiple_text, "kimi-for-coding"),
      "Cluster 1: T cells\nCluster 2: B cells"
    )
    expect_error(
      processor$extract_response_content(only_thinking, "kimi-for-coding"),
      "Unexpected response format from Kimi Messages API"
    )
  },
  get_logger = quiet_kimi_logger)
})

test_that("KimiProcessor surfaces Anthropic-mode HTTP errors", {
  testthat::with_mocked_bindings({
    testthat::with_mocked_bindings({
      processor <- KimiProcessor$new("https://api.kimi.com/coding/v1/messages")
      error <- tryCatch(
        processor$make_api_call("genes", "kimi-for-coding", "test-key"),
        error = identity
      )
      expect_s3_class(error, "mllm_api_error")
      expect_identical(error$status_code, 401L)
      expect_false(error$retryable)
      expect_identical(
        conditionMessage(error),
        "Kimi API request failed: invalid api key"
      )
    },
    POST = function(...) {
      make_kimi_json_response(
        list(error = list(message = "invalid api key")),
        status_code = 401L
      )
    },
    .package = "httr")
  },
  get_logger = quiet_kimi_logger)
})

#' @keywords internal
"_PACKAGE"

#' @importFrom R6 R6Class
#' @importFrom digest digest
#' @importFrom jsonlite toJSON
#' @importFrom tools R_user_dir
#' @importFrom utils head

#' @title Cell Type Annotation with Multi-LLM Framework
#' @name annotate_cell_types
#'
#' @description
#' A comprehensive function for automated cell type annotation using multiple Large Language Models (LLMs).
#' This function supports both Seurat's differential gene expression results and custom gene lists as input.
#' It implements a sophisticated annotation pipeline that leverages state-of-the-art LLMs to identify
#' cell types based on marker gene expression patterns.
#'
#
#'   - A data frame from Seurat's FindAllMarkers() function containing differential gene expression results
#'     (must have columns: 'cluster', 'gene', and 'avg_log2FC'). The function will select the top genes
#'     based on avg_log2FC for each cluster.
#'   - A list where each element has a 'genes' field containing marker genes for a cluster.
#'     This can be in one of these formats:
#'     * Named with cluster IDs: list("0" = list(genes = c(...)), "1" = list(genes = c(...)))
#'     * Named with cell type names: list(t_cells = list(genes = c(...)), b_cells = list(genes = c(...)))
#'     * Unnamed list: list(list(genes = c(...)), list(genes = c(...)))
#'   - Cluster IDs are preserved as-is. The function does not modify or re-index cluster IDs.
#
#'   'mouse brain'). This helps provide context for more accurate annotations.
#
#'   - OpenAI: 'gpt-5.5', 'gpt-5.4', 'gpt-5.4-mini'
#'   - Anthropic: 'claude-opus-4-7', 'claude-opus-4-6', 'claude-sonnet-4-6', 'claude-haiku-4-5-20251001'
#'   - DeepSeek: 'deepseek-v4-flash', 'deepseek-v4-pro'
#'   - Google: 'gemini-3.1-pro-preview', 'gemini-3-flash-preview', 'gemini-3.1-flash-lite'
#'   - Alibaba: 'qwen3.6-max-preview', 'qwen3.6-plus', 'qwen3.6-flash'
#'   - Stepfun: 'step-3.5-flash', 'step-3.5-flash-2603', 'step-3'
#'   - Zhipu/Z.AI: 'glm-5.1', 'glm-5-turbo', 'glm-5'
#'   - MiniMax: 'MiniMax-M2.7', 'MiniMax-M2.7-highspeed', 'MiniMax-M2.5'
#'   - X.AI: 'grok-4.3', 'grok-4.3-latest', 'grok-latest'
#'   - OpenRouter: Provides access to models from multiple providers through a single API. Format: 'provider/model-name'
#'     - OpenAI models: 'openai/gpt-5.5', 'openai/gpt-5.4-mini'
#'     - Anthropic models: 'anthropic/claude-opus-4.7', 'anthropic/claude-sonnet-4.6'
#'     - Google models: 'google/gemini-3.1-pro-preview', 'google/gemini-3-flash-preview'
#'     - X.AI models: 'x-ai/grok-4.3'
#'     - Stepfun models: 'stepfun/step-3.5-flash'
#
#'   Each provider requires a specific API key format and authentication method:
#'
#'   - OpenAI: "sk-..." (obtain from OpenAI platform)
#'   - Anthropic: "sk-ant-..." (obtain from Anthropic console)
#'   - Google: A Google API key for Gemini models (obtain from Google AI)
#'   - DeepSeek: API key from DeepSeek platform
#'   - Qwen: API key from Alibaba Cloud
#'   - Stepfun: API key from Stepfun AI
#'   - Zhipu: API key from Zhipu AI
#'   - MiniMax: API key from MiniMax
#'   - X.AI: API key for Grok models
#'   - OpenRouter: "sk-or-..." (obtain from OpenRouter)
#'     OpenRouter provides access to multiple models through a single API key
#'
#'   The API key can be provided directly or stored in environment variables:
#'   ```r
#'   # Direct API key
#'   result <- annotate_cell_types(input, tissue_name, model="gpt-5.5",
#'                                api_key="sk-...")
#'
#'   # Using environment variables
#'   Sys.setenv(OPENAI_API_KEY="sk-...")
#'   Sys.setenv(ANTHROPIC_API_KEY="sk-ant-...")
#'   Sys.setenv(OPENROUTER_API_KEY="sk-or-...")
#'
#'   # Then use with environment variables
#'   result <- annotate_cell_types(input, tissue_name, model="claude-sonnet-4-6",
#'                                api_key=Sys.getenv("ANTHROPIC_API_KEY"))
#'   ```
#'
#'   If NA, returns the generated prompt without making an API call, which is useful for
#'   reviewing the prompt before sending it to the API.
#
#'   when input is from Seurat's FindAllMarkers(). Default: 10
#
#
#'   - A single character string: Applied to all providers (e.g., "https://api.proxy.com/v1")
#'   - A named list: Provider-specific URLs (e.g., list(openai = "https://openai-proxy.com/v1",
#'     anthropic = "https://anthropic-proxy.com/v1")). This is useful for:
#'     * Users accessing international APIs through proxies
#'     * Enterprise users with internal API gateways
#'     * Development/testing with local or alternative endpoints
#'   If NULL (default), uses official API endpoints for each provider.
#'
#' @param input Either a data frame from Seurat's FindAllMarkers() containing columns 'cluster', 'gene', and 'avg_log2FC', or a list with 'genes' field for each cluster
#' @param tissue_name Required tissue context (e.g., 'human PBMC', 'mouse brain') for more accurate annotations
#' @param model Model name to use. Default: 'gpt-5.5'. See details for supported models
#' @param api_key API key for the selected model provider as a non-empty character scalar.
#'   If \code{NA}, returns prompt only.
#' @param top_gene_count Number of top genes to use per cluster when input is from Seurat. Default: 10
#' @param debug Logical indicating whether to enable debug output. Default: FALSE
#' @param base_urls Optional base URLs for API endpoints. Can be a string or named list for custom endpoints
#' @param return_reasoning Logical. If \code{TRUE}, returns a structured list per cluster containing
#'   \code{cell_type}, \code{marker_genes}, and \code{gene_expression} fields instead of plain labels.
#'   Default: \code{FALSE}.
#'
#' @return When `api_key` is provided and `return_reasoning` is \code{FALSE}, the provider response
#'   split by newline as a character vector. When `return_reasoning` is \code{TRUE}, a named list
#'   where each element is a list with \code{cell_type}, \code{marker_genes}, and
#'   \code{gene_expression}. When `api_key` is \code{NA}, the generated prompt string.
#'
#' @examples
#' # Example 1: Using custom gene lists, returning prompt only (no API call)
#' annotate_cell_types(
#'   input = list(
#'     t_cells = list(genes = c('CD3D', 'CD3E', 'CD3G', 'CD28')),
#'     b_cells = list(genes = c('CD19', 'CD79A', 'CD79B', 'MS4A1')),
#'     monocytes = list(genes = c('CD14', 'CD68', 'CSF1R', 'FCGR3A'))
#'   ),
#'   tissue_name = 'human PBMC',
#'   model = 'gpt-5.5',
#'   api_key = NA  # Returns prompt only without making API call
#' )
#'
#' # Example 2: Using with Seurat pipeline and OpenAI model
#' \dontrun{
#' library(Seurat)
#'
#' # Load example data
#' data("pbmc_small")
#'
#' # Find marker genes
#' all.markers <- FindAllMarkers(
#'   object = pbmc_small,
#'   only.pos = TRUE,
#'   min.pct = 0.25,
#'   logfc.threshold = 0.25
#' )
#'
#' # Set API key in environment variable (recommended approach)
#' Sys.setenv(OPENAI_API_KEY = "your-openai-api-key")
#'
#' # Get cell type annotations using OpenAI model
#' openai_annotations <- annotate_cell_types(
#'   input = all.markers,
#'   tissue_name = 'human PBMC',
#'   model = 'gpt-5.5',
#'   api_key = Sys.getenv("OPENAI_API_KEY"),
#'   top_gene_count = 15
#' )
#'
#' # Example 3: Using Anthropic Claude model
#' Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-api-key")
#'
#' claude_annotations <- annotate_cell_types(
#'   input = all.markers,
#'   tissue_name = 'human PBMC',
#'   model = 'claude-opus-4-7',
#'   api_key = Sys.getenv("ANTHROPIC_API_KEY"),
#'   top_gene_count = 15
#' )
#'
#' # Example 4: Using OpenRouter to access multiple models
#' Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key")
#'
#' # Access OpenAI models through OpenRouter
#' openrouter_gpt4_annotations <- annotate_cell_types(
#'   input = all.markers,
#'   tissue_name = 'human PBMC',
#'   model = 'openai/gpt-5.5',  # Note the provider/model format
#'   api_key = Sys.getenv("OPENROUTER_API_KEY"),
#'   top_gene_count = 15
#' )
#'
#' # Access Anthropic models through OpenRouter
#' openrouter_claude_annotations <- annotate_cell_types(
#'   input = all.markers,
#'   tissue_name = 'human PBMC',
#'   model = 'anthropic/claude-opus-4.6',  # Note the provider/model format
#'   api_key = Sys.getenv("OPENROUTER_API_KEY"),
#'   top_gene_count = 15
#' )
#'
#' # Example 5: Using with mouse brain data
#' mouse_annotations <- annotate_cell_types(
#'   input = mouse_markers,  # Your mouse marker genes
#'   tissue_name = 'mouse brain',  # Specify correct tissue for context
#'   model = 'gpt-5.5',
#'   api_key = Sys.getenv("OPENAI_API_KEY"),
#'   top_gene_count = 20,  # Use more genes for complex tissues
#'   debug = TRUE  # Enable debug output
#' )
#' }
#'
#' @seealso
#' * [Seurat::FindAllMarkers()]
#' * [mLLMCelltype::get_provider()]
#' @export
annotate_cell_types <- function(input,
                               tissue_name,
                               model = 'gpt-5.5',
                               api_key = NA,
                               top_gene_count = 10,
                               debug = FALSE,
                               base_urls = NULL,
                               return_reasoning = FALSE) {

  tissue_name <- tryCatch(
    .normalize_required_string(tissue_name, "tissue_name"),
    error = function(e) {
      stop("tissue_name is required. Specify the tissue type (e.g., 'human PBMC', 'mouse brain').")
    }
  )
  model <- .normalize_required_string(model, "model")
  top_gene_count <- .normalize_top_gene_count(top_gene_count)
  debug <- .normalize_flag(debug, "debug")
  return_reasoning <- .normalize_flag(return_reasoning, "return_reasoning")

  # Determine provider from model name
  provider <- get_provider(model)

  # Log model and provider information
  log_info("Processing input with model and provider", list(
    model = model, provider = provider,
    custom_url = !is.null(resolve_provider_base_url(provider, base_urls)),
    return_reasoning = return_reasoning
  ))

  # Generate prompt using the dedicated function
  if (return_reasoning) {
    prompt_result <- create_reasoning_annotation_prompt(input, tissue_name, top_gene_count)
  } else {
    prompt_result <- create_annotation_prompt(input, tissue_name, top_gene_count)
  }
  prompt <- prompt_result$prompt

  # If debug mode is enabled, temporarily enable console debug output
  # so the log_debug() calls below become visible to the user.
  if (debug) {
    .logger <- get_logger()
    .old_level <- .logger$log_level
    .old_console <- .logger$enable_console
    .logger$set_level("DEBUG")
    .logger$enable_console <- TRUE
    on.exit({
      .logger$set_level(.old_level)
      .logger$enable_console <- .old_console
    }, add = TRUE)
  }

  # Log gene lists (visible on console only when debug = TRUE)
  log_debug("Gene lists for each cluster")
  cluster_ids <- names(prompt_result$gene_lists)
  for (id in cluster_ids) {
    log_debug("Cluster gene list", list(cluster = id, genes = prompt_result$gene_lists[[id]]))
  }

  log_debug("Generated prompt", list(prompt = prompt))

  # If no API key, return prompt
  if (length(api_key) == 1 && is.na(api_key)) {
    return(prompt)
  }

  if (!is.character(api_key) || length(api_key) != 1 || is.na(api_key) ||
      !nzchar(trimws(api_key))) {
    stop("api_key must be a non-empty character scalar, or NA to return prompt only")
  }
  api_key <- trimws(api_key)

  # Delegate to get_model_response which handles provider dispatch.
  # For reasoning mode, preserve the raw response string so JSON structure
  # (including trailing commas) is not altered by line normalization.
  result <- get_model_response(
    prompt, model, api_key, base_urls,
    normalize = !return_reasoning
  )

  if (return_reasoning) {
    cluster_ids <- names(prompt_result$gene_lists)
    return(tryCatch(
      parse_reasoning_annotations(result, cluster_ids),
      error = function(e) {
        log_warn(
          "Failed to parse reasoning response; returning Unknown for all clusters",
          list(error = e$message)
        )
        unknown_record <- list(
          cell_type = "Unknown",
          marker_genes = "",
          gene_expression = ""
        )
        stats::setNames(rep(list(unknown_record), length(cluster_ids)), cluster_ids)
      }
    ))
  }

  logger <- get_logger()
  if (!is.null(logger$log_model_response) && is.function(logger$log_model_response)) {
    logger$log_model_response(
      provider = provider,
      model = model,
      response = result,
      stage = "annotation"
    )
  } else {
    # Fallback for compatibility with mocked logger objects in tests
    response_text <- if (is.character(result)) paste(result, collapse = "\n") else as.character(result)
    preview <- if (nchar(response_text) > 180) paste0(substr(response_text, 1, 180), "...") else response_text
    log_info("Model response received", list(
      provider = provider,
      model = model,
      response_chars = nchar(response_text),
      response_lines = length(strsplit(response_text, "\n", fixed = TRUE)[[1]]),
      response_preview = preview
    ))
  }

  return(result)
}

# Tests for annotate_cell_types return_reasoning functionality

test_that("annotate_cell_types return_reasoning=TRUE parses JSON array response", {
  json_response <- paste0(
    '[{"cluster_id": "0", "cell_type": "T cells", ',
    '"marker_genes": "CD3D, CD3E", ',
    '"gene_expression": "CD3D is highly expressed in T cells"}, ',
    '{"cluster_id": "1", "cell_type": "B cells", ',
    '"marker_genes": "CD19, CD79A", ',
    '"gene_expression": "CD19 is highly expressed in B cells"}]'
  )

  result <- testthat::with_mocked_bindings({
    annotate_cell_types(
      input = list(
        "0" = list(genes = c("CD3D", "CD3E")),
        "1" = list(genes = c("CD19", "CD79A"))
      ),
      tissue_name = "PBMC",
      model = "gpt-5.5",
      api_key = "test-key",
      return_reasoning = TRUE
    )
  },
  get_model_response = function(prompt, model, api_key, base_urls, normalize = TRUE) json_response,
  log_info = function(...) NULL,
  log_debug = function(...) NULL)

  expect_type(result, "list")
  expect_named(result, c("0", "1"))
  expect_identical(result[["0"]]$cell_type, "T cells")
  expect_identical(result[["0"]]$marker_genes, "CD3D, CD3E")
  expect_match(result[["0"]]$gene_expression, "CD3D")
  expect_identical(result[["1"]]$cell_type, "B cells")
})


test_that("annotate_cell_types return_reasoning=TRUE parses JSON object response", {
  json_response <- paste0(
    '{"0": {"cell_type": "T cells", "marker_genes": "CD3D", ',
    '"gene_expression": "CD3D in T cells"}, ',
    '"1": {"cell_type": "B cells", "marker_genes": "CD19", ',
    '"gene_expression": "CD19 in B cells"}}'
  )

  result <- testthat::with_mocked_bindings({
    annotate_cell_types(
      input = list(
        "0" = list(genes = c("CD3D")),
        "1" = list(genes = c("CD19"))
      ),
      tissue_name = "PBMC",
      model = "gpt-5.5",
      api_key = "test-key",
      return_reasoning = TRUE
    )
  },
  get_model_response = function(prompt, model, api_key, base_urls, normalize = TRUE) json_response,
  log_info = function(...) NULL,
  log_debug = function(...) NULL)

  expect_identical(result[["0"]]$cell_type, "T cells")
  expect_identical(result[["1"]]$cell_type, "B cells")
})


test_that("annotate_cell_types return_reasoning preserves marker gene case", {
  json_response <- '[{"cluster_id": "0", "cell_type": "T cells", "marker_genes": "Cd3d, CD3E", "gene_expression": "..."}]'

  result <- testthat::with_mocked_bindings({
    annotate_cell_types(
      input = list("0" = list(genes = c("Cd3d", "CD3E"))),
      tissue_name = "PBMC",
      model = "gpt-5.5",
      api_key = "test-key",
      return_reasoning = TRUE
    )
  },
  get_model_response = function(prompt, model, api_key, base_urls, normalize = TRUE) json_response,
  log_info = function(...) NULL,
  log_debug = function(...) NULL)

  expect_identical(result[["0"]]$marker_genes, "Cd3d, CD3E")
})


test_that("annotate_cell_types return_reasoning falls back to Unknown on invalid JSON", {
  result <- testthat::with_mocked_bindings({
    annotate_cell_types(
      input = list("0" = list(genes = c("CD3D"))),
      tissue_name = "PBMC",
      model = "gpt-5.5",
      api_key = "test-key",
      return_reasoning = TRUE
    )
  },
  get_model_response = function(prompt, model, api_key, base_urls, normalize = TRUE) "not valid json",
  log_info = function(...) NULL,
  log_debug = function(...) NULL)

  expect_identical(result[["0"]]$cell_type, "Unknown")
  expect_identical(result[["0"]]$marker_genes, "")
  expect_identical(result[["0"]]$gene_expression, "")
})


test_that("annotate_cell_types return_reasoning=FALSE behavior is unchanged", {
  result <- testthat::with_mocked_bindings({
    annotate_cell_types(
      input = list("0" = list(genes = c("CD3D"))),
      tissue_name = "PBMC",
      model = "gpt-5.5",
      api_key = "test-key",
      return_reasoning = FALSE
    )
  },
  get_model_response = function(prompt, model, api_key, base_urls, normalize = TRUE) c("T cells"),
  log_info = function(...) NULL,
  log_debug = function(...) NULL)

  expect_identical(result, c("T cells"))
})


test_that("annotate_cell_types return_reasoning=TRUE parses markdown-fenced multi-line JSON", {
  json_response <- paste(
    "```json",
    '[',
    '  {',
    '    "cluster_id": "0",',
    '    "cell_type": "T cells",',
    '    "marker_genes": "CD3D, CD3E",',
    '    "gene_expression": "CD3D and CD3E are T cell markers"',
    '  },',
    '  {',
    '    "cluster_id": "1",',
    '    "cell_type": "B cells",',
    '    "marker_genes": "CD19, CD79A",',
    '    "gene_expression": "CD19 and CD79A are B cell markers"',
    '  }',
    ']',
    "```",
    sep = "\n"
  )

  result <- testthat::with_mocked_bindings({
    annotate_cell_types(
      input = list(
        "0" = list(genes = c("CD3D", "CD3E")),
        "1" = list(genes = c("CD19", "CD79A"))
      ),
      tissue_name = "PBMC",
      model = "kimi-k2.6",
      api_key = "test-key",
      return_reasoning = TRUE
    )
  },
  get_model_response = function(prompt, model, api_key, base_urls, normalize = TRUE) json_response,
  log_info = function(...) NULL,
  log_debug = function(...) NULL)

  expect_identical(result[["0"]]$cell_type, "T cells")
  expect_identical(result[["0"]]$marker_genes, "CD3D, CD3E")
  expect_match(result[["0"]]$gene_expression, "CD3D")
  expect_identical(result[["1"]]$cell_type, "B cells")
  expect_identical(result[["1"]]$marker_genes, "CD19, CD79A")
})


test_that("annotate_cell_types return_reasoning=TRUE with api_key=NA returns prompt", {
  prompt <- annotate_cell_types(
    input = list("0" = list(genes = c("CD3D"))),
    tissue_name = "PBMC",
    model = "gpt-5.5",
    api_key = NA,
    return_reasoning = TRUE
  )

  expect_type(prompt, "character")
  expect_match(prompt, "0: CD3D", fixed = TRUE)
  expect_match(prompt, "CD3D", fixed = TRUE)
  expect_match(prompt, "JSON object", fixed = TRUE)
  expect_match(prompt, "marker_genes", fixed = TRUE)
  expect_match(prompt, "gene_expression", fixed = TRUE)
})

# =============================================================================
# HELPER FUNCTIONS
# =============================================================================

# Generate accepted aliases for one cluster ID.
prediction_cluster_aliases <- function(cluster_id) {
  normalized <- trimws(as.character(cluster_id))
  stripped <- trimws(sub("^cluster[ _]+", "", normalized, ignore.case = TRUE))
  unique(c(
    normalized,
    stripped,
    if (nzchar(stripped)) paste("Cluster", stripped) else character(0),
    if (nzchar(stripped)) paste0("Cluster_", stripped) else character(0)
  ))
}

resolve_prediction_cluster_id <- function(raw_cluster_id, all_clusters = NULL) {
  normalized <- trimws(as.character(raw_cluster_id))
  if (!nzchar(normalized)) {
    return(NULL)
  }

  if (is.null(all_clusters)) {
    stripped <- trimws(sub("^cluster[ _]+", "", normalized, ignore.case = TRUE))
    return(if (nzchar(stripped)) stripped else normalized)
  }

  target_clusters <- unique(as.character(all_clusters))
  exact_index <- match(normalized, target_clusters)
  if (!is.na(exact_index)) {
    return(target_clusters[[exact_index]])
  }

  normalized_lower <- tolower(normalized)
  matches <- target_clusters[vapply(target_clusters, function(cluster_id) {
    normalized_lower %in% tolower(prediction_cluster_aliases(cluster_id))
  }, logical(1))]
  if (length(matches) == 1) {
    return(matches[[1]])
  }
  NULL
}

parse_prediction_line <- function(line) {
  colon_match <- regexec(
    "^\\s*(?:[-*]\\s*)?(.+?)\\s*[:\uFF1A]\\s*(.*?)\\s*$",
    line,
    perl = TRUE
  )
  colon_parts <- regmatches(line, colon_match)[[1]]
  if (length(colon_parts) == 3) {
    return(list(cluster_id = colon_parts[[2]], annotation = colon_parts[[3]]))
  }

  numbered_match <- regexec(
    "^\\s*(\\d+)\\s*[.)-]\\s+(.+?)\\s*$",
    line,
    perl = TRUE
  )
  numbered_parts <- regmatches(line, numbered_match)[[1]]
  if (length(numbered_parts) == 3) {
    return(list(cluster_id = numbered_parts[[2]], annotation = numbered_parts[[3]]))
  }
  NULL
}

#' Whether a label key denotes a cluster id (bare number or "Cluster N")
#'
#' @param key Parsed label key.
#' @return \code{TRUE} when the key is an explicit cluster reference.
#' @keywords internal
looks_like_cluster_ref <- function(key) {
  stripped <- trimws(sub("(?i)^cluster[ _]*", "", trimws(as.character(key)), perl = TRUE))
  grepl("^[0-9]+$", stripped)
}

#' Extract one line's annotation content for positional mapping
#'
#' Returns \code{NULL} only for lines that do not occupy a cluster slot: an
#' empty-value preamble/header (\code{"Notes:"}) and an explicit label for an
#' unrequested/nonexistent cluster (\code{"Cluster 999: Noise"}). Every other
#' line -- including a sentinel such as \code{"Unknown"} -- keeps its slot so
#' the line->cluster correspondence is preserved (the caller leaves a
#' non-real slot unassigned). A numbered-list ordinal (\code{"1. T cells"}) or
#' a resolvable label prefix (\code{"0: T cells"}) is stripped to its content;
#' a colon that is part of the annotation (\code{"Neurons: excitatory"}) is
#' kept whole.
#'
#' @param line Single response line.
#' @param all_clusters Character vector of requested cluster IDs.
#' @return Annotation content string, or \code{NULL} to drop the line.
#' @keywords internal
extract_positional_annotation <- function(line, all_clusters) {
  # Numbered-list ordinal ("1. T cells", "2) B cells"): the leading number is a
  # list index, not a cluster id -> keep the content after it. The delimiter
  # class excludes "-" so a hyphenated cell type ("5 - HT neurons") is not
  # mistaken for a numbered item.
  numbered <- regmatches(
    line,
    regexec("^\\s*(?:[-*]\\s*)?\\d+\\s*[.)]\\s+(.+?)\\s*$", line, perl = TRUE)
  )[[1]]
  if (length(numbered) == 2) {
    return(numbered[[2]])
  }

  parsed <- parse_prediction_line(line)
  content <- line
  if (!is.null(parsed)) {
    if (!nzchar(trimws(parsed$annotation))) {
      return(NULL)  # preamble/header line such as "Here are the annotations:"
    }
    if (!is.null(resolve_prediction_cluster_id(parsed$cluster_id, all_clusters))) {
      content <- trimws(parsed$annotation)
    } else if (looks_like_cluster_ref(parsed$cluster_id) && grepl(":", line, fixed = TRUE)) {
      return(NULL)  # explicit colon label for an unrequested/nonexistent cluster
    }
    # else: part of the annotation ("Neurons: excitatory", "5 - HT neurons")
  }
  content
}

#' Parse text-format model predictions into a named list
#'
#' Reads a response as explicit cluster labels (keyed by resolved cluster ID,
#' so out-of-order labels land correctly) when at least one label resolves,
#' unless an unresolved cluster-reference key coexists with incomplete coverage
#' -- that signals list ordinals misaligned against the requested clusters and
#' falls through to positional mapping. Positional mapping preserves the
#' line->cluster slot correspondence so a mid-list "Unknown" does not shift
#' later clusters, and a stray "Summary:"/"Note:" line does not hijack the parse.
#'
#' @param model_preds Character vector of prediction lines from a model
#' @param all_clusters Optional character vector of cluster IDs for positional fallback
#' @return Named list mapping cluster IDs to cell type annotations
#' @keywords internal
parse_text_predictions <- function(model_preds, all_clusters = NULL) {
  if (!is.character(model_preds)) {
    stop("model_preds must be a character vector")
  }

  lines <- unlist(strsplit(model_preds, "\n", fixed = TRUE), use.names = FALSE)
  lines <- trimws(lines[!is.na(lines) & nzchar(trimws(lines))])

  # Phase 1: read the response as explicit cluster labels, keyed by resolved
  # cluster ID. The labeled path is used when at least one label resolves,
  # UNLESS an unresolved cluster-reference key coexists with incomplete coverage
  # -- that signals list ordinals misaligned against the requested clusters
  # (e.g. a 1-based numbered list on 0-based clusters), which must fall through
  # to positional mapping. A stray word-keyed colon line (Summary:, Note:) is
  # ignored; a hallucinated out-of-range label does not spoil an otherwise
  # fully-covered labeled response.
  labeled <- list()
  resolved_clusters <- character(0)
  has_unresolved_cluster_ref <- FALSE
  for (line in lines) {
    parsed <- parse_prediction_line(line)
    if (is.null(parsed) || !nzchar(trimws(parsed$annotation))) {
      next
    }
    cluster_id <- resolve_prediction_cluster_id(parsed$cluster_id, all_clusters)
    if (!is.null(cluster_id)) {
      resolved_clusters <- union(resolved_clusters, cluster_id)
      if (is.null(labeled[[cluster_id]]) &&
          is_real_cell_type_annotation(parsed$annotation)) {
        labeled[[cluster_id]] <- trimws(parsed$annotation)
      }
    } else if (looks_like_cluster_ref(parsed$cluster_id)) {
      has_unresolved_cluster_ref <- TRUE
    }
  }

  n_clusters <- length(unique(as.character(all_clusters)))
  is_labeled <- length(resolved_clusters) > 0 &&
    !(has_unresolved_cluster_ref && length(resolved_clusters) < n_clusters)
  if (is_labeled) {
    # Return in requested-cluster order (deterministic; matches the Python twin)
    # so out-of-order labels do not surface in response order.
    if (!is.null(all_clusters)) {
      labeled <- labeled[intersect(unique(as.character(all_clusters)), names(labeled))]
    }
    return(labeled)
  }

  # Phase 2: positional fallback. Extract each line's annotation content,
  # dropping only non-slot lines (preamble, unrequested-cluster labels). Every
  # remaining line keeps its slot; a slot whose content is not a real cell type
  # (e.g. "Unknown") is left unassigned so the alignment is preserved.
  if (is.null(all_clusters)) {
    return(labeled)
  }
  target_clusters <- unique(as.character(all_clusters))
  contents <- character(0)
  for (line in lines) {
    content <- extract_positional_annotation(line, all_clusters)
    if (!is.null(content)) {
      contents <- c(contents, content)
    }
  }

  model_structured <- list()
  for (index in seq_len(min(length(target_clusters), length(contents)))) {
    if (is_real_cell_type_annotation(contents[[index]])) {
      model_structured[[target_clusters[[index]]]] <- contents[[index]]
    }
  }
  model_structured
}

structure_model_predictions <- function(model_preds, all_clusters) {
  target_clusters <- unique(as.character(all_clusters))
  prediction_names <- names(model_preds)
  has_explicit_names <- !is.null(prediction_names) &&
    any(!is.na(prediction_names) & nzchar(trimws(prediction_names)))

  if (has_explicit_names && (is.list(model_preds) || is.character(model_preds))) {
    structured <- list()
    for (index in seq_along(model_preds)) {
      raw_name <- prediction_names[[index]]
      if (is.na(raw_name) || !nzchar(trimws(raw_name))) {
        next
      }

      cluster_id <- resolve_prediction_cluster_id(raw_name, target_clusters)
      annotation <- model_preds[[index]]
      is_valid_annotation <- is_real_cell_type_annotation(annotation)
      if (!is.null(cluster_id) &&
          is_valid_annotation &&
          is.null(structured[[cluster_id]])) {
        structured[[cluster_id]] <- trimws(annotation)
      }
    }
    return(structured)
  }

  if (is.character(model_preds)) {
    return(parse_text_predictions(model_preds, target_clusters))
  }

  list()
}

structure_cluster_predictions <- function(initial_predictions, all_clusters,
                                          cluster_id) {
  cluster_id <- .normalize_required_string(as.character(cluster_id), "cluster_id")
  structured_predictions <- list()

  for (model_name in names(initial_predictions)) {
    parsed <- structure_model_predictions(
      initial_predictions[[model_name]],
      all_clusters
    )
    structured_predictions[[model_name]] <- if (!is.null(parsed[[cluster_id]])) {
      parsed[[cluster_id]]
    } else {
      "Prediction_Missing"
    }
  }

  structured_predictions
}

build_discussion_cache_context <- function(input, cluster_id, models, api_keys,
                                           initial_predictions,
                                           max_discussion_rounds,
                                           controversy_threshold,
                                           entropy_threshold,
                                           consensus_check_model,
                                           base_urls) {
  all_cluster_ids <- canonical_cluster_ids(input)
  current_predictions <- structure_cluster_predictions(
    initial_predictions,
    all_cluster_ids,
    cluster_id
  )
  request_models <- unique(c(
    models,
    prepare_models_list(consensus_check_model)
  ))
  request_context <- lapply(request_models, function(model) {
    credential <- .resolve_model_api_key(model, api_keys)
    base_url <- if (credential$source == "none") {
      NULL
    } else {
      resolve_provider_base_url(credential$provider, base_urls)
    }
    model_config <- NULL
    model_name <- tolower(trimws(model))
    if (exists(model_name, envir = custom_models, inherits = FALSE)) {
      model_config <- get(model_name, envir = custom_models)$config
    }

    list(
      model = model,
      provider = credential$provider,
      credential_source = credential$source,
      base_url = base_url,
      model_config = model_config
    )
  })

  list(
    initial_predictions = current_predictions,
    max_discussion_rounds = max_discussion_rounds,
    controversy_threshold = controversy_threshold,
    entropy_threshold = entropy_threshold,
    consensus_check_model = consensus_check_model,
    requests = request_context
  )
}

align_model_predictions <- function(model_preds, all_clusters) {
  target_clusters <- unique(as.character(all_clusters))
  structured <- structure_model_predictions(model_preds, target_clusters)
  aligned <- vapply(target_clusters, function(cluster_id) {
    annotation <- structured[[cluster_id]]
    if (!is_real_cell_type_annotation(annotation)) NA_character_ else annotation
  }, character(1))
  names(aligned) <- target_clusters
  aligned
}

#' Get initial predictions from all models
#'
#' This function retrieves initial cell type predictions from all specified models.
#' It is an internal helper function used by the interactive_consensus_annotation function.
#'
#' @keywords internal
get_initial_predictions <- function(input, tissue_name, models, api_keys, top_gene_count, base_urls = NULL) {
  log_info("Phase 1: Getting initial predictions from all models...", list(
    models_count = length(models),
    models = models
  ))
  message("\nPhase 1: Getting initial predictions from all models...")

  # Initialize tracking variables
  individual_predictions <- list()
  successful_models <- character(0)

  # Get predictions from each model
  for (model in models) {
    provider <- tryCatch({
      get_provider(model)
    }, error = function(e) {
      log_warn("Failed to resolve provider for model", list(model = model, error = e$message))
      "unknown"
    })

    api_key <- tryCatch({
      get_api_key(model, api_keys)
    }, error = function(e) {
      log_warn("Failed to resolve API key", list(model = model, error = e$message))
      NULL
    })

    if (is.null(api_key)) {
      warning_msg <- sprintf("No API key found for model '%s' (provider: %s). This model will be skipped.",
                            model, provider)
      warning(warning_msg)
      log_warn(warning_msg, list(model = model, provider = provider))
      next
    }

    tryCatch({
      predictions <- annotate_cell_types(
        input = input,
        tissue_name = tissue_name,
        model = model,
        api_key = api_key,
        top_gene_count = top_gene_count,
        base_urls = base_urls
      )
      individual_predictions[[model]] <- predictions
      successful_models <- c(successful_models, model)
    }, error = function(e) {
      warning_msg <- sprintf("Failed to get predictions from %s: %s", model, e$message)
      warning(warning_msg)
      log_warn(warning_msg, list(model = model, error = e$message))
    })
  }

  if (length(successful_models) == 0) {
    stop("No models successfully completed predictions. Please check API keys and model availability.")
  }

  return(list(
    individual_predictions = individual_predictions,
    successful_models = successful_models
  ))
}

#' Identify controversial clusters based on consensus analysis
#'
#
#
#
#
#
#' @keywords internal
identify_controversial_clusters <- function(input, individual_predictions, controversy_threshold, entropy_threshold, api_keys, consensus_check_model = NULL, base_urls = NULL) {
  clusters <- canonical_cluster_ids(input)

  log_info("Phase 2: Identifying controversial clusters...", list(
    clusters_count = length(clusters),
    entropy_threshold = entropy_threshold,
    controversy_threshold = controversy_threshold
  ))
  message("\nPhase 2: Identifying controversial clusters...")

  # Initialize consensus tracking
  consensus_results <- list()
  controversial_clusters <- character(0)
  final_annotations <- list()

  # Get all cluster IDs for positional fallback in text parsing
  all_clusters <- as.character(clusters)

  # Restructure individual_predictions to be indexed by cluster_id
  structured_predictions <- list()

  for (model_name in names(individual_predictions)) {
    model_preds <- individual_predictions[[model_name]]
    structured_predictions[[model_name]] <- structure_model_predictions(
      model_preds,
      all_clusters
    )
  }

  for (cluster_id in clusters) {
    # Use original cluster ID for log output, no conversion needed
    log_info(sprintf("Analyzing cluster %s...", cluster_id), list(cluster_id = cluster_id))
    message(sprintf("\nAnalyzing cluster %s...", cluster_id))

    # Get predictions for this cluster from each model
    cluster_predictions <- sapply(structured_predictions, function(x) {
      if (is.null(x[[as.character(cluster_id)]])) NA else x[[as.character(cluster_id)]]
    })
    valid_predictions <- cluster_predictions[!is.na(cluster_predictions)]

    if (length(valid_predictions) == 0) {
      log_warn(sprintf("No valid predictions for cluster %s. Marking as controversial.", cluster_id), list(cluster_id = cluster_id))
      message(sprintf("No valid predictions for cluster %s. Marking as controversial.", cluster_id))
      controversial_clusters <- c(controversial_clusters, as.character(cluster_id))
      next
    }

    # Calculate agreement score
    initial_consensus <- check_consensus(valid_predictions, api_keys, controversy_threshold, entropy_threshold, consensus_check_model, base_urls)
    consensus_results[[as.character(cluster_id)]] <- initial_consensus

    # If no consensus is reached or the consensus metrics indicate high uncertainty, mark it as controversial.
    # Use both consensus proportion and entropy for decision making
    if (!initial_consensus$reached ||
        initial_consensus$consensus_proportion < controversy_threshold ||
        initial_consensus$entropy > entropy_threshold) {

      log_info(sprintf("Cluster %s marked as controversial", cluster_id), list(
        cluster_id = cluster_id,
        reached_consensus = initial_consensus$reached,
        consensus_proportion = initial_consensus$consensus_proportion,
        entropy = initial_consensus$entropy
      ))

      message(sprintf("Cluster %s marked as controversial (reached: %s, consensus proportion: %.2f, entropy: %.2f)",
                     cluster_id, initial_consensus$reached,
                     initial_consensus$consensus_proportion, initial_consensus$entropy))

      controversial_clusters <- c(controversial_clusters, as.character(cluster_id))
    } else {
      # Process non-controversial clusters
      final_annotations[[as.character(cluster_id)]] <- select_best_prediction(initial_consensus, valid_predictions)

      log_info(sprintf("Consensus reached for cluster %s", cluster_id), list(
        cluster_id = cluster_id,
        consensus_proportion = initial_consensus$consensus_proportion,
        entropy = initial_consensus$entropy,
        selected_cell_type = final_annotations[[as.character(cluster_id)]]
      ))

      message(sprintf("Consensus reached for cluster %s (consensus proportion: %.2f, entropy: %.2f, selected: %s)",
                     cluster_id, initial_consensus$consensus_proportion,
                     initial_consensus$entropy, final_annotations[[as.character(cluster_id)]]))
    }
  }

  return(list(
    consensus_results = consensus_results,
    controversial_clusters = controversial_clusters,
    final_annotations = final_annotations
  ))
}

#' Select the best prediction from consensus results
#'
#
#
#
#' @keywords internal
select_best_prediction <- function(consensus_result, valid_predictions) {
  majority <- clean_annotation(consensus_result$majority_prediction)

  # Accept the consensus check result if it is a real cell type
  if (!identical(majority, "Unknown")) {
    return(majority)
  }

  # Fallback: pick the most frequent real prediction from the models
  cleaned_predictions <- vapply(
    as.list(valid_predictions),
    clean_annotation,
    character(1)
  )
  real_predictions <- cleaned_predictions[cleaned_predictions != "Unknown"]
  if (length(real_predictions) == 0) {
    return("Unknown")
  }

  prediction_counts <- table(real_predictions)
  max_count <- max(prediction_counts)
  most_common <- names(prediction_counts[prediction_counts == max_count])

  if (length(most_common) == 1) {
    return(most_common)
  }
  # Preserve caller model priority when the vote is tied.
  real_predictions[real_predictions %in% most_common][[1]]
}

#' Process controversial clusters through discussion
#'
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#' @keywords internal
process_controversial_clusters <- function(controversial_clusters, input, tissue_name,
                                          successful_models, api_keys, individual_predictions,
                                          top_gene_count, controversy_threshold, entropy_threshold, max_discussion_rounds,
                                          cache_manager, use_cache, consensus_check_model = NULL, force_rerun = FALSE,
                                          base_urls = NULL) {

  if (length(controversial_clusters) == 0) {
    log_info("No controversial clusters found. All clusters have reached consensus.")
    message("\nNo controversial clusters found. All clusters have reached consensus.")
    return(list(
      discussion_logs = list(),
      final_annotations = list()
    ))
  }

  log_info(sprintf("Phase 3: Starting discussions for %d controversial clusters...",
                   length(controversial_clusters)), list(
    controversial_count = length(controversial_clusters),
    clusters = controversial_clusters
  ))
  message(sprintf("\nPhase 3: Starting discussions for %d controversial clusters...",
                 length(controversial_clusters)))

  discussion_logs <- list()
  final_annotations <- list()

  for (cluster_id in controversial_clusters) {
    # Ensure cluster_id is a string type
    char_cluster_id <- as.character(cluster_id)
    log_info(sprintf("Starting discussion for cluster %s...", char_cluster_id), list(
      cluster_id = char_cluster_id
    ))
    message(sprintf("\nStarting discussion for cluster %s...", char_cluster_id))

    # Generate cache key once (reused for both lookup and save)
    cache_key <- if (use_cache) {
      discussion_context <- build_discussion_cache_context(
        input = input,
        cluster_id = char_cluster_id,
        models = successful_models,
        api_keys = api_keys,
        initial_predictions = individual_predictions,
        max_discussion_rounds = max_discussion_rounds,
        controversy_threshold = controversy_threshold,
        entropy_threshold = entropy_threshold,
        consensus_check_model = consensus_check_model,
        base_urls = base_urls
      )
      cache_manager$generate_key(
        input,
        successful_models,
        char_cluster_id,
        tissue_name,
        top_gene_count,
        discussion_context = discussion_context
      )
    } else {
      NULL
    }

    # Check cache
    cached_result <- NULL
    if (use_cache && !force_rerun) {
      log_debug(sprintf("Cache check for cluster %s", char_cluster_id),
                list(cluster_id = char_cluster_id, cache_key = cache_key))

      cached_result <- cache_manager$load_from_cache(cache_key)
      if (!is.null(cached_result) && is_valid_consensus_cache_data(cached_result)) {
        log_info(sprintf("Loading cached result for cluster %s", char_cluster_id), list(
          cluster_id = char_cluster_id,
          cache_key = cache_key
        ))
        message(sprintf("Loading cached result for cluster %s", char_cluster_id))
      } else if (!is.null(cached_result)) {
        log_warn("Ignoring invalid cached discussion result", list(
          cluster_id = char_cluster_id,
          cache_key = cache_key
        ))
        cached_result <- NULL
      }
    } else if (force_rerun) {
      log_info(sprintf("Force rerun enabled, skipping cache for cluster %s", char_cluster_id))
    }

    # Use cached results or perform discussion
    if (!is.null(cached_result)) {
      discussion_result <- cached_result$discussion_log
      final_annotation <- cached_result$annotation

      log_info(sprintf("Using cached result for cluster %s", char_cluster_id), list(
        cluster_id = char_cluster_id
      ))
      message(sprintf("Using cached result for cluster %s", char_cluster_id))
    } else {
      # Perform discussion
      discussion_result <- facilitate_cluster_discussion(
        cluster_id = char_cluster_id,
        input = input,
        tissue_name = tissue_name,
        models = successful_models,
        api_keys = api_keys,
        initial_predictions = individual_predictions,
        top_gene_count = top_gene_count,
        max_rounds = max_discussion_rounds,
        controversy_threshold = controversy_threshold,
        entropy_threshold = entropy_threshold,
        consensus_check_model = consensus_check_model,
        base_urls = base_urls
      )

      # Find the last round that has a consensus_result (the last round may
      # lack one if it broke early due to insufficient valid responses)
      final_prediction <- NULL
      for (r in rev(seq_along(discussion_result$rounds))) {
        cr <- discussion_result$rounds[[r]]$consensus_result
        if (!is.null(cr)) {
          final_prediction <- cr$majority_prediction
          break
        }
      }

      # Extract and clean majority_prediction
      final_annotation <- clean_annotation(final_prediction)

      # Save to cache
      if (use_cache) {
        cache_data <- list(
          annotation = final_annotation,
          discussion_log = discussion_result,
          is_controversial = TRUE
        )
        cache_saved <- cache_manager$save_to_cache(cache_key, cache_data)
        if (isTRUE(cache_saved)) {
          log_info(sprintf("Saved result to cache for cluster %s", char_cluster_id), list(
            cluster_id = char_cluster_id
          ))
        }
      }
    }

    # Ensure cluster_id in discussion_result is a string type
    if (!is.null(discussion_result) && !is.character(discussion_result$cluster_id)) {
      discussion_result$cluster_id <- char_cluster_id
    }

    discussion_logs[[char_cluster_id]] <- discussion_result
    final_annotations[[char_cluster_id]] <- final_annotation

    log_info(sprintf("Completed discussion for cluster %s", char_cluster_id), list(
      cluster_id = char_cluster_id
    ))
    message(sprintf("Completed discussion for cluster %s", char_cluster_id))
  }

  return(list(
    discussion_logs = discussion_logs,
    final_annotations = final_annotations
  ))
}

#' Clean annotation text by removing prefixes and extra whitespace
#'
#
#
#' @keywords internal
clean_annotation <- function(annotation) {
  is_text_scalar <- is.character(annotation) &&
    length(annotation) == 1 &&
    !is.na(annotation)
  if (!is_text_scalar) {
    return("Unknown")
  }

  # Remove numbered prefixes like "1. ", "1: ", "1- ", etc.
  annotation <- gsub("^\\d+[\\.:\\-\\s]+\\s*", "", annotation)
  # Remove "CELL TYPE:" prefix
  annotation <- gsub(
    "^CELL\\s*TYPE[\\s:]*",
    "",
    annotation,
    ignore.case = TRUE
  )
  # Final trim of whitespace
  annotation <- trimws(annotation)

  # Normalize sentinel values to a user-friendly fallback
  if (!is_real_cell_type_annotation(annotation)) {
    return("Unknown")
  }

  return(annotation)
}

#' Combine results from all phases of consensus annotation
#'
#
#
#
#
#' @keywords internal
combine_results <- function(initial_results, controversy_results, discussion_results) {
  # Start with non-controversial cluster annotations
  final_annotations <- lapply(
    controversy_results$final_annotations,
    clean_annotation
  )

  # Merge controversial cluster annotations (already cleaned by process_controversial_clusters)
  for (cluster_id in names(discussion_results$final_annotations)) {
    char_cluster_id <- as.character(cluster_id)
    annotation <- discussion_results$final_annotations[[char_cluster_id]]
    final_annotations[[char_cluster_id]] <- clean_annotation(annotation)
  }

  result <- list(
    initial_results = list(
      individual_predictions = initial_results$individual_predictions,
      consensus_results = controversy_results$consensus_results,
      controversial_clusters = controversy_results$controversial_clusters
    ),
    final_annotations = final_annotations,
    controversial_clusters = controversy_results$controversial_clusters,
    discussion_logs = discussion_results$discussion_logs,
    session_id = get_logger()$session_id
  )

  # Backward-compatible aliases.
  result$voting_results <- result$initial_results
  result$discussion_results <- result$discussion_logs
  result$final_consensus <- result$final_annotations

  result
}

# =============================================================================
# MAIN FUNCTION
# =============================================================================

# Constants are now defined as function parameters

#' Interactive consensus building for cell type annotation
#'
#' This function implements an interactive voting and discussion mechanism where multiple LLMs
#' collaborate to reach a consensus on cell type annotations, particularly focusing on
#' clusters with low agreement. The process includes:
#' 1. Initial voting by all LLMs
#' 2. Identification of controversial clusters
#' 3. Detailed discussion for controversial clusters
#' 4. Final summary by a designated LLM (default: Claude)
#'
#' @param input Either a data frame from Seurat's FindAllMarkers() function containing 
#'   differential gene expression results (must have columns: 'cluster', 'gene', 
#'   and 'avg_log2FC'), or a list where each element is either a character vector
#'   of genes or a list containing a `genes` field.
#' @param tissue_name Character string specifying the tissue type for context-aware
#'   cell type annotation (e.g., 'human PBMC', 'mouse brain'). Required.
#' @param models Character vector of model names to use for consensus annotation. 
#'   Minimum 2 models required. Supports models from OpenAI, Anthropic, DeepSeek, 
#'   Google, Alibaba, Stepfun, Zhipu, MiniMax, X.AI, and OpenRouter.
#' @param api_keys Named, non-empty list of API keys. Can use provider names as keys
#'   (e.g., "openai", "anthropic") or model names as keys (e.g., "gpt-5").
#' @param top_gene_count Integer specifying the number of top marker genes to use 
#'   for annotation per cluster (default: 10).
#' @param controversy_threshold Numeric value between 0 and 1 for consensus proportion 
#'   threshold. Clusters below this threshold are considered controversial (default: 0.7).
#' @param entropy_threshold Numeric value for entropy threshold. Higher entropy 
#'   indicates more disagreement among models (default: 1.0).
#' @param max_discussion_rounds Integer specifying maximum number of discussion rounds 
#'   for controversial clusters (default: 3).
#' @param consensus_check_model Character string specifying which model to use for
#'   consensus checking. If NULL, uses the first model that succeeds during initial annotation.
#' @param log_dir Character scalar specifying directory for log files (default: "logs").
#'   This function reinitializes the session logger with this directory at the start
#'   of each call.
#' @param cache_dir Character string or NULL. Cache directory for storing results. 
#'   NULL uses system cache, "local" uses current directory, "temp" uses temporary 
#'   directory, or specify custom path.
#' @param use_cache Logical indicating whether to use caching (default: TRUE).
#' @param base_urls Named list or character string specifying custom API base URLs. 
#'   Useful for proxies or alternative endpoints. If NULL, uses official endpoints.
#' @param clusters_to_analyze Character or numeric vector specifying which clusters 
#'   to analyze. If NULL (default), all clusters are analyzed.
#' @param force_rerun Logical indicating whether to force rerun of all specified 
#'   clusters, ignoring cache. Only affects controversial cluster discussions 
#'   (default: FALSE).
#'
#' @return A list containing:
#'   \itemize{
#'     \item \code{initial_results}: Initial voting results, consensus checks, and controversial cluster IDs
#'     \item \code{final_annotations}: Final annotations keyed by cluster ID
#'     \item \code{controversial_clusters}: Clusters identified as controversial
#'     \item \code{discussion_logs}: Detailed discussion logs for controversial clusters
#'     \item \code{session_id}: Logger session identifier
#'     \item \code{voting_results}: Backward-compatible alias of \code{initial_results}
#'     \item \code{discussion_results}: Backward-compatible alias of \code{discussion_logs}
#'     \item \code{final_consensus}: Backward-compatible alias of \code{final_annotations}
#'   }
#' @importFrom stats setNames
#' @export
interactive_consensus_annotation <- function(input,
                                           tissue_name,
                                           models = c("claude-opus-4-7",
                                                     "gpt-5.5",
                                                     "gemini-3.1-pro-preview",
                                                     "deepseek-v4-flash",
                                                     "grok-4.3"),
                                           api_keys,
                                           top_gene_count = 10,
                                           controversy_threshold = 0.7,
                                           entropy_threshold = 1.0,
                                           max_discussion_rounds = 3,
                                           consensus_check_model = NULL,
                                           log_dir = "logs",
                                           cache_dir = NULL,
                                           use_cache = TRUE,
                                           base_urls = NULL,
                                           clusters_to_analyze = NULL,
                                           force_rerun = FALSE) {
  tissue_name <- .normalize_required_string(tissue_name, "tissue_name")
  models <- .normalize_model_vector(models, minimum_count = 2L)
  api_keys <- .normalize_api_keys(api_keys)
  top_gene_count <- .normalize_top_gene_count(top_gene_count)
  controversy_threshold <- .normalize_probability(
    controversy_threshold,
    "controversy_threshold"
  )
  entropy_threshold <- .normalize_nonnegative_number(
    entropy_threshold,
    "entropy_threshold"
  )
  max_discussion_rounds <- .normalize_positive_integer(
    max_discussion_rounds,
    "max_discussion_rounds"
  )
  log_dir <- .normalize_required_string(log_dir, "log_dir")
  use_cache <- .normalize_flag(use_cache, "use_cache")
  force_rerun <- .normalize_flag(force_rerun, "force_rerun")
  if (!is.null(consensus_check_model)) {
    consensus_check_model <- .normalize_required_string(
      consensus_check_model,
      "consensus_check_model"
    )
  }

  if (!is.null(clusters_to_analyze)) {
    selector_is_valid <- (is.character(clusters_to_analyze) ||
      is.numeric(clusters_to_analyze)) &&
      length(clusters_to_analyze) > 0 &&
      !anyNA(clusters_to_analyze)
    if (!selector_is_valid) {
      stop("clusters_to_analyze must be a non-empty character or numeric vector")
    }
    clusters_to_analyze <- unique(trimws(as.character(clusters_to_analyze)))
    if (any(!nzchar(clusters_to_analyze))) {
      stop("clusters_to_analyze must not contain empty cluster IDs")
    }
  }

  cluster_name_map <- NULL

  # Normalize list input to a canonical cluster->genes mapping to keep
  # contract consistent with annotate_cell_types/create_annotation_prompt.
  if (is.list(input) && !is.data.frame(input)) {
    original_names <- names(input)
    normalized_input <- normalize_cluster_gene_list(input)
    if (!is.null(original_names)) {
      cluster_name_map <- setNames(names(normalized_input), original_names)
    }
    input <- lapply(normalized_input, function(genes) list(genes = genes))
  }

  prompt_metadata <- create_annotation_prompt(input, tissue_name, top_gene_count)
  available_clusters <- names(prompt_metadata$gene_lists)
  if (is.list(input) && !is.data.frame(input)) {
    input <- input[available_clusters]
  }

  initialize_logger(log_dir)

  # Initialize cache manager
  cache_manager <- CacheManager$new(cache_dir)
  
  # Get actual cache directory path (important!)
  actual_cache_dir <- cache_manager$get_cache_dir()

  # Log cache settings - use actual path
  if (use_cache && !force_rerun) {
    cache_msg <- sprintf("Cache enabled. Using cache directory: %s", actual_cache_dir)
    log_info(cache_msg, list(cache_dir = actual_cache_dir))
    message(cache_msg)
  } else if (force_rerun) {
    log_info("Force rerun enabled, cache will be ignored for controversial clusters")
    message("Force rerun enabled. Cache will be ignored for controversial clusters.")
  } else {
    log_info("Cache disabled")
    message("Cache disabled.")
  }

  # Filter clusters if clusters_to_analyze is specified
  if (!is.null(clusters_to_analyze)) {
    # If list input names were normalized, accept either original or normalized IDs.
    if (!is.null(cluster_name_map)) {
      mapped_ids <- ifelse(
        clusters_to_analyze %in% names(cluster_name_map),
        cluster_name_map[clusters_to_analyze],
        clusters_to_analyze
      )
      clusters_to_analyze <- as.character(mapped_ids)
    }
    
    # Check which requested clusters exist
    valid_clusters <- clusters_to_analyze[clusters_to_analyze %in% available_clusters]
    invalid_clusters <- clusters_to_analyze[!clusters_to_analyze %in% available_clusters]
    
    # Warn about non-existent clusters
    if (length(invalid_clusters) > 0) {
      warning(sprintf("The following cluster IDs were not found in the input: %s",
                     paste(invalid_clusters, collapse = ", ")))
      log_warn("Specified cluster IDs not found in input", list(invalid_clusters = invalid_clusters))
    }
    
    # Stop if no valid clusters
    if (length(valid_clusters) == 0) {
      stop("None of the specified clusters exist in the input data.")
    }
    
    # Filter input based on type
    if (is.list(input) && !is.data.frame(input)) {
      # For list input, subset by names
      input <- input[valid_clusters]
    } else {
      # For dataframe input, filter rows
      normalized_input_clusters <- trimws(as.character(input$cluster))
      input <- input[normalized_input_clusters %in% valid_clusters, , drop = FALSE]
    }
    
    # Log the filtering
    log_info(sprintf("Filtered to analyze %d clusters: %s", 
                    length(valid_clusters), 
                    paste(valid_clusters, collapse = ", ")))
    message(sprintf("Analyzing %d specified clusters: %s", 
                   length(valid_clusters), 
                   paste(valid_clusters, collapse = ", ")))
  }

  # Phase 1: Get initial predictions from all models
  initial_results <- get_initial_predictions(
    input = input,
    tissue_name = tissue_name,
    models = models,
    api_keys = api_keys,
    top_gene_count = top_gene_count,
    base_urls = base_urls
  )

  # Phase 2: Identify controversial clusters
  # If consensus_check_model is NULL, use the first model that succeeded in
  # Phase 1 (not models[1], which may have failed)
  if (is.null(consensus_check_model)) {
    consensus_check_model <- initial_results$successful_models[1]
    log_info("No consensus_check_model specified, using first successful model",
             list(consensus_check_model = consensus_check_model))
  }

  controversy_results <- identify_controversial_clusters(
    input = input,
    individual_predictions = initial_results$individual_predictions,
    controversy_threshold = controversy_threshold,
    entropy_threshold = entropy_threshold,
    api_keys = api_keys,
    consensus_check_model = consensus_check_model,
    base_urls = base_urls
  )

  # Phase 3: Process controversial clusters through discussion
  discussion_results <- process_controversial_clusters(
    controversial_clusters = controversy_results$controversial_clusters,
    input = input,
    tissue_name = tissue_name,
    successful_models = initial_results$successful_models,
    api_keys = api_keys,
    individual_predictions = initial_results$individual_predictions,
    top_gene_count = top_gene_count,
    controversy_threshold = controversy_threshold,
    entropy_threshold = entropy_threshold,
    max_discussion_rounds = max_discussion_rounds,
    cache_manager = cache_manager,
    use_cache = use_cache,
    consensus_check_model = consensus_check_model,
    force_rerun = force_rerun,
    base_urls = base_urls
  )

  # Combine results from all phases
  final_results <- combine_results(
    initial_results = initial_results,
    controversy_results = controversy_results,
    discussion_results = discussion_results
  )

  # Print summary of consensus building process
  print_consensus_summary(final_results)

  # Return results
  return(final_results)
}

test_that("API request timeout defaults to 120 seconds", {
  old_option <- options(mLLMCelltype.api_timeout = NULL)
  on.exit(options(old_option), add = TRUE)

  expect_identical(get_api_request_timeout(), 120)
})

test_that("API request timeout accepts a positive numeric override", {
  old_option <- options(mLLMCelltype.api_timeout = 300)
  on.exit(options(old_option), add = TRUE)

  expect_identical(get_api_request_timeout(), 300)
})

test_that("API request timeout rejects invalid overrides", {
  old_option <- options(mLLMCelltype.api_timeout = 0)
  on.exit(options(old_option), add = TRUE)

  expect_error(
    get_api_request_timeout(),
    "must be one positive number of seconds",
    fixed = TRUE
  )
})

---
title: "Why Choose Consensus? The Scientific Foundation of Multi-LLM Annotation"
description: "Overview of the consensus-based approach to cell type annotation, including its scientific basis, methodology, and trade-offs."
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Why Choose Consensus? The Scientific Foundation of Multi-LLM Annotation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Why Choose Consensus? The Scientific Foundation of Multi-LLM Annotation

Multi-LLM consensus can improve annotation accuracy by combining the strengths of diverse AI models while reducing the impact of individual model limitations (see Yang et al., 2025).

## The Challenge with Single-Model Approaches

Traditional single-model annotation systems face inherent limitations:

### Accuracy Limitations
- **Single-point failure**: One model's bias affects all results
- **Limited perspective**: Each model has unique strengths and blind spots
- **Inconsistent performance**: Varies across cell types and tissues

### Reliability Issues
- **Model hallucinations**: Confident but incorrect predictions
- **Lack of uncertainty**: Difficult to identify questionable annotations
- **Reproducibility challenges**: Different model versions may yield different results

## The Consensus Approach: Inspired by Scientific Peer Review

mLLMCelltype's consensus framework is analogous to the peer review process in scientific publishing.

### The Scientific Parallel

Just as scientific papers benefit from multiple expert reviewers, cell annotations can benefit from multiple AI models:

| Scientific Peer Review | mLLMCelltype Consensus |
|------------------------|------------------------|
| Multiple expert reviewers | Multiple LLM models |
| Diverse perspectives | Different training approaches |
| Debate and discussion | Structured deliberation |
| Consensus building | Agreement quantification |
| Quality assurance | Uncertainty metrics |

### How It Works

**1. Error Detection Through Cross-Validation**
- Models check each other's work
- Individual model biases can be averaged out
- Outlier predictions are identified

**2. Transparent Uncertainty Quantification**
- **Consensus Proportion (CP)**: Measures inter-model agreement
- **Shannon Entropy**: Quantifies prediction uncertainty
- **Controversy Detection**: Automatically identifies clusters requiring expert review

## Why Multiple Perspectives Help

Cell type annotation involves:

- **Marker gene interpretation**: Different models may have different strengths across gene families
- **Context understanding**: Various models may capture different biological contexts
- **Rare cell types**: Ensemble approaches can improve detection of uncommon populations
- **Batch effects**: Multiple models may provide robustness against technical artifacts

For benchmark results, see Yang et al. (2025):

Yang, C., Zhang, X., & Chen, J. (2025). Large Language Model Consensus Substantially Improves the Cell Type Annotation Accuracy for scRNA-seq Data. *bioRxiv*. https://doi.org/10.1101/2025.04.10.647852

## Cost Considerations

The two-stage approach can reduce API calls when models agree early:

- **Stage 1**: Initial consensus check -- clusters where models agree skip further processing
- **Stage 2**: Deliberation only for clusters without initial agreement
- **Caching**: Results can be reused across similar analyses

This means the cost overhead of using multiple models is partially offset by skipping deliberation for clear cases.

## Technical Implementation

### The Three-Stage Process

**Stage 1: Independent Analysis**
Each LLM analyzes marker genes and provides:
- Cell type predictions
- Confidence scores
- Reasoning chains

**Stage 2: Consensus Building**
The system:
- Compares predictions across models
- Identifies areas of agreement and disagreement
- Calculates uncertainty metrics

**Stage 3: Deliberation (when needed)**
For controversial clusters:
- Models share their reasoning
- Structured debate occurs
- Final consensus emerges

### Quality Metrics

- **Semantic similarity analysis**: Ensures meaningful disagreements are detected
- **Evidence-based reasoning**: All predictions include supporting evidence
- **Iterative refinement**: Multiple rounds of discussion when needed

## When to Choose Consensus

**Consensus may be preferable when:**
- Uncertainty quantification is needed
- Datasets involve novel or complex tissues
- Results will be published or used in downstream analyses
- Identifying low-confidence annotations is important

**Consider alternatives when:**
- Quick exploratory analysis is the goal
- Datasets are well-characterized with clear markers
- API budget is very limited
- Proof-of-concept work in early stages

## Quick Start Example

```r
library(mLLMCelltype)

# Load marker genes from your single-cell data, then run consensus annotation
api_keys <- list(
  openai = Sys.getenv("OPENAI_API_KEY"),
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY")
)

results <- interactive_consensus_annotation(
  input = marker_data,
  tissue_name = "human PBMC",
  models = c("gpt-5.5", "claude-sonnet-4-6", "gemini-3.1-pro-preview"),
  api_keys = api_keys
)

```

### Understanding Your Results

- **High consensus (CP > 0.8)**: Reliable annotations
- **Medium consensus (0.5 < CP < 0.8)**: Review recommended
- **Low consensus (CP < 0.5)**: Expert validation needed

## Summary

The consensus approach provides a framework for combining multiple LLM predictions with built-in uncertainty quantification. As new models become available, the framework can incorporate them without changes to the overall methodology.

## Learn More

- [Getting Started Guide](https://cafferyang.com/mLLMCelltype/articles/getting-started.html)
- [Consensus vs Single-Agent Methods](https://cafferyang.com/mLLMCelltype/articles/vs-single-agent.html)
- [Advanced Features](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html)
- [API Reference](https://cafferyang.com/mLLMCelltype/reference/index.html)

---
title: "mLLMCelltype: Overview and Quick Reference"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{mLLMCelltype: Overview and Quick Reference}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE, purl=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

```{r setup, purl=FALSE}
library(mLLMCelltype)
```

## Overview

mLLMCelltype is an R package that leverages various large language models (LLMs) for automated cell type annotation in single-cell RNA sequencing data. It implements a consensus-based approach where multiple LLMs collaborate to provide more reliable annotations. For more details, please refer to our paper (doi:10.1038/s42003-026-10420-8).

## Installation

You can install the released version of mLLMCelltype from CRAN:

```{r installation, purl=FALSE}
install.packages("mLLMCelltype")
```

Or install the development version from GitHub:

```{r installation-dev, purl=FALSE}
# install.packages("devtools")
devtools::install_github("cafferychen777/mLLMCelltype")
```

## Setting Up API Keys

Before using mLLMCelltype, you need to set up API keys for the LLMs you want to use. The package supports multiple LLM providers including OpenAI, Anthropic, Google (Gemini), and X.AI (Grok).

```{r api-keys, purl=FALSE}
# Set API keys (recommended to use .Renviron or .env file)
# OpenAI API key
Sys.setenv(OPENAI_API_KEY = "your-openai-api-key")

# Anthropic API key
Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-api-key")

# Google API key
Sys.setenv(GOOGLE_API_KEY = "your-google-api-key")

# X.AI API key
Sys.setenv(XAI_API_KEY = "your-xai-api-key")

# Alternatively, use .env file
# dotenv::load_dot_env()
```

## Basic Usage

### Annotating Cell Types with Seurat Object

The main function of mLLMCelltype is to annotate cell types in a Seurat object:

```{r basic-usage, purl=FALSE}
# Load required packages
library(mLLMCelltype)
library(Seurat)
library(dplyr)

# Load your preprocessed Seurat object
# pbmc <- readRDS("your_seurat_object.rds")

# Find marker genes for each cluster
# pbmc_markers <- FindAllMarkers(pbmc,
#                           only.pos = TRUE,
#                           min.pct = 0.25,
#                           logfc.threshold = 0.25)

# Cache will be automatically managed using system cache directory

# Run LLMCelltype annotation with multiple LLM models
consensus_results <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",  # provide tissue context
  models = c(
    "claude-sonnet-4-6",  # Anthropic
    "gpt-5.5",                   # OpenAI
    "gemini-3.1-pro-preview"            # Google
  ),
  api_keys = list(
    anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
    openai = Sys.getenv("OPENAI_API_KEY"),
    gemini = Sys.getenv("GOOGLE_API_KEY")
  ),
  top_gene_count = 10,
  controversy_threshold = 1.0,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6",  # Use a reliable model for consensus checking
  cache_dir = NULL  # Uses default system cache directory
)

# Add annotations to Seurat object
# Get cell type annotations from consensus_results$final_annotations
cluster_to_celltype_map <- consensus_results$final_annotations

# Create new cell type identifier column
cell_types <- as.character(Idents(pbmc))
for (cluster_id in names(cluster_to_celltype_map)) {
  cell_types[cell_types == cluster_id] <- cluster_to_celltype_map[[cluster_id]]
}

# Add cell types to Seurat object
pbmc$mLLM_cell_type <- cell_types
```

### Visualizing Results

```{r visualization, purl=FALSE}
# Plot UMAP with cell type annotations
library(Seurat)
library(ggplot2)
library(cowplot)

# Basic visualization
p1 <- DimPlot(pbmc, group.by = "mLLM_cell_type", label = TRUE) +
  ggtitle("mLLMCelltype Consensus Annotations")

# Visualize consensus proportion (confidence metric)
pbmc$consensus_proportion <- 0
for (cluster_id in names(consensus_results$initial_results$consensus_results)) {
  cluster_cells <- which(Idents(pbmc) == cluster_id)
  pbmc$consensus_proportion[cluster_cells] <-
    consensus_results$initial_results$consensus_results[[cluster_id]]$consensus_proportion
}

p2 <- FeaturePlot(pbmc, features = "consensus_proportion",
                 min.cutoff = 0, max.cutoff = 1) +
  scale_color_gradientn(colors = c("blue", "green", "red")) +
  ggtitle("Consensus Proportion") +
  theme(legend.position = "right")

```

## Supported Models

mLLMCelltype supports various LLM models:

- **OpenAI**: gpt-5.5, gpt-5.4, gpt-5.4-mini
- **Anthropic**: claude-opus-4-7, claude-sonnet-4-6, claude-haiku-4-5-20251001
- **DeepSeek**: deepseek-v4-flash, deepseek-v4-pro
- **Google**: gemini-3.1-pro-preview, gemini-3-flash-preview, gemini-3.1-flash-lite
- **Qwen**: qwen3.6-plus, qwen3.6-flash, qwen3.6-max-preview
- **Stepfun**: step-3.5-flash, step-3.5-flash-2603, step-3
- **Zhipu/Z.AI**: glm-5.1, glm-5, glm-5-turbo
- **MiniMax**: MiniMax-M2.7, MiniMax-M2.7-highspeed, MiniMax-M2.5
- **Grok**: grok-4.3, grok-4.3-latest, grok-latest
- **OpenRouter**: Access to models from multiple providers through a single API

## Advanced Usage

### Using a Single LLM Model

If you only want to use a single LLM model instead of the consensus approach:

```{r single-model, purl=FALSE}
# Run annotation with a single model
single_model_results <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10
)

# Inspect the provider response lines. Parse or review this text before
# mapping annotations back onto a Seurat object.
cat(single_model_results)
```

### Customizing Consensus Parameters

```{r custom-consensus, purl=FALSE}
# Customize consensus parameters
custom_consensus_results <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
  api_keys = list(
    anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
    openai = Sys.getenv("OPENAI_API_KEY"),
    gemini = Sys.getenv("GOOGLE_API_KEY")
  ),
  top_gene_count = 15,                # Number of top marker genes to use
  controversy_threshold = 0.7,        # Threshold for controversy detection
  entropy_threshold = 0.7,            # Entropy threshold for controversy detection
  max_discussion_rounds = 2,          # Maximum rounds of discussion
  consensus_check_model = "claude-sonnet-4-6",  # Use a reliable model for consensus checking
  cache_dir = NULL  # Uses default system cache directory
)
```

### Using Custom Providers

You can register custom LLM providers:

```{r custom-provider, purl=FALSE}
# Register a custom provider
register_custom_provider(
  provider_name = "my_custom_provider",
  process_fn = my_process_function,
  description = "My custom LLM provider"
)

# Register a model served by that provider
register_custom_model(
  model_name = "my-model-1",
  provider_name = "my_custom_provider",
  model_config = list()
)

# Use the custom model
results <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "my-model-1",
  api_key = Sys.getenv("MY_CUSTOM_PROVIDER_API_KEY")
)
```

## Caching Results

mLLMCelltype includes a caching system to save API costs:

```{r caching, purl=FALSE}
# Enable caching
Sys.setenv(MLLM_CACHE_ENABLED = "TRUE")

# Set cache directory
Sys.setenv(MLLM_CACHE_DIR = "path/to/cache")

# Clear cache if needed
mllmcelltype_clear_cache()
```

## Conclusion

mLLMCelltype is a framework for automated cell type annotation using large language models. By combining predictions from multiple models, it aims to produce more reliable annotations than single-model approaches.

For bug reports or feature requests, please visit: [https://github.com/cafferychen777/mLLMCelltype/issues](https://github.com/cafferychen777/mLLMCelltype/issues).

---
title: "Consensus vs Single-Agent: A Methodology Comparison"
description: "Comparison of multi-LLM consensus and single-agent approaches for cell type annotation, covering architecture, trade-offs, and practical considerations."
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Consensus vs Single-Agent: A Methodology Comparison}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Consensus vs Single-Agent: A Methodology Comparison

This vignette describes the architectural differences between multi-LLM consensus and single-agent approaches for cell type annotation, along with their respective trade-offs.

## Architectural Overview

### Single-Agent Systems

Single-agent approaches use specialized AI systems with predefined roles:

- **Specialized roles**: Different agents handle specific annotation tasks
- **Sequential processing**: Agents work in a pipeline fashion
- **Role-based validation**: Quality control through dedicated validation agents
- **Structured workflows**: Fixed processing sequences

### Multi-LLM Consensus Systems

Consensus frameworks leverage multiple independent models:

- **Parallel processing**: Multiple models analyze simultaneously
- **Collective decision-making**: Annotations are determined by agreement across models
- **Iterative refinement**: Discussion rounds for clusters where models disagree
- **Adaptive complexity**: More discussion is allocated to difficult cases

## Methodological Differences

### Single-Agent Approach

**Strengths:**
- Clear role definition: Each agent has specific responsibilities
- Streamlined workflows: Predictable processing pipelines
- Focused optimization: Agents can be fine-tuned for specific tasks
- Lower initial complexity: Easier to implement and understand

**Limitations:**
- Sequential bottlenecks: Failure in one agent affects the entire pipeline
- Limited model diversity: Typically relies on one underlying LLM family
- Rigid processing: Difficult to adapt to edge cases
- Single point of failure: Agent malfunction can compromise results

### Consensus Approach

**Strengths:**
- Error correction: Multiple models can catch each other's mistakes
- Model diversity: Leverages different training approaches and strengths
- Adaptive processing: More resources allocated to difficult cases
- Transparent uncertainty: Clear metrics for prediction confidence

**Challenges:**
- Initial complexity: Requires coordination between multiple models
- Resource coordination: Must manage multiple API calls efficiently
- Consensus building: Additional time for deliberation processes
- Model compatibility: Ensuring different models work together effectively

## Performance

For benchmark results comparing the two approaches, see Yang et al. (2025):

Yang, C., Zhang, X., & Chen, J. (2025). Large Language Model Consensus Substantially Improves the Cell Type Annotation Accuracy for scRNA-seq Data. *bioRxiv*. https://doi.org/10.1101/2025.04.10.647852

## Cost and Resource Trade-offs

| Aspect | Single-Agent | Consensus |
|--------|-------------|-----------|
| **API calls per cluster** | Fewer | More (multiple models) |
| **Cost per run** | Lower | Higher per run |
| **Two-stage optimization** | N/A | Reduces calls when models agree early |
| **Scalability** | Good | Good, with caching support |

The two-stage consensus approach in mLLMCelltype can reduce API calls when models agree early, since only clusters without initial consensus proceed to the deliberation stage.

## Practical Considerations

### When single-agent approaches may suffice:
- Standardized datasets with well-characterized tissues
- High-throughput screening of many similar datasets
- Limited API budget
- Straightforward annotation tasks

### When consensus approaches may be preferable:
- Novel biological contexts where model agreement provides additional confidence
- Work intended for publication, where uncertainty quantification is useful
- Complex tissues with many similar cell types
- Cases where identifying uncertain annotations is important

## Hybrid Approaches

Advanced workflows can combine both approaches:

1. **Initial screening**: Single-agent for clear cases
2. **Consensus validation**: Multi-model for uncertain cases
3. **Expert review**: Human validation for critical decisions

This tiered approach balances cost and thoroughness.

## Summary

Both methodologies have distinct strengths. Single-agent systems are simpler and less expensive per run, while consensus approaches provide uncertainty quantification and cross-model validation. The choice depends on the specific requirements of accuracy, cost, and biological complexity for a given project.

## Next Steps

- [Why Choose Consensus?](https://cafferyang.com/mLLMCelltype/articles/why-consensus.html) - Details on the consensus methodology
- [Getting Started Guide](https://cafferyang.com/mLLMCelltype/articles/getting-started.html) - Practical implementation tutorial
- [Performance Benchmarks](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html) - Detailed comparisons
- [API Reference](https://cafferyang.com/mLLMCelltype/reference/index.html) - Technical documentation

---
title: "Usage Tutorial"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Usage Tutorial}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

# Usage Tutorial

This tutorial provides detailed instructions for using mLLMCelltype for cell type annotation in single-cell RNA sequencing data. We'll cover various usage scenarios, parameter configurations, and integration with Seurat.

## Comprehensive Function Parameters

### annotate_cell_types()

The main function for cell type annotation with a single model:

```{r}
library(mLLMCelltype)

results <- annotate_cell_types(
  input,                # Marker gene data (data frame, list, or file path)
  tissue_name,          # Tissue name (e.g., "human PBMC", "mouse brain")
  model,                # LLM model to use
  api_key = NA,         # API key (if not set in environment, NA returns prompt only)
  top_gene_count = 10,  # Number of top genes per cluster to use
  debug = FALSE         # Whether to print debugging information
)
```

### interactive_consensus_annotation()

Function for creating consensus annotations from multiple models through interactive discussion:

```{r eval=FALSE}
consensus_results <- interactive_consensus_annotation(
  input,                # Original marker gene data (Seurat FindAllMarkers result or list of genes)
  tissue_name = NULL,   # Optional tissue name
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),  # Models to use
  api_keys,             # Named list of API keys
  top_gene_count = 10,  # Number of top genes to use
  controversy_threshold = 0.7,  # Threshold for identifying controversial clusters
  entropy_threshold = 1.0,  # Entropy threshold for controversial clusters
  max_discussion_rounds = 3,  # Maximum discussion rounds
  consensus_check_model = NULL,  # Model to use for consensus checking (see recommendations below)
  log_dir = "logs",     # Directory for logs
  cache_dir = NULL,  # Uses default system cache directory
  use_cache = TRUE      # Whether to use cache
)
```

**Important Note on `consensus_check_model`**: This parameter is used for evaluating semantic similarity, calculating consensus metrics, and moderating discussions. We recommend using capable models such as:
- `claude-sonnet-4-6` (Anthropic)
- `claude-opus-4-1-20250805` (Anthropic)
- `o1` (OpenAI)
- `gpt-5.5` (OpenAI)
- `gemini-3.1-pro-preview` (Google)

See the main README for detailed recommendations and examples.

## Detailed Usage Scenarios

### Scenario 1: Basic Annotation with a Single Model

For quick exploration or when API usage is a concern:

```{r}
# Load example data
library(Seurat)
data("pbmc_small")

# Find markers
pbmc_markers <- FindAllMarkers(pbmc_small, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)

# Run annotation with a single model
results <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10
)

# Add annotations to Seurat object
pbmc_small$cell_type_claude <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = names(results),
  to = results
)

# Visualize
DimPlot(pbmc_small, group.by = "cell_type_claude", label = TRUE)
```

### Scenario 2: Multi-Model Consensus for High Accuracy

For publication-quality annotations with uncertainty quantification:

```{r}
# Define multiple models to use
models <- c(
  "claude-sonnet-4-6",  # Anthropic
  "gpt-5.5",                      # OpenAI
  "gemini-3.1-pro-preview",              # Google
  "grok-4.3"                       # X.AI
)

# API keys for different providers
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  openai = Sys.getenv("OPENAI_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY"),
  grok = Sys.getenv("GROK_API_KEY")
)

# Run annotation with multiple models
results <- list()
for (model in models) {
  provider <- get_provider(model)
  api_key <- api_keys[[provider]]

  results[[model]] <- annotate_cell_types(
    input = pbmc_markers,
    tissue_name = "human PBMC",
    model = model,
    api_key = api_key,
    top_gene_count = 10
  )
}

# Create consensus
consensus_results <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = models,  # Use all the models defined above
  api_keys = api_keys,
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6"
)

# View consensus results
# You can access the final annotations with consensus_results$final_annotations

# Add consensus annotations and metrics to Seurat object
pbmc_small$cell_type_consensus <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = names(consensus_results$final_annotations),
  to = consensus_results$final_annotations
)

# Extract consensus metrics from the consensus results
consensus_metrics <- lapply(names(consensus_results$initial_results$consensus_results), function(cluster_id) {
  metrics <- consensus_results$initial_results$consensus_results[[cluster_id]]
  return(list(
    cluster = cluster_id,
    consensus_proportion = metrics$consensus_proportion,
    entropy = metrics$entropy
  ))
})

# Convert to data frame for easier handling
metrics_df <- do.call(rbind, lapply(consensus_metrics, data.frame))

# Add consensus proportion to Seurat object
pbmc_small$consensus_proportion <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = metrics_df$cluster,
  to = metrics_df$consensus_proportion
)

# Add entropy to Seurat object
pbmc_small$shannon_entropy <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = metrics_df$cluster,
  to = metrics_df$entropy
)
```

### Scenario 2b: Using Free OpenRouter Models

For users with limited API credits or budget constraints:

```{r}
# Set OpenRouter API key
openrouter_api_key <- Sys.getenv("OPENROUTER_API_KEY")

# Define free OpenRouter models to use
free_models <- c(
  "meta-llama/llama-4-maverick:free",                # Meta Llama 4 Maverick (free)
  "meta-llama/llama-3.3-70b-instruct:free",    # Meta Llama 3.3 70B (free)
  "deepseek/deepseek-v4-pro:free",             # DeepSeek V4 Pro (free)
  "meta-llama/llama-3.3-70b-instruct:free"          # Meta Llama 3.3 70B (free)
)

# Run annotation with free OpenRouter models
free_results <- list()
for (model in free_models) {
  free_results[[model]] <- annotate_cell_types(
    input = pbmc_markers,
    tissue_name = "human PBMC",
    model = model,  # OpenRouter models are automatically detected by format: 'provider/model-name:free'
    api_key = openrouter_api_key,
    top_gene_count = 10
  )
}

# Create consensus with free models
free_consensus_results <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = free_models,  # Use all the free models defined above
  api_keys = list("openrouter" = openrouter_api_key),
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "meta-llama/llama-4-maverick:free"  # Use a free model for consensus checking
)

# View free model consensus results
# You can access the final annotations with free_consensus_results$final_annotations

# Add free model consensus annotations to Seurat object
pbmc_small$free_model_consensus <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = names(free_consensus_results$final_annotations),
  to = free_consensus_results$final_annotations
)

# Compare paid vs. free model results
comparison <- data.frame(
  cluster = names(consensus_results$final_annotations),
  paid_models = consensus_results$final_annotations,
  free_models = free_consensus_results$final_annotations,
  agreement = consensus_results$final_annotations == free_consensus_results$final_annotations
)
print(comparison)
```

### Scenario 3: Working with CSV Files

For users who prefer working with files:

```{r}
# Save markers to CSV
write.csv(pbmc_markers, "pbmc_markers.csv", row.names = FALSE)

# Run annotation using the CSV file
results <- annotate_cell_types(
  input = "pbmc_markers.csv",
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

### Scenario 4: Custom Caching

For better control over caching behavior:

```{r}
# Note: The annotate_cell_types function does not have built-in caching.
# If you need caching, you can implement it separately.

# Run annotation
results <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10,
  debug = FALSE
)

# If you need custom caching, you can implement it using your own cache manager
# This is just a conceptual example and not part of the actual package
# cache_manager <- YourCacheManager$new(cache_dir = "path/to/cache")
# cache_manager$clear_cache()
```

## Model Selection Guide

mLLMCelltype supports a wide range of LLM models. Here's a guide to help you choose:

### High Performance Models

For the most accurate annotations:

- **Anthropic Claude Opus 4.7** (`claude-opus-4-7`)
- **Anthropic Claude Sonnet 4.6** (`claude-sonnet-4-6`)
- **OpenAI GPT-5.5** (`gpt-5.5`)
- **Google Gemini 3.1 Pro Preview** (`gemini-3.1-pro-preview`)

### Balanced Performance/Cost Models

For good results with lower API costs:

- **X.AI Grok 4.3** (`grok-4.3`): Competitive performance at lower cost
- **DeepSeek V4 Flash** (`deepseek-v4-flash`): Good performance for specialized tissues

### Economy Models

For preliminary exploration or large datasets:

- **Qwen 3.6 Flash** (`qwen3.6-flash`): Good performance for the cost
- **Zhipu/Z.AI GLM-5-Turbo** (`glm-5-turbo`): Economical option with decent performance
- **MiniMax M2.5** (`MiniMax-M2.5`): Cost-effective for initial exploration

### Free Models via OpenRouter

For users with limited API credits or budget constraints:

- **Meta Llama 4 Maverick** (`meta-llama/llama-4-maverick:free`): Most reliable and fast, recommended for consensus checking
- **Meta Llama 3.3 70B** (`meta-llama/llama-3.3-70b-instruct:free`): Good performance with consistent formatting
- **DeepSeek V4 Pro** (`deepseek/deepseek-v4-pro:free`): Reasoning model (free)

Based on our testing, we recommend the following free models:

- `meta-llama/llama-4-maverick:free`: Most reliable and fast, recommended for consensus checking
- `meta-llama/llama-3.3-70b-instruct:free`: Good performance with consistent formatting
- `deepseek/deepseek-v4-pro:free`: Reliable with good response time

Some models may have limitations:

- `deepseek/deepseek-v4-pro:free`: May occasionally return empty results
- `thudm/glm-z1-9b:free`: May return localized error messages ("non-character parameter") when used for consensus checking

These free models are accessed through OpenRouter and don't consume credits, but may have limitations compared to paid models. Use the `:free` suffix in the model name to access them.

```{r}
# Example of using a free model via OpenRouter
# First, set your OpenRouter API key
Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key")

# Then use a free model with the :free suffix
free_model_results <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "meta-llama/llama-4-maverick:free",  # Note the :free suffix
  api_key = Sys.getenv("OPENROUTER_API_KEY")
  # No need to specify provider - it's automatically detected from the model name format
)
```

## Integration with Seurat Workflow

Here's a complete example of integrating mLLMCelltype with a Seurat workflow:

```{r}
library(Seurat)
library(mLLMCelltype)
library(ggplot2)

# Load data
data("pbmc_small")

# Standard Seurat preprocessing
pbmc_small <- NormalizeData(pbmc_small)
pbmc_small <- FindVariableFeatures(pbmc_small)
pbmc_small <- ScaleData(pbmc_small)
pbmc_small <- RunPCA(pbmc_small)
pbmc_small <- FindNeighbors(pbmc_small)
pbmc_small <- FindClusters(pbmc_small, resolution = 0.5)
pbmc_small <- RunUMAP(pbmc_small, dims = 1:10)

# Find markers for each cluster
pbmc_markers <- FindAllMarkers(pbmc_small, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)

# Define models to use
models <- c(
  "claude-sonnet-4-6",
  "gpt-5.5",
  "gemini-3.1-pro-preview"
)

# API keys
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  openai = Sys.getenv("OPENAI_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY")
)

# Run annotation with multiple models
results <- list()
for (model in models) {
  provider <- get_provider(model)
  api_key <- api_keys[[provider]]

  results[[model]] <- annotate_cell_types(
    input = pbmc_markers,
    tissue_name = "human PBMC",
    model = model,
    api_key = api_key,
    top_gene_count = 10
  )

  # Add individual model results to Seurat object
  column_name <- paste0("cell_type_", gsub("[^a-zA-Z0-9]", "_", model))
  pbmc_small[[column_name]] <- plyr::mapvalues(
    x = as.character(Idents(pbmc_small)),
    from = names(results[[model]]),
    to = results[[model]]
  )
}

# Create consensus
consensus_results <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = models,  # Use all the models defined above
  api_keys = api_keys,
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6"
)

# Add consensus results to Seurat object
pbmc_small$cell_type_consensus <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = names(consensus_results$final_annotations),
  to = consensus_results$final_annotations
)

# Extract consensus metrics from the consensus results
consensus_metrics <- lapply(names(consensus_results$initial_results$consensus_results), function(cluster_id) {
  metrics <- consensus_results$initial_results$consensus_results[[cluster_id]]
  return(list(
    cluster = cluster_id,
    consensus_proportion = metrics$consensus_proportion,
    entropy = metrics$entropy
  ))
})

# Convert to data frame for easier handling
metrics_df <- do.call(rbind, lapply(consensus_metrics, data.frame))

# Add consensus proportion to Seurat object
pbmc_small$consensus_proportion <- as.numeric(plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = metrics_df$cluster,
  to = metrics_df$consensus_proportion
))

# Add entropy to Seurat object
pbmc_small$shannon_entropy <- as.numeric(plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = metrics_df$cluster,
  to = metrics_df$entropy
))

# Visualize results
p1 <- DimPlot(pbmc_small, group.by = "cell_type_consensus", label = TRUE, repel = TRUE) +
  ggtitle("Cell Type Annotations") +
  theme(plot.title = element_text(hjust = 0.5))

p2 <- FeaturePlot(pbmc_small, features = "consensus_proportion", cols = c("yellow", "green", "blue")) +
  ggtitle("Consensus Proportion") +
  theme(plot.title = element_text(hjust = 0.5))

p3 <- FeaturePlot(pbmc_small, features = "shannon_entropy", cols = c("red", "orange")) +
  ggtitle("Shannon Entropy") +
  theme(plot.title = element_text(hjust = 0.5))

# Combine plots
p1 | p2 | p3
```

## Advanced Parameter Tuning

### Adjusting top_gene_count

The `top_gene_count` parameter controls how many top marker genes per cluster are used for annotation:

```{r}
# Using more genes (better for well-characterized tissues)
results_more_genes <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 20  # Using more genes
)

# Using fewer genes (better for noisy data)
results_fewer_genes <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 5   # Using fewer genes
)
```

### Adjusting controversy_threshold

The `controversy_threshold` parameter in the `interactive_consensus_annotation` function controls which clusters are considered controversial and require discussion:

```{r eval=FALSE}
# Example of using interactive_consensus_annotation with different controversy thresholds
# Lower threshold (more clusters will be discussed)
consensus_results_low_threshold <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3-flash-preview"),
  api_keys = list(
    "anthropic" = Sys.getenv("ANTHROPIC_API_KEY"),
    "openai" = Sys.getenv("OPENAI_API_KEY"),
    "gemini" = Sys.getenv("GEMINI_API_KEY")
  ),
  controversy_threshold = 0.3  # Lower threshold - more clusters will be discussed
)

# Higher threshold (fewer clusters will be discussed)
consensus_results_high_threshold <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3-flash-preview"),
  api_keys = list(
    "anthropic" = Sys.getenv("ANTHROPIC_API_KEY"),
    "openai" = Sys.getenv("OPENAI_API_KEY"),
    "gemini" = Sys.getenv("GEMINI_API_KEY")
  ),
  controversy_threshold = 0.7  # Higher threshold - fewer clusters will be discussed
)
```

## Performance Considerations

### API Rate Limits and Costs

Different LLM providers have different rate limits and pricing:

- **OpenAI**: Higher rate limits but can be more expensive
- **Anthropic**: Good balance of rate limits and cost
- **Google**: Competitive pricing with good rate limits
- **Others**: Generally lower cost but may have stricter rate limits

To manage costs and rate limits:

1. Use caching to avoid redundant API calls
2. Start with a single model for exploration
3. Use more economical models for initial testing
4. Reserve multi-model consensus for final analysis

### Execution Time

Typical execution times:

- Single model annotation: 5-30 seconds per cluster
- Multi-model consensus: 1-5 minutes for a typical dataset
- Discussion process: Additional 1-3 minutes per controversial cluster

To improve performance:

1. Use a smaller `top_gene_count` for faster execution
2. Enable caching to reuse results
3. Use a higher `controversy_threshold` to reduce the number of clusters that require discussion

## Troubleshooting

### Common Issues with OpenRouter

If you encounter "No auth credentials found" errors with OpenRouter:

- Verify your API key is correct and active
- Ensure you're using the correct format for the model name (e.g., `provider/model-name:free`)
- Try a different free model, as some models may have access restrictions

### Non-English Model Error Messages

If you see "non-character parameter" errors:

- This typically occurs when using non-English language models like `thudm/glm-z1-9b:free` for consensus checking
- Switch to an English-optimized model like `meta-llama/llama-4-maverick:free` for consensus checking

### Empty Results

If a model returns empty results:

- Try increasing the timeout or retry with the same model
- Some models like `deepseek/deepseek-v4-pro:free` may occasionally return empty results
- Switch to a more reliable model if the problem persists

## Next Steps

Now that you understand the detailed usage of mLLMCelltype, you can explore:

- [Consensus Annotation Principles](https://cafferyang.com/mLLMCelltype/articles/consensus-principles.html): Learn about the technical principles
- [Visualization Guide](https://cafferyang.com/mLLMCelltype/articles/visualization-guide.html): Create publication-ready visualizations
- [FAQ](https://cafferyang.com/mLLMCelltype/articles/faq.html): Find answers to common questions
- [Advanced Features](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html): Explore hierarchical annotation and other advanced features

---
title: "Advanced Features & Case Studies"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Advanced Features & Case Studies}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

This article explores advanced features of mLLMCelltype and presents practical examples demonstrating its application in various research contexts.

## Hierarchical Cell Type Annotation

### Understanding Hierarchical Annotation

Cell types often exist in hierarchical relationships. For example, T cells can be further classified into CD4+ T cells, CD8+ T cells, regulatory T cells, etc. mLLMCelltype can be used in a multi-step workflow to capture these hierarchical relationships.

### Implementing Hierarchical Annotation

Here's a practical approach to perform hierarchical annotation:

```{r}
library(mLLMCelltype)
library(Seurat)
library(dplyr)

# Step 1: Perform initial high-level annotation
high_level_results <- annotate_cell_types(
  input = marker_data,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10
)

# Step 2: Add high-level annotations to Seurat object
seurat_obj$high_level_celltype <- plyr::mapvalues(
  x = as.character(Idents(seurat_obj)),
  from = names(high_level_results),
  to = high_level_results
)

# Step 3: Subset T cells for further annotation
t_cells <- subset(seurat_obj, high_level_celltype == "T cells")

# Step 4: Find markers within T cells
t_cell_markers <- FindAllMarkers(t_cells, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)

# Step 5: Perform T cell subtype annotation
t_cell_subtypes <- annotate_cell_types(
  input = t_cell_markers,
  tissue_name = "human PBMC T cells",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10
)

# Step 6: Add T cell subtypes back to original object
t_cell_barcodes <- WhichCells(t_cells)
seurat_obj$detailed_celltype <- seurat_obj$high_level_celltype
seurat_obj$detailed_celltype[t_cell_barcodes] <- plyr::mapvalues(
  x = as.character(Idents(t_cells)),
  from = names(t_cell_subtypes),
  to = paste0("T cells: ", t_cell_subtypes)
)
```

### Validating Hierarchical Annotations

After creating hierarchical annotations, it's important to validate the consistency between levels:

```{r}
# Create a simple function to check parent-child consistency
validate_hierarchy <- function(high_level, detailed_level) {
  # Extract parent type from detailed annotation (before the colon)
  parent_from_detailed <- sapply(strsplit(detailed_level, ": "), function(x) x[1])

  # Check if parent matches high-level annotation
  consistent <- parent_from_detailed == high_level

  # Return consistency check results
  data.frame(
    high_level = high_level,
    detailed_level = detailed_level,
    consistent = consistent
  )
}

# Apply validation
hierarchy_validation <- validate_hierarchy(
  seurat_obj$high_level_celltype,
  seurat_obj$detailed_celltype
)

# Identify inconsistencies
inconsistencies <- hierarchy_validation[!hierarchy_validation$consistent, ]
print(inconsistencies)
```

## Handling Noisy Input Data

### Strategies for Noisy Marker Genes

Real-world scRNA-seq data often contains noise. Here are practical strategies for handling noisy input:

#### 1. Adjust the top_gene_count parameter

For noisy datasets, using fewer top genes can help focus on the strongest signals:

```{r}
# For noisy data, use fewer top genes
results_fewer_genes <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 5  # Use fewer genes to focus on strongest signals
)
```

#### 2. Apply stricter filtering for marker genes

Pre-filtering marker genes with stricter thresholds can improve annotation quality:

```{r}
# Apply stricter filtering to marker genes
filtered_markers <- marker_data %>%
  filter(p_val_adj < 0.01, avg_log2FC > 1.0)  # Stricter thresholds

# Annotate with filtered markers
results_filtered <- annotate_cell_types(
  input = filtered_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

#### 3. Use multi-model consensus

The consensus approach can help overcome noise by combining predictions from multiple models:

```{r}
# Set up API keys
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  openai = Sys.getenv("OPENAI_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY")
)

# Define multiple models to use
models <- c(
  "claude-sonnet-4-6",
  "gpt-5.5",
  "gemini-3.1-pro-preview"
)

# Create consensus using interactive_consensus_annotation
consensus_results <- interactive_consensus_annotation(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  models = models,
  api_keys = api_keys,
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6"
)
```

### Handling Data with Batch Effects

When working with data affected by batch effects, you can:

#### 1. Use the consensus approach with a lower controversy threshold

```{r}
# For data with batch effects, use consensus with lower threshold
batch_consensus <- interactive_consensus_annotation(
  input = marker_data,  # Your marker gene data with batch effects
  tissue_name = "mouse brain",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
  api_keys = api_keys,
  controversy_threshold = 0.4,  # Lower threshold to discuss more clusters
  entropy_threshold = 0.8  # Lower entropy threshold
)
```

#### 2. Include batch information in the tissue context

```{r}
# Include batch information in the tissue context
batch_aware_results <- annotate_cell_types(
  input = marker_data,  # Your marker gene data with batch effects
  tissue_name = "mouse brain with technical batch effects",  # Include batch context
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

## Incorporating Domain Knowledge

### Using Tissue Context

One of the key features of mLLMCelltype is the ability to incorporate domain knowledge through the `tissue_name` parameter. This provides important context to the LLM:

```{r}
# Basic annotation without specific tissue context
basic_results <- annotate_cell_types(
  input = marker_data,
  tissue_name = "human sample",  # Generic context
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

# Annotation with specific tissue context
specific_results <- annotate_cell_types(
  input = marker_data,
  tissue_name = "human fetal liver at 20 weeks gestation",  # Detailed context
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

### Creating Custom Prompts

For advanced use cases, you can create and modify the annotation prompt directly:

```{r}
# Create a custom annotation prompt
custom_prompt <- create_annotation_prompt(
  input = marker_data,
  tissue_name = "human PBMC",
  top_gene_count = 10
)

# Modify the prompt to include additional context
modified_prompt <- paste0(
  custom_prompt$prompt,
  "\n\nAdditional context: This sample is from a patient with rheumatoid arthritis. ",
  "Previous studies have identified activated T cells, B cells, and CXCR4-high monocytes in this condition."
)

# Use the modified prompt directly
custom_results <- get_model_response(
  prompt = modified_prompt,
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

### Combining with External Resources

You can enhance your annotation workflow by combining mLLMCelltype with other R packages and resources:

```{r}
library(Seurat)
library(dplyr)

# Example: Using CellMarker database information to validate annotations
# This is a conceptual example - implementation would depend on your specific needs

# 1. Get annotations with mLLMCelltype
annotations <- annotate_cell_types(
  input = marker_data,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

# 2. Compare with known marker genes (conceptual)
# In a real workflow, you would query a database or use a reference dataset
known_markers <- list(
  "T cells" = c("CD3D", "CD3E", "CD3G"),
  "B cells" = c("CD19", "MS4A1", "CD79A"),
  "Monocytes" = c("CD14", "LYZ", "CSF1R")
)

# 3. Validate annotations against known markers
# This is a simplified example of how you might validate annotations
validate_annotations <- function(annotations, marker_data, known_markers) {
  validation_results <- list()

  for (i in 1:length(annotations)) {
    cluster_id <- i
    predicted_type <- annotations[i]

    # Get markers for this cluster
    cluster_markers <- marker_data %>%
      filter(cluster == cluster_id) %>%
      arrange(desc(avg_log2FC)) %>%
      pull(gene) %>%
      head(20)

    # Check overlap with known markers for this cell type
    if (predicted_type %in% names(known_markers)) {
      expected_markers <- known_markers[[predicted_type]]
      overlap <- intersect(cluster_markers, expected_markers)

      validation_results[[i]] <- list(
        cluster = cluster_id,
        predicted_type = predicted_type,
        overlap_count = length(overlap),
        overlap_genes = paste(overlap, collapse = ", "),
        confidence = length(overlap) / length(expected_markers)
      )
    } else {
      validation_results[[i]] <- list(
        cluster = cluster_id,
        predicted_type = predicted_type,
        overlap_count = 0,
        overlap_genes = "",
        confidence = 0
      )
    }
  }

  return(validation_results)
}

# This is a conceptual example of how you might validate annotations
# validation_results <- validate_annotations(annotations, marker_data, known_markers)
```

## Practical Case Studies

### Case Study 1: PBMC Dataset Analysis

This example demonstrates a complete workflow for analyzing a PBMC dataset:

```{r}
library(Seurat)
library(mLLMCelltype)
library(ggplot2)
library(dplyr)

# Load example PBMC data
# In a real workflow, you would use your own data
data("pbmc_small")  # Example dataset from Seurat

# Find marker genes
pbmc_markers <- FindAllMarkers(pbmc_small,
                              only.pos = TRUE,
                              min.pct = 0.25,
                              logfc.threshold = 0.25)

# Set up API keys
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  openai = Sys.getenv("OPENAI_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY")
)

# Use consensus annotation
consensus_results <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
  api_keys = api_keys,
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6"
)

# Add results to Seurat object
pbmc_small$cell_type <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = names(consensus_results$final_annotations),
  to = consensus_results$final_annotations
)

# Visualize results
# In a real workflow, you would create a UMAP or t-SNE plot
# DimPlot(pbmc_small, group.by = "cell_type", label = TRUE) +
#   ggtitle("PBMC Cell Types")
```

### Case Study 2: Identifying Rare Cell Types

When working with datasets containing rare cell populations, you can adjust parameters to improve detection:

```{r}
# For rare cell types, use these strategies:

# 1. Increase the number of marker genes considered
rare_cell_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human bone marrow",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 20  # Use more genes for rare cell types
)

# 2. Use consensus with lower thresholds to discuss more clusters
rare_cell_consensus <- interactive_consensus_annotation(
  input = marker_data,  # Your marker gene data
  tissue_name = "human bone marrow",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
  api_keys = api_keys,
  controversy_threshold = 0.4,  # Lower threshold to discuss more clusters
  entropy_threshold = 0.8,  # Lower entropy threshold
  consensus_check_model = "claude-sonnet-4-6"
)

# 3. Provide more specific tissue context
specific_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human bone marrow with expected rare plasma cells and basophils",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

### Case Study 3: Cross-Species Comparison

mLLMCelltype can be used to compare cell types across different species:

```{r}
# Example workflow for cross-species comparison

# 1. Annotate human and mouse datasets separately
# (Assuming you have marker data for both species)
human_annotations <- annotate_cell_types(
  input = human_marker_data,  # Your human marker data
  tissue_name = "human brain cortex",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

mouse_annotations <- annotate_cell_types(
  input = mouse_marker_data,  # Your mouse marker data
  tissue_name = "mouse brain cortex",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

# 2. Compare annotations
# This is a conceptual example - in a real workflow, you would:
# - Map annotations to Seurat objects
# - Calculate proportions
# - Create comparison visualizations
# - Identify conserved and species-specific cell types

# Example comparison function (conceptual)
compare_species_annotations <- function(human_annotations, mouse_annotations) {
  # Get unique cell types from both species
  human_types <- unique(human_annotations)
  mouse_types <- unique(mouse_annotations)

  # Find common cell types
  common_types <- intersect(human_types, mouse_types)

  # Find species-specific cell types
  human_specific <- setdiff(human_types, mouse_types)
  mouse_specific <- setdiff(mouse_types, human_types)

  # Return comparison results
  list(
    common_types = common_types,
    human_specific = human_specific,
    mouse_specific = mouse_specific
  )
}

# This is a conceptual example
# comparison <- compare_species_annotations(human_annotations, mouse_annotations)
```

## Performance Considerations

### API Cost Management

When using mLLMCelltype, it's important to consider the costs associated with API calls to different LLM providers:

```{r}
# Example of cost-efficient model selection
# Choose models based on your specific needs and budget

# For initial exploration or smaller datasets
# Use more affordable models
affordable_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  model = "claude-haiku-4-5-20251001",  # More affordable model
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

# For final analysis or challenging datasets
# Use larger models
premium_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",  # Larger model
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

# Use OpenRouter for access to free models
openrouter_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  model = "meta-llama/llama-3.3-70b-instruct:free",  # Free model via OpenRouter
  api_key = Sys.getenv("OPENROUTER_API_KEY")
)
```

### Optimizing Runtime

To optimize runtime when working with large datasets:

```{r}
# 1. Use caching with interactive_consensus_annotation
consensus_with_cache <- interactive_consensus_annotation(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  models = c("claude-sonnet-4-6", "gpt-5.5"),
  api_keys = api_keys,
  use_cache = TRUE,  # Enable caching
  cache_dir = NULL  # Uses default system cache directory
)

# 2. Process clusters in batches
# This is a conceptual example - implementation would depend on your workflow
process_in_batches <- function(marker_data, batch_size = 5) {
  # Get unique clusters
  clusters <- unique(marker_data$cluster)

  # Process in batches
  results <- list()
  for (i in seq(1, length(clusters), by = batch_size)) {
    # Get current batch of clusters
    batch_clusters <- clusters[i:min(i + batch_size - 1, length(clusters))]

    # Filter marker data for current batch
    batch_data <- marker_data %>% filter(cluster %in% batch_clusters)

    # Process batch
    batch_results <- annotate_cell_types(
      input = batch_data,
      tissue_name = "human PBMC",
      model = "claude-sonnet-4-6",
      api_key = Sys.getenv("ANTHROPIC_API_KEY")
    )

    # Store results
    results <- c(results, batch_results)
  }

  return(results)
}

# 3. Use faster models for initial exploration
fast_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  model = "claude-haiku-4-5-20251001",  # Faster model
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

## Advanced Customization

### Custom Processing Functions

For advanced users, mLLMCelltype allows you to register custom providers and models:

```{r}
# Define a custom processing function
# This function must accept prompt, model, and api_key parameters
custom_process_fn <- function(prompt, model, api_key) {
  # Custom implementation to process prompts and get responses
  # This is a simplified example
  cat("Processing prompt with custom provider\n")
  cat("Model:", model, "\n")

  # In a real implementation, you would make API calls here
  # For example:
  # response <- httr::POST(
  #   url = "https://api.custom-provider.com/v1/chat/completions",
  #   body = list(prompt = prompt, model = model),
  #   httr::add_headers(Authorization = paste("Bearer", api_key)),
  #   encode = "json"
  # )
  # result <- httr::content(response)$choices[[1]]$text

  # For this example, just return a fixed response
  result <- "T cells"
  return(result)
}

# Register the custom provider
register_custom_provider(
  provider_name = "custom_provider",
  process_fn = custom_process_fn,
  description = "My custom LLM provider"
)

# Register a custom model
register_custom_model(
  model_name = "custom-model",
  provider_name = "custom_provider",
  model_config = list(
    temperature = 0.7,
    max_tokens = 2000
  )
)

# Use the custom model
# custom_results <- annotate_cell_types(
#   input = marker_data,
#   tissue_name = "human PBMC",
#   model = "custom-model",
#   api_key = "your-custom-api-key"
# )
```

### Using the Unified Logging System

mLLMCelltype provides a comprehensive unified logging system with structured output, performance monitoring, and multi-level logging:

```{r}
# Configure the global logger (recommended approach)
configure_logger(level = "INFO", console_output = TRUE, json_format = TRUE)

# Use simple logging functions
log_info("Starting analysis of cluster 0", list(
  cluster_id = "0",
  tissue_name = "human PBMC",
  marker_genes = c("CD3D", "CD3E", "CD2", "IL7R", "LTB")
))

# Log API calls with performance tracking
log_info("API call completed", list(
  provider = "anthropic",
  model = "claude-sonnet-4-6",
  duration_seconds = 2.34,
  success = TRUE
))

# Log warnings and errors
log_warn("Model response had unusual format", list(
  model = "gpt-5.5",
  response_length = 50
))

log_error("API call failed", list(
  provider = "openai",
  error = "Rate limit exceeded"
))

# Alternatively, create a custom logger instance
custom_logger <- UnifiedLogger$new(
  base_dir = "custom_logs",
  level = "DEBUG",
  console_output = TRUE,
  json_format = TRUE
)

# Use the custom logger
custom_logger$info("Custom log message", list(analysis_step = "preprocessing"))
custom_logger$debug("Detailed debugging info", list(variable_state = "initialized"))

# Get performance summary
performance <- get_logger()$get_performance_summary()
print(performance)
```

### Using the CacheManager

The `CacheManager` class helps optimize performance by caching consensus results:

```{r}
# Create a cache manager
cache_manager <- CacheManager$new(cache_dir = NULL)

# Generate a cache key
cache_key <- cache_manager$generate_key(
  input = marker_data,
  models = c("claude-sonnet-4-6", "gpt-5.5"),
  cluster_id = "0"
)

# Check if results exist in cache
if (cache_manager$has_cache(cache_key)) {
  # Load from cache
  cached_results <- cache_manager$load_from_cache(cache_key)
} else {
  # Process and save to cache
  # results <- process_cluster(...)
  # cache_manager$save_to_cache(cache_key, results)
}

# Get cache statistics
cache_stats <- cache_manager$get_cache_stats()

# Clear cache (with confirmation)
# cache_manager$clear_cache(confirm = TRUE)
```

### Cache Management

mLLMCelltype provides convenient functions for managing cache directories:

```{r}
# Check cache location
mllmcelltype_cache_dir()

# Use local cache
mllmcelltype_cache_dir("local")

# Clear cache
mllmcelltype_clear_cache()
```

## Next Steps

Now that you've explored the advanced features of mLLMCelltype, you can:

- [Contribute to the project](https://cafferyang.com/mLLMCelltype/articles/contributing-guide.html): Learn how to contribute to mLLMCelltype
- [Review the version history](https://cafferyang.com/mLLMCelltype/news/index.html): Explore the development history of the package
- [Return to the introduction](https://cafferyang.com/mLLMCelltype/articles/introduction.html): Review the basic concepts

---
title: "Installation Guide"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Installation Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

# Installation Guide

This guide provides detailed instructions for installing and configuring mLLMCelltype for cell type annotation in single-cell RNA sequencing data.

## System Requirements

Before installing mLLMCelltype, ensure your system meets the following requirements:

- **R version**: 4.0.0 or higher
- **Memory**: At least 8GB RAM recommended (more for large datasets)
- **Operating System**: Windows, macOS, or Linux
- **Internet Connection**: Required for API calls to LLM providers

## Installing the R Package

### Installation from CRAN (Recommended)

mLLMCelltype is now available on CRAN. You can install it directly using:

```{r}
# Install from CRAN
install.packages("mLLMCelltype")
```

This will install the stable version of mLLMCelltype with all required dependencies.

### Installation from GitHub (Development Version)

To install the latest development version from GitHub:

```{r}
# Install devtools if not already installed
if (!requireNamespace("devtools", quietly = TRUE)) {
  install.packages("devtools")
}

# Install mLLMCelltype development version
devtools::install_github("cafferychen777/mLLMCelltype", subdir = "R")
```

### Installation from a Local Source

If you have downloaded the source code or need to install from a local copy:

```{r}
# Assuming the package is in the current working directory
devtools::install_local("path/to/mLLMCelltype/R")
```

## Dependencies

mLLMCelltype depends on several R packages that will be automatically installed during the installation process. The main dependencies include:

- **dplyr**: For data manipulation
- **httr**: For API requests
- **jsonlite**: For JSON parsing
- **R6**: For object-oriented programming
- **digest**: For caching mechanisms
- **magrittr**: For pipe operations

For visualization and integration with single-cell analysis workflows, the following packages are recommended but not required:

- **Seurat**: For integration with Seurat objects
- **ggplot2**: For visualization
- **SCpubr**: For publication-ready visualizations

## API Keys Setup

mLLMCelltype requires API keys to access different LLM providers. You will need to obtain API keys for at least one of the supported providers:

### Obtaining API Keys

1. **OpenAI (GPT-4o/4.1)**
   - Visit OpenAI Platform
   - Create an account or log in
   - Navigate to API keys section
   - Create a new API key

2. **Anthropic (Claude-3.7/3.5)**
   - Visit Anthropic Console
   - Create an account or log in
   - Generate an API key

3. **Google (Gemini-2.0/2.5)**
   - Visit [Google AI Studio](https://makersuite.google.com/)
   - Create a Google account or log in
   - Generate an API key

4. **Other Providers**
   - Similar processes apply for DeepSeek, Qwen, Zhipu, MiniMax, Stepfun, and Grok
   - Visit their respective websites to obtain API keys

### Setting Up API Keys

There are three ways to set up your API keys:

#### 1. Environment Variables

Create a `.env` file in your project directory with your API keys:

```
# API Keys for different LLM models
OPENAI_API_KEY=your-openai-key
ANTHROPIC_API_KEY=your-anthropic-key
GEMINI_API_KEY=your-gemini-key
DEEPSEEK_API_KEY=your-deepseek-key
QWEN_API_KEY=your-qwen-key
ZHIPU_API_KEY=your-zhipu-key
STEPFUN_API_KEY=your-stepfun-key
MINIMAX_API_KEY=your-minimax-key
GROK_API_KEY=your-grok-key
OPENROUTER_API_KEY=your-openrouter-key
```

Then load the environment variables in your R script:

```{r}
library(dotenv)
dotenv::load_dot_env()
```

#### 2. Direct Specification in Function Calls

You can directly provide API keys in function calls:

```{r}
library(mLLMCelltype)

results <- annotate_cell_types(
  input = your_marker_data,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = "your-anthropic-key",
  top_gene_count = 10
)
```

#### 3. R Environment Variables

Set API keys as R environment variables:

```{r}
Sys.setenv(OPENAI_API_KEY = "your-openai-key")
Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-key")
# Set other API keys as needed
```

## Verifying Installation

To verify that mLLMCelltype is installed correctly and API keys are set up properly:

```{r}
library(mLLMCelltype)

# Check if the package is loaded correctly
packageVersion("mLLMCelltype")

# Verify API key setup for a specific provider
api_key <- get_api_key("anthropic")
if (!is.null(api_key) && api_key != "") {
  cat("Anthropic API key is set up correctly\n")
} else {
  cat("Anthropic API key is not set up\n")
}
```

## Common Installation Issues

### Package Installation Failures

If you encounter issues during installation:

1. **Check R version**: Ensure you're using R 4.0.0 or higher
2. **Update devtools**: Run `install.packages("devtools")` to ensure you have the latest version
3. **Check dependencies**: Some dependencies might require system libraries on Linux

### API Connection Issues

If you encounter issues connecting to LLM APIs:

1. **Verify API keys**: Ensure your API keys are correct and have not expired
2. **Check internet connection**: Ensure you have a stable internet connection
3. **Proxy settings**: If you're behind a proxy, configure R to use your proxy settings

```{r}
# Example of setting proxy for httr
httr::set_config(httr::use_proxy(url = "proxy_url", port = proxy_port))
```

### Memory Limitations

For large datasets, you might encounter memory issues:

1. **Increase R memory limit**: Use `memory.limit(size = 16000)` on Windows to increase available memory
2. **Process data in batches**: Consider processing large datasets in smaller batches

## Next Steps

Now that you have installed mLLMCelltype, you can proceed to:

- [Getting Started](https://cafferyang.com/mLLMCelltype/articles/getting-started.html): Learn the basics of using mLLMCelltype
- [Usage Tutorial](https://cafferyang.com/mLLMCelltype/articles/usage-tutorial.html): Explore more advanced usage scenarios
- [Visualization Guide](https://cafferyang.com/mLLMCelltype/articles/visualization-guide.html): Learn how to visualize your results

If you encounter any issues not covered in this guide, please [open an issue](https://github.com/cafferychen777/mLLMCelltype/issues) on our GitHub repository.

---
title: "Installation Guide"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Installation Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE, purl=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

# Installation Guide

This guide provides detailed instructions for installing and configuring mLLMCelltype for cell type annotation in single-cell RNA sequencing data.

## System Requirements

Before installing mLLMCelltype, ensure your system meets the following requirements:

- **R version**: 4.0.0 or higher
- **Memory**: At least 8GB RAM recommended (more for large datasets)
- **Operating System**: Windows, macOS, or Linux
- **Internet Connection**: Required for API calls to LLM providers

## Installing the R Package

### Installation from CRAN (Recommended)

mLLMCelltype is now available on CRAN. You can install it directly using:

```{r, purl=FALSE}
# Install from CRAN
install.packages("mLLMCelltype")
```

This will install the stable version of mLLMCelltype with all required dependencies.

### Installation from GitHub (Development Version)

To install the latest development version from GitHub:

```{r, purl=FALSE}
# Install devtools if not already installed
if (!requireNamespace("devtools", quietly = TRUE)) {
  install.packages("devtools")
}

# Install mLLMCelltype development version
devtools::install_github("cafferychen777/mLLMCelltype", subdir = "R")
```

### Installation from a Local Source

If you have downloaded the source code or need to install from a local copy:

```{r, purl=FALSE}
# Assuming the package is in the current working directory
devtools::install_local("path/to/mLLMCelltype/R")
```

## Dependencies

mLLMCelltype depends on several R packages that will be automatically installed during the installation process. The main dependencies include:

- **dplyr**: For data manipulation
- **httr**: For API requests
- **jsonlite**: For JSON parsing
- **R6**: For object-oriented programming
- **digest**: For caching mechanisms
- **magrittr**: For pipe operations

For visualization and integration with single-cell analysis workflows, the following packages are recommended but not required:

- **Seurat**: For integration with Seurat objects
- **ggplot2**: For visualization
- **SCpubr**: For publication-ready visualizations

## API Keys Setup

mLLMCelltype requires API keys to access different LLM providers. You will need to obtain API keys for at least one of the supported providers:

### Obtaining API Keys

1. **OpenAI (GPT-4o/4.1)**
   - Visit OpenAI Platform
   - Create an account or log in
   - Navigate to API keys section
   - Create a new API key

2. **Anthropic (Claude-3.7/3.5)**
   - Visit Anthropic Console
   - Create an account or log in
   - Generate an API key

3. **Google (Gemini-2.0/2.5)**
   - Visit [Google AI Studio](https://makersuite.google.com/)
   - Create a Google account or log in
   - Generate an API key

4. **Other Providers**
   - Similar processes apply for DeepSeek, Qwen, Zhipu, MiniMax, Stepfun, and Grok
   - Visit their respective websites to obtain API keys

### Setting Up API Keys

There are three ways to set up your API keys:

#### 1. Environment Variables

Create a `.env` file in your project directory with your API keys:

```
# API Keys for different LLM models
OPENAI_API_KEY=your-openai-key
ANTHROPIC_API_KEY=your-anthropic-key
GEMINI_API_KEY=your-gemini-key
DEEPSEEK_API_KEY=your-deepseek-key
QWEN_API_KEY=your-qwen-key
ZHIPU_API_KEY=your-zhipu-key
STEPFUN_API_KEY=your-stepfun-key
MINIMAX_API_KEY=your-minimax-key
GROK_API_KEY=your-grok-key
OPENROUTER_API_KEY=your-openrouter-key
```

Then load the environment variables in your R script:

```{r, purl=FALSE}
library(dotenv)
dotenv::load_dot_env()
```

#### 2. Direct Specification in Function Calls

You can directly provide API keys in function calls:

```{r, purl=FALSE}
library(mLLMCelltype)

results <- annotate_cell_types(
  input = your_marker_data,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = "your-anthropic-key",
  top_gene_count = 10
)
```

#### 3. R Environment Variables

Set API keys as R environment variables:

```{r, purl=FALSE}
Sys.setenv(OPENAI_API_KEY = "your-openai-key")
Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-key")
# Set other API keys as needed
```

## Verifying Installation

To verify that mLLMCelltype is installed correctly and API keys are set up properly:

```{r, purl=FALSE}
library(mLLMCelltype)

# Check if the package is loaded correctly
packageVersion("mLLMCelltype")

# Verify API key setup for a specific provider
api_key <- get_api_key("anthropic")
if (!is.null(api_key) && api_key != "") {
  cat("Anthropic API key is set up correctly\n")
} else {
  cat("Anthropic API key is not set up\n")
}
```

## Common Installation Issues

### Package Installation Failures

If you encounter issues during installation:

1. **Check R version**: Ensure you're using R 4.0.0 or higher
2. **Update devtools**: Run `install.packages("devtools")` to ensure you have the latest version
3. **Check dependencies**: Some dependencies might require system libraries on Linux

### API Connection Issues

If you encounter issues connecting to LLM APIs:

1. **Verify API keys**: Ensure your API keys are correct and have not expired
2. **Check internet connection**: Ensure you have a stable internet connection
3. **Proxy settings**: If you're behind a proxy, configure R to use your proxy settings

```{r, purl=FALSE}
# Example of setting proxy for httr
httr::set_config(httr::use_proxy(url = "proxy_url", port = proxy_port))
```

### Memory Limitations

For large datasets, you might encounter memory issues:

1. **Increase R memory limit**: Use `memory.limit(size = 16000)` on Windows to increase available memory
2. **Process data in batches**: Consider processing large datasets in smaller batches

## Next Steps

Now that you have installed mLLMCelltype, you can proceed to:

- [Getting Started](https://cafferyang.com/mLLMCelltype/articles/getting-started.html): Learn the basics of using mLLMCelltype
- [Usage Tutorial](https://cafferyang.com/mLLMCelltype/articles/usage-tutorial.html): Explore more advanced usage scenarios
- [Visualization Guide](https://cafferyang.com/mLLMCelltype/articles/visualization-guide.html): Learn how to visualize your results

If you encounter any issues not covered in this guide, please [open an issue](https://github.com/cafferychen777/mLLMCelltype/issues) on our GitHub repository.

---
title: "Advanced Features & Case Studies"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Advanced Features & Case Studies}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE, purl=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

This article explores advanced features of mLLMCelltype and presents practical examples demonstrating its application in various research contexts.

## Hierarchical Cell Type Annotation

### Understanding Hierarchical Annotation

Cell types often exist in hierarchical relationships. For example, T cells can be further classified into CD4+ T cells, CD8+ T cells, regulatory T cells, etc. mLLMCelltype can be used in a multi-step workflow to capture these hierarchical relationships.

### Implementing Hierarchical Annotation

Here's a practical approach to perform hierarchical annotation:

```{r, purl=FALSE}
library(mLLMCelltype)
library(Seurat)
library(dplyr)

# Step 1: Perform initial high-level annotation
high_level_results <- annotate_cell_types(
  input = marker_data,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10
)

# Step 2: Add high-level annotations to Seurat object
seurat_obj$high_level_celltype <- plyr::mapvalues(
  x = as.character(Idents(seurat_obj)),
  from = names(high_level_results),
  to = high_level_results
)

# Step 3: Subset T cells for further annotation
t_cells <- subset(seurat_obj, high_level_celltype == "T cells")

# Step 4: Find markers within T cells
t_cell_markers <- FindAllMarkers(t_cells, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)

# Step 5: Perform T cell subtype annotation
t_cell_subtypes <- annotate_cell_types(
  input = t_cell_markers,
  tissue_name = "human PBMC T cells",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10
)

# Step 6: Add T cell subtypes back to original object
t_cell_barcodes <- WhichCells(t_cells)
seurat_obj$detailed_celltype <- seurat_obj$high_level_celltype
seurat_obj$detailed_celltype[t_cell_barcodes] <- plyr::mapvalues(
  x = as.character(Idents(t_cells)),
  from = names(t_cell_subtypes),
  to = paste0("T cells: ", t_cell_subtypes)
)
```

### Validating Hierarchical Annotations

After creating hierarchical annotations, it's important to validate the consistency between levels:

```{r, purl=FALSE}
# Create a simple function to check parent-child consistency
validate_hierarchy <- function(high_level, detailed_level) {
  # Extract parent type from detailed annotation (before the colon)
  parent_from_detailed <- sapply(strsplit(detailed_level, ": "), function(x) x[1])

  # Check if parent matches high-level annotation
  consistent <- parent_from_detailed == high_level

  # Return consistency check results
  data.frame(
    high_level = high_level,
    detailed_level = detailed_level,
    consistent = consistent
  )
}

# Apply validation
hierarchy_validation <- validate_hierarchy(
  seurat_obj$high_level_celltype,
  seurat_obj$detailed_celltype
)

# Identify inconsistencies
inconsistencies <- hierarchy_validation[!hierarchy_validation$consistent, ]
print(inconsistencies)
```

## Handling Noisy Input Data

### Strategies for Noisy Marker Genes

Real-world scRNA-seq data often contains noise. Here are practical strategies for handling noisy input:

#### 1. Adjust the top_gene_count parameter

For noisy datasets, using fewer top genes can help focus on the strongest signals:

```{r, purl=FALSE}
# For noisy data, use fewer top genes
results_fewer_genes <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 5  # Use fewer genes to focus on strongest signals
)
```

#### 2. Apply stricter filtering for marker genes

Pre-filtering marker genes with stricter thresholds can improve annotation quality:

```{r, purl=FALSE}
# Apply stricter filtering to marker genes
filtered_markers <- marker_data %>%
  filter(p_val_adj < 0.01, avg_log2FC > 1.0)  # Stricter thresholds

# Annotate with filtered markers
results_filtered <- annotate_cell_types(
  input = filtered_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

#### 3. Use multi-model consensus

The consensus approach can help overcome noise by combining predictions from multiple models:

```{r, purl=FALSE}
# Set up API keys
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  openai = Sys.getenv("OPENAI_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY")
)

# Define multiple models to use
models <- c(
  "claude-sonnet-4-6",
  "gpt-5.5",
  "gemini-3.1-pro-preview"
)

# Create consensus using interactive_consensus_annotation
consensus_results <- interactive_consensus_annotation(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  models = models,
  api_keys = api_keys,
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6"
)
```

### Handling Data with Batch Effects

When working with data affected by batch effects, you can:

#### 1. Use the consensus approach with a lower controversy threshold

```{r, purl=FALSE}
# For data with batch effects, use consensus with lower threshold
batch_consensus <- interactive_consensus_annotation(
  input = marker_data,  # Your marker gene data with batch effects
  tissue_name = "mouse brain",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
  api_keys = api_keys,
  controversy_threshold = 0.4,  # Lower threshold to discuss more clusters
  entropy_threshold = 0.8  # Lower entropy threshold
)
```

#### 2. Include batch information in the tissue context

```{r, purl=FALSE}
# Include batch information in the tissue context
batch_aware_results <- annotate_cell_types(
  input = marker_data,  # Your marker gene data with batch effects
  tissue_name = "mouse brain with technical batch effects",  # Include batch context
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

## Incorporating Domain Knowledge

### Using Tissue Context

One of the key features of mLLMCelltype is the ability to incorporate domain knowledge through the `tissue_name` parameter. This provides important context to the LLM:

```{r, purl=FALSE}
# Basic annotation without specific tissue context
basic_results <- annotate_cell_types(
  input = marker_data,
  tissue_name = "human sample",  # Generic context
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

# Annotation with specific tissue context
specific_results <- annotate_cell_types(
  input = marker_data,
  tissue_name = "human fetal liver at 20 weeks gestation",  # Detailed context
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

### Creating Custom Prompts

For advanced use cases, you can create and modify the annotation prompt directly:

```{r, purl=FALSE}
# Create a custom annotation prompt
custom_prompt <- create_annotation_prompt(
  input = marker_data,
  tissue_name = "human PBMC",
  top_gene_count = 10
)

# Modify the prompt to include additional context
modified_prompt <- paste0(
  custom_prompt$prompt,
  "\n\nAdditional context: This sample is from a patient with rheumatoid arthritis. ",
  "Previous studies have identified activated T cells, B cells, and CXCR4-high monocytes in this condition."
)

# Use the modified prompt directly
custom_results <- get_model_response(
  prompt = modified_prompt,
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

### Combining with External Resources

You can enhance your annotation workflow by combining mLLMCelltype with other R packages and resources:

```{r, purl=FALSE}
library(Seurat)
library(dplyr)

# Example: Using CellMarker database information to validate annotations
# This is a conceptual example - implementation would depend on your specific needs

# 1. Get annotations with mLLMCelltype
annotations <- annotate_cell_types(
  input = marker_data,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

# 2. Compare with known marker genes (conceptual)
# In a real workflow, you would query a database or use a reference dataset
known_markers <- list(
  "T cells" = c("CD3D", "CD3E", "CD3G"),
  "B cells" = c("CD19", "MS4A1", "CD79A"),
  "Monocytes" = c("CD14", "LYZ", "CSF1R")
)

# 3. Validate annotations against known markers
# This is a simplified example of how you might validate annotations
validate_annotations <- function(annotations, marker_data, known_markers) {
  validation_results <- list()

  for (i in 1:length(annotations)) {
    cluster_id <- i
    predicted_type <- annotations[i]

    # Get markers for this cluster
    cluster_markers <- marker_data %>%
      filter(cluster == cluster_id) %>%
      arrange(desc(avg_log2FC)) %>%
      pull(gene) %>%
      head(20)

    # Check overlap with known markers for this cell type
    if (predicted_type %in% names(known_markers)) {
      expected_markers <- known_markers[[predicted_type]]
      overlap <- intersect(cluster_markers, expected_markers)

      validation_results[[i]] <- list(
        cluster = cluster_id,
        predicted_type = predicted_type,
        overlap_count = length(overlap),
        overlap_genes = paste(overlap, collapse = ", "),
        confidence = length(overlap) / length(expected_markers)
      )
    } else {
      validation_results[[i]] <- list(
        cluster = cluster_id,
        predicted_type = predicted_type,
        overlap_count = 0,
        overlap_genes = "",
        confidence = 0
      )
    }
  }

  return(validation_results)
}

# This is a conceptual example of how you might validate annotations
# validation_results <- validate_annotations(annotations, marker_data, known_markers)
```

## Practical Case Studies

### Case Study 1: PBMC Dataset Analysis

This example demonstrates a complete workflow for analyzing a PBMC dataset:

```{r, purl=FALSE}
library(Seurat)
library(mLLMCelltype)
library(ggplot2)
library(dplyr)

# Load example PBMC data
# In a real workflow, you would use your own data
data("pbmc_small")  # Example dataset from Seurat

# Find marker genes
pbmc_markers <- FindAllMarkers(pbmc_small,
                              only.pos = TRUE,
                              min.pct = 0.25,
                              logfc.threshold = 0.25)

# Set up API keys
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  openai = Sys.getenv("OPENAI_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY")
)

# Use consensus annotation
consensus_results <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
  api_keys = api_keys,
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6"
)

# Add results to Seurat object
pbmc_small$cell_type <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = names(consensus_results$final_annotations),
  to = consensus_results$final_annotations
)

# Visualize results
# In a real workflow, you would create a UMAP or t-SNE plot
# DimPlot(pbmc_small, group.by = "cell_type", label = TRUE) +
#   ggtitle("PBMC Cell Types")
```

### Case Study 2: Identifying Rare Cell Types

When working with datasets containing rare cell populations, you can adjust parameters to improve detection:

```{r, purl=FALSE}
# For rare cell types, use these strategies:

# 1. Increase the number of marker genes considered
rare_cell_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human bone marrow",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 20  # Use more genes for rare cell types
)

# 2. Use consensus with lower thresholds to discuss more clusters
rare_cell_consensus <- interactive_consensus_annotation(
  input = marker_data,  # Your marker gene data
  tissue_name = "human bone marrow",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),
  api_keys = api_keys,
  controversy_threshold = 0.4,  # Lower threshold to discuss more clusters
  entropy_threshold = 0.8,  # Lower entropy threshold
  consensus_check_model = "claude-sonnet-4-6"
)

# 3. Provide more specific tissue context
specific_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human bone marrow with expected rare plasma cells and basophils",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

### Case Study 3: Cross-Species Comparison

mLLMCelltype can be used to compare cell types across different species:

```{r, purl=FALSE}
# Example workflow for cross-species comparison

# 1. Annotate human and mouse datasets separately
# (Assuming you have marker data for both species)
human_annotations <- annotate_cell_types(
  input = human_marker_data,  # Your human marker data
  tissue_name = "human brain cortex",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

mouse_annotations <- annotate_cell_types(
  input = mouse_marker_data,  # Your mouse marker data
  tissue_name = "mouse brain cortex",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

# 2. Compare annotations
# This is a conceptual example - in a real workflow, you would:
# - Map annotations to Seurat objects
# - Calculate proportions
# - Create comparison visualizations
# - Identify conserved and species-specific cell types

# Example comparison function (conceptual)
compare_species_annotations <- function(human_annotations, mouse_annotations) {
  # Get unique cell types from both species
  human_types <- unique(human_annotations)
  mouse_types <- unique(mouse_annotations)

  # Find common cell types
  common_types <- intersect(human_types, mouse_types)

  # Find species-specific cell types
  human_specific <- setdiff(human_types, mouse_types)
  mouse_specific <- setdiff(mouse_types, human_types)

  # Return comparison results
  list(
    common_types = common_types,
    human_specific = human_specific,
    mouse_specific = mouse_specific
  )
}

# This is a conceptual example
# comparison <- compare_species_annotations(human_annotations, mouse_annotations)
```

## Performance Considerations

### API Cost Management

When using mLLMCelltype, it's important to consider the costs associated with API calls to different LLM providers:

```{r, purl=FALSE}
# Example of cost-efficient model selection
# Choose models based on your specific needs and budget

# For initial exploration or smaller datasets
# Use more affordable models
affordable_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  model = "claude-haiku-4-5-20251001",  # More affordable model
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

# For final analysis or challenging datasets
# Use larger models
premium_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",  # Larger model
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)

# Use OpenRouter for access to free models
openrouter_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  model = "meta-llama/llama-3.3-70b-instruct:free",  # Free model via OpenRouter
  api_key = Sys.getenv("OPENROUTER_API_KEY")
)
```

### Optimizing Runtime

To optimize runtime when working with large datasets:

```{r, purl=FALSE}
# 1. Use caching with interactive_consensus_annotation
consensus_with_cache <- interactive_consensus_annotation(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  models = c("claude-sonnet-4-6", "gpt-5.5"),
  api_keys = api_keys,
  use_cache = TRUE,  # Enable caching
  cache_dir = NULL  # Uses default system cache directory
)

# 2. Process clusters in batches
# This is a conceptual example - implementation would depend on your workflow
process_in_batches <- function(marker_data, batch_size = 5) {
  # Get unique clusters
  clusters <- unique(marker_data$cluster)

  # Process in batches
  results <- list()
  for (i in seq(1, length(clusters), by = batch_size)) {
    # Get current batch of clusters
    batch_clusters <- clusters[i:min(i + batch_size - 1, length(clusters))]

    # Filter marker data for current batch
    batch_data <- marker_data %>% filter(cluster %in% batch_clusters)

    # Process batch
    batch_results <- annotate_cell_types(
      input = batch_data,
      tissue_name = "human PBMC",
      model = "claude-sonnet-4-6",
      api_key = Sys.getenv("ANTHROPIC_API_KEY")
    )

    # Store results
    results <- c(results, batch_results)
  }

  return(results)
}

# 3. Use faster models for initial exploration
fast_annotation <- annotate_cell_types(
  input = marker_data,  # Your marker gene data
  tissue_name = "human PBMC",
  model = "claude-haiku-4-5-20251001",  # Faster model
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

## Advanced Customization

### Custom Processing Functions

For advanced users, mLLMCelltype allows you to register custom providers and models:

```{r, purl=FALSE}
# Define a custom processing function
# This function must accept prompt, model, and api_key parameters
custom_process_fn <- function(prompt, model, api_key) {
  # Custom implementation to process prompts and get responses
  # This is a simplified example
  cat("Processing prompt with custom provider\n")
  cat("Model:", model, "\n")

  # In a real implementation, you would make API calls here
  # For example:
  # response <- httr::POST(
  #   url = "https://api.custom-provider.com/v1/chat/completions",
  #   body = list(prompt = prompt, model = model),
  #   httr::add_headers(Authorization = paste("Bearer", api_key)),
  #   encode = "json"
  # )
  # result <- httr::content(response)$choices[[1]]$text

  # For this example, just return a fixed response
  result <- "T cells"
  return(result)
}

# Register the custom provider
register_custom_provider(
  provider_name = "custom_provider",
  process_fn = custom_process_fn,
  description = "My custom LLM provider"
)

# Register a custom model
register_custom_model(
  model_name = "custom-model",
  provider_name = "custom_provider",
  model_config = list(
    temperature = 0.7,
    max_tokens = 2000
  )
)

# Use the custom model
# custom_results <- annotate_cell_types(
#   input = marker_data,
#   tissue_name = "human PBMC",
#   model = "custom-model",
#   api_key = "your-custom-api-key"
# )
```

### Using the Unified Logging System

mLLMCelltype provides a comprehensive unified logging system with structured output, performance monitoring, and multi-level logging:

```{r, purl=FALSE}
# Configure the global logger (recommended approach)
configure_logger(level = "INFO", console_output = TRUE, json_format = TRUE)

# Use simple logging functions
log_info("Starting analysis of cluster 0", list(
  cluster_id = "0",
  tissue_name = "human PBMC",
  marker_genes = c("CD3D", "CD3E", "CD2", "IL7R", "LTB")
))

# Log API calls with performance tracking
log_info("API call completed", list(
  provider = "anthropic",
  model = "claude-sonnet-4-6",
  duration_seconds = 2.34,
  success = TRUE
))

# Log warnings and errors
log_warn("Model response had unusual format", list(
  model = "gpt-5.5",
  response_length = 50
))

log_error("API call failed", list(
  provider = "openai",
  error = "Rate limit exceeded"
))

# Alternatively, create a custom logger instance
custom_logger <- UnifiedLogger$new(
  base_dir = "custom_logs",
  level = "DEBUG",
  console_output = TRUE,
  json_format = TRUE
)

# Use the custom logger
custom_logger$info("Custom log message", list(analysis_step = "preprocessing"))
custom_logger$debug("Detailed debugging info", list(variable_state = "initialized"))

# Get performance summary
performance <- get_logger()$get_performance_summary()
print(performance)
```

### Using the CacheManager

The `CacheManager` class helps optimize performance by caching consensus results:

```{r, purl=FALSE}
# Create a cache manager
cache_manager <- CacheManager$new(cache_dir = NULL)

# Generate a cache key
cache_key <- cache_manager$generate_key(
  input = marker_data,
  models = c("claude-sonnet-4-6", "gpt-5.5"),
  cluster_id = "0"
)

# Check if results exist in cache
if (cache_manager$has_cache(cache_key)) {
  # Load from cache
  cached_results <- cache_manager$load_from_cache(cache_key)
} else {
  # Process and save to cache
  # results <- process_cluster(...)
  # cache_manager$save_to_cache(cache_key, results)
}

# Get cache statistics
cache_stats <- cache_manager$get_cache_stats()

# Clear cache (with confirmation)
# cache_manager$clear_cache(confirm = TRUE)
```

### Cache Management

mLLMCelltype provides convenient functions for managing cache directories:

```{r, purl=FALSE}
# Check cache location
mllmcelltype_cache_dir()

# Use local cache
mllmcelltype_cache_dir("local")

# Clear cache
mllmcelltype_clear_cache()
```

## Next Steps

Now that you've explored the advanced features of mLLMCelltype, you can:

- [Contribute to the project](https://cafferyang.com/mLLMCelltype/articles/contributing-guide.html): Learn how to contribute to mLLMCelltype
- [Review the version history](https://cafferyang.com/mLLMCelltype/news/index.html): Explore the development history of the package
- [Return to the introduction](https://cafferyang.com/mLLMCelltype/articles/introduction.html): Review the basic concepts

---
title: "Contributing Guide"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Contributing Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

# Contributing to mLLMCelltype

Thank you for your interest in contributing to mLLMCelltype! This guide will help you understand how to contribute to the project effectively.

## Getting Started

### Fork and Clone the Repository

1. Fork the [mLLMCelltype repository](https://github.com/cafferychen777/mLLMCelltype) on GitHub
2. Clone your fork to your local machine:

```bash
git clone https://github.com/YOUR-USERNAME/mLLMCelltype.git
cd mLLMCelltype
```

3. Add the original repository as an upstream remote:

```bash
git remote add upstream https://github.com/cafferychen777/mLLMCelltype.git
```

### Setting Up the Development Environment

For R package development:

```r
# Install required packages for development
install.packages(c("devtools", "roxygen2", "testthat", "knitr", "rmarkdown"))

# Install the package in development mode
devtools::install_dev("R")
```

## Project Structure

The mLLMCelltype project has the following structure:

```
mLLMCelltype/
├── R/                  # R package source code
│   ├── R/              # R functions
│   ├── man/            # Documentation
│   ├── tests/          # Unit tests
│   ├── vignettes/      # Package vignettes
│   └── DESCRIPTION     # Package metadata
├── python/             # Python package source code
├── .github/            # GitHub workflows and templates
├── assets/             # Images and other assets
├── examples/           # Example notebooks and scripts
└── README.md           # Project overview
```

## Development Workflow

### Creating a New Feature

1. Create a new branch for your feature:

```bash
git checkout -b feature/your-feature-name
```

2. Make your changes to the codebase
3. Add and commit your changes:

```bash
git add .
git commit -m "Add your descriptive commit message here"
```

4. Push your changes to your fork:

```bash
git push origin feature/your-feature-name
```

5. Create a pull request from your fork to the main repository

### Code Style Guidelines

#### R Code Style

We follow the [tidyverse style guide](https://style.tidyverse.org/) for R code:

- Use snake_case for variable and function names
- Use spaces around operators and after commas
- Use 2 spaces for indentation
- Limit line length to 80 characters
- Use roxygen2 for documentation

Example of properly formatted R code:

```r
#' Annotate Cell Types
#'
#' This function annotates cell types based on marker genes.
#'
#' @param input A data frame containing marker genes.
#' @param tissue_name The name of the tissue.
#' @param model The LLM model to use.
#' @param api_key The API key for the LLM provider.
#'
#' @return A vector of cell type annotations.
#' @export
annotate_cell_types <- function(input, tissue_name, model, api_key) {
  # Function implementation
  results <- process_markers(input, top_n = 10)

  for (i in seq_along(results)) {
    if (is_valid_result(results[i])) {
      results[i] <- clean_result(results[i])
    }
  }

  return(results)
}
```

#### Documentation Guidelines

All functions should be documented using roxygen2 with the following sections:

- Title (first line)
- Description (paragraph after title)
- @param for each parameter
- @return for the return value
- @examples for usage examples
- @export if the function should be exported

### Testing

We use the testthat package for testing. Tests should be placed in the `R/tests/testthat/` directory.

To run tests:

```r
devtools::test()
```

Example test file (`test-annotate_cell_types.R`):

```r
context("Cell type annotation")

test_that("annotate_cell_types returns expected format", {
  # Setup test data
  test_markers <- data.frame(
    cluster = c(0, 0, 1, 1),
    gene = c("CD3D", "CD3E", "CD19", "MS4A1"),
    avg_log2FC = c(2.5, 2.3, 3.1, 2.8),
    p_val_adj = c(0.001, 0.002, 0.001, 0.003)
  )

  # Mock the API response
  mockery::stub(
    annotate_cell_types,
    "get_model_response",
    function(...) c("T cells", "B cells")
  )

  # Run the function
  result <- annotate_cell_types(
    input = test_markers,
    tissue_name = "test tissue",
    model = "test-model",
    api_key = "test-key"
  )

  # Assertions
  expect_is(result, "character")
  expect_length(result, 2)
  expect_equal(result, c("T cells", "B cells"))
})
```

## Contributing Areas

### Adding Support for New LLM Models

To add support for a new LLM model:

1. Identify the model provider and API endpoint
2. Create a new processing function in `R/R/process_[provider].R`
3. Update the `get_provider()` function in `R/R/get_provider.R`
4. Add the model to the supported models list
5. Create tests for the new model
6. Update documentation

Example of adding a new model:

```r
# In process_newprovider.R
process_newprovider <- function(prompt, api_key) {
  # Implementation for the new provider
  url <- "https://api.newprovider.com/v1/completions"

  headers <- c(
    "Content-Type" = "application/json",
    "Authorization" = paste("Bearer", api_key)
  )

  body <- list(
    model = "newprovider-model",
    prompt = prompt,
    max_tokens = 1000,
    temperature = 0.1
  )

  # Make API request using httr
  response <- httr::POST(
    url = url,
    httr::add_headers(.headers = headers),
    body = jsonlite::toJSON(body, auto_unbox = TRUE),
    encode = "json"
  )

  # Check for HTTP errors
  httr::stop_for_status(response)

  # Parse the response
  content <- httr::content(response, "text", encoding = "UTF-8")
  parsed_response <- jsonlite::fromJSON(content)
  result <- parsed_response$choices[[1]]$text

  return(result)
}

# In get_provider.R
get_provider <- function(model) {
  # Add to the model mapping
  model_mapping <- list(
    # Existing models...
    "newprovider-model" = "newprovider"
  )

  provider <- model_mapping[[model]]
  if (is.null(provider)) {
    stop("Unsupported model: ", model)
  }

  return(provider)
}
```

### Improving Documentation

Documentation improvements are always welcome:

1. Update function documentation with roxygen2
2. Improve vignettes with more examples and explanations
3. Add tutorials for specific use cases
4. Fix typos and clarify existing documentation

### Adding New Features

Some ideas for new features:

1. Integration with additional single-cell analysis frameworks
2. Support for spatial transcriptomics data
3. Interactive visualization tools
4. Batch processing for large datasets
5. Performance optimizations

### Reporting Issues

When reporting issues, please include:

1. A minimal reproducible example
2. The version of mLLMCelltype you're using
3. The error message or unexpected behavior
4. Your R session information (`sessionInfo()`)

## Pull Request Process

1. Ensure your code follows the style guidelines
2. Add or update tests as necessary
3. Update documentation to reflect your changes
4. Ensure all tests pass
5. Submit your pull request with a clear description of the changes

## Code Review Process

All pull requests will be reviewed by the maintainers. The review process includes:

1. Checking that the code follows style guidelines
2. Verifying that tests pass
3. Ensuring documentation is updated
4. Evaluating the overall design and implementation

## Release Process

mLLMCelltype follows semantic versioning (MAJOR.MINOR.PATCH):

- MAJOR version for incompatible API changes
- MINOR version for new functionality in a backward-compatible manner
- PATCH version for backward-compatible bug fixes

## Community Guidelines

### Code of Conduct

We follow a code of conduct to ensure a welcoming and inclusive community:

- Be respectful and inclusive
- Be collaborative
- Be open to feedback
- Focus on the best solution for the community

### Communication Channels

- GitHub Issues: For bug reports, feature requests, and discussions
- GitHub Discussions: For general questions and community discussions
- Pull Requests: For code contributions

## Acknowledgment

Contributors will be acknowledged in the package documentation and README.

## License

By contributing to mLLMCelltype, you agree that your contributions will be licensed under the same license as the project (MIT License).

## Next Steps

Now that you know how to contribute to mLLMCelltype, you can:

- [Review the version history](https://cafferyang.com/mLLMCelltype/articles/version-history.html) to understand recent changes
- [Explore advanced features](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html) to identify areas for improvement
- [Check the FAQ](https://cafferyang.com/mLLMCelltype/articles/faq.html) to see common questions that might need better documentation

---
title: "Frequently Asked Questions"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Frequently Asked Questions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

# Frequently Asked Questions

This document addresses common questions about using mLLMCelltype for cell type annotation in single-cell RNA sequencing data.

## General Questions

### What makes mLLMCelltype different from other cell type annotation tools?

mLLMCelltype differs from traditional cell type annotation tools in several key ways:

1. **No reference dataset required**: Unlike reference-based methods, mLLMCelltype doesn't require a pre-existing reference dataset.

2. **Multi-model consensus**: mLLMCelltype leverages multiple large language models to achieve more reliable annotations than any single model could provide.

3. **Transparent reasoning**: The package provides complete reasoning chains for annotations, making the process interpretable and transparent.

4. **Uncertainty quantification**: mLLMCelltype provides explicit uncertainty metrics (consensus proportion and Shannon entropy) to identify ambiguous cell populations.

5. **Structured deliberation**: For controversial clusters, mLLMCelltype initiates a structured discussion process among models to reach a more reliable consensus.

### Which tissues and species does mLLMCelltype support?

mLLMCelltype can annotate cell types from virtually any tissue and species, as it relies on the biological knowledge embedded in large language models rather than pre-defined reference datasets. However, performance may vary depending on how well-characterized the tissue is in the scientific literature.

The package has been extensively tested on:
- Human tissues (PBMC, bone marrow, brain, lung, liver, kidney, etc.)
- Mouse tissues (brain, lung, kidney, etc.)
- Other model organisms (zebrafish, fruit fly, etc.)

For very specialized or poorly characterized tissues, the uncertainty metrics will help identify clusters that may require expert review.

### How accurate is mLLMCelltype compared to other methods?

In our benchmarks (Yang et al., 2025; see our [paper](https://doi.org/10.1101/2025.04.10.647852)), the consensus approach showed improvements over both traditional annotation methods and single-LLM approaches:

- Compared to reference-based methods (e.g., SingleR, Seurat label transfer), mLLMCelltype showed comparable or better performance without requiring a reference dataset.
- Compared to marker-based methods (e.g., SCINA, CellAssign), mLLMCelltype demonstrated higher accuracy and flexibility.
- Compared to single-LLM approaches, the consensus mechanism showed accuracy improvements on tested datasets.

The accuracy advantage is particularly pronounced for rare cell types and tissues with limited reference data.

## Technical Questions

### How does mLLMCelltype handle cluster indices?

mLLMCelltype preserves your original cluster IDs as-is. Whether your clusters are numbered 0, 1, 2 (Seurat default) or 1, 2, 3 (R convention), or use custom names like "t_cells", the package will use them directly without modification. The returned annotations use the same cluster IDs as your input.

### What is the recommended number of marker genes per cluster?

The default setting uses the top 10 marker genes per cluster, which works well for most scenarios. However, you can adjust this using the `top_gene_count` parameter:

- For well-characterized cell types: 5-10 marker genes is usually sufficient
- For rare or poorly characterized cell types: 10-20 marker genes may be beneficial
- For noisy data: Fewer genes (5-7) might give better results by focusing on the strongest signals

The optimal number depends on the quality of your marker genes and the complexity of the tissue. We recommend starting with the default of 10 and adjusting based on the results.

### How does caching work in mLLMCelltype?

mLLMCelltype implements a caching system to avoid redundant API calls, which saves time and reduces costs:

- `interactive_consensus_annotation()` uses caching by default (`use_cache = TRUE`)
- `annotate_cell_types()` supports direct caching through `use_cache` and `cache_dir`
- The cache is based on a hash of the input data, model, cluster ID, tissue context, and marker-gene count
- You can specify a custom cache directory using the `cache_dir` parameter

To clear the cache:

```r
cache_manager <- CacheManager$new()
cache_manager$clear_cache()
```

For one-off prompt generation, set `api_key = NA`; no cache entry is created because no provider call is made.

### How does mLLMCelltype handle rate limits and API errors?

The package includes error handling for API calls:

- Detailed error messages: When an API call fails, the error message includes details to help diagnose the issue

If you're processing many clusters, you might encounter rate limits. In this case:

1. Reduce the number of models used for initial annotation
2. Process batches of clusters separately with pauses between batches
3. Consider implementing your own retry mechanism if needed

## Performance and Optimization

### How long does it take to run mLLMCelltype?

The runtime depends on several factors:

- **Number of clusters**: Each cluster requires separate API calls
- **Number of models**: More models means more API calls
- **Discussion process**: Controversial clusters require additional API calls for discussion
- **API response times**: Different providers have different response times
- **Network conditions**: Internet speed and reliability affect performance

Typical runtimes:
- Single model, 10 clusters: 1-2 minutes
- Multi-model consensus (3 models), 10 clusters: 3-5 minutes
- Multi-model consensus with discussion, 10 clusters: 5-10 minutes

To optimize runtime:
- Keep `use_cache = TRUE` and set a persistent `cache_dir` for repeated runs
- Start with fewer models for initial exploration
- Use a higher `controversy_threshold` to reduce the number of controversial clusters
- Process large datasets in batches

### What are the API costs associated with using mLLMCelltype?

The API costs depend on the models you use and the number of clusters:

- **OpenAI models** (GPT-5.5, etc.): $0.01-0.05 per cluster for annotation
- **Anthropic models** (Claude Opus 4.7, Sonnet 4.6, etc.): $0.01-0.03 per cluster for annotation
- **Google models** (Gemini 3.1, etc.): $0.001-0.01 per cluster for annotation
- **Other models**: Generally lower cost
- **OpenRouter free models**: $0.00 (free with `:free` suffix)

For a typical dataset with 10-20 clusters:
- Single model annotation: $0.10-1.00 total
- Multi-model consensus (3 models): $0.30-3.00 total
- With discussion process: Additional $0.10-1.00
- Using OpenRouter free models: $0.00 total

To reduce costs:
- Keep caching enabled to avoid redundant API calls
- Start with more economical models
- Use fewer models for initial exploration
- Reserve multi-model consensus for final analysis
- Consider using OpenRouter free models (see below)

### How can I use OpenRouter free models?

OpenRouter provides access to several high-quality models for free:

1. **Sign up for an OpenRouter account** at [openrouter.ai](https://openrouter.ai/keys)

2. **Get your API key** from the OpenRouter dashboard

3. **Use models with the `:free` suffix**:
```r
# Set your OpenRouter API key
Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key")

# Use a free model
results <- annotate_cell_types(
  input = marker_data,
  tissue_name = "human PBMC",
  model = "meta-llama/llama-4-maverick:free",  # Note the :free suffix
  api_key = Sys.getenv("OPENROUTER_API_KEY")
  # No need to specify provider - it's automatically detected from the model name format
)
```

4. **Recommended free models** (Updated Oct 2025):
   - `meta-llama/llama-4-maverick:free` - Meta Llama 4 Maverick (256K context, best performance)
   - `deepseek/deepseek-v4-pro:free` - DeepSeek V4 Pro
   - `meta-llama/llama-3.3-70b-instruct:free` - Meta Llama 3.3 70B (reliable)
   - `venice/uncensored:free` - Venice Uncensored (new model)
   - `minimax/minimax-m2:free` - MiniMax M2 (optimized for coding)
   - `z-ai/glm-4.5-air:free` - GLM 4.5 Air (lightweight)

**Important Notes about Free Models** (Updated Oct 2025):
- **Daily limits reduced**: 50 requests/day for free accounts (down from 200)
- **Accounts with $10+ credits**: 1000 requests/day
- **Rate limit**: 20 requests/minute for all accounts
- **Some models removed**: NVIDIA Nemotron and others have exited the free tier
- **Availability may change**: Verify current free models at https://openrouter.ai/models?q=free
- **For production use**: Consider using paid models for better reliability

### How can I improve the accuracy of annotations?

To get the most accurate annotations:

1. **Use multiple high-quality models**: Include diverse, high-performing models like Claude Opus 4.7, GPT-5.5, and Gemini 3.1

2. **Provide good marker genes**: Use robust differential expression analysis to identify strong marker genes

3. **Specify the correct tissue**: Always provide the correct tissue name to give models the proper context

4. **Review uncertainty metrics**: Pay attention to consensus proportion and Shannon entropy to identify clusters that may need manual review

5. **Examine discussion logs**: For controversial clusters, review the discussion logs to understand the reasoning

6. **Iterate if needed**: If results are unsatisfactory, try adjusting parameters or providing additional context

## Troubleshooting

### Why am I getting different results with the same input?

There are several possible reasons for getting different results with the same input:

1. **Model updates**: LLMs are regularly updated, which can change their outputs

2. **Temperature/sampling**: Some randomness is inherent in LLM outputs

3. **Context window limitations**: Different runs might include slightly different context

4. **API changes**: Providers may change how their APIs work

To ensure reproducibility:
- Keep `use_cache = TRUE` and set a persistent `cache_dir` to reuse results
- Specify model versions explicitly when available
- Save and document your results
- Consider saving the raw API responses for future reference

### I'm getting an error about invalid cluster indices. What should I do?

If you see an error about invalid cluster indices, check that your cluster column contains valid values. mLLMCelltype accepts any cluster IDs (numeric or character) and preserves them as-is. Common issues:

1. Ensure the `cluster` column exists in your data frame
2. Check for `NA` values in the cluster column
3. If using non-standard column names, rename to `cluster` before calling the function

### How do I handle "API key not found" errors?

If you get an error about missing API keys:

1. **Check environment variables**: Ensure your API keys are set correctly in your `.env` file or environment

2. **Provide keys directly**: Pass the API key directly to the function:
```r
results <- annotate_cell_types(..., api_key = "your-api-key")
```

3. **Check provider name**: Make sure you're using the correct provider name for your API key:
```r
# Set API key for a specific provider
Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-key")
```

4. **Verify key validity**: Check if your API key is still valid by testing it directly with the provider's API

### Why are some cell types not being correctly identified?

If specific cell types are not being correctly identified:

1. **Check marker genes**: Ensure the marker genes for these cell types are strong and specific

2. **Provide more context**: Specify the tissue type accurately to give models the right context

3. **Use more models**: Different models have different strengths; using multiple models improves coverage

4. **Increase marker count**: Try increasing `top_gene_count` to include more marker genes

5. **Review discussion logs**: For controversial clusters, examine the discussion to understand the reasoning

6. **Consider rare cell types**: Some cell types may be poorly represented in the training data of LLMs

## Integration with Other Tools

### How does mLLMCelltype integrate with Seurat?

mLLMCelltype integrates with Seurat:

1. **Input**: You can directly use Seurat's `FindAllMarkers()` output as input

2. **Output**: Annotation results can be easily added to your Seurat object:
```r
seurat_obj$cell_type_consensus <- plyr::mapvalues(
  x = as.character(Idents(seurat_obj)),
  from = names(consensus_results$final_annotations),
  to = consensus_results$final_annotations
)
```

3. **Visualization**: Use Seurat's visualization functions with the added annotations:
```r
DimPlot(seurat_obj, group.by = "cell_type_consensus", label = TRUE)
```

### Can I use mLLMCelltype with Scanpy/AnnData in R?

Yes, you can use mLLMCelltype with Scanpy/AnnData objects in R:

1. **Extract marker genes**: Export marker genes from your Scanpy analysis to a CSV file

2. **Run mLLMCelltype**: Use the CSV file as input to mLLMCelltype

3. **Import results**: Add the annotation results back to your AnnData object

Alternatively, you can use the Python version of mLLMCelltype for direct integration with Scanpy.

### How can I combine mLLMCelltype with traditional annotation methods?

mLLMCelltype can be used alongside traditional annotation methods:

1. **Complementary approach**: Use both methods and compare results

2. **Validation**: Use mLLMCelltype to validate annotations from reference-based methods

3. **Hybrid approach**: Use reference-based methods for well-characterized cell types and mLLMCelltype for novel or rare cell types

4. **Ensemble method**: Create a consensus between mLLMCelltype and traditional methods

## Advanced Usage

### How can I customize the prompts used by mLLMCelltype?

While mLLMCelltype uses carefully designed prompts, advanced users can customize them:

```r
# Create a custom annotation prompt
custom_prompt <- create_annotation_prompt(
  input = your_markers,
  tissue_name = "your_tissue",
  top_gene_count = 10
)

# Add custom context before sending the prompt directly
prompt_with_context <- paste0(
  custom_prompt$prompt,
  "\n\nAdditional context: Also consider developmental stage and activation state."
)

response <- get_model_response(
  prompt = prompt_with_context,
  model = "claude-sonnet-4-6",
  api_key = your_api_key
)
```

### Can I add my own custom LLM models?

Yes, you can register custom models and providers:

```r
# Register a custom provider
register_custom_provider(
  provider_name = "my_provider",
  process_fn = function(prompt, model, api_key) {
    # Custom implementation that returns the model response text
  },
  description = "My custom LLM provider"
)

# Register a custom model
register_custom_model(
  model_name = "my-custom-model",
  provider_name = "my_provider",
  model_config = list()
)

# Use the custom model
results <- annotate_cell_types(
  input = your_markers,
  tissue_name = "your_tissue",
  model = "my-custom-model",
  api_key = your_api_key
)
```

### How can I contribute to mLLMCelltype?

We welcome contributions! Here are some ways to contribute:

1. **Report issues**: Report bugs or suggest features on our [GitHub repository](https://github.com/cafferychen777/mLLMCelltype/issues)

2. **Improve documentation**: Help us improve documentation and examples

3. **Add new models**: Implement support for new LLM models

4. **Share benchmarks**: Share your benchmarking results with different tissues and species

5. **Develop new features**: Contribute code for new features or improvements

See our [Contributing Guide](https://cafferyang.com/mLLMCelltype/articles/contributing-guide.html) for more details.

## Next Steps

Now that you have answers to common questions, you can explore:

- [Advanced Features](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html): Learn about hierarchical annotation and other advanced features
- [Contributing Guide](https://cafferyang.com/mLLMCelltype/articles/contributing-guide.html): Find out how to contribute to the project
- [Version History](https://cafferyang.com/mLLMCelltype/articles/version-history.html): Review the development history of mLLMCelltype

---
title: "Getting Started with mLLMCelltype"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with mLLMCelltype}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

# Getting Started with mLLMCelltype

This guide provides a quick introduction to using mLLMCelltype for cell type annotation in single-cell RNA sequencing data. We'll cover the basic workflow, input data requirements, and a simple example to get you started.

## Basic Workflow

The mLLMCelltype workflow consists of these main steps:

1. **Prepare marker gene data** for each cluster
2. **Run annotation** using one or multiple LLMs
3. **Create consensus** from multiple model predictions (optional)
4. **Integrate results** with your Seurat or Scanpy object
5. **Visualize results** with uncertainty metrics

## Loading the Package and Setting Up API Keys

First, load the mLLMCelltype package:

```{r}
library(mLLMCelltype)
```

### Setting Up API Keys

Before using mLLMCelltype, you need to set up API keys for the LLM providers you plan to use:

```{r}
# Set API keys as environment variables
Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-api-key")  # For Claude models
Sys.setenv(OPENAI_API_KEY = "your-openai-api-key")        # For GPT models
Sys.setenv(GEMINI_API_KEY = "your-gemini-api-key")        # For Gemini models
Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key") # For OpenRouter models
```

You can obtain API keys from:
- Anthropic: https://console.anthropic.com/
- OpenAI: https://platform.openai.com/
- Google (Gemini): https://ai.google.dev/
- OpenRouter: https://openrouter.ai/keys

Alternatively, you can provide API keys directly in function calls:

```{r}
results <- annotate_cell_types(
  input = markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = "your-anthropic-api-key",  # Direct API key
  top_gene_count = 10
)
```

## Input Data Requirements

mLLMCelltype accepts marker gene data in several formats:

### 1. Data Frame Format

A data frame with the following columns:
- `cluster`: Cluster ID (preserved as-is from your data)
- `gene`: Gene name/symbol
- `avg_log2FC` or similar metric: Log fold change
- `p_val_adj` or similar metric: Adjusted p-value

Example:

```{r}
# Example marker data frame
markers_df <- data.frame(
  cluster = c(0, 0, 0, 1, 1, 1),
  gene = c("CD3D", "CD3E", "CD2", "CD14", "LYZ", "CST3"),
  avg_log2FC = c(2.5, 2.3, 2.1, 3.1, 2.8, 2.5),
  p_val_adj = c(0.001, 0.001, 0.002, 0.0001, 0.0002, 0.0005)
)
```

### 2. Seurat FindMarkers Output

You can directly use the output from Seurat's `FindAllMarkers()` function:

```{r}
# Assuming you have a Seurat object named 'seurat_obj'
library(Seurat)
all_markers <- FindAllMarkers(seurat_obj, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)
```

### 3. CSV File Path

A path to a CSV file containing marker gene data:

```{r}
# Path to your CSV file
markers_file <- "path/to/markers.csv"
```

### 4. List Format

A list where each element contains marker genes for a cluster:

```{r}
# Example marker list
markers_list <- list(
  "0" = c("CD3D", "CD3E", "CD2", "IL7R", "LTB"),
  "1" = c("CD14", "LYZ", "CST3", "MS4A7", "FCGR3A")
)
```

## Function Parameters

The `annotate_cell_types` function has the following parameters:

| Parameter | Description | Default Value |
|-----------|-------------|---------------|
| `input` | Marker gene data (data frame, list, or file path) | (required) |
| `tissue_name` | Tissue name (e.g., "human PBMC", "mouse brain") | `NULL` |
| `model` | LLM model to use | `"gpt-5.5"` |
| `api_key` | API key (if not set in environment) | `NA` |
| `top_gene_count` | Number of top genes per cluster to use | `10` |
| `debug` | Whether to print debugging information | `FALSE` |

Note: If `api_key` is set to `NA`, the function will return the generated prompt without making an API call, which is useful for reviewing the prompt before sending it to the API.

## Basic Usage Example

Here's a simple example using a single LLM model for annotation:

```{r}
# Example marker data
markers <- data.frame(
  cluster = c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1),
  gene = c("CD3D", "CD3E", "CD2", "IL7R", "LTB", "CD14", "LYZ", "CST3", "MS4A7", "FCGR3A"),
  avg_log2FC = c(2.5, 2.3, 2.1, 1.8, 1.7, 3.1, 2.8, 2.5, 2.2, 2.0),
  p_val_adj = c(0.001, 0.001, 0.002, 0.003, 0.005, 0.0001, 0.0002, 0.0005, 0.001, 0.002)
)

# Run annotation with a single model
results <- annotate_cell_types(
  input = markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10,
  debug = FALSE  # Set to TRUE for more detailed output
)

# Print results
print(results)
```

### Example Output

When using a single model like Claude, the output will be a character vector with one annotation per cluster:

```r
> print(results)
[1] "0: T cells"   "1: Monocytes"
```

## Multi-Model Consensus Example

For more reliable annotations, you can use multiple models and create a consensus:

```{r}
# Define models to use
models <- c(
  "claude-sonnet-4-6",  # Anthropic
  "gpt-5.5",                      # OpenAI
  "gemini-3.1-pro-preview"               # Google
)

# API keys for different providers
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  openai = Sys.getenv("OPENAI_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY")
)

# Run annotation with multiple models
results <- list()
for (model in models) {
  provider <- get_provider(model)
  api_key <- api_keys[[provider]]

  results[[model]] <- annotate_cell_types(
    input = markers,
    tissue_name = "human PBMC",
    model = model,
    api_key = api_key,
    top_gene_count = 10
  )
}

# Create consensus
consensus_results <- interactive_consensus_annotation(
  input = markers,
  tissue_name = "human PBMC",
  models = models,  # Use all the models defined above
  api_keys = api_keys,
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6"
)
```

### Consensus Output Example

The function automatically prints a summary upon completion:

```r
>
Consensus Summary:
-----------------
Total clusters: 2
Controversial clusters: 0
Consensus achieved for all clusters

Cluster 0:
  Final annotation: T cells
  Consensus proportion: 1.0
  Entropy: 0.0
  Model predictions:
    - claude-sonnet-4-6: T cells
    - gpt-5.5: T cells
    - gemini-3.1-pro-preview: T cells

Cluster 1:
  Final annotation: Monocytes
  Consensus proportion: 1.0
  Entropy: 0.0
  Model predictions:
    - claude-sonnet-4-6: Monocytes
    - gpt-5.5: Monocytes
    - gemini-3.1-pro-preview: Monocytes
```

## Integrating with Seurat

To add the annotations to your Seurat object:

```{r}
# Assuming you have a Seurat object named 'seurat_obj' and consensus results
library(Seurat)

# Add consensus annotations to Seurat object
seurat_obj$cell_type_consensus <- plyr::mapvalues(
  x = as.character(Idents(seurat_obj)),
  from = names(consensus_results$final_annotations),
  to = consensus_results$final_annotations
)

# Extract consensus metrics from the consensus results
# Note: These metrics are available in the consensus_results$initial_results$consensus_results
consensus_metrics <- lapply(names(consensus_results$initial_results$consensus_results), function(cluster_id) {
  metrics <- consensus_results$initial_results$consensus_results[[cluster_id]]
  return(list(
    cluster = cluster_id,
    consensus_proportion = metrics$consensus_proportion,
    entropy = metrics$entropy
  ))
})

# Convert to data frame for easier handling
metrics_df <- do.call(rbind, lapply(consensus_metrics, data.frame))

# Add consensus proportion to Seurat object
seurat_obj$consensus_proportion <- plyr::mapvalues(
  x = as.character(Idents(seurat_obj)),
  from = metrics_df$cluster,
  to = metrics_df$consensus_proportion
)

# Add entropy to Seurat object
seurat_obj$entropy <- plyr::mapvalues(
  x = as.character(Idents(seurat_obj)),
  from = metrics_df$cluster,
  to = metrics_df$entropy
)
```

## Basic Visualization

Here's a simple visualization of the results using Seurat:

```{r}
# Plot UMAP with cell type annotations
DimPlot(seurat_obj, group.by = "cell_type_consensus", label = TRUE, repel = TRUE) +
  ggtitle("Cell Type Annotations") +
  theme(plot.title = element_text(hjust = 0.5))
```

## Understanding the Output

The output of `annotate_cell_types()` is a vector of cell type annotations, where each element corresponds to a cluster.

The output of `interactive_consensus_annotation()` is a list containing:

- `final_annotations`: Final consensus cell type annotations
- `initial_results`: Initial predictions from each model
- `controversial_clusters`: List of clusters that required discussion
- `discussion_logs`: Detailed logs of the discussion process
- `session_id`: Unique identifier for the annotation session

### Understanding Uncertainty Metrics

When using consensus annotation, two key metrics help evaluate the reliability of annotations:

- **Consensus Proportion**: Ranges from 0 to 1, indicating the proportion of models that agree on the final annotation. Higher values indicate stronger agreement.
- **Entropy**: Measures the uncertainty in model predictions. Lower values indicate more certainty. An entropy of 0 means all models agree perfectly.

Clusters with low consensus proportion or high entropy may require manual review.

## Using OpenRouter Free Models

If you don't have access to paid API keys, you can use OpenRouter's free models:

```{r}
# Set OpenRouter API key
Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key")

# Use a free model
free_results <- annotate_cell_types(
  input = markers,
  tissue_name = "human PBMC",
  model = "meta-llama/llama-4-maverick:free",  # Note the :free suffix
  api_key = Sys.getenv("OPENROUTER_API_KEY"),
  top_gene_count = 10
)

# Print results
print(free_results)
```

Available free models (Updated Oct 2025):

- `meta-llama/llama-4-maverick:free` - Meta Llama 4 Maverick (256K context, best performance)
- `deepseek/deepseek-v4-pro:free` - DeepSeek V4 Pro
- `meta-llama/llama-3.3-70b-instruct:free` - Meta Llama 3.3 70B (reliable)
- `venice/uncensored:free` - Venice Uncensored (new model)
- `z-ai/glm-4.5-air:free` - GLM 4.5 Air (lightweight)

**Important**: OpenRouter reduced free tier limits in 2025:
- **Free accounts**: 50 requests/day (down from 200), 20 requests/minute
- **Accounts with $10+ credits**: 1000 requests/day
- **Some models removed**: NVIDIA Nemotron and others have exited the free tier
- **For production use**: Consider using paid models for better reliability

## Troubleshooting

### Common Issues

1. **API Key Not Found**:

   ```r
   Error: No auth credentials found
   ```

   **Solution**: Ensure you've set the correct API key environment variable or provided it directly in the function call.

2. **Rate Limiting**:

   ```r
   Error: Rate limit exceeded
   ```

   **Solution**: Wait a few minutes before trying again, or reduce the number of API calls by processing fewer clusters at once.

3. **Invalid Model Name**:

   ```r
   Error: Unsupported model: [model_name]
   ```

   **Solution**: Check that you're using a supported model name and that it's spelled correctly.

4. **Network Issues**:

   ```r
   Error: Could not connect to API
   ```

   **Solution**: Check your internet connection and try again. If the problem persists, the API service might be down.

## Next Steps

Now that you understand the basics of mLLMCelltype, you can explore:

- [Usage Tutorial](https://cafferyang.com/mLLMCelltype/articles/usage-tutorial.html): More detailed usage examples
- [Consensus Annotation Principles](https://cafferyang.com/mLLMCelltype/articles/consensus-principles.html): Learn about the consensus mechanism
- [Visualization Guide](https://cafferyang.com/mLLMCelltype/articles/visualization-guide.html): Create publication-ready visualizations
- [Advanced Features](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html): Explore hierarchical annotation and other advanced features
- [FAQ](https://cafferyang.com/mLLMCelltype/articles/faq.html): Answers to common questions

If you encounter any issues, check the [FAQ](https://cafferyang.com/mLLMCelltype/articles/faq.html) or [open an issue](https://github.com/cafferychen777/mLLMCelltype/issues) on our GitHub repository.

---
title: "Usage Tutorial"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Usage Tutorial}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE, purl=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

# Usage Tutorial

This tutorial provides detailed instructions for using mLLMCelltype for cell type annotation in single-cell RNA sequencing data. We'll cover various usage scenarios, parameter configurations, and integration with Seurat.

## Comprehensive Function Parameters

### annotate_cell_types()

The main function for cell type annotation with a single model:

```{r, purl=FALSE}
library(mLLMCelltype)

results <- annotate_cell_types(
  input,                # Marker gene data (data frame, list, or file path)
  tissue_name,          # Tissue name (e.g., "human PBMC", "mouse brain")
  model,                # LLM model to use
  api_key = NA,         # API key (if not set in environment, NA returns prompt only)
  top_gene_count = 10,  # Number of top genes per cluster to use
  debug = FALSE         # Whether to print debugging information
)
```

### interactive_consensus_annotation()

Function for creating consensus annotations from multiple models through interactive discussion:

```{r eval=FALSE, purl=FALSE}
consensus_results <- interactive_consensus_annotation(
  input,                # Original marker gene data (Seurat FindAllMarkers result or list of genes)
  tissue_name = NULL,   # Optional tissue name
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3.1-pro-preview"),  # Models to use
  api_keys,             # Named list of API keys
  top_gene_count = 10,  # Number of top genes to use
  controversy_threshold = 0.7,  # Threshold for identifying controversial clusters
  entropy_threshold = 1.0,  # Entropy threshold for controversial clusters
  max_discussion_rounds = 3,  # Maximum discussion rounds
  consensus_check_model = NULL,  # Model to use for consensus checking (see recommendations below)
  log_dir = "logs",     # Directory for logs
  cache_dir = NULL,  # Uses default system cache directory
  use_cache = TRUE      # Whether to use cache
)
```

**Important Note on `consensus_check_model`**: This parameter is used for evaluating semantic similarity, calculating consensus metrics, and moderating discussions. We recommend using capable models such as:
- `claude-sonnet-4-6` (Anthropic)
- `claude-opus-4-1-20250805` (Anthropic)
- `o1` (OpenAI)
- `gpt-5.5` (OpenAI)
- `gemini-3.1-pro-preview` (Google)

See the main README for detailed recommendations and examples.

## Detailed Usage Scenarios

### Scenario 1: Basic Annotation with a Single Model

For quick exploration or when API usage is a concern:

```{r, purl=FALSE}
# Load example data
library(Seurat)
data("pbmc_small")

# Find markers
pbmc_markers <- FindAllMarkers(pbmc_small, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)

# Run annotation with a single model
results <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10
)

# Add annotations to Seurat object
pbmc_small$cell_type_claude <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = names(results),
  to = results
)

# Visualize
DimPlot(pbmc_small, group.by = "cell_type_claude", label = TRUE)
```

### Scenario 2: Multi-Model Consensus for High Accuracy

For publication-quality annotations with uncertainty quantification:

```{r, purl=FALSE}
# Define multiple models to use
models <- c(
  "claude-sonnet-4-6",  # Anthropic
  "gpt-5.5",                      # OpenAI
  "gemini-3.1-pro-preview",              # Google
  "grok-4.3"                       # X.AI
)

# API keys for different providers
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  openai = Sys.getenv("OPENAI_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY"),
  grok = Sys.getenv("GROK_API_KEY")
)

# Run annotation with multiple models
results <- list()
for (model in models) {
  provider <- get_provider(model)
  api_key <- api_keys[[provider]]

  results[[model]] <- annotate_cell_types(
    input = pbmc_markers,
    tissue_name = "human PBMC",
    model = model,
    api_key = api_key,
    top_gene_count = 10
  )
}

# Create consensus
consensus_results <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = models,  # Use all the models defined above
  api_keys = api_keys,
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6"
)

# View consensus results
# You can access the final annotations with consensus_results$final_annotations

# Add consensus annotations and metrics to Seurat object
pbmc_small$cell_type_consensus <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = names(consensus_results$final_annotations),
  to = consensus_results$final_annotations
)

# Extract consensus metrics from the consensus results
consensus_metrics <- lapply(names(consensus_results$initial_results$consensus_results), function(cluster_id) {
  metrics <- consensus_results$initial_results$consensus_results[[cluster_id]]
  return(list(
    cluster = cluster_id,
    consensus_proportion = metrics$consensus_proportion,
    entropy = metrics$entropy
  ))
})

# Convert to data frame for easier handling
metrics_df <- do.call(rbind, lapply(consensus_metrics, data.frame))

# Add consensus proportion to Seurat object
pbmc_small$consensus_proportion <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = metrics_df$cluster,
  to = metrics_df$consensus_proportion
)

# Add entropy to Seurat object
pbmc_small$shannon_entropy <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = metrics_df$cluster,
  to = metrics_df$entropy
)
```

### Scenario 2b: Using Free OpenRouter Models

For users with limited API credits or budget constraints:

```{r, purl=FALSE}
# Set OpenRouter API key
openrouter_api_key <- Sys.getenv("OPENROUTER_API_KEY")

# Define free OpenRouter models to use
free_models <- c(
  "meta-llama/llama-4-maverick:free",                # Meta Llama 4 Maverick (free)
  "meta-llama/llama-3.3-70b-instruct:free",    # Meta Llama 3.3 70B (free)
  "deepseek/deepseek-v4-pro:free",             # DeepSeek V4 Pro (free)
  "meta-llama/llama-3.3-70b-instruct:free"          # Meta Llama 3.3 70B (free)
)

# Run annotation with free OpenRouter models
free_results <- list()
for (model in free_models) {
  free_results[[model]] <- annotate_cell_types(
    input = pbmc_markers,
    tissue_name = "human PBMC",
    model = model,  # OpenRouter models are automatically detected by format: 'provider/model-name:free'
    api_key = openrouter_api_key,
    top_gene_count = 10
  )
}

# Create consensus with free models
free_consensus_results <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = free_models,  # Use all the free models defined above
  api_keys = list("openrouter" = openrouter_api_key),
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "meta-llama/llama-4-maverick:free"  # Use a free model for consensus checking
)

# View free model consensus results
# You can access the final annotations with free_consensus_results$final_annotations

# Add free model consensus annotations to Seurat object
pbmc_small$free_model_consensus <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = names(free_consensus_results$final_annotations),
  to = free_consensus_results$final_annotations
)

# Compare paid vs. free model results
comparison <- data.frame(
  cluster = names(consensus_results$final_annotations),
  paid_models = consensus_results$final_annotations,
  free_models = free_consensus_results$final_annotations,
  agreement = consensus_results$final_annotations == free_consensus_results$final_annotations
)
print(comparison)
```

### Scenario 3: Working with CSV Files

For users who prefer working with files:

```{r, purl=FALSE}
# Save markers to CSV
write.csv(pbmc_markers, "pbmc_markers.csv", row.names = FALSE)

# Run annotation using the CSV file
results <- annotate_cell_types(
  input = "pbmc_markers.csv",
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY")
)
```

### Scenario 4: Custom Caching

For better control over caching behavior:

```{r, purl=FALSE}
# Note: The annotate_cell_types function does not have built-in caching.
# If you need caching, you can implement it separately.

# Run annotation
results <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10,
  debug = FALSE
)

# If you need custom caching, you can implement it using your own cache manager
# This is just a conceptual example and not part of the actual package
# cache_manager <- YourCacheManager$new(cache_dir = "path/to/cache")
# cache_manager$clear_cache()
```

## Model Selection Guide

mLLMCelltype supports a wide range of LLM models. Here's a guide to help you choose:

### High Performance Models

For the most accurate annotations:

- **Anthropic Claude Opus 4.7** (`claude-opus-4-7`)
- **Anthropic Claude Sonnet 4.6** (`claude-sonnet-4-6`)
- **OpenAI GPT-5.5** (`gpt-5.5`)
- **Google Gemini 3.1 Pro Preview** (`gemini-3.1-pro-preview`)

### Balanced Performance/Cost Models

For good results with lower API costs:

- **X.AI Grok 4.3** (`grok-4.3`): Competitive performance at lower cost
- **DeepSeek V4 Flash** (`deepseek-v4-flash`): Good performance for specialized tissues

### Economy Models

For preliminary exploration or large datasets:

- **Qwen 3.6 Flash** (`qwen3.6-flash`): Good performance for the cost
- **Zhipu/Z.AI GLM-5-Turbo** (`glm-5-turbo`): Economical option with decent performance
- **MiniMax M2.5** (`MiniMax-M2.5`): Cost-effective for initial exploration

### Free Models via OpenRouter

For users with limited API credits or budget constraints:

- **Meta Llama 4 Maverick** (`meta-llama/llama-4-maverick:free`): Most reliable and fast, recommended for consensus checking
- **Meta Llama 3.3 70B** (`meta-llama/llama-3.3-70b-instruct:free`): Good performance with consistent formatting
- **DeepSeek V4 Pro** (`deepseek/deepseek-v4-pro:free`): Reasoning model (free)

Based on our testing, we recommend the following free models:

- `meta-llama/llama-4-maverick:free`: Most reliable and fast, recommended for consensus checking
- `meta-llama/llama-3.3-70b-instruct:free`: Good performance with consistent formatting
- `deepseek/deepseek-v4-pro:free`: Reliable with good response time

Some models may have limitations:

- `deepseek/deepseek-v4-pro:free`: May occasionally return empty results
- `thudm/glm-z1-9b:free`: May return localized error messages ("non-character parameter") when used for consensus checking

These free models are accessed through OpenRouter and don't consume credits, but may have limitations compared to paid models. Use the `:free` suffix in the model name to access them.

```{r, purl=FALSE}
# Example of using a free model via OpenRouter
# First, set your OpenRouter API key
Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key")

# Then use a free model with the :free suffix
free_model_results <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "meta-llama/llama-4-maverick:free",  # Note the :free suffix
  api_key = Sys.getenv("OPENROUTER_API_KEY")
  # No need to specify provider - it's automatically detected from the model name format
)
```

## Integration with Seurat Workflow

Here's a complete example of integrating mLLMCelltype with a Seurat workflow:

```{r, purl=FALSE}
library(Seurat)
library(mLLMCelltype)
library(ggplot2)

# Load data
data("pbmc_small")

# Standard Seurat preprocessing
pbmc_small <- NormalizeData(pbmc_small)
pbmc_small <- FindVariableFeatures(pbmc_small)
pbmc_small <- ScaleData(pbmc_small)
pbmc_small <- RunPCA(pbmc_small)
pbmc_small <- FindNeighbors(pbmc_small)
pbmc_small <- FindClusters(pbmc_small, resolution = 0.5)
pbmc_small <- RunUMAP(pbmc_small, dims = 1:10)

# Find markers for each cluster
pbmc_markers <- FindAllMarkers(pbmc_small, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)

# Define models to use
models <- c(
  "claude-sonnet-4-6",
  "gpt-5.5",
  "gemini-3.1-pro-preview"
)

# API keys
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  openai = Sys.getenv("OPENAI_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY")
)

# Run annotation with multiple models
results <- list()
for (model in models) {
  provider <- get_provider(model)
  api_key <- api_keys[[provider]]

  results[[model]] <- annotate_cell_types(
    input = pbmc_markers,
    tissue_name = "human PBMC",
    model = model,
    api_key = api_key,
    top_gene_count = 10
  )

  # Add individual model results to Seurat object
  column_name <- paste0("cell_type_", gsub("[^a-zA-Z0-9]", "_", model))
  pbmc_small[[column_name]] <- plyr::mapvalues(
    x = as.character(Idents(pbmc_small)),
    from = names(results[[model]]),
    to = results[[model]]
  )
}

# Create consensus
consensus_results <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = models,  # Use all the models defined above
  api_keys = api_keys,
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6"
)

# Add consensus results to Seurat object
pbmc_small$cell_type_consensus <- plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = names(consensus_results$final_annotations),
  to = consensus_results$final_annotations
)

# Extract consensus metrics from the consensus results
consensus_metrics <- lapply(names(consensus_results$initial_results$consensus_results), function(cluster_id) {
  metrics <- consensus_results$initial_results$consensus_results[[cluster_id]]
  return(list(
    cluster = cluster_id,
    consensus_proportion = metrics$consensus_proportion,
    entropy = metrics$entropy
  ))
})

# Convert to data frame for easier handling
metrics_df <- do.call(rbind, lapply(consensus_metrics, data.frame))

# Add consensus proportion to Seurat object
pbmc_small$consensus_proportion <- as.numeric(plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = metrics_df$cluster,
  to = metrics_df$consensus_proportion
))

# Add entropy to Seurat object
pbmc_small$shannon_entropy <- as.numeric(plyr::mapvalues(
  x = as.character(Idents(pbmc_small)),
  from = metrics_df$cluster,
  to = metrics_df$entropy
))

# Visualize results
p1 <- DimPlot(pbmc_small, group.by = "cell_type_consensus", label = TRUE, repel = TRUE) +
  ggtitle("Cell Type Annotations") +
  theme(plot.title = element_text(hjust = 0.5))

p2 <- FeaturePlot(pbmc_small, features = "consensus_proportion", cols = c("yellow", "green", "blue")) +
  ggtitle("Consensus Proportion") +
  theme(plot.title = element_text(hjust = 0.5))

p3 <- FeaturePlot(pbmc_small, features = "shannon_entropy", cols = c("red", "orange")) +
  ggtitle("Shannon Entropy") +
  theme(plot.title = element_text(hjust = 0.5))

# Combine plots
p1 | p2 | p3
```

## Advanced Parameter Tuning

### Adjusting top_gene_count

The `top_gene_count` parameter controls how many top marker genes per cluster are used for annotation:

```{r, purl=FALSE}
# Using more genes (better for well-characterized tissues)
results_more_genes <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 20  # Using more genes
)

# Using fewer genes (better for noisy data)
results_fewer_genes <- annotate_cell_types(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 5   # Using fewer genes
)
```

### Adjusting controversy_threshold

The `controversy_threshold` parameter in the `interactive_consensus_annotation` function controls which clusters are considered controversial and require discussion:

```{r eval=FALSE, purl=FALSE}
# Example of using interactive_consensus_annotation with different controversy thresholds
# Lower threshold (more clusters will be discussed)
consensus_results_low_threshold <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3-flash-preview"),
  api_keys = list(
    "anthropic" = Sys.getenv("ANTHROPIC_API_KEY"),
    "openai" = Sys.getenv("OPENAI_API_KEY"),
    "gemini" = Sys.getenv("GEMINI_API_KEY")
  ),
  controversy_threshold = 0.3  # Lower threshold - more clusters will be discussed
)

# Higher threshold (fewer clusters will be discussed)
consensus_results_high_threshold <- interactive_consensus_annotation(
  input = pbmc_markers,
  tissue_name = "human PBMC",
  models = c("claude-sonnet-4-6", "gpt-5.5", "gemini-3-flash-preview"),
  api_keys = list(
    "anthropic" = Sys.getenv("ANTHROPIC_API_KEY"),
    "openai" = Sys.getenv("OPENAI_API_KEY"),
    "gemini" = Sys.getenv("GEMINI_API_KEY")
  ),
  controversy_threshold = 0.7  # Higher threshold - fewer clusters will be discussed
)
```

## Performance Considerations

### API Rate Limits and Costs

Different LLM providers have different rate limits and pricing:

- **OpenAI**: Higher rate limits but can be more expensive
- **Anthropic**: Good balance of rate limits and cost
- **Google**: Competitive pricing with good rate limits
- **Others**: Generally lower cost but may have stricter rate limits

To manage costs and rate limits:

1. Use caching to avoid redundant API calls
2. Start with a single model for exploration
3. Use more economical models for initial testing
4. Reserve multi-model consensus for final analysis

### Execution Time

Typical execution times:

- Single model annotation: 5-30 seconds per cluster
- Multi-model consensus: 1-5 minutes for a typical dataset
- Discussion process: Additional 1-3 minutes per controversial cluster

To improve performance:

1. Use a smaller `top_gene_count` for faster execution
2. Enable caching to reuse results
3. Use a higher `controversy_threshold` to reduce the number of clusters that require discussion

## Troubleshooting

### Common Issues with OpenRouter

If you encounter "No auth credentials found" errors with OpenRouter:

- Verify your API key is correct and active
- Ensure you're using the correct format for the model name (e.g., `provider/model-name:free`)
- Try a different free model, as some models may have access restrictions

### Non-English Model Error Messages

If you see "non-character parameter" errors:

- This typically occurs when using non-English language models like `thudm/glm-z1-9b:free` for consensus checking
- Switch to an English-optimized model like `meta-llama/llama-4-maverick:free` for consensus checking

### Empty Results

If a model returns empty results:

- Try increasing the timeout or retry with the same model
- Some models like `deepseek/deepseek-v4-pro:free` may occasionally return empty results
- Switch to a more reliable model if the problem persists

## Next Steps

Now that you understand the detailed usage of mLLMCelltype, you can explore:

- [Consensus Annotation Principles](https://cafferyang.com/mLLMCelltype/articles/consensus-principles.html): Learn about the technical principles
- [Visualization Guide](https://cafferyang.com/mLLMCelltype/articles/visualization-guide.html): Create publication-ready visualizations
- [FAQ](https://cafferyang.com/mLLMCelltype/articles/faq.html): Find answers to common questions
- [Advanced Features](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html): Explore hierarchical annotation and other advanced features

---
title: "Why Choose Consensus? The Scientific Foundation of Multi-LLM Annotation"
description: "Overview of the consensus-based approach to cell type annotation, including its scientific basis, methodology, and trade-offs."
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Why Choose Consensus? The Scientific Foundation of Multi-LLM Annotation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Why Choose Consensus? The Scientific Foundation of Multi-LLM Annotation

Multi-LLM consensus can improve annotation accuracy by combining the strengths of diverse AI models while reducing the impact of individual model limitations (see Yang et al., 2026).

## The Challenge with Single-Model Approaches

Traditional single-model annotation systems face inherent limitations:

### Accuracy Limitations
- **Single-point failure**: One model's bias affects all results
- **Limited perspective**: Each model has unique strengths and blind spots
- **Inconsistent performance**: Varies across cell types and tissues

### Reliability Issues
- **Model hallucinations**: Confident but incorrect predictions
- **Lack of uncertainty**: Difficult to identify questionable annotations
- **Reproducibility challenges**: Different model versions may yield different results

## The Consensus Approach: Inspired by Scientific Peer Review

mLLMCelltype's consensus framework is analogous to the peer review process in scientific publishing.

### The Scientific Parallel

Just as scientific papers benefit from multiple expert reviewers, cell annotations can benefit from multiple AI models:

| Scientific Peer Review | mLLMCelltype Consensus |
|------------------------|------------------------|
| Multiple expert reviewers | Multiple LLM models |
| Diverse perspectives | Different training approaches |
| Debate and discussion | Structured deliberation |
| Consensus building | Agreement quantification |
| Quality assurance | Uncertainty metrics |

### How It Works

**1. Error Detection Through Cross-Validation**
- Models check each other's work
- Individual model biases can be averaged out
- Outlier predictions are identified

**2. Transparent Uncertainty Quantification**
- **Consensus Proportion (CP)**: Measures inter-model agreement
- **Shannon Entropy**: Quantifies prediction uncertainty
- **Controversy Detection**: Automatically identifies clusters requiring expert review

## Why Multiple Perspectives Help

Cell type annotation involves:

- **Marker gene interpretation**: Different models may have different strengths across gene families
- **Context understanding**: Various models may capture different biological contexts
- **Rare cell types**: Ensemble approaches can improve detection of uncommon populations
- **Batch effects**: Multiple models may provide robustness against technical artifacts

For benchmark results, see Yang et al. (2026):

Yang, C., Zhang, X., & Chen, J. (2026). Large language model consensus substantially improves the cell type annotation accuracy for scRNA-seq data. *Communications Biology*. https://doi.org/10.1038/s42003-026-10420-8

## Cost Considerations

The two-stage approach can reduce API calls when models agree early:

- **Stage 1**: Initial consensus check -- clusters where models agree skip further processing
- **Stage 2**: Deliberation only for clusters without initial agreement
- **Caching**: Results can be reused across similar analyses

This means the cost overhead of using multiple models is partially offset by skipping deliberation for clear cases.

## Technical Implementation

### The Three-Stage Process

**Stage 1: Independent Analysis**
Each LLM analyzes marker genes and provides:
- Cell type predictions
- Confidence scores
- Reasoning chains

**Stage 2: Consensus Building**
The system:
- Compares predictions across models
- Identifies areas of agreement and disagreement
- Calculates uncertainty metrics

**Stage 3: Deliberation (when needed)**
For controversial clusters:
- Models share their reasoning
- Structured debate occurs
- Final consensus emerges

### Quality Metrics

- **Semantic similarity analysis**: Ensures meaningful disagreements are detected
- **Evidence-based reasoning**: All predictions include supporting evidence
- **Iterative refinement**: Multiple rounds of discussion when needed

## When to Choose Consensus

**Consensus may be preferable when:**
- Uncertainty quantification is needed
- Datasets involve novel or complex tissues
- Results will be published or used in downstream analyses
- Identifying low-confidence annotations is important

**Consider alternatives when:**
- Quick exploratory analysis is the goal
- Datasets are well-characterized with clear markers
- API budget is very limited
- Proof-of-concept work in early stages

## Quick Start Example

```r
library(mLLMCelltype)

# Load marker genes from your single-cell data, then run consensus annotation
api_keys <- list(
  openai = Sys.getenv("OPENAI_API_KEY"),
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY")
)

results <- interactive_consensus_annotation(
  input = marker_data,
  tissue_name = "human PBMC",
  models = c("gpt-5.5", "claude-sonnet-4-6", "gemini-3.1-pro-preview"),
  api_keys = api_keys
)

```

### Understanding Your Results

- **High consensus (CP > 0.8)**: Reliable annotations
- **Medium consensus (0.5 < CP < 0.8)**: Review recommended
- **Low consensus (CP < 0.5)**: Expert validation needed

## Summary

The consensus approach provides a framework for combining multiple LLM predictions with built-in uncertainty quantification. As new models become available, the framework can incorporate them without changes to the overall methodology.

## Learn More

- [Getting Started Guide](https://cafferyang.com/mLLMCelltype/articles/getting-started.html)
- [Consensus vs Single-Agent Methods](https://cafferyang.com/mLLMCelltype/articles/vs-single-agent.html)
- [Advanced Features](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html)
- [API Reference](https://cafferyang.com/mLLMCelltype/reference/index.html)

---
title: "Getting Started with mLLMCelltype"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with mLLMCelltype}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE, purl=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

# Getting Started with mLLMCelltype

This guide provides a quick introduction to using mLLMCelltype for cell type annotation in single-cell RNA sequencing data. We'll cover the basic workflow, input data requirements, and a simple example to get you started.

## Basic Workflow

The mLLMCelltype workflow consists of these main steps:

1. **Prepare marker gene data** for each cluster
2. **Run annotation** using one or multiple LLMs
3. **Create consensus** from multiple model predictions (optional)
4. **Integrate results** with your Seurat or Scanpy object
5. **Visualize results** with uncertainty metrics

## Loading the Package and Setting Up API Keys

First, load the mLLMCelltype package:

```{r, purl=FALSE}
library(mLLMCelltype)
```

### Setting Up API Keys

Before using mLLMCelltype, you need to set up API keys for the LLM providers you plan to use:

```{r, purl=FALSE}
# Set API keys as environment variables
Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-api-key")  # For Claude models
Sys.setenv(OPENAI_API_KEY = "your-openai-api-key")        # For GPT models
Sys.setenv(GEMINI_API_KEY = "your-gemini-api-key")        # For Gemini models
Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key") # For OpenRouter models
```

You can obtain API keys from:
- Anthropic: https://platform.claude.com/
- OpenAI: https://platform.openai.com/
- Google (Gemini): https://ai.google.dev/
- OpenRouter: https://openrouter.ai/keys

Alternatively, you can provide API keys directly in function calls:

```{r, purl=FALSE}
results <- annotate_cell_types(
  input = markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = "your-anthropic-api-key",  # Direct API key
  top_gene_count = 10
)
```

## Input Data Requirements

mLLMCelltype accepts marker gene data in several formats:

### 1. Data Frame Format

A data frame with the following columns:
- `cluster`: Cluster ID (preserved as-is from your data)
- `gene`: Gene name/symbol
- `avg_log2FC` or similar metric: Log fold change
- `p_val_adj` or similar metric: Adjusted p-value

Example:

```{r, purl=FALSE}
# Example marker data frame
markers_df <- data.frame(
  cluster = c(0, 0, 0, 1, 1, 1),
  gene = c("CD3D", "CD3E", "CD2", "CD14", "LYZ", "CST3"),
  avg_log2FC = c(2.5, 2.3, 2.1, 3.1, 2.8, 2.5),
  p_val_adj = c(0.001, 0.001, 0.002, 0.0001, 0.0002, 0.0005)
)
```

### 2. Seurat FindMarkers Output

You can directly use the output from Seurat's `FindAllMarkers()` function:

```{r, purl=FALSE}
# Assuming you have a Seurat object named 'seurat_obj'
library(Seurat)
all_markers <- FindAllMarkers(seurat_obj, only.pos = TRUE, min.pct = 0.25, logfc.threshold = 0.25)
```

### 3. CSV File Path

A path to a CSV file containing marker gene data:

```{r, purl=FALSE}
# Path to your CSV file
markers_file <- "path/to/markers.csv"
```

### 4. List Format

A list where each element contains marker genes for a cluster:

```{r, purl=FALSE}
# Example marker list
markers_list <- list(
  "0" = c("CD3D", "CD3E", "CD2", "IL7R", "LTB"),
  "1" = c("CD14", "LYZ", "CST3", "MS4A7", "FCGR3A")
)
```

## Function Parameters

The `annotate_cell_types` function has the following parameters:

| Parameter | Description | Default Value |
|-----------|-------------|---------------|
| `input` | Marker gene data (data frame, list, or file path) | (required) |
| `tissue_name` | Tissue name (e.g., "human PBMC", "mouse brain") | `NULL` |
| `model` | LLM model to use | `"gpt-5.5"` |
| `api_key` | API key (if not set in environment) | `NA` |
| `top_gene_count` | Number of top genes per cluster to use | `10` |
| `debug` | Whether to print debugging information | `FALSE` |

Note: If `api_key` is set to `NA`, the function will return the generated prompt without making an API call, which is useful for reviewing the prompt before sending it to the API.

## Basic Usage Example

Here's a simple example using a single LLM model for annotation:

```{r, purl=FALSE}
# Example marker data
markers <- data.frame(
  cluster = c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1),
  gene = c("CD3D", "CD3E", "CD2", "IL7R", "LTB", "CD14", "LYZ", "CST3", "MS4A7", "FCGR3A"),
  avg_log2FC = c(2.5, 2.3, 2.1, 1.8, 1.7, 3.1, 2.8, 2.5, 2.2, 2.0),
  p_val_adj = c(0.001, 0.001, 0.002, 0.003, 0.005, 0.0001, 0.0002, 0.0005, 0.001, 0.002)
)

# Run annotation with a single model
results <- annotate_cell_types(
  input = markers,
  tissue_name = "human PBMC",
  model = "claude-sonnet-4-6",
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 10,
  debug = FALSE  # Set to TRUE for more detailed output
)

# Print results
print(results)
```

### Example Output

When using a single model like Claude, the output will be a character vector with one annotation per cluster:

```r
> print(results)
[1] "0: T cells"   "1: Monocytes"
```

## Multi-Model Consensus Example

For more reliable annotations, you can use multiple models and create a consensus:

```{r, purl=FALSE}
# Define models to use
models <- c(
  "claude-sonnet-4-6",  # Anthropic
  "gpt-5.5",                      # OpenAI
  "gemini-3.1-pro-preview"               # Google
)

# API keys for different providers
api_keys <- list(
  anthropic = Sys.getenv("ANTHROPIC_API_KEY"),
  openai = Sys.getenv("OPENAI_API_KEY"),
  gemini = Sys.getenv("GEMINI_API_KEY")
)

# Run annotation with multiple models
results <- list()
for (model in models) {
  provider <- get_provider(model)
  api_key <- api_keys[[provider]]

  results[[model]] <- annotate_cell_types(
    input = markers,
    tissue_name = "human PBMC",
    model = model,
    api_key = api_key,
    top_gene_count = 10
  )
}

# Create consensus
consensus_results <- interactive_consensus_annotation(
  input = markers,
  tissue_name = "human PBMC",
  models = models,  # Use all the models defined above
  api_keys = api_keys,
  controversy_threshold = 0.7,
  entropy_threshold = 1.0,
  consensus_check_model = "claude-sonnet-4-6"
)
```

### Consensus Output Example

The function automatically prints a summary upon completion:

```r
>
Consensus Summary:
-----------------
Total clusters: 2
Controversial clusters: 0
Consensus achieved for all clusters

Cluster 0:
  Final annotation: T cells
  Consensus proportion: 1.0
  Entropy: 0.0
  Model predictions:
    - claude-sonnet-4-6: T cells
    - gpt-5.5: T cells
    - gemini-3.1-pro-preview: T cells

Cluster 1:
  Final annotation: Monocytes
  Consensus proportion: 1.0
  Entropy: 0.0
  Model predictions:
    - claude-sonnet-4-6: Monocytes
    - gpt-5.5: Monocytes
    - gemini-3.1-pro-preview: Monocytes
```

## Integrating with Seurat

To add the annotations to your Seurat object:

```{r, purl=FALSE}
# Assuming you have a Seurat object named 'seurat_obj' and consensus results
library(Seurat)

# Add consensus annotations to Seurat object
seurat_obj$cell_type_consensus <- plyr::mapvalues(
  x = as.character(Idents(seurat_obj)),
  from = names(consensus_results$final_annotations),
  to = consensus_results$final_annotations
)

# Extract consensus metrics from the consensus results
# Note: These metrics are available in the consensus_results$initial_results$consensus_results
consensus_metrics <- lapply(names(consensus_results$initial_results$consensus_results), function(cluster_id) {
  metrics <- consensus_results$initial_results$consensus_results[[cluster_id]]
  return(list(
    cluster = cluster_id,
    consensus_proportion = metrics$consensus_proportion,
    entropy = metrics$entropy
  ))
})

# Convert to data frame for easier handling
metrics_df <- do.call(rbind, lapply(consensus_metrics, data.frame))

# Add consensus proportion to Seurat object
seurat_obj$consensus_proportion <- plyr::mapvalues(
  x = as.character(Idents(seurat_obj)),
  from = metrics_df$cluster,
  to = metrics_df$consensus_proportion
)

# Add entropy to Seurat object
seurat_obj$entropy <- plyr::mapvalues(
  x = as.character(Idents(seurat_obj)),
  from = metrics_df$cluster,
  to = metrics_df$entropy
)
```

## Basic Visualization

Here's a simple visualization of the results using Seurat:

```{r, purl=FALSE}
# Plot UMAP with cell type annotations
DimPlot(seurat_obj, group.by = "cell_type_consensus", label = TRUE, repel = TRUE) +
  ggtitle("Cell Type Annotations") +
  theme(plot.title = element_text(hjust = 0.5))
```

## Understanding the Output

The output of `annotate_cell_types()` is a vector of cell type annotations, where each element corresponds to a cluster.

The output of `interactive_consensus_annotation()` is a list containing:

- `final_annotations`: Final consensus cell type annotations
- `initial_results`: Initial predictions from each model
- `controversial_clusters`: List of clusters that required discussion
- `discussion_logs`: Detailed logs of the discussion process
- `session_id`: Unique identifier for the annotation session

### Understanding Uncertainty Metrics

When using consensus annotation, two key metrics help evaluate the reliability of annotations:

- **Consensus Proportion**: Ranges from 0 to 1, indicating the proportion of models that agree on the final annotation. Higher values indicate stronger agreement.
- **Entropy**: Measures the uncertainty in model predictions. Lower values indicate more certainty. An entropy of 0 means all models agree perfectly.

Clusters with low consensus proportion or high entropy may require manual review.

## Using OpenRouter Free Models

If you don't have access to paid API keys, you can use OpenRouter's free models:

```{r, purl=FALSE}
# Set OpenRouter API key
Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key")

# Use a free model
free_results <- annotate_cell_types(
  input = markers,
  tissue_name = "human PBMC",
  model = "meta-llama/llama-4-maverick:free",  # Note the :free suffix
  api_key = Sys.getenv("OPENROUTER_API_KEY"),
  top_gene_count = 10
)

# Print results
print(free_results)
```

Available free models (Updated Oct 2025):

- `meta-llama/llama-4-maverick:free` - Meta Llama 4 Maverick (256K context, best performance)
- `deepseek/deepseek-v4-pro:free` - DeepSeek V4 Pro
- `meta-llama/llama-3.3-70b-instruct:free` - Meta Llama 3.3 70B (reliable)
- `venice/uncensored:free` - Venice Uncensored (new model)
- `z-ai/glm-4.5-air:free` - GLM 4.5 Air (lightweight)

**Important**: OpenRouter reduced free tier limits in 2025:
- **Free accounts**: 50 requests/day (down from 200), 20 requests/minute
- **Accounts with $10+ credits**: 1000 requests/day
- **Some models removed**: NVIDIA Nemotron and others have exited the free tier
- **For production use**: Consider using paid models for better reliability

## Troubleshooting

### Common Issues

1. **API Key Not Found**:

   ```r
   Error: No auth credentials found
   ```

   **Solution**: Ensure you've set the correct API key environment variable or provided it directly in the function call.

2. **Rate Limiting**:

   ```r
   Error: Rate limit exceeded
   ```

   **Solution**: Wait a few minutes before trying again, or reduce the number of API calls by processing fewer clusters at once.

3. **Invalid Model Name**:

   ```r
   Error: Unsupported model: [model_name]
   ```

   **Solution**: Check that you're using a supported model name and that it's spelled correctly.

4. **Network Issues**:

   ```r
   Error: Could not connect to API
   ```

   **Solution**: Check your internet connection and try again. If the problem persists, the API service might be down.

## Next Steps

Now that you understand the basics of mLLMCelltype, you can explore:

- [Usage Tutorial](https://cafferyang.com/mLLMCelltype/articles/usage-tutorial.html): More detailed usage examples
- [Consensus Annotation Principles](https://cafferyang.com/mLLMCelltype/articles/consensus-principles.html): Learn about the consensus mechanism
- [Visualization Guide](https://cafferyang.com/mLLMCelltype/articles/visualization-guide.html): Create publication-ready visualizations
- [Advanced Features](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html): Explore hierarchical annotation and other advanced features
- [FAQ](https://cafferyang.com/mLLMCelltype/articles/faq.html): Answers to common questions

If you encounter any issues, check the [FAQ](https://cafferyang.com/mLLMCelltype/articles/faq.html) or [open an issue](https://github.com/cafferychen777/mLLMCelltype/issues) on our GitHub repository.

---
title: "Contributing Guide"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Contributing Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE, purl=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

# Contributing to mLLMCelltype

Thank you for your interest in contributing to mLLMCelltype! This guide will help you understand how to contribute to the project effectively.

## Getting Started

### Fork and Clone the Repository

1. Fork the [mLLMCelltype repository](https://github.com/cafferychen777/mLLMCelltype) on GitHub
2. Clone your fork to your local machine:

```bash
git clone https://github.com/YOUR-USERNAME/mLLMCelltype.git
cd mLLMCelltype
```

3. Add the original repository as an upstream remote:

```bash
git remote add upstream https://github.com/cafferychen777/mLLMCelltype.git
```

### Setting Up the Development Environment

For R package development:

```r
# Install required packages for development
install.packages(c("devtools", "roxygen2", "testthat", "knitr", "rmarkdown"))

# Install the package in development mode
devtools::install_dev("R")
```

## Project Structure

The mLLMCelltype project has the following structure:

```
mLLMCelltype/
├── R/                  # R package source code
│   ├── R/              # R functions
│   ├── man/            # Documentation
│   ├── tests/          # Unit tests
│   ├── vignettes/      # Package vignettes
│   └── DESCRIPTION     # Package metadata
├── python/             # Python package source code
├── .github/            # GitHub workflows and templates
├── assets/             # Images and other assets
├── examples/           # Example notebooks and scripts
└── README.md           # Project overview
```

## Development Workflow

### Creating a New Feature

1. Create a new branch for your feature:

```bash
git checkout -b feature/your-feature-name
```

2. Make your changes to the codebase
3. Add and commit your changes:

```bash
git add .
git commit -m "Add your descriptive commit message here"
```

4. Push your changes to your fork:

```bash
git push origin feature/your-feature-name
```

5. Create a pull request from your fork to the main repository

### Code Style Guidelines

#### R Code Style

We follow the [tidyverse style guide](https://style.tidyverse.org/) for R code:

- Use snake_case for variable and function names
- Use spaces around operators and after commas
- Use 2 spaces for indentation
- Limit line length to 80 characters
- Use roxygen2 for documentation

Example of properly formatted R code:

```r
#' Annotate Cell Types
#'
#' This function annotates cell types based on marker genes.
#'
#' @param input A data frame containing marker genes.
#' @param tissue_name The name of the tissue.
#' @param model The LLM model to use.
#' @param api_key The API key for the LLM provider.
#'
#' @return A vector of cell type annotations.
#' @export
annotate_cell_types <- function(input, tissue_name, model, api_key) {
  # Function implementation
  results <- process_markers(input, top_n = 10)

  for (i in seq_along(results)) {
    if (is_valid_result(results[i])) {
      results[i] <- clean_result(results[i])
    }
  }

  return(results)
}
```

#### Documentation Guidelines

All functions should be documented using roxygen2 with the following sections:

- Title (first line)
- Description (paragraph after title)
- @param for each parameter
- @return for the return value
- @examples for usage examples
- @export if the function should be exported

### Testing

We use the testthat package for testing. Tests should be placed in the `R/tests/testthat/` directory.

To run tests:

```r
devtools::test()
```

Example test file (`test-annotate_cell_types.R`):

```r
context("Cell type annotation")

test_that("annotate_cell_types returns expected format", {
  # Setup test data
  test_markers <- data.frame(
    cluster = c(0, 0, 1, 1),
    gene = c("CD3D", "CD3E", "CD19", "MS4A1"),
    avg_log2FC = c(2.5, 2.3, 3.1, 2.8),
    p_val_adj = c(0.001, 0.002, 0.001, 0.003)
  )

  # Mock the API response
  mockery::stub(
    annotate_cell_types,
    "get_model_response",
    function(...) c("T cells", "B cells")
  )

  # Run the function
  result <- annotate_cell_types(
    input = test_markers,
    tissue_name = "test tissue",
    model = "test-model",
    api_key = "test-key"
  )

  # Assertions
  expect_is(result, "character")
  expect_length(result, 2)
  expect_equal(result, c("T cells", "B cells"))
})
```

## Contributing Areas

### Adding Support for New LLM Models

To add support for a new LLM model:

1. Identify the model provider and API endpoint
2. Create a new processing function in `R/R/process_[provider].R`
3. Update the `get_provider()` function in `R/R/get_provider.R`
4. Add the model to the supported models list
5. Create tests for the new model
6. Update documentation

Example of adding a new model:

```r
# In process_newprovider.R
process_newprovider <- function(prompt, api_key) {
  # Implementation for the new provider
  url <- "https://api.newprovider.com/v1/completions"

  headers <- c(
    "Content-Type" = "application/json",
    "Authorization" = paste("Bearer", api_key)
  )

  body <- list(
    model = "newprovider-model",
    prompt = prompt,
    max_tokens = 1000,
    temperature = 0.1
  )

  # Make API request using httr
  response <- httr::POST(
    url = url,
    httr::add_headers(.headers = headers),
    body = jsonlite::toJSON(body, auto_unbox = TRUE),
    encode = "json"
  )

  # Check for HTTP errors
  httr::stop_for_status(response)

  # Parse the response
  content <- httr::content(response, "text", encoding = "UTF-8")
  parsed_response <- jsonlite::fromJSON(content)
  result <- parsed_response$choices[[1]]$text

  return(result)
}

# In get_provider.R
get_provider <- function(model) {
  # Add to the model mapping
  model_mapping <- list(
    # Existing models...
    "newprovider-model" = "newprovider"
  )

  provider <- model_mapping[[model]]
  if (is.null(provider)) {
    stop("Unsupported model: ", model)
  }

  return(provider)
}
```

### Improving Documentation

Documentation improvements are always welcome:

1. Update function documentation with roxygen2
2. Improve vignettes with more examples and explanations
3. Add tutorials for specific use cases
4. Fix typos and clarify existing documentation

### Adding New Features

Some ideas for new features:

1. Integration with additional single-cell analysis frameworks
2. Support for spatial transcriptomics data
3. Interactive visualization tools
4. Batch processing for large datasets
5. Performance optimizations

### Reporting Issues

When reporting issues, please include:

1. A minimal reproducible example
2. The version of mLLMCelltype you're using
3. The error message or unexpected behavior
4. Your R session information (`sessionInfo()`)

## Pull Request Process

1. Ensure your code follows the style guidelines
2. Add or update tests as necessary
3. Update documentation to reflect your changes
4. Ensure all tests pass
5. Submit your pull request with a clear description of the changes

## Code Review Process

All pull requests will be reviewed by the maintainers. The review process includes:

1. Checking that the code follows style guidelines
2. Verifying that tests pass
3. Ensuring documentation is updated
4. Evaluating the overall design and implementation

## Release Process

mLLMCelltype follows semantic versioning (MAJOR.MINOR.PATCH):

- MAJOR version for incompatible API changes
- MINOR version for new functionality in a backward-compatible manner
- PATCH version for backward-compatible bug fixes

## Community Guidelines

### Code of Conduct

We follow a code of conduct to ensure a welcoming and inclusive community:

- Be respectful and inclusive
- Be collaborative
- Be open to feedback
- Focus on the best solution for the community

### Communication Channels

- GitHub Issues: For bug reports, feature requests, and discussions
- GitHub Discussions: For general questions and community discussions
- Pull Requests: For code contributions

## Acknowledgment

Contributors will be acknowledged in the package documentation and README.

## License

By contributing to mLLMCelltype, you agree that your contributions will be licensed under the same license as the project (MIT License).

## Next Steps

Now that you know how to contribute to mLLMCelltype, you can:

- [Review the version history](https://cafferyang.com/mLLMCelltype/articles/version-history.html) to understand recent changes
- [Explore advanced features](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html) to identify areas for improvement
- [Check the FAQ](https://cafferyang.com/mLLMCelltype/articles/faq.html) to see common questions that might need better documentation

---
title: "Consensus vs Single-Agent: A Methodology Comparison"
description: "Comparison of multi-LLM consensus and single-agent approaches for cell type annotation, covering architecture, trade-offs, and practical considerations."
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Consensus vs Single-Agent: A Methodology Comparison}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Consensus vs Single-Agent: A Methodology Comparison

This vignette describes the architectural differences between multi-LLM consensus and single-agent approaches for cell type annotation, along with their respective trade-offs.

## Architectural Overview

### Single-Agent Systems

Single-agent approaches use specialized AI systems with predefined roles:

- **Specialized roles**: Different agents handle specific annotation tasks
- **Sequential processing**: Agents work in a pipeline fashion
- **Role-based validation**: Quality control through dedicated validation agents
- **Structured workflows**: Fixed processing sequences

### Multi-LLM Consensus Systems

Consensus frameworks leverage multiple independent models:

- **Parallel processing**: Multiple models analyze simultaneously
- **Collective decision-making**: Annotations are determined by agreement across models
- **Iterative refinement**: Discussion rounds for clusters where models disagree
- **Adaptive complexity**: More discussion is allocated to difficult cases

## Methodological Differences

### Single-Agent Approach

**Strengths:**
- Clear role definition: Each agent has specific responsibilities
- Streamlined workflows: Predictable processing pipelines
- Focused optimization: Agents can be fine-tuned for specific tasks
- Lower initial complexity: Easier to implement and understand

**Limitations:**
- Sequential bottlenecks: Failure in one agent affects the entire pipeline
- Limited model diversity: Typically relies on one underlying LLM family
- Rigid processing: Difficult to adapt to edge cases
- Single point of failure: Agent malfunction can compromise results

### Consensus Approach

**Strengths:**
- Error correction: Multiple models can catch each other's mistakes
- Model diversity: Leverages different training approaches and strengths
- Adaptive processing: More resources allocated to difficult cases
- Transparent uncertainty: Clear metrics for prediction confidence

**Challenges:**
- Initial complexity: Requires coordination between multiple models
- Resource coordination: Must manage multiple API calls efficiently
- Consensus building: Additional time for deliberation processes
- Model compatibility: Ensuring different models work together effectively

## Performance

For benchmark results comparing the two approaches, see Yang et al. (2026):

Yang, C., Zhang, X., & Chen, J. (2026). Large language model consensus substantially improves the cell type annotation accuracy for scRNA-seq data. *Communications Biology*. https://doi.org/10.1038/s42003-026-10420-8

## Cost and Resource Trade-offs

| Aspect | Single-Agent | Consensus |
|--------|-------------|-----------|
| **API calls per cluster** | Fewer | More (multiple models) |
| **Cost per run** | Lower | Higher per run |
| **Two-stage optimization** | N/A | Reduces calls when models agree early |
| **Scalability** | Good | Good, with caching support |

The two-stage consensus approach in mLLMCelltype can reduce API calls when models agree early, since only clusters without initial consensus proceed to the deliberation stage.

## Practical Considerations

### When single-agent approaches may suffice:
- Standardized datasets with well-characterized tissues
- High-throughput screening of many similar datasets
- Limited API budget
- Straightforward annotation tasks

### When consensus approaches may be preferable:
- Novel biological contexts where model agreement provides additional confidence
- Work intended for publication, where uncertainty quantification is useful
- Complex tissues with many similar cell types
- Cases where identifying uncertain annotations is important

## Hybrid Approaches

Advanced workflows can combine both approaches:

1. **Initial screening**: Single-agent for clear cases
2. **Consensus validation**: Multi-model for uncertain cases
3. **Expert review**: Human validation for critical decisions

This tiered approach balances cost and thoroughness.

## Summary

Both methodologies have distinct strengths. Single-agent systems are simpler and less expensive per run, while consensus approaches provide uncertainty quantification and cross-model validation. The choice depends on the specific requirements of accuracy, cost, and biological complexity for a given project.

## Next Steps

- [Why Choose Consensus?](https://cafferyang.com/mLLMCelltype/articles/why-consensus.html) - Details on the consensus methodology
- [Getting Started Guide](https://cafferyang.com/mLLMCelltype/articles/getting-started.html) - Practical implementation tutorial
- [Performance Benchmarks](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html) - Detailed comparisons
- [API Reference](https://cafferyang.com/mLLMCelltype/reference/index.html) - Technical documentation

---
title: "Consensus Annotation Principles"
author: "Chen Yang"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Consensus Annotation Principles}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE, purl=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  eval = FALSE
)
```

# Consensus Annotation Principles

This article explains the technical principles behind mLLMCelltype's consensus annotation approach. Understanding these principles will help you make better use of the package and interpret the results more effectively.

## The Multi-LLM Consensus Architecture

### Why Multiple Models?

Single large language models (LLMs) can produce impressive results for cell type annotation, but they also have limitations:

1. **Knowledge gaps**: No single model has perfect knowledge of all cell types across all tissues
2. **Hallucinations**: LLMs can sometimes generate plausible-sounding but incorrect annotations
3. **Biases**: Each model has biases based on its training data and architecture
4. **Inconsistency**: The same model may give different answers to the same question

The multi-LLM consensus approach addresses these limitations by leveraging the complementary strengths of different models. This is similar to how a panel of experts might collaborate to reach a more reliable conclusion than any single expert could provide alone.

### Model Diversity

mLLMCelltype deliberately incorporates models with different architectures and training data:

- **Anthropic Claude models**: Known for careful reasoning and biological knowledge
- **OpenAI GPT models**: Strong general knowledge and pattern recognition
- **Google Gemini models**: Good at integrating information from multiple sources
- **X.AI Grok models**: Newer architecture with different training approach
- **Other models**: Provide additional diversity in reasoning approaches

This diversity is crucial for the consensus mechanism to work effectively. Models with different "perspectives" can catch each other's errors and provide complementary insights.

## The Structured Deliberation Process

The consensus formation in mLLMCelltype follows a structured deliberation process:

### 1. Initial Independent Annotation

Each model independently annotates the cell clusters based on the marker genes:

```{r, purl=FALSE}
# Conceptual representation of the initial annotation process
initial_results <- list()
for (model in models) {
  initial_results[[model]] <- annotate_cell_types(
    input = marker_data,
    tissue_name = tissue_name,
    model = model,
    api_key = api_keys[[get_provider(model)]]
  )
}
```

This step ensures that each model forms its own opinion without being influenced by others, similar to how jurors might form initial opinions before deliberation.

### 2. Identification of Controversial Clusters

The system identifies clusters where there is significant disagreement among models:

```{r, purl=FALSE}
# Conceptual representation of controversial cluster identification
controversial_clusters <- identify_controversial_clusters(
  initial_results,
  threshold = discussion_threshold
)
```

A cluster is considered "controversial" if the proportion of models agreeing with the most common annotation is below a certain threshold (default: 0.5). This means that if less than half of the models agree on the cell type, the cluster requires further discussion.

### 3. Structured Discussion for Controversial Clusters

For controversial clusters, the system initiates a structured discussion process:

```{r, purl=FALSE}
# Conceptual representation of the discussion process
discussion_results <- facilitate_cluster_discussion(
  controversial_clusters,
  initial_results,
  marker_data,
  tissue_name,
  discussion_model,
  api_key
)
```

This discussion follows a specific format:

1. **Initial positions**: Each model's initial annotation and reasoning are presented
2. **Evidence evaluation**: The discussion model evaluates the evidence for each proposed cell type
3. **Counter-arguments**: Potential weaknesses in each argument are identified
4. **Synthesis**: The discussion model synthesizes the arguments to reach a conclusion

This structured approach mimics how human experts might deliberate on a difficult case, considering multiple perspectives and critically evaluating the evidence.

### 4. Final Consensus Formation

After discussion, the system forms a final consensus for all clusters:

```{r, purl=FALSE}
# Conceptual representation of consensus formation
final_annotations <- combine_results(
  initial_results,
  discussion_results,
  controversial_clusters
)
```

For non-controversial clusters, the most common annotation among the initial results is used. For controversial clusters, the result from the structured discussion is used.

## Uncertainty Quantification

A key feature of mLLMCelltype is its transparent uncertainty quantification:

### Consensus Proportion

The consensus proportion measures the level of agreement among models:

```{r, purl=FALSE}
# Conceptual calculation of consensus proportion
consensus_proportion <- sapply(clusters, function(cluster) {
  annotations <- sapply(models, function(model) initial_results[[model]][cluster])
  most_common <- names(which.max(table(annotations)))
  sum(annotations == most_common) / length(annotations)
})
```

This metric ranges from 0 to 1:
- **1.0**: Perfect agreement (all models agree)
- **0.5**: Moderate agreement (half of the models agree)
- **< 0.5**: Low agreement (less than half of the models agree)

The consensus proportion helps identify which annotations are more reliable and which might require further investigation.

### Shannon Entropy

Shannon entropy quantifies the uncertainty in the annotations:

```{r, purl=FALSE}
# Conceptual calculation of Shannon entropy
shannon_entropy <- sapply(clusters, function(cluster) {
  annotations <- sapply(models, function(model) initial_results[[model]][cluster])
  p <- table(annotations) / length(annotations)
  -sum(p * log2(p))
})
```

Shannon entropy is a measure from information theory:
- **0**: No uncertainty (all models give the same answer)
- **Higher values**: More uncertainty (models give diverse answers)

Unlike consensus proportion, Shannon entropy captures the full distribution of annotations, not just the most common one. This makes it particularly useful for identifying clusters with high uncertainty.

## Hallucination Reduction Mechanisms

mLLMCelltype incorporates several mechanisms to reduce hallucinations:

### Cross-Model Verification

By requiring multiple independent models to agree, the system naturally filters out many hallucinations. A hallucinated annotation from one model is unlikely to be independently hallucinated by other models.

### Evidence-Based Reasoning

The structured discussion process explicitly requires models to ground their annotations in the marker gene evidence. This reduces the likelihood of hallucinations that aren't supported by the data.

### Critical Evaluation

During the discussion process, models critically evaluate each other's reasoning. This helps identify and correct potential hallucinations or reasoning errors.

## Robustness to Input Noise

mLLMCelltype is designed to be robust to noise in the input data:

### Collective Error Correction

Even if some marker genes are noisy or misleading, the consensus approach can still reach the correct conclusion if enough models can identify the true signal in the data.

### Focus on Strong Signals

By using the `top_gene_count` parameter, the system focuses on the strongest marker genes, which are less likely to be noise.

### Uncertainty Flagging

When the input data is too noisy to make a reliable annotation, the system will show low consensus proportion and high Shannon entropy, flagging the cluster for human review.

## Technical Implementation Details

### Prompt Engineering

The prompts used in mLLMCelltype are carefully designed to:

1. **Structure the reasoning process**: Guide models through a step-by-step analysis
2. **Enforce evidence-based reasoning**: Require models to cite specific marker genes
3. **Encourage critical thinking**: Ask models to consider alternative explanations
4. **Standardize output format**: Ensure consistent, parseable responses

Here's a simplified example of the annotation prompt structure:

```
You are an expert in single-cell RNA sequencing analysis.

TASK:
Identify the cell type for a cluster based on its marker genes.

MARKER GENES:
[List of marker genes with fold changes and p-values]

TISSUE:
[Tissue name]

STEPS:
1. Analyze the marker genes and their expression levels
2. Identify key cell type-specific markers
3. Consider multiple possible cell types
4. Determine the most likely cell type based on the evidence
5. Provide your reasoning

OUTPUT FORMAT:
Cell Type: [Your answer]
Reasoning: [Your step-by-step reasoning]
Confidence: [High/Medium/Low]
```

### Discussion Orchestration

The discussion process is orchestrated by a single "discussion model" (typically Claude) that:

1. Presents each model's initial annotation and reasoning
2. Evaluates the evidence for each proposed cell type
3. Identifies potential weaknesses in each argument
4. Synthesizes the arguments to reach a conclusion

This approach allows for a coherent discussion while still incorporating the diverse perspectives of multiple models.

## Comparison with Other Approaches

### vs. Single LLM Annotation

Compared to using a single LLM:
- **Advantages**: Higher accuracy, uncertainty quantification, reduced hallucinations
- **Disadvantages**: Higher computational cost, more complex implementation, requires multiple API keys

### vs. Traditional Annotation Methods

Compared to traditional methods (e.g., reference-based, marker-based):
- **Advantages**: No reference dataset required, more flexible, captures rare cell types, provides reasoning
- **Disadvantages**: Depends on LLM knowledge, potentially higher cost, requires internet connection

### vs. Human Expert Annotation

Compared to human expert annotation:
- **Advantages**: Faster, more scalable, consistent methodology, transparent reasoning
- **Disadvantages**: May miss novel cell types not in literature, lacks domain-specific expertise for very specialized tissues

## Practical Implications

Understanding these principles has several practical implications for using mLLMCelltype:

1. **Model selection matters**: Including diverse, high-quality models improves consensus
2. **Uncertainty metrics are valuable**: Pay attention to consensus proportion and Shannon entropy
3. **Discussion logs provide insight**: Review discussion logs for controversial clusters
4. **Input quality affects results**: Better marker gene data leads to more reliable annotations

## Next Steps

Now that you understand the technical principles behind mLLMCelltype, you can explore:

- [Visualization Guide](https://cafferyang.com/mLLMCelltype/articles/visualization-guide.html): Learn how to visualize consensus and uncertainty
- [FAQ](https://cafferyang.com/mLLMCelltype/articles/faq.html): Find answers to common questions
- [Advanced Features](https://cafferyang.com/mLLMCelltype/articles/advanced-features.html): Explore hierarchical annotation and other advanced features
- [Contributing Guide](https://cafferyang.com/mLLMCelltype/articles/contributing-guide.html): Learn how to contribute to the project

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_consensus.R
\name{execute_consensus_check}
\alias{execute_consensus_check}
\title{Execute consensus check across ordered model candidates}
\usage{
execute_consensus_check(
  formatted_responses,
  api_keys,
  models_to_try,
  base_urls = NULL
)
}
\arguments{
\item{formatted_responses}{Prompt containing model responses}

\item{api_keys}{Named API key list}

\item{models_to_try}{Ordered model candidates}

\item{base_urls}{Optional provider base URLs}
}
\value{
A list containing success status and response text
}
\description{
Execute consensus check across ordered model candidates
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus_annotation.R
\name{combine_results}
\alias{combine_results}
\title{Combine results from all phases of consensus annotation}
\usage{
combine_results(initial_results, controversy_results, discussion_results)
}
\description{
Combine results from all phases of consensus annotation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_consensus.R
\name{extract_labeled_value}
\alias{extract_labeled_value}
\title{Extract numeric value from line containing a label}
\usage{
extract_labeled_value(lines, label_pattern)
}
\description{
Extract numeric value from line containing a label
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_consensus.R
\name{parse_consensus_response}
\alias{parse_consensus_response}
\title{Parse consensus response from model}
\usage{
parse_consensus_response(response)
}
\description{
Parse consensus response from model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_consensus.R
\name{check_consensus}
\alias{check_consensus}
\title{Check if consensus is reached among models}
\usage{
check_consensus(
  round_responses,
  api_keys = NULL,
  controversy_threshold = 2/3,
  entropy_threshold = 1,
  consensus_check_model = NULL,
  base_urls = NULL
)
}
\description{
Check if consensus is reached among models
}
\note{
This function uses create_consensus_check_prompt from prompt_templates.R
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_utils.R
\name{get_api_key}
\alias{get_api_key}
\title{Get an API key for a model}
\usage{
get_api_key(model, api_keys)
}
\arguments{
\item{model}{Model name to get API key for}

\item{api_keys}{Named list of API keys with provider or model names as keys}
}
\value{
A trimmed API key string, or \code{NULL} when no valid key is configured.
}
\description{
Retrieves a configured API key by checking the model's provider name first,
followed by the exact model name.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus_annotation.R
\name{clean_annotation}
\alias{clean_annotation}
\title{Clean annotation text by removing prefixes and extra whitespace}
\usage{
clean_annotation(annotation)
}
\description{
Clean annotation text by removing prefixes and extra whitespace
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url_utils.R
\name{resolve_provider_base_url}
\alias{resolve_provider_base_url}
\title{Resolve provider-specific base URL}
\usage{
resolve_provider_base_url(provider, base_urls)
}
\arguments{
\item{provider}{Provider name (e.g., "openai", "anthropic")}

\item{base_urls}{User-provided base URLs: NULL, a single string, or a named list}
}
\value{
Resolved and normalized base URL, or NULL if not specified
}
\description{
This is the single entry point for all base URL resolution. It resolves
the appropriate URL and normalizes it (strips trailing slashes).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus_annotation.R
\name{parse_text_predictions}
\alias{parse_text_predictions}
\title{Parse text-format model predictions into a named list}
\usage{
parse_text_predictions(model_preds, all_clusters = NULL)
}
\arguments{
\item{model_preds}{Character vector of prediction lines from a model}

\item{all_clusters}{Optional character vector of cluster IDs for positional fallback}
}
\value{
Named list mapping cluster IDs to cell type annotations
}
\description{
Reads a response as explicit cluster labels (keyed by resolved cluster ID,
so out-of-order labels land correctly) when at least one label resolves,
unless an unresolved cluster-reference key coexists with incomplete coverage
-- that signals list ordinals misaligned against the requested clusters and
falls through to positional mapping. Positional mapping preserves the
line->cluster slot correspondence so a mid-list "Unknown" does not shift
later clusters, and a stray "Summary:"/"Note:" line does not hijack the parse.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prompt_templates.R
\name{create_standardization_prompt}
\alias{create_standardization_prompt}
\title{Create prompt for standardizing cell type names}
\usage{
create_standardization_prompt(all_cell_types)
}
\description{
Create prompt for standardizing cell type names
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_consensus.R
\name{calculate_simple_consensus}
\alias{calculate_simple_consensus}
\title{Calculate simple consensus without LLM}
\usage{
calculate_simple_consensus(round_responses)
}
\description{
Calculate simple consensus without LLM
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_consensus.R
\name{parse_standard_format}
\alias{parse_standard_format}
\title{Parse standard 4-line consensus response format}
\usage{
parse_standard_format(result_lines)
}
\description{
Parse standard 4-line consensus response format
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unified_logger.R
\name{UnifiedLogger}
\alias{UnifiedLogger}
\title{Unified Logger for mLLMCelltype Package}
\description{
Unified Logger for mLLMCelltype Package

Unified Logger for mLLMCelltype Package
}
\details{
This logger provides centralized, multi-level logging with structured output,
log rotation, and performance monitoring capabilities.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{log_dir}}{Directory for storing log files}

\item{\code{log_level}}{Current logging level}

\item{\code{session_id}}{Unique identifier for the current session}

\item{\code{max_log_size}}{Maximum log file size in MB (default: 10MB)}

\item{\code{max_log_files}}{Maximum number of log files to keep (default: 5)}

\item{\code{enable_console}}{Whether to output to console (default: FALSE)}

\item{\code{enable_json}}{Whether to use JSON format (default: TRUE)}

\item{\code{performance_stats}}{Performance monitoring statistics}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-UnifiedLogger-new}{\code{UnifiedLogger$new()}}
\item \href{#method-UnifiedLogger-debug}{\code{UnifiedLogger$debug()}}
\item \href{#method-UnifiedLogger-info}{\code{UnifiedLogger$info()}}
\item \href{#method-UnifiedLogger-warn}{\code{UnifiedLogger$warn()}}
\item \href{#method-UnifiedLogger-error}{\code{UnifiedLogger$error()}}
\item \href{#method-UnifiedLogger-log_api_call}{\code{UnifiedLogger$log_api_call()}}
\item \href{#method-UnifiedLogger-log_api_request_response}{\code{UnifiedLogger$log_api_request_response()}}
\item \href{#method-UnifiedLogger-log_cache_operation}{\code{UnifiedLogger$log_cache_operation()}}
\item \href{#method-UnifiedLogger-log_cluster_progress}{\code{UnifiedLogger$log_cluster_progress()}}
\item \href{#method-UnifiedLogger-log_discussion}{\code{UnifiedLogger$log_discussion()}}
\item \href{#method-UnifiedLogger-log_model_response}{\code{UnifiedLogger$log_model_response()}}
\item \href{#method-UnifiedLogger-get_performance_summary}{\code{UnifiedLogger$get_performance_summary()}}
\item \href{#method-UnifiedLogger-cleanup_logs}{\code{UnifiedLogger$cleanup_logs()}}
\item \href{#method-UnifiedLogger-set_level}{\code{UnifiedLogger$set_level()}}
\item \href{#method-UnifiedLogger-clone}{\code{UnifiedLogger$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-new"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-new}{}}}
\subsection{Method \code{new()}}{
Initialize the unified logger
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$new(
  base_dir = "logs",
  level = "INFO",
  max_size = 10,
  max_files = 5,
  console_output = FALSE,
  json_format = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{base_dir}}{Directory for log files}

\item{\code{level}}{Minimum log level}

\item{\code{max_size}}{Maximum main log size in megabytes}

\item{\code{max_files}}{Maximum number of retained main logs}

\item{\code{console_output}}{Whether to mirror logs to the console}

\item{\code{json_format}}{Whether main logs use JSON format}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-debug"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-debug}{}}}
\subsection{Method \code{debug()}}{
Log a debug message
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$debug(message, context = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{message}}{Log message}

\item{\code{context}}{Optional structured context}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-info"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-info}{}}}
\subsection{Method \code{info()}}{
Log an info message
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$info(message, context = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{message}}{Log message}

\item{\code{context}}{Optional structured context}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-warn"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-warn}{}}}
\subsection{Method \code{warn()}}{
Log a warning message
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$warn(message, context = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{message}}{Log message}

\item{\code{context}}{Optional structured context}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-error"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-error}{}}}
\subsection{Method \code{error()}}{
Log an error message
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$error(message, context = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{message}}{Log message}

\item{\code{context}}{Optional structured context}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-log_api_call"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-log_api_call}{}}}
\subsection{Method \code{log_api_call()}}{
Log API call performance
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$log_api_call(
  provider,
  model,
  duration,
  success = TRUE,
  tokens = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{provider}}{Provider identifier}

\item{\code{model}}{Model identifier}

\item{\code{duration}}{Request duration in seconds}

\item{\code{success}}{Whether the request succeeded}

\item{\code{tokens}}{Optional token usage metadata}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-log_api_request_response"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-log_api_request_response}{}}}
\subsection{Method \code{log_api_request_response()}}{
Log complete API request and response for debugging and audit
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$log_api_request_response(
  provider,
  model,
  prompt_content,
  response_content,
  request_metadata = NULL,
  response_metadata = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{provider}}{Provider identifier}

\item{\code{model}}{Model identifier}

\item{\code{prompt_content}}{Request prompt}

\item{\code{response_content}}{Provider response or error text}

\item{\code{request_metadata}}{Optional request metadata}

\item{\code{response_metadata}}{Optional response metadata}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-log_cache_operation"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-log_cache_operation}{}}}
\subsection{Method \code{log_cache_operation()}}{
Log cache operations
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$log_cache_operation(operation, key, size = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{operation}}{Cache operation name}

\item{\code{key}}{Cache key}

\item{\code{size}}{Optional cache object size in bytes}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-log_cluster_progress"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-log_cluster_progress}{}}}
\subsection{Method \code{log_cluster_progress()}}{
Log cluster annotation progress
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$log_cluster_progress(cluster_id, stage, progress = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cluster_id}}{Cluster identifier}

\item{\code{stage}}{Processing stage}

\item{\code{progress}}{Optional progress value}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-log_discussion"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-log_discussion}{}}}
\subsection{Method \code{log_discussion()}}{
Log detailed cluster discussion with complete model conversations
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$log_discussion(cluster_id, event_type, data = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cluster_id}}{Cluster identifier}

\item{\code{event_type}}{Discussion event type}

\item{\code{data}}{Optional event payload}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-log_model_response"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-log_model_response}{}}}
\subsection{Method \code{log_model_response()}}{
Log model response with concise summary in main log and full text in file
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$log_model_response(
  provider,
  model,
  response,
  stage = "annotation",
  cluster_id = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{provider}}{Provider identifier}

\item{\code{model}}{Model identifier}

\item{\code{response}}{Model response}

\item{\code{stage}}{Processing stage}

\item{\code{cluster_id}}{Optional cluster identifier}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-get_performance_summary"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-get_performance_summary}{}}}
\subsection{Method \code{get_performance_summary()}}{
Get performance summary
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$get_performance_summary()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-cleanup_logs"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-cleanup_logs}{}}}
\subsection{Method \code{cleanup_logs()}}{
Clean up old log files
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$cleanup_logs(force = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{force}}{Whether to remove every main log file}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-set_level"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-set_level}{}}}
\subsection{Method \code{set_level()}}{
Set logging level
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$set_level(level)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{level}}{Minimum log level}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-UnifiedLogger-clone"></a>}}
\if{latex}{\out{\hypertarget{method-UnifiedLogger-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{UnifiedLogger$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus_annotation.R
\name{process_controversial_clusters}
\alias{process_controversial_clusters}
\title{Process controversial clusters through discussion}
\usage{
process_controversial_clusters(
  controversial_clusters,
  input,
  tissue_name,
  successful_models,
  api_keys,
  individual_predictions,
  top_gene_count,
  controversy_threshold,
  entropy_threshold,
  max_discussion_rounds,
  cache_manager,
  use_cache,
  consensus_check_model = NULL,
  force_rerun = FALSE,
  base_urls = NULL
)
}
\description{
Process controversial clusters through discussion
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prompt_templates.R
\name{create_reasoning_annotation_prompt}
\alias{create_reasoning_annotation_prompt}
\title{Create reasoning-aware prompt for cell type annotation}
\usage{
create_reasoning_annotation_prompt(input, tissue_name, top_gene_count = 10)
}
\arguments{
\item{input}{Either a data frame from Seurat's FindAllMarkers() or a list for each cluster
where each element is either a character vector of genes or a list containing a \code{genes} field.}

\item{tissue_name}{Tissue context for the annotation (e.g., 'human PBMC', 'mouse brain')}

\item{top_gene_count}{Number of top genes to use per cluster when input is from Seurat. Default: 10}
}
\value{
A list with \code{prompt} (formatted prompt text), \code{expected_count}
(number of clusters), and \code{gene_lists} (cluster ID to marker genes mapping).
}
\description{
Create reasoning-aware prompt for cell type annotation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_provider.R
\name{get_provider}
\alias{get_provider}
\title{Determine provider from model name}
\usage{
get_provider(model)
}
\arguments{
\item{model}{Character string specifying the model name (e.g., "gpt-5.5", "claude-opus-4-7").}
}
\value{
Character string of the provider name (e.g., "openai", "anthropic").
}
\description{
This function determines the appropriate provider (e.g., OpenAI, Anthropic, Google, OpenRouter) based on the model name.
Uses prefix-based matching for efficient and maintainable provider detection.
New models following existing naming conventions are automatically supported.
}
\details{
Supported providers and model prefixes:
\itemize{
\item OpenAI: gpt-\emph{, o1}, o3*, o4*, chatgpt-\emph{, codex-} (e.g., 'gpt-5.5', 'gpt-5.4-mini')
\item Anthropic: claude-* (e.g., 'claude-opus-4-7', 'claude-sonnet-4-6')
\item DeepSeek: deepseek-* (e.g., 'deepseek-v4-flash', 'deepseek-v4-pro')
\item Google: gemini-* (e.g., 'gemini-3.1-pro-preview', 'gemini-3-flash-preview')
\item Qwen: qwen*, qwq-* (e.g., 'qwen3.6-plus', 'qwen3.6-flash')
\item Stepfun: step-* (e.g., 'step-3.5-flash', 'step-3')
\item Zhipu: glm-\emph{, chatglm} (e.g., 'glm-5.1', 'glm-5-turbo')
\item MiniMax: minimax-* (e.g., 'MiniMax-M2.7', 'MiniMax-M2.5')
\item Grok: grok-* (e.g., 'grok-4.3', 'grok-4.3-latest')
\item Kimi (Moonshot AI Open Platform): kimi-\emph{, moonshot-} (e.g., 'kimi-k2.6', 'moonshot-v1-8k')
\item OpenRouter: Any model with '/' in the name (e.g., 'openai/gpt-5.5', 'anthropic/claude-opus-4.7')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cell_type_annotation.R
\docType{package}
\name{mLLMCelltype-package}
\alias{mLLMCelltype}
\alias{mLLMCelltype-package}
\title{mLLMCelltype: Cell Type Annotation Using Large Language Models}
\description{
Automated cell type annotation for single-cell RNA sequencing data using consensus predictions from multiple large language models. Integrates with Seurat objects and provides uncertainty quantification for annotations. Supports various LLM providers including OpenAI, Anthropic, and Google. For details see Yang et al. (2026) \doi{10.1038/s42003-026-10420-8}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://cafferyang.com/mLLMCelltype/}
  \item Report bugs at \url{https://github.com/cafferychen777/mLLMCelltype/issues}
}

}
\author{
\strong{Maintainer}: Chen Yang \email{cafferychen777@tamu.edu} [copyright holder]

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/base_api_processor.R
\name{BaseAPIProcessor}
\alias{BaseAPIProcessor}
\title{Base API Processor Class}
\description{
Base API Processor Class

Base API Processor Class
}
\details{
Abstract base class for API processors that provides common functionality
including unified logging, error handling, input processing, and response validation.
This eliminates code duplication across all provider-specific processors.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{provider_name}}{Name of the API provider}

\item{\code{logger}}{Unified logger instance}

\item{\code{base_url}}{Custom base URL for API endpoints}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-BaseAPIProcessor-new}{\code{BaseAPIProcessor$new()}}
\item \href{#method-BaseAPIProcessor-process_request}{\code{BaseAPIProcessor$process_request()}}
\item \href{#method-BaseAPIProcessor-get_api_url}{\code{BaseAPIProcessor$get_api_url()}}
\item \href{#method-BaseAPIProcessor-get_default_api_url}{\code{BaseAPIProcessor$get_default_api_url()}}
\item \href{#method-BaseAPIProcessor-make_api_call}{\code{BaseAPIProcessor$make_api_call()}}
\item \href{#method-BaseAPIProcessor-extract_response_content}{\code{BaseAPIProcessor$extract_response_content()}}
\item \href{#method-BaseAPIProcessor-extract_usage}{\code{BaseAPIProcessor$extract_usage()}}
\item \href{#method-BaseAPIProcessor-clone}{\code{BaseAPIProcessor$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BaseAPIProcessor-new"></a>}}
\if{latex}{\out{\hypertarget{method-BaseAPIProcessor-new}{}}}
\subsection{Method \code{new()}}{
Initialize the base API processor
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BaseAPIProcessor$new(provider_name, base_url = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{provider_name}}{Provider identifier used for logging and dispatch}

\item{\code{base_url}}{Optional custom API endpoint}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BaseAPIProcessor-process_request"></a>}}
\if{latex}{\out{\hypertarget{method-BaseAPIProcessor-process_request}{}}}
\subsection{Method \code{process_request()}}{
Main entry point for processing API requests
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BaseAPIProcessor$process_request(prompt, model, api_key, normalize = TRUE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{Prompt text to send}

\item{\code{model}}{Model identifier}

\item{\code{api_key}}{Provider API key}

\item{\code{normalize}}{Logical. If \code{TRUE} (default), the response is
normalized into non-empty trimmed lines. If \code{FALSE}, the raw
response string is returned instead. Set to \code{FALSE} when the
caller needs the original text (e.g., JSON parsing in reasoning mode).}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BaseAPIProcessor-get_api_url"></a>}}
\if{latex}{\out{\hypertarget{method-BaseAPIProcessor-get_api_url}{}}}
\subsection{Method \code{get_api_url()}}{
Get the API URL to use for requests
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BaseAPIProcessor$get_api_url()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BaseAPIProcessor-get_default_api_url"></a>}}
\if{latex}{\out{\hypertarget{method-BaseAPIProcessor-get_default_api_url}{}}}
\subsection{Method \code{get_default_api_url()}}{
Abstract method to be implemented by subclasses for getting default API URL
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BaseAPIProcessor$get_default_api_url()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BaseAPIProcessor-make_api_call"></a>}}
\if{latex}{\out{\hypertarget{method-BaseAPIProcessor-make_api_call}{}}}
\subsection{Method \code{make_api_call()}}{
Abstract method to be implemented by subclasses for making the actual API call
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BaseAPIProcessor$make_api_call(chunk_content, model, api_key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{chunk_content}}{Prompt text to send}

\item{\code{model}}{Model identifier}

\item{\code{api_key}}{Provider API key}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BaseAPIProcessor-extract_response_content"></a>}}
\if{latex}{\out{\hypertarget{method-BaseAPIProcessor-extract_response_content}{}}}
\subsection{Method \code{extract_response_content()}}{
Abstract method to be implemented by subclasses for extracting content from response
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BaseAPIProcessor$extract_response_content(response, model)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{response}}{HTTP response object}

\item{\code{model}}{Model identifier}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BaseAPIProcessor-extract_usage"></a>}}
\if{latex}{\out{\hypertarget{method-BaseAPIProcessor-extract_usage}{}}}
\subsection{Method \code{extract_usage()}}{
Extract normalized token usage from a provider response
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BaseAPIProcessor$extract_usage(response)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{response}}{HTTP response object
Make API call and extract response content}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BaseAPIProcessor-clone"></a>}}
\if{latex}{\out{\hypertarget{method-BaseAPIProcessor-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BaseAPIProcessor$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prompt_templates.R
\name{create_discussion_prompt}
\alias{create_discussion_prompt}
\title{Create prompt for additional discussion rounds}
\usage{
create_discussion_prompt(
  cluster_id,
  cluster_genes,
  tissue_name,
  previous_rounds,
  round_number
)
}
\description{
Create prompt for additional discussion rounds
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facilitate_cluster_discussion.R
\name{filter_valid_responses}
\alias{filter_valid_responses}
\title{Filter out error responses from model round responses}
\usage{
filter_valid_responses(responses, cluster_id, round = NULL)
}
\description{
Filter out error responses from model round responses
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_consensus.R
\name{prepare_models_list}
\alias{prepare_models_list}
\title{Prepare list of models to try for consensus checking}
\usage{
prepare_models_list(consensus_check_model = NULL)
}
\description{
Prepare list of models to try for consensus checking
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deepseek_processor.R
\name{DeepSeekProcessor}
\alias{DeepSeekProcessor}
\title{DeepSeek API Processor}
\description{
DeepSeek API Processor

DeepSeek API Processor
}
\details{
Concrete implementation of BaseAPIProcessor for DeepSeek models.
Handles DeepSeek-specific API calls, authentication, and response parsing.
}
\section{Super class}{
\code{\link[mLLMCelltype:BaseAPIProcessor]{mLLMCelltype::BaseAPIProcessor}} -> \code{DeepSeekProcessor}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-DeepSeekProcessor-new}{\code{DeepSeekProcessor$new()}}
\item \href{#method-DeepSeekProcessor-get_default_api_url}{\code{DeepSeekProcessor$get_default_api_url()}}
\item \href{#method-DeepSeekProcessor-make_api_call}{\code{DeepSeekProcessor$make_api_call()}}
\item \href{#method-DeepSeekProcessor-extract_response_content}{\code{DeepSeekProcessor$extract_response_content()}}
\item \href{#method-DeepSeekProcessor-clone}{\code{DeepSeekProcessor$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="mLLMCelltype" data-topic="BaseAPIProcessor" data-id="extract_usage"><a href='../../mLLMCelltype/html/BaseAPIProcessor.html#method-BaseAPIProcessor-extract_usage'><code>mLLMCelltype::BaseAPIProcessor$extract_usage()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mLLMCelltype" data-topic="BaseAPIProcessor" data-id="get_api_url"><a href='../../mLLMCelltype/html/BaseAPIProcessor.html#method-BaseAPIProcessor-get_api_url'><code>mLLMCelltype::BaseAPIProcessor$get_api_url()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mLLMCelltype" data-topic="BaseAPIProcessor" data-id="process_request"><a href='../../mLLMCelltype/html/BaseAPIProcessor.html#method-BaseAPIProcessor-process_request'><code>mLLMCelltype::BaseAPIProcessor$process_request()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DeepSeekProcessor-new"></a>}}
\if{latex}{\out{\hypertarget{method-DeepSeekProcessor-new}{}}}
\subsection{Method \code{new()}}{
Initialize DeepSeek processor
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DeepSeekProcessor$new(base_url = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{base_url}}{Optional custom API endpoint}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DeepSeekProcessor-get_default_api_url"></a>}}
\if{latex}{\out{\hypertarget{method-DeepSeekProcessor-get_default_api_url}{}}}
\subsection{Method \code{get_default_api_url()}}{
Get default DeepSeek API URL
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DeepSeekProcessor$get_default_api_url()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DeepSeekProcessor-make_api_call"></a>}}
\if{latex}{\out{\hypertarget{method-DeepSeekProcessor-make_api_call}{}}}
\subsection{Method \code{make_api_call()}}{
Make API call to DeepSeek
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DeepSeekProcessor$make_api_call(chunk_content, model, api_key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{chunk_content}}{Prompt text to send}

\item{\code{model}}{Model identifier}

\item{\code{api_key}}{DeepSeek API key}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DeepSeekProcessor-extract_response_content"></a>}}
\if{latex}{\out{\hypertarget{method-DeepSeekProcessor-extract_response_content}{}}}
\subsection{Method \code{extract_response_content()}}{
Extract response content from DeepSeek API response
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DeepSeekProcessor$extract_response_content(response, model)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{response}}{HTTP response object}

\item{\code{model}}{Model identifier}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DeepSeekProcessor-clone"></a>}}
\if{latex}{\out{\hypertarget{method-DeepSeekProcessor-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DeepSeekProcessor$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_consensus.R
\name{find_majority_prediction}
\alias{find_majority_prediction}
\title{Find majority prediction from response lines}
\usage{
find_majority_prediction(lines)
}
\description{
Find majority prediction from response lines
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_model_predictions.R
\name{compare_model_predictions}
\alias{compare_model_predictions}
\title{Compare predictions from different models}
\usage{
compare_model_predictions(
  input,
  tissue_name,
  models = c("claude-opus-4-7", "gpt-5.5", "gemini-3.1-pro-preview", "deepseek-v4-flash",
    "qwen3.6-plus", "grok-4.3"),
  api_keys,
  top_gene_count = 10,
  consensus_threshold = 0.5,
  base_urls = NULL
)
}
\arguments{
\item{input}{Either a data frame from Seurat's FindAllMarkers() containing columns 'cluster', 'gene', and 'avg_log2FC', or a list with 'genes' field for each cluster}

\item{tissue_name}{Tissue context (e.g., 'human PBMC', 'mouse brain') for more accurate annotations}

\item{models}{Vector of model names to use for comparison. Default includes top models from each provider}

\item{api_keys}{Named list of API keys for the models, with provider or model names as keys.
Every model in \code{models} must resolve to a non-NULL API key.}

\item{top_gene_count}{Number of top genes to use per cluster when input is from Seurat. Default: 10}

\item{consensus_threshold}{Minimum agreement threshold for consensus (0-1). Default: 0.5.
Consensus is only evaluated when at least two non-missing model predictions are available for a cluster.}

\item{base_urls}{Optional base URLs for API endpoints. Can be a string or named list for provider-specific custom endpoints.}
}
\value{
List containing individual model predictions and consensus analysis
If a cluster has fewer than two valid predictions after alignment/padding,
its consensus-related outputs are \code{NA}.
}
\description{
This function runs the same input through multiple models and compares their predictions.
It provides both individual predictions and a consensus analysis.
}
\note{
This function uses create_standardization_prompt from prompt_templates.R
Supported models:
\itemize{
\item OpenAI: 'gpt-5.5', 'gpt-5.4', 'gpt-5.4-mini'
\item Anthropic: 'claude-opus-4-7', 'claude-opus-4-6', 'claude-sonnet-4-6', 'claude-haiku-4-5-20251001'
\item DeepSeek: 'deepseek-v4-flash', 'deepseek-v4-pro'
\item Google: 'gemini-3.1-pro-preview', 'gemini-3-flash-preview', 'gemini-3.1-flash-lite'
\item Alibaba: 'qwen3.6-max-preview', 'qwen3.6-plus', 'qwen3.6-flash'
\item Stepfun: 'step-3.5-flash', 'step-3.5-flash-2603', 'step-3'
\item Zhipu/Z.AI: 'glm-5.1', 'glm-5-turbo', 'glm-5'
\item MiniMax: 'MiniMax-M2.7', 'MiniMax-M2.7-highspeed', 'MiniMax-M2.5'
\item X.AI: 'grok-4.3', 'grok-4.3-latest', 'grok-latest'
\item OpenRouter: Provides access to models from multiple providers through a single API. Format: 'provider/model-name'
\itemize{
\item OpenAI models: 'openai/gpt-5.5', 'openai/gpt-5.4-mini'
\item Anthropic models: 'anthropic/claude-opus-4.7', 'anthropic/claude-sonnet-4.6'
\item Google models: 'google/gemini-3.1-pro-preview', 'google/gemini-3-flash-preview'
\item X.AI models: 'x-ai/grok-4.3'
\item Stepfun models: 'stepfun/step-3.5-flash'
}
}
\enumerate{
\item With provider names as keys: \code{list("openai" = "sk-...", "anthropic" = "sk-ant-...", "openrouter" = "sk-or-...")}
\item With model names as keys: \code{list("gpt-5.5" = "sk-...", "claude-sonnet-4-6" = "sk-ant-...")}
}

The system first tries to find the API key using the provider name. If not found, it then tries using the model name.
Example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{api_keys <- list(
  "openai" = Sys.getenv("OPENAI_API_KEY"),
  "anthropic" = Sys.getenv("ANTHROPIC_API_KEY"),
  "openrouter" = Sys.getenv("OPENROUTER_API_KEY"),
  "claude-opus-4-7" = "your-claude-opus-key"
)
}\if{html}{\out{</div>}}
}
\examples{
\dontrun{
# Compare predictions using different models
api_keys <- list(
  "claude-sonnet-4-6" = "your-anthropic-key",
  "deepseek-v4-pro" = "your-deepseek-key",
  "gemini-3.1-pro-preview" = "your-gemini-key",
  "qwen3.6-plus" = "your-qwen-key"
)

results <- compare_model_predictions(
  input = list(gs1=c('CD4','CD3D'), gs2='CD14'),
  tissue_name = 'PBMC',
  api_keys = api_keys
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_model_response.R
\name{get_model_response}
\alias{get_model_response}
\title{Get response from a specific model}
\usage{
get_model_response(prompt, model, api_key, base_urls = NULL, normalize = TRUE)
}
\arguments{
\item{prompt}{Non-empty prompt string}

\item{model}{Non-empty model name}

\item{api_key}{Non-empty API key}

\item{base_urls}{Optional shared or provider-specific base URL configuration}

\item{normalize}{Logical; if \code{TRUE} (default) the response is normalized
into non-empty lines, otherwise the raw response string is returned.}
}
\value{
Provider response as a character vector
}
\description{
Get response from a specific model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_consensus.R
\name{normalize_annotation}
\alias{normalize_annotation}
\title{Normalize annotation for comparison}
\usage{
normalize_annotation(annotation)
}
\description{
Normalize annotation for comparison
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prompt_templates.R
\name{create_consensus_check_prompt}
\alias{create_consensus_check_prompt}
\title{Create prompt for checking consensus among model predictions}
\usage{
create_consensus_check_prompt(
  round_responses,
  controversy_threshold = 2/3,
  entropy_threshold = 1
)
}
\description{
Create prompt for checking consensus among model predictions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prompt_templates.R
\name{create_initial_discussion_prompt}
\alias{create_initial_discussion_prompt}
\title{Create prompt for the initial round of discussion}
\usage{
create_initial_discussion_prompt(
  cluster_id,
  cluster_genes,
  tissue_name,
  initial_predictions
)
}
\description{
Create prompt for the initial round of discussion
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus_annotation.R
\name{identify_controversial_clusters}
\alias{identify_controversial_clusters}
\title{Identify controversial clusters based on consensus analysis}
\usage{
identify_controversial_clusters(
  input,
  individual_predictions,
  controversy_threshold,
  entropy_threshold,
  api_keys,
  consensus_check_model = NULL,
  base_urls = NULL
)
}
\description{
Identify controversial clusters based on consensus analysis
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cell_type_annotation.R
\name{annotate_cell_types}
\alias{annotate_cell_types}
\title{Cell Type Annotation with Multi-LLM Framework}
\usage{
annotate_cell_types(
  input,
  tissue_name,
  model = "gpt-5.5",
  api_key = NA,
  top_gene_count = 10,
  debug = FALSE,
  base_urls = NULL,
  return_reasoning = FALSE
)
}
\arguments{
\item{input}{Either a data frame from Seurat's FindAllMarkers() containing columns 'cluster', 'gene', and 'avg_log2FC', or a list with 'genes' field for each cluster}

\item{tissue_name}{Required tissue context (e.g., 'human PBMC', 'mouse brain') for more accurate annotations}

\item{model}{Model name to use. Default: 'gpt-5.5'. See details for supported models}

\item{api_key}{API key for the selected model provider as a non-empty character scalar.
If \code{NA}, returns prompt only.}

\item{top_gene_count}{Number of top genes to use per cluster when input is from Seurat. Default: 10}

\item{debug}{Logical indicating whether to enable debug output. Default: FALSE}

\item{base_urls}{Optional base URLs for API endpoints. Can be a string or named list for custom endpoints}

\item{return_reasoning}{Logical. If \code{TRUE}, returns a structured list per cluster containing
\code{cell_type}, \code{marker_genes}, and \code{gene_expression} fields instead of plain labels.
Default: \code{FALSE}.}
}
\value{
When \code{api_key} is provided and \code{return_reasoning} is \code{FALSE}, the provider response
split by newline as a character vector. When \code{return_reasoning} is \code{TRUE}, a named list
where each element is a list with \code{cell_type}, \code{marker_genes}, and
\code{gene_expression}. When \code{api_key} is \code{NA}, the generated prompt string.
}
\description{
A comprehensive function for automated cell type annotation using multiple Large Language Models (LLMs).
This function supports both Seurat's differential gene expression results and custom gene lists as input.
It implements a sophisticated annotation pipeline that leverages state-of-the-art LLMs to identify
cell types based on marker gene expression patterns.
\itemize{
\item A data frame from Seurat's FindAllMarkers() function containing differential gene expression results
(must have columns: 'cluster', 'gene', and 'avg_log2FC'). The function will select the top genes
based on avg_log2FC for each cluster.
\item A list where each element has a 'genes' field containing marker genes for a cluster.
This can be in one of these formats:
\itemize{
\item Named with cluster IDs: list("0" = list(genes = c(...)), "1" = list(genes = c(...)))
\item Named with cell type names: list(t_cells = list(genes = c(...)), b_cells = list(genes = c(...)))
\item Unnamed list: list(list(genes = c(...)), list(genes = c(...)))
}
\item Cluster IDs are preserved as-is. The function does not modify or re-index cluster IDs.
'mouse brain'). This helps provide context for more accurate annotations.
\item OpenAI: 'gpt-5.5', 'gpt-5.4', 'gpt-5.4-mini'
\item Anthropic: 'claude-opus-4-7', 'claude-opus-4-6', 'claude-sonnet-4-6', 'claude-haiku-4-5-20251001'
\item DeepSeek: 'deepseek-v4-flash', 'deepseek-v4-pro'
\item Google: 'gemini-3.1-pro-preview', 'gemini-3-flash-preview', 'gemini-3.1-flash-lite'
\item Alibaba: 'qwen3.6-max-preview', 'qwen3.6-plus', 'qwen3.6-flash'
\item Stepfun: 'step-3.5-flash', 'step-3.5-flash-2603', 'step-3'
\item Zhipu/Z.AI: 'glm-5.1', 'glm-5-turbo', 'glm-5'
\item MiniMax: 'MiniMax-M2.7', 'MiniMax-M2.7-highspeed', 'MiniMax-M2.5'
\item X.AI: 'grok-4.3', 'grok-4.3-latest', 'grok-latest'
\item OpenRouter: Provides access to models from multiple providers through a single API. Format: 'provider/model-name'
\itemize{
\item OpenAI models: 'openai/gpt-5.5', 'openai/gpt-5.4-mini'
\item Anthropic models: 'anthropic/claude-opus-4.7', 'anthropic/claude-sonnet-4.6'
\item Google models: 'google/gemini-3.1-pro-preview', 'google/gemini-3-flash-preview'
\item X.AI models: 'x-ai/grok-4.3'
\item Stepfun models: 'stepfun/step-3.5-flash'
Each provider requires a specific API key format and authentication method:
}
\item OpenAI: "sk-..." (obtain from OpenAI platform)
\item Anthropic: "sk-ant-..." (obtain from Anthropic console)
\item Google: A Google API key for Gemini models (obtain from Google AI)
\item DeepSeek: API key from DeepSeek platform
\item Qwen: API key from Alibaba Cloud
\item Stepfun: API key from Stepfun AI
\item Zhipu: API key from Zhipu AI
\item MiniMax: API key from MiniMax
\item X.AI: API key for Grok models
\item OpenRouter: "sk-or-..." (obtain from OpenRouter)
OpenRouter provides access to multiple models through a single API key
}

The API key can be provided directly or stored in environment variables:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# Direct API key
result <- annotate_cell_types(input, tissue_name, model="gpt-5.5",
                             api_key="sk-...")

# Using environment variables
Sys.setenv(OPENAI_API_KEY="sk-...")
Sys.setenv(ANTHROPIC_API_KEY="sk-ant-...")
Sys.setenv(OPENROUTER_API_KEY="sk-or-...")

# Then use with environment variables
result <- annotate_cell_types(input, tissue_name, model="claude-sonnet-4-6",
                             api_key=Sys.getenv("ANTHROPIC_API_KEY"))
}\if{html}{\out{</div>}}

If NA, returns the generated prompt without making an API call, which is useful for
reviewing the prompt before sending it to the API.
when input is from Seurat's FindAllMarkers(). Default: 10
\itemize{
\item A single character string: Applied to all providers (e.g., "https://api.proxy.com/v1")
\item A named list: Provider-specific URLs (e.g., list(openai = "https://openai-proxy.com/v1",
anthropic = "https://anthropic-proxy.com/v1")). This is useful for:
\itemize{
\item Users accessing international APIs through proxies
\item Enterprise users with internal API gateways
\item Development/testing with local or alternative endpoints
If NULL (default), uses official API endpoints for each provider.
}
}
}
\examples{
# Example 1: Using custom gene lists, returning prompt only (no API call)
annotate_cell_types(
  input = list(
    t_cells = list(genes = c('CD3D', 'CD3E', 'CD3G', 'CD28')),
    b_cells = list(genes = c('CD19', 'CD79A', 'CD79B', 'MS4A1')),
    monocytes = list(genes = c('CD14', 'CD68', 'CSF1R', 'FCGR3A'))
  ),
  tissue_name = 'human PBMC',
  model = 'gpt-5.5',
  api_key = NA  # Returns prompt only without making API call
)

# Example 2: Using with Seurat pipeline and OpenAI model
\dontrun{
library(Seurat)

# Load example data
data("pbmc_small")

# Find marker genes
all.markers <- FindAllMarkers(
  object = pbmc_small,
  only.pos = TRUE,
  min.pct = 0.25,
  logfc.threshold = 0.25
)

# Set API key in environment variable (recommended approach)
Sys.setenv(OPENAI_API_KEY = "your-openai-api-key")

# Get cell type annotations using OpenAI model
openai_annotations <- annotate_cell_types(
  input = all.markers,
  tissue_name = 'human PBMC',
  model = 'gpt-5.5',
  api_key = Sys.getenv("OPENAI_API_KEY"),
  top_gene_count = 15
)

# Example 3: Using Anthropic Claude model
Sys.setenv(ANTHROPIC_API_KEY = "your-anthropic-api-key")

claude_annotations <- annotate_cell_types(
  input = all.markers,
  tissue_name = 'human PBMC',
  model = 'claude-opus-4-7',
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  top_gene_count = 15
)

# Example 4: Using OpenRouter to access multiple models
Sys.setenv(OPENROUTER_API_KEY = "your-openrouter-api-key")

# Access OpenAI models through OpenRouter
openrouter_gpt4_annotations <- annotate_cell_types(
  input = all.markers,
  tissue_name = 'human PBMC',
  model = 'openai/gpt-5.5',  # Note the provider/model format
  api_key = Sys.getenv("OPENROUTER_API_KEY"),
  top_gene_count = 15
)

# Access Anthropic models through OpenRouter
openrouter_claude_annotations <- annotate_cell_types(
  input = all.markers,
  tissue_name = 'human PBMC',
  model = 'anthropic/claude-opus-4.6',  # Note the provider/model format
  api_key = Sys.getenv("OPENROUTER_API_KEY"),
  top_gene_count = 15
)

# Example 5: Using with mouse brain data
mouse_annotations <- annotate_cell_types(
  input = mouse_markers,  # Your mouse marker genes
  tissue_name = 'mouse brain',  # Specify correct tissue for context
  model = 'gpt-5.5',
  api_key = Sys.getenv("OPENAI_API_KEY"),
  top_gene_count = 20,  # Use more genes for complex tissues
  debug = TRUE  # Enable debug output
)
}

}
\seealso{
\itemize{
\item \code{\link[Seurat:FindAllMarkers]{Seurat::FindAllMarkers()}}
\item \code{\link[=get_provider]{get_provider()}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gemini_processor.R
\name{GeminiProcessor}
\alias{GeminiProcessor}
\title{Gemini API Processor}
\description{
Gemini API Processor

Gemini API Processor
}
\details{
Concrete implementation of BaseAPIProcessor for Gemini models.
Handles Gemini-specific API calls, authentication, and response parsing.
}
\section{Super class}{
\code{\link[mLLMCelltype:BaseAPIProcessor]{mLLMCelltype::BaseAPIProcessor}} -> \code{GeminiProcessor}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-GeminiProcessor-new}{\code{GeminiProcessor$new()}}
\item \href{#method-GeminiProcessor-get_default_api_url}{\code{GeminiProcessor$get_default_api_url()}}
\item \href{#method-GeminiProcessor-get_api_url_for_model}{\code{GeminiProcessor$get_api_url_for_model()}}
\item \href{#method-GeminiProcessor-make_api_call}{\code{GeminiProcessor$make_api_call()}}
\item \href{#method-GeminiProcessor-extract_response_content}{\code{GeminiProcessor$extract_response_content()}}
\item \href{#method-GeminiProcessor-extract_usage}{\code{GeminiProcessor$extract_usage()}}
\item \href{#method-GeminiProcessor-clone}{\code{GeminiProcessor$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="mLLMCelltype" data-topic="BaseAPIProcessor" data-id="get_api_url"><a href='../../mLLMCelltype/html/BaseAPIProcessor.html#method-BaseAPIProcessor-get_api_url'><code>mLLMCelltype::BaseAPIProcessor$get_api_url()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mLLMCelltype" data-topic="BaseAPIProcessor" data-id="process_request"><a href='../../mLLMCelltype/html/BaseAPIProcessor.html#method-BaseAPIProcessor-process_request'><code>mLLMCelltype::BaseAPIProcessor$process_request()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GeminiProcessor-new"></a>}}
\if{latex}{\out{\hypertarget{method-GeminiProcessor-new}{}}}
\subsection{Method \code{new()}}{
Initialize Gemini processor
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GeminiProcessor$new(base_url = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{base_url}}{Optional custom API endpoint}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GeminiProcessor-get_default_api_url"></a>}}
\if{latex}{\out{\hypertarget{method-GeminiProcessor-get_default_api_url}{}}}
\subsection{Method \code{get_default_api_url()}}{
Get default Gemini API URL template
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GeminiProcessor$get_default_api_url()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GeminiProcessor-get_api_url_for_model"></a>}}
\if{latex}{\out{\hypertarget{method-GeminiProcessor-get_api_url_for_model}{}}}
\subsection{Method \code{get_api_url_for_model()}}{
Get API URL for specific model
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GeminiProcessor$get_api_url_for_model(model)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{model}}{Model identifier}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GeminiProcessor-make_api_call"></a>}}
\if{latex}{\out{\hypertarget{method-GeminiProcessor-make_api_call}{}}}
\subsection{Method \code{make_api_call()}}{
Make API call to Gemini
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GeminiProcessor$make_api_call(chunk_content, model, api_key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{chunk_content}}{Prompt text to send}

\item{\code{model}}{Model identifier}

\item{\code{api_key}}{Gemini API key}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GeminiProcessor-extract_response_content"></a>}}
\if{latex}{\out{\hypertarget{method-GeminiProcessor-extract_response_content}{}}}
\subsection{Method \code{extract_response_content()}}{
Extract response content from Gemini API response
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GeminiProcessor$extract_response_content(response, model)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{response}}{HTTP response object}

\item{\code{model}}{Model identifier}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GeminiProcessor-extract_usage"></a>}}
\if{latex}{\out{\hypertarget{method-GeminiProcessor-extract_usage}{}}}
\subsection{Method \code{extract_usage()}}{
Extract normalized Gemini token usage
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GeminiProcessor$extract_usage(response)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{response}}{HTTP response object}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GeminiProcessor-clone"></a>}}
\if{latex}{\out{\hypertarget{method-GeminiProcessor-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GeminiProcessor$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{.onLoad}
\alias{.onLoad}
\title{Package load message}
\usage{
.onLoad(libname, pkgname)
}
\description{
Package load message
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus_annotation.R
\name{extract_positional_annotation}
\alias{extract_positional_annotation}
\title{Extract one line's annotation content for positional mapping}
\usage{
extract_positional_annotation(line, all_clusters)
}
\arguments{
\item{line}{Single response line.}

\item{all_clusters}{Character vector of requested cluster IDs.}
}
\value{
Annotation content string, or \code{NULL} to drop the line.
}
\description{
Returns \code{NULL} only for lines that do not occupy a cluster slot: an
empty-value preamble/header (\code{"Notes:"}) and an explicit label for an
unrequested/nonexistent cluster (\code{"Cluster 999: Noise"}). Every other
line -- including a sentinel such as \code{"Unknown"} -- keeps its slot so
the line->cluster correspondence is preserved (the caller leaves a
non-real slot unassigned). A numbered-list ordinal (\code{"1. T cells"}) or
a resolvable label prefix (\code{"0: T cells"}) is stripped to its content;
a colon that is part of the annotation (\code{"Neurons: excitatory"}) is
kept whole.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_model_manager.R
\name{register_custom_provider}
\alias{register_custom_provider}
\title{Register a custom LLM provider}
\usage{
register_custom_provider(provider_name, process_fn, description = NULL)
}
\arguments{
\item{provider_name}{Unique name for the custom provider}

\item{process_fn}{Function that processes LLM requests. Must accept parameters:
prompt, model, api_key; may optionally accept model_config and base_url}

\item{description}{Optional description of the provider}
}
\value{
Invisible NULL
}
\description{
Register a custom LLM provider
}
\examples{
\dontrun{
register_custom_provider(
  provider_name = "my_provider",
  process_fn = function(prompt, model, api_key) {
    # Custom implementation
    response <- httr::POST(
      url = "your_api_endpoint",
      body = list(prompt = prompt),
      encode = "json"
    )
    return(httr::content(response)$choices[[1]]$text)
  }
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache_manager.R
\name{CacheManager}
\alias{CacheManager}
\title{Cache Manager Class}
\description{
Manages caching of consensus analysis results
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{cache_dir}}{Directory to store cache files. Options:
\itemize{
\item NULL (default): Uses system cache directory
\item "local": Uses .mllmcelltype_cache in current directory
\item "temp": Uses temporary directory
\item Custom path: Any other string is used as directory path
}}

\item{\code{cache_version}}{Current cache version}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CacheManager-new}{\code{CacheManager$new()}}
\item \href{#method-CacheManager-get_cache_dir}{\code{CacheManager$get_cache_dir()}}
\item \href{#method-CacheManager-generate_key}{\code{CacheManager$generate_key()}}
\item \href{#method-CacheManager-save_to_cache}{\code{CacheManager$save_to_cache()}}
\item \href{#method-CacheManager-load_from_cache}{\code{CacheManager$load_from_cache()}}
\item \href{#method-CacheManager-has_cache}{\code{CacheManager$has_cache()}}
\item \href{#method-CacheManager-get_cache_stats}{\code{CacheManager$get_cache_stats()}}
\item \href{#method-CacheManager-clear_cache}{\code{CacheManager$clear_cache()}}
\item \href{#method-CacheManager-validate_cache}{\code{CacheManager$validate_cache()}}
\item \href{#method-CacheManager-clone}{\code{CacheManager$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CacheManager-new"></a>}}
\if{latex}{\out{\hypertarget{method-CacheManager-new}{}}}
\subsection{Method \code{new()}}{
Initialize cache manager
\itemize{
\item NULL (default): Uses system cache directory via \code{tools::R_user_dir()}
\item "local": Uses .mllmcelltype_cache in current directory
\item "temp": Uses temporary directory (cleared on R restart)
\item Custom path: Any other string is used as directory path
}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CacheManager$new(cache_dir = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cache_dir}}{Cache directory selector or custom path}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CacheManager-get_cache_dir"></a>}}
\if{latex}{\out{\hypertarget{method-CacheManager-get_cache_dir}{}}}
\subsection{Method \code{get_cache_dir()}}{
Get actual cache directory path
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CacheManager$get_cache_dir()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CacheManager-generate_key"></a>}}
\if{latex}{\out{\hypertarget{method-CacheManager-generate_key}{}}}
\subsection{Method \code{generate_key()}}{
Generate cache key from input parameters (improved version)
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CacheManager$generate_key(
  input,
  models,
  cluster_id,
  tissue_name = "",
  top_gene_count = 10,
  discussion_context = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{input}}{Marker gene input}

\item{\code{models}}{Model identifiers contributing to the result}

\item{\code{cluster_id}}{Cluster identifier}

\item{\code{tissue_name}}{Tissue context included in the cache key}

\item{\code{top_gene_count}}{Number of marker genes used}

\item{\code{discussion_context}}{Optional normalized inputs that affect a discussion}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CacheManager-save_to_cache"></a>}}
\if{latex}{\out{\hypertarget{method-CacheManager-save_to_cache}{}}}
\subsection{Method \code{save_to_cache()}}{
Save results to cache
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CacheManager$save_to_cache(key, data)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{Valid cache key}

\item{\code{data}}{Result object to persist}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CacheManager-load_from_cache"></a>}}
\if{latex}{\out{\hypertarget{method-CacheManager-load_from_cache}{}}}
\subsection{Method \code{load_from_cache()}}{
Load results from cache
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CacheManager$load_from_cache(key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{Valid cache key}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CacheManager-has_cache"></a>}}
\if{latex}{\out{\hypertarget{method-CacheManager-has_cache}{}}}
\subsection{Method \code{has_cache()}}{
Check if results exist in cache
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CacheManager$has_cache(key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{Valid cache key}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CacheManager-get_cache_stats"></a>}}
\if{latex}{\out{\hypertarget{method-CacheManager-get_cache_stats}{}}}
\subsection{Method \code{get_cache_stats()}}{
Get cache statistics
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CacheManager$get_cache_stats()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CacheManager-clear_cache"></a>}}
\if{latex}{\out{\hypertarget{method-CacheManager-clear_cache}{}}}
\subsection{Method \code{clear_cache()}}{
Clear all cache
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CacheManager$clear_cache(confirm = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{confirm}}{Whether deletion is explicitly confirmed}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CacheManager-validate_cache"></a>}}
\if{latex}{\out{\hypertarget{method-CacheManager-validate_cache}{}}}
\subsection{Method \code{validate_cache()}}{
Validate cache content
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CacheManager$validate_cache(key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{Valid cache key
Return deterministic paths for regular RDS cache files
Resolve a validated cache key to a path inside the cache directory
Write one R object to a path; separated for deterministic failure testing
Persist cache data through a same-directory temporary file
Create stable hash from genes list
Create stable hash from models list
Create stable hash from tissue_name and top_gene_count
Create stable hash from cluster ID}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CacheManager-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CacheManager-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CacheManager$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_consensus_summary.R
\name{print_consensus_summary}
\alias{print_consensus_summary}
\title{Print summary of consensus results}
\usage{
print_consensus_summary(results)
}
\description{
This function prints a detailed summary of the consensus building process,
including initial predictions from all models, uncertainty metrics, and final consensus
for each controversial cluster.
}
\details{
\itemize{
\item initial_results: A list containing individual_predictions, consensus_results, and controversial_clusters
\item final_annotations: A list of final cell type annotations for each cluster
\item controversial_clusters: A character vector of cluster IDs that were controversial
\item discussion_logs: A list of discussion logs for each controversial cluster
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unified_logger.R
\name{configure_logger}
\alias{configure_logger}
\title{Set global logger configuration}
\usage{
configure_logger(level = "INFO", console_output = FALSE, json_format = TRUE)
}
\arguments{
\item{level}{Logging level: "DEBUG", "INFO", "WARN", or "ERROR". Default: "INFO"}

\item{console_output}{Whether to enable console output. Default: FALSE}

\item{json_format}{Whether to use JSON format for log messages. Default: TRUE}
}
\value{
Invisible logger object
}
\description{
Set global logger configuration
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unified_logger.R
\name{logging_functions}
\alias{logging_functions}
\alias{log_debug}
\alias{log_info}
\alias{log_warn}
\alias{log_error}
\title{Convenience functions for logging}
\usage{
log_debug(message, context = NULL)

log_info(message, context = NULL)

log_warn(message, context = NULL)

log_error(message, context = NULL)
}
\arguments{
\item{message}{Log message string}

\item{context}{Optional context information (list or character)}
}
\value{
Invisible NULL
}
\description{
Convenience functions for logging
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_model_predictions.R
\name{standardize_cell_type_names}
\alias{standardize_cell_type_names}
\title{Standardize cell type names using a language model}
\usage{
standardize_cell_type_names(
  predictions,
  models,
  api_keys,
  standardization_model = NULL,
  base_urls = NULL
)
}
\description{
This function takes predictions from multiple models and standardizes the cell type
nomenclature to ensure consistent naming across different models' outputs.
}
\details{
\enumerate{
\item With provider names as keys: \code{list("openai" = "sk-...", "anthropic" = "sk-ant-...", "openrouter" = "sk-or-...")}
\item With model names as keys: \code{list("gpt-5.5" = "sk-...", "claude-sonnet-4-6" = "sk-ant-...")}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facilitate_cluster_discussion.R
\name{facilitate_cluster_discussion}
\alias{facilitate_cluster_discussion}
\title{Facilitate discussion for a controversial cluster}
\usage{
facilitate_cluster_discussion(
  cluster_id,
  input,
  tissue_name,
  models,
  api_keys,
  initial_predictions,
  top_gene_count,
  max_rounds = 3,
  controversy_threshold = 0.7,
  entropy_threshold = 1,
  consensus_check_model = NULL,
  base_urls = NULL
)
}
\description{
Facilitate discussion for a controversial cluster
}
\note{
This function uses create_initial_discussion_prompt and create_discussion_prompt from prompt_templates.R
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prompt_templates.R
\name{normalize_cluster_gene_list}
\alias{normalize_cluster_gene_list}
\title{Normalize list input into a canonical cluster->genes mapping}
\usage{
normalize_cluster_gene_list(input)
}
\arguments{
\item{input}{List input for cluster annotation}
}
\value{
Named list of character vectors (cluster_id -> genes)
}
\description{
For list input, each element can be either:
\enumerate{
\item a list containing a \code{genes} field, or
\item a character vector of genes.
}
}
\details{
Naming rules:
\itemize{
\item unnamed lists are assigned 0-based IDs ("0", "1", ...)
\item numeric names are preserved as-is (e.g., "1", "2", "3" stays unchanged)
\item non-numeric names are preserved as-is
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_model_manager.R
\name{register_custom_model}
\alias{register_custom_model}
\title{Register a custom model for a provider}
\usage{
register_custom_model(model_name, provider_name, model_config = list())
}
\arguments{
\item{model_name}{Unique name for the custom model}

\item{provider_name}{Name of the provider this model belongs to}

\item{model_config}{List of configuration parameters for the model (e.g., temperature, max_tokens)}
}
\value{
Invisible TRUE on success
}
\description{
Register a custom model for a provider
}
\examples{
\dontrun{
register_custom_model(
  model_name = "my_model",
  provider_name = "my_provider",
  model_config = list(
    temperature = 0.7,
    max_tokens = 2000
  )
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unified_logger.R
\name{initialize_logger}
\alias{initialize_logger}
\title{Reinitialize global logger with a specific directory}
\usage{
initialize_logger(log_dir = "logs")
}
\arguments{
\item{log_dir}{Directory for log files}
}
\value{
Invisible logger object
}
\description{
Preserves the current logger configuration (level, size, retention, console/json)
while changing the log directory for a new annotation session.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_model_manager.R
\name{process_custom}
\alias{process_custom}
\title{Process request using custom provider}
\usage{
process_custom(prompt, model, api_key, base_url = NULL)
}
\description{
Process request using custom provider
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_consensus.R
\name{parse_flexible_format}
\alias{parse_flexible_format}
\title{Parse flexible format consensus response}
\usage{
parse_flexible_format(lines)
}
\description{
Parse flexible format consensus response
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus_annotation.R
\name{interactive_consensus_annotation}
\alias{interactive_consensus_annotation}
\title{Interactive consensus building for cell type annotation}
\usage{
interactive_consensus_annotation(
  input,
  tissue_name,
  models = c("claude-opus-4-7", "gpt-5.5", "gemini-3.1-pro-preview", "deepseek-v4-flash",
    "grok-4.3"),
  api_keys,
  top_gene_count = 10,
  controversy_threshold = 0.7,
  entropy_threshold = 1,
  max_discussion_rounds = 3,
  consensus_check_model = NULL,
  log_dir = "logs",
  cache_dir = NULL,
  use_cache = TRUE,
  base_urls = NULL,
  clusters_to_analyze = NULL,
  force_rerun = FALSE
)
}
\arguments{
\item{input}{Either a data frame from Seurat's FindAllMarkers() function containing
differential gene expression results (must have columns: 'cluster', 'gene',
and 'avg_log2FC'), or a list where each element is either a character vector
of genes or a list containing a \code{genes} field.}

\item{tissue_name}{Character string specifying the tissue type for context-aware
cell type annotation (e.g., 'human PBMC', 'mouse brain'). Required.}

\item{models}{Character vector of model names to use for consensus annotation.
Minimum 2 models required. Supports models from OpenAI, Anthropic, DeepSeek,
Google, Alibaba, Stepfun, Zhipu, MiniMax, X.AI, and OpenRouter.}

\item{api_keys}{Named, non-empty list of API keys. Can use provider names as keys
(e.g., "openai", "anthropic") or model names as keys (e.g., "gpt-5").}

\item{top_gene_count}{Integer specifying the number of top marker genes to use
for annotation per cluster (default: 10).}

\item{controversy_threshold}{Numeric value between 0 and 1 for consensus proportion
threshold. Clusters below this threshold are considered controversial (default: 0.7).}

\item{entropy_threshold}{Numeric value for entropy threshold. Higher entropy
indicates more disagreement among models (default: 1.0).}

\item{max_discussion_rounds}{Integer specifying maximum number of discussion rounds
for controversial clusters (default: 3).}

\item{consensus_check_model}{Character string specifying which model to use for
consensus checking. If NULL, uses the first model that succeeds during initial annotation.}

\item{log_dir}{Character scalar specifying directory for log files (default: "logs").
This function reinitializes the session logger with this directory at the start
of each call.}

\item{cache_dir}{Character string or NULL. Cache directory for storing results.
NULL uses system cache, "local" uses current directory, "temp" uses temporary
directory, or specify custom path.}

\item{use_cache}{Logical indicating whether to use caching (default: TRUE).}

\item{base_urls}{Named list or character string specifying custom API base URLs.
Useful for proxies or alternative endpoints. If NULL, uses official endpoints.}

\item{clusters_to_analyze}{Character or numeric vector specifying which clusters
to analyze. If NULL (default), all clusters are analyzed.}

\item{force_rerun}{Logical indicating whether to force rerun of all specified
clusters, ignoring cache. Only affects controversial cluster discussions
(default: FALSE).}
}
\value{
A list containing:
\itemize{
\item \code{initial_results}: Initial voting results, consensus checks, and controversial cluster IDs
\item \code{final_annotations}: Final annotations keyed by cluster ID
\item \code{controversial_clusters}: Clusters identified as controversial
\item \code{discussion_logs}: Detailed discussion logs for controversial clusters
\item \code{session_id}: Logger session identifier
\item \code{voting_results}: Backward-compatible alias of \code{initial_results}
\item \code{discussion_results}: Backward-compatible alias of \code{discussion_logs}
\item \code{final_consensus}: Backward-compatible alias of \code{final_annotations}
}
}
\description{
This function implements an interactive voting and discussion mechanism where multiple LLMs
collaborate to reach a consensus on cell type annotations, particularly focusing on
clusters with low agreement. The process includes:
\enumerate{
\item Initial voting by all LLMs
\item Identification of controversial clusters
\item Detailed discussion for controversial clusters
\item Final summary by a designated LLM (default: Claude)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus_annotation.R
\name{looks_like_cluster_ref}
\alias{looks_like_cluster_ref}
\title{Whether a label key denotes a cluster id (bare number or "Cluster N")}
\usage{
looks_like_cluster_ref(key)
}
\arguments{
\item{key}{Parsed label key.}
}
\value{
\code{TRUE} when the key is an explicit cluster reference.
}
\description{
Whether a label key denotes a cluster id (bare number or "Cluster N")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_model_manager.R
\name{list_custom_providers}
\alias{list_custom_providers}
\title{Get list of registered custom providers}
\usage{
list_custom_providers()
}
\description{
Get list of registered custom providers
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache_utils.R
\name{mllmcelltype_clear_cache}
\alias{mllmcelltype_clear_cache}
\title{Clear mLLMCelltype cache}
\usage{
mllmcelltype_clear_cache(cache_dir = NULL)
}
\arguments{
\item{cache_dir}{Cache directory specification. NULL uses system default, "local" uses current dir, "temp" uses temp dir, or custom path}
}
\value{
Invisible NULL
}
\description{
Clear the mLLMCelltype cache
}
\examples{
\dontrun{
mllmcelltype_clear_cache()
mllmcelltype_clear_cache("local")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unified_logger.R
\name{get_logger}
\alias{get_logger}
\title{Get the global logger instance}
\usage{
get_logger()
}
\description{
Get the global logger instance
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus_annotation.R
\name{get_initial_predictions}
\alias{get_initial_predictions}
\title{Get initial predictions from all models}
\usage{
get_initial_predictions(
  input,
  tissue_name,
  models,
  api_keys,
  top_gene_count,
  base_urls = NULL
)
}
\description{
This function retrieves initial cell type predictions from all specified models.
It is an internal helper function used by the interactive_consensus_annotation function.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensus_annotation.R
\name{select_best_prediction}
\alias{select_best_prediction}
\title{Select the best prediction from consensus results}
\usage{
select_best_prediction(consensus_result, valid_predictions)
}
\description{
Select the best prediction from consensus results
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache_utils.R
\name{mllmcelltype_cache_dir}
\alias{mllmcelltype_cache_dir}
\title{Get mLLMCelltype cache location}
\usage{
mllmcelltype_cache_dir(cache_dir = NULL)
}
\arguments{
\item{cache_dir}{Cache directory specification. NULL uses system default, "local" uses current dir, "temp" uses temp dir, or custom path}
}
\value{
Invisible cache directory path
}
\description{
Display the cache directory location
}
\examples{
\dontrun{
mllmcelltype_cache_dir()
mllmcelltype_cache_dir("local")
}
}
